/* Generated by Cython 0.29.22 */

/* BEGIN: Cython Metadata
{
    "distutils": {
        "depends": [],
        "include_dirs": [
            "/usr/include/gdal"
        ],
        "language": "c++",
        "libraries": [
            "gdal"
        ],
        "library_dirs": [
            "/usr"
        ],
        "name": "pycoalescence.necsim.necsim",
        "sources": [
            "pycoalescence/necsim/necsim.pyx",
            "pycoalescence/lib/necsim/neutral_analytical.cpp",
            "pycoalescence/lib/necsim/ActivityMap.cpp",
            "pycoalescence/lib/necsim/AnalyticalSpeciesAbundancesHandler.cpp",
            "pycoalescence/lib/necsim/Cell.cpp",
            "pycoalescence/lib/necsim/DataMask.cpp",
            "pycoalescence/lib/necsim/DataPoint.cpp",
            "pycoalescence/lib/necsim/GillespieCalculator.cpp",
            "pycoalescence/lib/necsim/Logger.cpp",
            "pycoalescence/lib/necsim/MapLocation.cpp",
            "pycoalescence/lib/necsim/ProtractedTree.cpp",
            "pycoalescence/lib/necsim/SimulatedSpeciesAbundancesHandler.cpp",
            "pycoalescence/lib/necsim/SpeciationCommands.cpp",
            "pycoalescence/lib/necsim/SpeciesAbundancesHandler.cpp",
            "pycoalescence/lib/necsim/SpeciesList.cpp",
            "pycoalescence/lib/necsim/TreeNode.cpp",
            "pycoalescence/lib/necsim/double_comparison.cpp",
            "pycoalescence/lib/necsim/setup.cpp",
            "pycoalescence/lib/necsim/SimulateDispersal.cpp",
            "pycoalescence/lib/necsim/ConfigParser.cpp",
            "pycoalescence/lib/necsim/DispersalCoordinator.cpp",
            "pycoalescence/lib/necsim/Landscape.cpp",
            "pycoalescence/lib/necsim/LogFile.cpp",
            "pycoalescence/lib/necsim/Metacommunity.cpp",
            "pycoalescence/lib/necsim/SQLiteHandler.cpp",
            "pycoalescence/lib/necsim/SpatialTree.cpp",
            "pycoalescence/lib/necsim/Tree.cpp",
            "pycoalescence/lib/necsim/cpl_custom_handler.cpp",
            "pycoalescence/lib/necsim/file_system.cpp",
            "pycoalescence/lib/necsim/parameters.cpp",
            "pycoalescence/lib/necsim/Community.cpp",
            "pycoalescence/lib/necsim/ProtractedSpatialTree.cpp",
            "pycoalescence/lib/PyLogger.cpp",
            "pycoalescence/lib/PyLogging.cpp",
            "pycoalescence/lib/LandscapeMetricsCalculator.cpp"
        ]
    },
    "module_name": "pycoalescence.necsim.necsim"
}
END: Cython Metadata */

#define PY_SSIZE_T_CLEAN
#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02060000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03030000)
    #error Cython requires Python 2.6+ or Python 3.3+.
#else
#define CYTHON_ABI "0_29_22"
#define CYTHON_HEX_VERSION 0x001D16F0
#define CYTHON_FUTURE_DIVISION 1
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef HAVE_LONG_LONG
  #if PY_VERSION_HEX >= 0x02070000
    #define HAVE_LONG_LONG
  #endif
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#ifdef PYPY_VERSION
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#elif defined(PYSTON_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 0
  #elif !defined(CYTHON_USE_PYTYPE_LOOKUP)
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #if PY_MAJOR_VERSION < 3
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 0
  #elif !defined(CYTHON_USE_PYLONG_INTERNALS)
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if PY_VERSION_HEX < 0x030300F0
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #ifndef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL 1
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT (PY_VERSION_HEX >= 0x03050000)
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE (PY_VERSION_HEX >= 0x030400a1)
  #endif
  #ifndef CYTHON_USE_DICT_VERSIONS
    #define CYTHON_USE_DICT_VERSIONS (PY_VERSION_HEX >= 0x030600B1)
  #endif
  #ifndef CYTHON_USE_EXC_INFO_STACK
    #define CYTHON_USE_EXC_INFO_STACK (PY_VERSION_HEX >= 0x030700A3)
  #endif
#endif
#if !defined(CYTHON_FAST_PYCCALL)
#define CYTHON_FAST_PYCCALL  (CYTHON_FAST_PYCALL && PY_VERSION_HEX >= 0x030600B1)
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #include "longintrepr.h"
  #undef SHIFT
  #undef BASE
  #undef MASK
  #ifdef SIZEOF_VOID_P
    enum { __pyx_check_sizeof_voidp = 1 / (int)(SIZEOF_VOID_P == sizeof(void*)) };
  #endif
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_MAYBE_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_MAYBE_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#ifdef _MSC_VER
    #ifndef _MSC_STDINT_H_
        #if _MSC_VER < 1300
           typedef unsigned char     uint8_t;
           typedef unsigned int      uint32_t;
        #else
           typedef unsigned __int8   uint8_t;
           typedef unsigned __int32  uint32_t;
        #endif
    #endif
#else
   #include <stdint.h>
#endif
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus) && __cplusplus >= 201103L
    #if __has_cpp_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH [[fallthrough]]
    #elif __has_cpp_attribute(clang::fallthrough)
      #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
    #elif __has_cpp_attribute(gnu::fallthrough)
      #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__ ) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif

#ifndef __cplusplus
  #error "Cython files generated with the C++ option must be compiled with a C++ compiler."
#endif
#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #else
    #define CYTHON_INLINE inline
  #endif
#endif
template<typename T>
void __Pyx_call_destructor(T& x) {
    x.~T();
}
template<typename T>
class __Pyx_FakeReference {
  public:
    __Pyx_FakeReference() : ptr(NULL) { }
    __Pyx_FakeReference(const T& ref) : ptr(const_cast<T*>(&ref)) { }
    T *operator->() { return ptr; }
    T *operator&() { return ptr; }
    operator T&() { return *ptr; }
    template<typename U> bool operator ==(U other) { return *ptr == other; }
    template<typename U> bool operator !=(U other) { return *ptr != other; }
  private:
    T *ptr;
};

#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x02070600 && !defined(Py_OptimizeFlag)
  #define Py_OptimizeFlag 0
#endif
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyClass_Type
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
#if PY_VERSION_HEX >= 0x030800A4 && PY_VERSION_HEX < 0x030800B2
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, 0, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#else
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#endif
  #define __Pyx_DefaultClassType PyType_Type
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef METH_STACKLESS
  #define METH_STACKLESS 0
#endif
#if PY_VERSION_HEX <= 0x030700A3 || !defined(METH_FASTCALL)
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #define __Pyx_PyCFunctionFast _PyCFunctionFast
  #define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
#endif
#if CYTHON_FAST_PYCCALL
#define __Pyx_PyFastCFunction_Check(func)\
    ((PyCFunction_Check(func) && (METH_FASTCALL == (PyCFunction_GET_FLAGS(func) & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)))))
#else
#define __Pyx_PyFastCFunction_Check(func) 0
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030400A1
  #define PyMem_RawMalloc(n)           PyMem_Malloc(n)
  #define PyMem_RawRealloc(p, n)       PyMem_Realloc(p, n)
  #define PyMem_RawFree(p)             PyMem_Free(p)
#endif
#if CYTHON_COMPILING_IN_PYSTON
  #define __Pyx_PyCode_HasFreeVars(co)  PyCode_HasFreeVars(co)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) PyFrame_SetLineNumber(frame, lineno)
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if !CYTHON_FAST_THREAD_STATE || PY_VERSION_HEX < 0x02070000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x03060000
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#elif PY_VERSION_HEX >= 0x03000000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_Current
#endif
#if PY_VERSION_HEX < 0x030700A2 && !defined(PyThread_tss_create) && !defined(Py_tss_NEEDS_INIT)
#include "pythread.h"
#define Py_tss_NEEDS_INIT 0
typedef int Py_tss_t;
static CYTHON_INLINE int PyThread_tss_create(Py_tss_t *key) {
  *key = PyThread_create_key();
  return 0;
}
static CYTHON_INLINE Py_tss_t * PyThread_tss_alloc(void) {
  Py_tss_t *key = (Py_tss_t *)PyObject_Malloc(sizeof(Py_tss_t));
  *key = Py_tss_NEEDS_INIT;
  return key;
}
static CYTHON_INLINE void PyThread_tss_free(Py_tss_t *key) {
  PyObject_Free(key);
}
static CYTHON_INLINE int PyThread_tss_is_created(Py_tss_t *key) {
  return *key != Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE void PyThread_tss_delete(Py_tss_t *key) {
  PyThread_delete_key(*key);
  *key = Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE int PyThread_tss_set(Py_tss_t *key, void *value) {
  return PyThread_set_key_value(*key, value);
}
static CYTHON_INLINE void * PyThread_tss_get(Py_tss_t *key) {
  return PyThread_get_key_value(*key);
}
#endif
#if CYTHON_COMPILING_IN_CPYTHON || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#if PY_MAJOR_VERSION >= 3 || CYTHON_FUTURE_DIVISION
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1 && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStr(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
#else
#define __Pyx_PyDict_GetItemStr(dict, name)  PyDict_GetItem(dict, name)
#endif
#if PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                              0 : _PyUnicode_Ready((PyObject *)(op)))
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         PyUnicode_KIND(u)
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, ch)
  #if defined(PyUnicode_IS_READY) && defined(PyUnicode_GET_SIZE)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
  #else
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_LENGTH(u))
  #endif
#else
  #define CYTHON_PEP393_ENABLED 0
  #define PyUnicode_1BYTE_KIND  1
  #define PyUnicode_2BYTE_KIND  2
  #define PyUnicode_4BYTE_KIND  4
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((sizeof(Py_UNICODE) == 2) ? 65535 : 1114111)
  #define __Pyx_PyUnicode_KIND(u)         (sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  (((void)(k)), ((Py_UNICODE*)d)[i] = ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_Contains)
  #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyByteArray_Check)
  #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Format)
  #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None || (PyString_Check(b) && !PyString_CheckExact(b)))) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None || (PyUnicode_Check(b) && !PyUnicode_CheckExact(b)))) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
#ifndef PyObject_Unicode
  #define PyObject_Unicode             PyObject_Str
#endif
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        (Py_TYPE(obj) == &PySet_Type)
#endif
#if PY_VERSION_HEX >= 0x030900A4
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_SET_REFCNT(obj, refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SET_SIZE(obj, size)
#else
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_REFCNT(obj) = (refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SIZE(obj) = (size)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
#else
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   PyInt_AsLong
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   PyInt_AsSsize_t
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyMethod_New(func, self, klass) ((self) ? ((void)(klass), PyMethod_New(func, self)) : __Pyx_NewRef(func))
#else
  #define __Pyx_PyMethod_New(func, self, klass) PyMethod_New(func, self, klass)
#endif
#if CYTHON_USE_ASYNC_SLOTS
  #if PY_VERSION_HEX >= 0x030500B1
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
    #define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
  #else
    #define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
  #endif
#else
  #define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef __Pyx_PyAsyncMethodsStruct
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
    } __Pyx_PyAsyncMethodsStruct;
#endif

#if defined(WIN32) || defined(MS_WINDOWS)
  #define _USE_MATH_DEFINES
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif

#define __PYX_MARK_ERR_POS(f_index, lineno) \
    { __pyx_filename = __pyx_f[f_index]; (void)__pyx_filename; __pyx_lineno = lineno; (void)__pyx_lineno; __pyx_clineno = __LINE__; (void)__pyx_clineno; }
#define __PYX_ERR(f_index, lineno, Ln_error) \
    { __PYX_MARK_ERR_POS(f_index, lineno) goto Ln_error; }

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__pycoalescence__necsim__necsim
#define __PYX_HAVE_API__pycoalescence__necsim__necsim
/* Early includes */
#include <string.h>
#include "ios"
#include "new"
#include "stdexcept"
#include "typeinfo"
#include <string>
#include <vector>
#include <memory>
#include "necsim/SpatialTree.h"
#include "necsim/Tree.h"
#include "necsim/ProtractedTree.h"
#include "necsim/ProtractedSpatialTree.h"
#include "necsim/SpecSimParameters.h"
#include "PyLogging.h"
#include "necsim/GenericTree.h"
#include "necsim/Community.h"
#include "necsim/Metacommunity.h"
#include "LandscapeMetricsCalculator.h"
#include "necsim/SimParameters.h"
#include "necsim/SimulateDispersal.h"
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_UTF8 1
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT (PY_MAJOR_VERSION >= 3 && __PYX_DEFAULT_STRING_ENCODING_IS_UTF8)
#define __PYX_DEFAULT_STRING_ENCODING "utf8"
#define __Pyx_PyObject_FromString __Pyx_PyUnicode_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
static CYTHON_INLINE int __Pyx_is_valid_index(Py_ssize_t i, Py_ssize_t limit) {
    return (size_t) i < (size_t) limit;
}
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
#define __Pyx_PyByteArray_FromString(s) PyByteArray_FromStringAndSize((const char*)s, strlen((const char*)s))
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyObject_AsWritableString(s)    ((char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const Py_UNICODE *u) {
    const Py_UNICODE *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#define __Pyx_PyUnicode_FromUnicode(u)       PyUnicode_FromUnicode(u, __Pyx_Py_UNICODE_strlen(u))
#define __Pyx_PyUnicode_FromUnicodeAndLength PyUnicode_FromUnicode
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
#if CYTHON_ASSUME_SAFE_MACROS
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#define __Pyx_PyNumber_Float(x) (PyFloat_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Float(x))
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c) + 1);
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }

static PyObject *__pyx_m = NULL;
static PyObject *__pyx_d;
static PyObject *__pyx_b;
static PyObject *__pyx_cython_runtime = NULL;
static PyObject *__pyx_empty_tuple;
static PyObject *__pyx_empty_bytes;
static PyObject *__pyx_empty_unicode;
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm= __FILE__;
static const char *__pyx_filename;


static const char *__pyx_f[] = {
  "pycoalescence/necsim/necsim.pyx",
  "stringsource",
};

/*--- Type declarations ---*/
struct __pyx_obj_13pycoalescence_6necsim_6necsim_CSpatialSimulation;
struct __pyx_obj_13pycoalescence_6necsim_6necsim_CNSESimulation;
struct __pyx_obj_13pycoalescence_6necsim_6necsim_CPNSESimulation;
struct __pyx_obj_13pycoalescence_6necsim_6necsim_CPSpatialSimulation;
struct __pyx_obj_13pycoalescence_6necsim_6necsim_CCommunity;
struct __pyx_obj_13pycoalescence_6necsim_6necsim_CMetacommunity;
struct __pyx_obj_13pycoalescence_6necsim_6necsim_CLandscapeMetricsCalculator;
struct __pyx_obj_13pycoalescence_6necsim_6necsim_CDispersalSimulator;

/* "pycoalescence/necsim/necsim.pyx":19
 * 
 * 
 * cdef class CSpatialSimulation:             # <<<<<<<<<<<<<<
 *     cdef GenericTree[SpatialTree] c_simulation  # Hold a C++ instance which we're wrapping
 *     cdef object logger
 */
struct __pyx_obj_13pycoalescence_6necsim_6necsim_CSpatialSimulation {
  PyObject_HEAD
  necsim::GenericTree<necsim::SpatialTree>  c_simulation;
  PyObject *logger;
};


/* "pycoalescence/necsim/necsim.pyx":68
 *         self.c_simulation.applyMultipleRates()
 * 
 * cdef class CNSESimulation:             # <<<<<<<<<<<<<<
 *     cdef GenericTree[Tree] c_simulation  # Hold a C++ instance which we're wrapping
 *     cdef object logger
 */
struct __pyx_obj_13pycoalescence_6necsim_6necsim_CNSESimulation {
  PyObject_HEAD
  necsim::GenericTree<necsim::Tree>  c_simulation;
  PyObject *logger;
};


/* "pycoalescence/necsim/necsim.pyx":117
 *         self.c_simulation.applyMultipleRates()
 * 
 * cdef class CPNSESimulation:             # <<<<<<<<<<<<<<
 *     cdef GenericTree[ProtractedTree] c_simulation  # Hold a C++ instance which we're wrapping
 *     cdef object logger
 */
struct __pyx_obj_13pycoalescence_6necsim_6necsim_CPNSESimulation {
  PyObject_HEAD
  necsim::GenericTree<necsim::ProtractedTree>  c_simulation;
  PyObject *logger;
};


/* "pycoalescence/necsim/necsim.pyx":167
 * 
 * 
 * cdef class CPSpatialSimulation:             # <<<<<<<<<<<<<<
 *     cdef GenericTree[ProtractedSpatialTree] c_simulation  # Hold a C++ instance which we're wrapping
 *     cdef object logger
 */
struct __pyx_obj_13pycoalescence_6necsim_6necsim_CPSpatialSimulation {
  PyObject_HEAD
  necsim::GenericTree<necsim::ProtractedSpatialTree>  c_simulation;
  PyObject *logger;
};


/* "pycoalescence/necsim/necsim.pyx":216
 *         self.c_simulation.applyMultipleRates()
 * 
 * cdef class CCommunity:             # <<<<<<<<<<<<<<
 *     cdef Community c_community  # Hold a C++ instance which we're wrapping
 *     cdef shared_ptr[SpecSimParameters] c_spec_parameters
 */
struct __pyx_obj_13pycoalescence_6necsim_6necsim_CCommunity {
  PyObject_HEAD
  necsim::Community c_community;
  std::shared_ptr<struct necsim::SpecSimParameters>  c_spec_parameters;
  PyObject *logger;
};


/* "pycoalescence/necsim/necsim.pyx":281
 *         self.c_community.speciateRemainingLineages(file_str)
 * 
 * cdef class CMetacommunity:             # <<<<<<<<<<<<<<
 *     cdef Metacommunity c_community  # Hold a C++ instance which we're wrapping
 *     cdef shared_ptr[SpecSimParameters] c_spec_parameters
 */
struct __pyx_obj_13pycoalescence_6necsim_6necsim_CMetacommunity {
  PyObject_HEAD
  necsim::Metacommunity c_community;
  std::shared_ptr<struct necsim::SpecSimParameters>  c_spec_parameters;
  PyObject *logger;
};


/* "pycoalescence/necsim/necsim.pyx":346
 *         self.c_community.speciateRemainingLineages(file_str)
 * 
 * cdef class CLandscapeMetricsCalculator:             # <<<<<<<<<<<<<<
 *     cdef LandscapeMetricsCalculator c_landscapes  # Hold a C++ instance which we're wrapping
 *     cdef object logger
 */
struct __pyx_obj_13pycoalescence_6necsim_6necsim_CLandscapeMetricsCalculator {
  PyObject_HEAD
  LandscapeMetricsCalculator c_landscapes;
  PyObject *logger;
};


/* "pycoalescence/necsim/necsim.pyx":370
 *         return self.c_landscapes.calculateClumpiness()
 * 
 * cdef class CDispersalSimulator:             # <<<<<<<<<<<<<<
 *     cdef SimulateDispersal c_simulator
 *     cdef shared_ptr[SimParameters] c_sim_parameters
 */
struct __pyx_obj_13pycoalescence_6necsim_6necsim_CDispersalSimulator {
  PyObject_HEAD
  necsim::SimulateDispersal c_simulator;
  std::shared_ptr<necsim::SimParameters>  c_sim_parameters;
  PyObject *logger;
  PyObject *output_database;
  bool has_imported_maps;
  bool needs_update;
  bool printing;
};


/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, int);
    void (*DECREF)(void*, PyObject*, int);
    void (*GOTREF)(void*, PyObject*, int);
    void (*GIVEREF)(void*, PyObject*, int);
    void* (*SetupContext)(const char*, int, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__)
#endif
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_XINCREF(r)  do { if((r) != NULL) {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) != NULL) {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) != NULL) {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) != NULL) {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyObjectGetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* RaiseDoubleKeywords.proto */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywords.proto */
static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject **argnames[],\
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args,\
    const char* function_name);

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* PyDictVersioning.proto */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
#define __PYX_DICT_VERSION_INIT  ((PY_UINT64_T) -1)
#define __PYX_GET_DICT_VERSION(dict)  (((PyDictObject*)(dict))->ma_version_tag)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)\
    (version_var) = __PYX_GET_DICT_VERSION(dict);\
    (cache_var) = (value);
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP) {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    if (likely(__PYX_GET_DICT_VERSION(DICT) == __pyx_dict_version)) {\
        (VAR) = __pyx_dict_cached_value;\
    } else {\
        (VAR) = __pyx_dict_cached_value = (LOOKUP);\
        __pyx_dict_version = __PYX_GET_DICT_VERSION(DICT);\
    }\
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj);
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj);
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version);
#else
#define __PYX_GET_DICT_VERSION(dict)  (0)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP)  (VAR) = (LOOKUP);
#endif

/* GetModuleGlobalName.proto */
#if CYTHON_USE_DICT_VERSIONS
#define __Pyx_GetModuleGlobalName(var, name)  {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    (var) = (likely(__pyx_dict_version == __PYX_GET_DICT_VERSION(__pyx_d))) ?\
        (likely(__pyx_dict_cached_value) ? __Pyx_NewRef(__pyx_dict_cached_value) : __Pyx_GetBuiltinName(name)) :\
        __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
}
#define __Pyx_GetModuleGlobalNameUncached(var, name)  {\
    PY_UINT64_T __pyx_dict_version;\
    PyObject *__pyx_dict_cached_value;\
    (var) = __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
}
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value);
#else
#define __Pyx_GetModuleGlobalName(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name);
#endif

/* PyFunctionFastCall.proto */
#if CYTHON_FAST_PYCALL
#define __Pyx_PyFunction_FastCall(func, args, nargs)\
    __Pyx_PyFunction_FastCallDict((func), (args), (nargs), NULL)
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs);
#else
#define __Pyx_PyFunction_FastCallDict(func, args, nargs, kwargs) _PyFunction_FastCallDict(func, args, nargs, kwargs)
#endif
#define __Pyx_BUILD_ASSERT_EXPR(cond)\
    (sizeof(char [1 - 2*!(cond)]) - 1)
#ifndef Py_MEMBER_SIZE
#define Py_MEMBER_SIZE(type, member) sizeof(((type *)0)->member)
#endif
  static size_t __pyx_pyframe_localsplus_offset = 0;
  #include "frameobject.h"
  #define __Pxy_PyFrame_Initialize_Offsets()\
    ((void)__Pyx_BUILD_ASSERT_EXPR(sizeof(PyFrameObject) == offsetof(PyFrameObject, f_localsplus) + Py_MEMBER_SIZE(PyFrameObject, f_localsplus)),\
     (void)(__pyx_pyframe_localsplus_offset = ((size_t)PyFrame_Type.tp_basicsize) - Py_MEMBER_SIZE(PyFrameObject, f_localsplus)))
  #define __Pyx_PyFrame_GetLocalsplus(frame)\
    (assert(__pyx_pyframe_localsplus_offset), (PyObject **)(((char *)(frame)) + __pyx_pyframe_localsplus_offset))
#endif

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* PyObjectCallMethO.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectCallNoArg.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func);
#else
#define __Pyx_PyObject_CallNoArg(func) __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL)
#endif

/* PyCFunctionFastCall.proto */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject *__Pyx_PyCFunction_FastCall(PyObject *func, PyObject **args, Py_ssize_t nargs);
#else
#define __Pyx_PyCFunction_FastCall(func, args, nargs)  (assert(0), NULL)
#endif

/* PyObjectCallOneArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* PyThreadStateGet.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#define __Pyx_PyErr_Occurred()  __pyx_tstate->curexc_type
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  PyErr_Occurred()
#endif

/* PyErrFetchRestore.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* RaiseException.proto */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* IncludeStringH.proto */
#include <string.h>

/* BytesEquals.proto */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals);

/* UnicodeEquals.proto */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals);

/* PyObjectCall2Args.proto */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2);

/* ListCompAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_ListComp_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len)) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        __Pyx_SET_SIZE(list, len + 1);
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_ListComp_Append(L,x) PyList_Append(L,x)
#endif

/* PyObject_GenericGetAttrNoDict.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttrNoDict PyObject_GenericGetAttr
#endif

/* PyObject_GenericGetAttr.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttr PyObject_GenericGetAttr
#endif

/* PyErrExceptionMatches.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* PyObjectGetAttrStrNoError.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name);

/* SetupReduce.proto */
static int __Pyx_setup_reduce(PyObject* type_obj);

/* Import.proto */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level);

/* ImportFrom.proto */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name);

/* CLineInTraceback.proto */
#ifdef CYTHON_CLINE_IN_TRACEBACK
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#else
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#endif

/* CodeObjectCache.proto */
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* None.proto */
#include <new>

/* GCCDiagnostics.proto */
#if defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 6))
#define __Pyx_HAS_GCC_DIAGNOSTIC
#endif

/* CppExceptionConversion.proto */
#ifndef __Pyx_CppExn2PyErr
#include <new>
#include <typeinfo>
#include <stdexcept>
#include <ios>
static void __Pyx_CppExn2PyErr() {
  try {
    if (PyErr_Occurred())
      ; // let the latest Python exn pass through and ignore the current one
    else
      throw;
  } catch (const std::bad_alloc& exn) {
    PyErr_SetString(PyExc_MemoryError, exn.what());
  } catch (const std::bad_cast& exn) {
    PyErr_SetString(PyExc_TypeError, exn.what());
  } catch (const std::bad_typeid& exn) {
    PyErr_SetString(PyExc_TypeError, exn.what());
  } catch (const std::domain_error& exn) {
    PyErr_SetString(PyExc_ValueError, exn.what());
  } catch (const std::invalid_argument& exn) {
    PyErr_SetString(PyExc_ValueError, exn.what());
  } catch (const std::ios_base::failure& exn) {
    PyErr_SetString(PyExc_IOError, exn.what());
  } catch (const std::out_of_range& exn) {
    PyErr_SetString(PyExc_IndexError, exn.what());
  } catch (const std::overflow_error& exn) {
    PyErr_SetString(PyExc_OverflowError, exn.what());
  } catch (const std::range_error& exn) {
    PyErr_SetString(PyExc_ArithmeticError, exn.what());
  } catch (const std::underflow_error& exn) {
    PyErr_SetString(PyExc_ArithmeticError, exn.what());
  } catch (const std::exception& exn) {
    PyErr_SetString(PyExc_RuntimeError, exn.what());
  }
  catch (...)
  {
    PyErr_SetString(PyExc_RuntimeError, "Unknown exception");
  }
}
#endif

/* CIntFromPy.proto */
static CYTHON_INLINE unsigned long __Pyx_PyInt_As_unsigned_long(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_unsigned_long(unsigned long value);

/* CIntFromPy.proto */
static CYTHON_INLINE size_t __Pyx_PyInt_As_size_t(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
#define __Pyx_PyErr_GivenExceptionMatches2(err, type1, type2) (PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2))
#endif
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(void);

/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);


/* Module declarations from 'libcpp' */

/* Module declarations from 'libc.string' */

/* Module declarations from 'libcpp.string' */

/* Module declarations from 'libcpp.vector' */

/* Module declarations from 'libcpp.memory' */

/* Module declarations from 'pycoalescence.necsim.necsim' */
static PyTypeObject *__pyx_ptype_13pycoalescence_6necsim_6necsim_CSpatialSimulation = 0;
static PyTypeObject *__pyx_ptype_13pycoalescence_6necsim_6necsim_CNSESimulation = 0;
static PyTypeObject *__pyx_ptype_13pycoalescence_6necsim_6necsim_CPNSESimulation = 0;
static PyTypeObject *__pyx_ptype_13pycoalescence_6necsim_6necsim_CPSpatialSimulation = 0;
static PyTypeObject *__pyx_ptype_13pycoalescence_6necsim_6necsim_CCommunity = 0;
static PyTypeObject *__pyx_ptype_13pycoalescence_6necsim_6necsim_CMetacommunity = 0;
static PyTypeObject *__pyx_ptype_13pycoalescence_6necsim_6necsim_CLandscapeMetricsCalculator = 0;
static PyTypeObject *__pyx_ptype_13pycoalescence_6necsim_6necsim_CDispersalSimulator = 0;
static std::string __pyx_convert_string_from_py_std__in_string(PyObject *); /*proto*/
static std::vector<double>  __pyx_convert_vector_from_py_double(PyObject *); /*proto*/
static std::vector<std::string>  __pyx_convert_vector_from_py_std_3a__3a_string(PyObject *); /*proto*/
static std::vector<unsigned long>  __pyx_convert_vector_from_py_unsigned_long(PyObject *); /*proto*/
static std::vector<long>  __pyx_convert_vector_from_py_long(PyObject *); /*proto*/
static std::vector<long double>  __pyx_convert_vector_from_py_long__double(PyObject *); /*proto*/
static PyObject *__pyx_convert_vector_to_py_unsigned_long(const std::vector<unsigned long>  &); /*proto*/
#define __Pyx_MODULE_NAME "pycoalescence.necsim.necsim"
extern int __pyx_module_is_main_pycoalescence__necsim__necsim;
int __pyx_module_is_main_pycoalescence__necsim__necsim = 0;

/* Implementation of 'pycoalescence.necsim.necsim' */
static PyObject *__pyx_builtin_RuntimeError;
static PyObject *__pyx_builtin_TypeError;
static PyObject *__pyx_builtin_range;
static const char __pyx_k_tau[] = "tau";
static const char __pyx_k_List[] = "List";
static const char __pyx_k_deme[] = "deme";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_name[] = "__name__";
static const char __pyx_k_none[] = "none";
static const char __pyx_k_seed[] = "seed";
static const char __pyx_k_task[] = "task";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_range[] = "range";
static const char __pyx_k_setup[] = "setup";
static const char __pyx_k_sigma[] = "sigma";
static const char __pyx_k_times[] = "times";
static const char __pyx_k_closed[] = "closed";
static const char __pyx_k_cutoff[] = "cutoff";
static const char __pyx_k_import[] = "__import__";
static const char __pyx_k_logger[] = "logger";
static const char __pyx_k_m_prob[] = "m_prob";
static const char __pyx_k_reduce[] = "__reduce__";
static const char __pyx_k_typing[] = "typing";
static const char __pyx_k_database[] = "database";
static const char __pyx_k_getstate[] = "__getstate__";
static const char __pyx_k_max_time[] = "max_time";
static const char __pyx_k_setstate[] = "__setstate__";
static const char __pyx_k_TypeError[] = "TypeError";
static const char __pyx_k_path_fine[] = "path_fine";
static const char __pyx_k_rate_fine[] = "rate_fine";
static const char __pyx_k_reduce_ex[] = "__reduce_ex__";
static const char __pyx_k_samples_x[] = "samples_x";
static const char __pyx_k_samples_y[] = "samples_y";
static const char __pyx_k_time_fine[] = "time_fine";
static const char __pyx_k_CCommunity[] = "CCommunity";
static const char __pyx_k_base_setup[] = "base_setup";
static const char __pyx_k_set_logger[] = "set_logger";
static const char __pyx_k_number_fine[] = "number_fine";
static const char __pyx_k_path_coarse[] = "path_coarse";
static const char __pyx_k_rate_coarse[] = "rate_coarse";
static const char __pyx_k_sample_file[] = "sample_file";
static const char __pyx_k_time_coarse[] = "time_coarse";
static const char __pyx_k_use_spatial[] = "use_spatial";
static const char __pyx_k_RuntimeError[] = "RuntimeError";
static const char __pyx_k_number_steps[] = "number_steps";
static const char __pyx_k_write_to_log[] = "write_to_log";
static const char __pyx_k_fine_map_file[] = "fine_map_file";
static const char __pyx_k_is_sequential[] = "is_sequential";
static const char __pyx_k_number_coarse[] = "number_coarse";
static const char __pyx_k_out_directory[] = "out_directory";
static const char __pyx_k_reduce_cython[] = "__reduce_cython__";
static const char __pyx_k_restrict_self[] = "restrict_self";
static const char __pyx_k_CMetacommunity[] = "CMetacommunity";
static const char __pyx_k_CNSESimulation[] = "CNSESimulation";
static const char __pyx_k_dispersal_file[] = "dispersal_file";
static const char __pyx_k_landscape_type[] = "landscape_type";
static const char __pyx_k_number_repeats[] = "number_repeats";
static const char __pyx_k_number_workers[] = "number_workers";
static const char __pyx_k_write_database[] = "write_database";
static const char __pyx_k_CPNSESimulation[] = "CPNSESimulation";
static const char __pyx_k_check_completed[] = "check_completed";
static const char __pyx_k_coarse_map_file[] = "coarse_map_file";
static const char __pyx_k_fine_map_x_size[] = "fine_map_x_size";
static const char __pyx_k_fine_map_y_size[] = "fine_map_y_size";
static const char __pyx_k_pause_directory[] = "pause_directory";
static const char __pyx_k_setstate_cython[] = "__setstate_cython__";
static const char __pyx_k_speciation_rate[] = "speciation_rate";
static const char __pyx_k_coarse_map_scale[] = "coarse_map_scale";
static const char __pyx_k_dispersal_method[] = "dispersal_method";
static const char __pyx_k_import_maps_base[] = "import_maps_base";
static const char __pyx_k_record_fragments[] = "record_fragments";
static const char __pyx_k_speciation_rates[] = "speciation_rates";
static const char __pyx_k_coarse_map_x_size[] = "coarse_map_x_size";
static const char __pyx_k_coarse_map_y_size[] = "coarse_map_y_size";
static const char __pyx_k_fine_map_x_offset[] = "fine_map_x_offset";
static const char __pyx_k_fine_map_y_offset[] = "fine_map_y_offset";
static const char __pyx_k_sample_map_x_size[] = "sample_map_x_size";
static const char __pyx_k_sample_map_y_size[] = "sample_map_y_size";
static const char __pyx_k_CSpatialSimulation[] = "CSpatialSimulation";
static const char __pyx_k_cline_in_traceback[] = "cline_in_traceback";
static const char __pyx_k_max_speciation_gen[] = "max_speciation_gen";
static const char __pyx_k_metacommunity_size[] = "metacommunity_size";
static const char __pyx_k_min_speciation_gen[] = "min_speciation_gen";
static const char __pyx_k_CDispersalSimulator[] = "CDispersalSimulator";
static const char __pyx_k_CPSpatialSimulation[] = "CPSpatialSimulation";
static const char __pyx_k_DISPERSAL_DISTANCES[] = "DISPERSAL_DISTANCES";
static const char __pyx_k_DISTANCES_TRAVELLED[] = "DISTANCES_TRAVELLED";
static const char __pyx_k_coarse_map_x_offset[] = "coarse_map_x_offset";
static const char __pyx_k_coarse_map_y_offset[] = "coarse_map_y_offset";
static const char __pyx_k_metacommunity_option[] = "metacommunity_option";
static const char __pyx_k_dispersal_relative_cost[] = "dispersal_relative_cost";
static const char __pyx_k_metacommunity_reference[] = "metacommunity_reference";
static const char __pyx_k_CLandscapeMetricsCalculator[] = "CLandscapeMetricsCalculator";
static const char __pyx_k_Couldn_t_find_paused_simulation[] = "Couldn't find paused simulation";
static const char __pyx_k_Maps_have_already_been_imported[] = "Maps have already been imported";
static const char __pyx_k_pycoalescence_system_operations[] = "pycoalescence.system_operations";
static const char __pyx_k_Maps_have_not_been_imported_cann[] = "Maps have not been imported - cannot start simulations.";
static const char __pyx_k_Output_database_has_not_been_set[] = "Output database has not been set.";
static const char __pyx_k_no_default___reduce___due_to_non[] = "no default __reduce__ due to non-trivial __cinit__";
static PyObject *__pyx_n_s_CCommunity;
static PyObject *__pyx_n_s_CDispersalSimulator;
static PyObject *__pyx_n_s_CLandscapeMetricsCalculator;
static PyObject *__pyx_n_s_CMetacommunity;
static PyObject *__pyx_n_s_CNSESimulation;
static PyObject *__pyx_n_s_CPNSESimulation;
static PyObject *__pyx_n_s_CPSpatialSimulation;
static PyObject *__pyx_n_s_CSpatialSimulation;
static PyObject *__pyx_kp_u_Couldn_t_find_paused_simulation;
static PyObject *__pyx_n_u_DISPERSAL_DISTANCES;
static PyObject *__pyx_n_u_DISTANCES_TRAVELLED;
static PyObject *__pyx_n_s_List;
static PyObject *__pyx_kp_u_Maps_have_already_been_imported;
static PyObject *__pyx_kp_u_Maps_have_not_been_imported_cann;
static PyObject *__pyx_kp_u_Output_database_has_not_been_set;
static PyObject *__pyx_n_s_RuntimeError;
static PyObject *__pyx_n_s_TypeError;
static PyObject *__pyx_n_s_base_setup;
static PyObject *__pyx_n_s_check_completed;
static PyObject *__pyx_n_s_cline_in_traceback;
static PyObject *__pyx_n_b_closed;
static PyObject *__pyx_n_s_coarse_map_file;
static PyObject *__pyx_n_s_coarse_map_scale;
static PyObject *__pyx_n_s_coarse_map_x_offset;
static PyObject *__pyx_n_s_coarse_map_x_size;
static PyObject *__pyx_n_s_coarse_map_y_offset;
static PyObject *__pyx_n_s_coarse_map_y_size;
static PyObject *__pyx_n_s_cutoff;
static PyObject *__pyx_n_s_database;
static PyObject *__pyx_n_s_deme;
static PyObject *__pyx_n_s_dispersal_file;
static PyObject *__pyx_n_s_dispersal_method;
static PyObject *__pyx_n_s_dispersal_relative_cost;
static PyObject *__pyx_n_s_fine_map_file;
static PyObject *__pyx_n_s_fine_map_x_offset;
static PyObject *__pyx_n_s_fine_map_x_size;
static PyObject *__pyx_n_s_fine_map_y_offset;
static PyObject *__pyx_n_s_fine_map_y_size;
static PyObject *__pyx_n_s_getstate;
static PyObject *__pyx_n_s_import;
static PyObject *__pyx_n_s_import_maps_base;
static PyObject *__pyx_n_s_is_sequential;
static PyObject *__pyx_n_s_landscape_type;
static PyObject *__pyx_n_s_logger;
static PyObject *__pyx_n_s_m_prob;
static PyObject *__pyx_n_s_main;
static PyObject *__pyx_n_s_max_speciation_gen;
static PyObject *__pyx_n_s_max_time;
static PyObject *__pyx_n_s_metacommunity_option;
static PyObject *__pyx_n_s_metacommunity_reference;
static PyObject *__pyx_n_s_metacommunity_size;
static PyObject *__pyx_n_s_min_speciation_gen;
static PyObject *__pyx_n_s_name;
static PyObject *__pyx_kp_s_no_default___reduce___due_to_non;
static PyObject *__pyx_n_u_none;
static PyObject *__pyx_n_s_number_coarse;
static PyObject *__pyx_n_s_number_fine;
static PyObject *__pyx_n_s_number_repeats;
static PyObject *__pyx_n_s_number_steps;
static PyObject *__pyx_n_s_number_workers;
static PyObject *__pyx_n_s_out_directory;
static PyObject *__pyx_n_s_path_coarse;
static PyObject *__pyx_n_s_path_fine;
static PyObject *__pyx_n_s_pause_directory;
static PyObject *__pyx_n_s_pycoalescence_system_operations;
static PyObject *__pyx_n_s_range;
static PyObject *__pyx_n_s_rate_coarse;
static PyObject *__pyx_n_s_rate_fine;
static PyObject *__pyx_n_s_record_fragments;
static PyObject *__pyx_n_s_reduce;
static PyObject *__pyx_n_s_reduce_cython;
static PyObject *__pyx_n_s_reduce_ex;
static PyObject *__pyx_n_s_restrict_self;
static PyObject *__pyx_n_s_sample_file;
static PyObject *__pyx_n_s_sample_map_x_size;
static PyObject *__pyx_n_s_sample_map_y_size;
static PyObject *__pyx_n_s_samples_x;
static PyObject *__pyx_n_s_samples_y;
static PyObject *__pyx_n_s_seed;
static PyObject *__pyx_n_s_set_logger;
static PyObject *__pyx_n_s_setstate;
static PyObject *__pyx_n_s_setstate_cython;
static PyObject *__pyx_n_s_setup;
static PyObject *__pyx_n_s_sigma;
static PyObject *__pyx_n_s_speciation_rate;
static PyObject *__pyx_n_s_speciation_rates;
static PyObject *__pyx_n_s_task;
static PyObject *__pyx_n_s_tau;
static PyObject *__pyx_n_s_test;
static PyObject *__pyx_n_s_time_coarse;
static PyObject *__pyx_n_s_time_fine;
static PyObject *__pyx_n_s_times;
static PyObject *__pyx_n_s_typing;
static PyObject *__pyx_n_s_use_spatial;
static PyObject *__pyx_n_s_write_database;
static PyObject *__pyx_n_s_write_to_log;
static int __pyx_pf_13pycoalescence_6necsim_6necsim_18CSpatialSimulation___cinit__(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CSpatialSimulation *__pyx_v_self, PyObject *__pyx_v_logger); /* proto */
static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_18CSpatialSimulation_2set_logger(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CSpatialSimulation *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_18CSpatialSimulation_4import_from_config(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CSpatialSimulation *__pyx_v_self, PyObject *__pyx_v_config_file); /* proto */
static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_18CSpatialSimulation_6import_from_config_string(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CSpatialSimulation *__pyx_v_self, PyObject *__pyx_v_config); /* proto */
static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_18CSpatialSimulation_8setup(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CSpatialSimulation *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_18CSpatialSimulation_10setup_resume(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CSpatialSimulation *__pyx_v_self, PyObject *__pyx_v_pause_directory, PyObject *__pyx_v_out_directory, PyObject *__pyx_v_seed, PyObject *__pyx_v_task, PyObject *__pyx_v_max_time); /* proto */
static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_18CSpatialSimulation_12add_gillespie(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CSpatialSimulation *__pyx_v_self, PyObject *__pyx_v_gillespie_threshold); /* proto */
static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_18CSpatialSimulation_14run(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CSpatialSimulation *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_18CSpatialSimulation_16apply_speciation_rates(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CSpatialSimulation *__pyx_v_self, PyObject *__pyx_v_speciation_rates); /* proto */
static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_18CSpatialSimulation_18__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_13pycoalescence_6necsim_6necsim_CSpatialSimulation *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_18CSpatialSimulation_20__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_13pycoalescence_6necsim_6necsim_CSpatialSimulation *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_13pycoalescence_6necsim_6necsim_14CNSESimulation___cinit__(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CNSESimulation *__pyx_v_self, PyObject *__pyx_v_logger); /* proto */
static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_14CNSESimulation_2set_logger(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CNSESimulation *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_14CNSESimulation_4import_from_config(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CNSESimulation *__pyx_v_self, PyObject *__pyx_v_config_file); /* proto */
static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_14CNSESimulation_6import_from_config_string(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CNSESimulation *__pyx_v_self, PyObject *__pyx_v_config); /* proto */
static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_14CNSESimulation_8setup(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CNSESimulation *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_14CNSESimulation_10setup_resume(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CNSESimulation *__pyx_v_self, PyObject *__pyx_v_pause_directory, PyObject *__pyx_v_out_directory, PyObject *__pyx_v_seed, PyObject *__pyx_v_task, PyObject *__pyx_v_max_time); /* proto */
static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_14CNSESimulation_12add_gillespie(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CNSESimulation *__pyx_v_self, PyObject *__pyx_v_gillespie_threshold); /* proto */
static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_14CNSESimulation_14run(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CNSESimulation *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_14CNSESimulation_16apply_speciation_rates(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CNSESimulation *__pyx_v_self, PyObject *__pyx_v_speciation_rates); /* proto */
static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_14CNSESimulation_18__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_13pycoalescence_6necsim_6necsim_CNSESimulation *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_14CNSESimulation_20__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_13pycoalescence_6necsim_6necsim_CNSESimulation *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_13pycoalescence_6necsim_6necsim_15CPNSESimulation___cinit__(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CPNSESimulation *__pyx_v_self, PyObject *__pyx_v_logger); /* proto */
static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_15CPNSESimulation_2set_logger(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CPNSESimulation *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_15CPNSESimulation_4import_from_config(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CPNSESimulation *__pyx_v_self, PyObject *__pyx_v_config_file); /* proto */
static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_15CPNSESimulation_6import_from_config_string(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CPNSESimulation *__pyx_v_self, PyObject *__pyx_v_config); /* proto */
static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_15CPNSESimulation_8setup(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CPNSESimulation *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_15CPNSESimulation_10setup_resume(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CPNSESimulation *__pyx_v_self, PyObject *__pyx_v_pause_directory, PyObject *__pyx_v_out_directory, PyObject *__pyx_v_seed, PyObject *__pyx_v_task, PyObject *__pyx_v_max_time); /* proto */
static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_15CPNSESimulation_12add_gillespie(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CPNSESimulation *__pyx_v_self, PyObject *__pyx_v_gillespie_threshold); /* proto */
static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_15CPNSESimulation_14run(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CPNSESimulation *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_15CPNSESimulation_16apply_speciation_rates(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CPNSESimulation *__pyx_v_self, PyObject *__pyx_v_speciation_rates); /* proto */
static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_15CPNSESimulation_18__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_13pycoalescence_6necsim_6necsim_CPNSESimulation *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_15CPNSESimulation_20__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_13pycoalescence_6necsim_6necsim_CPNSESimulation *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_13pycoalescence_6necsim_6necsim_19CPSpatialSimulation___cinit__(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CPSpatialSimulation *__pyx_v_self, PyObject *__pyx_v_logger); /* proto */
static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_19CPSpatialSimulation_2set_logger(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CPSpatialSimulation *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_19CPSpatialSimulation_4import_from_config(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CPSpatialSimulation *__pyx_v_self, PyObject *__pyx_v_config_file); /* proto */
static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_19CPSpatialSimulation_6import_from_config_string(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CPSpatialSimulation *__pyx_v_self, PyObject *__pyx_v_config); /* proto */
static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_19CPSpatialSimulation_8setup(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CPSpatialSimulation *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_19CPSpatialSimulation_10setup_resume(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CPSpatialSimulation *__pyx_v_self, PyObject *__pyx_v_pause_directory, PyObject *__pyx_v_out_directory, PyObject *__pyx_v_seed, PyObject *__pyx_v_task, PyObject *__pyx_v_max_time); /* proto */
static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_19CPSpatialSimulation_12add_gillespie(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CPSpatialSimulation *__pyx_v_self, PyObject *__pyx_v_gillespie_threshold); /* proto */
static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_19CPSpatialSimulation_14run(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CPSpatialSimulation *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_19CPSpatialSimulation_16apply_speciation_rates(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CPSpatialSimulation *__pyx_v_self, PyObject *__pyx_v_speciation_rates); /* proto */
static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_19CPSpatialSimulation_18__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_13pycoalescence_6necsim_6necsim_CPSpatialSimulation *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_19CPSpatialSimulation_20__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_13pycoalescence_6necsim_6necsim_CPSpatialSimulation *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_13pycoalescence_6necsim_6necsim_10CCommunity___cinit__(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CCommunity *__pyx_v_self, PyObject *__pyx_v_logger); /* proto */
static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_10CCommunity_2set_logger(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CCommunity *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_10CCommunity_4wipe_protracted_parameters(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CCommunity *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_10CCommunity_6add_protracted_parameters(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CCommunity *__pyx_v_self, double __pyx_v_min_speciation_gen, double __pyx_v_max_speciation_gen); /* proto */
static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_10CCommunity_8add_metacommunity_parameters(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CCommunity *__pyx_v_self, unsigned long __pyx_v_metacommunity_size, double __pyx_v_speciation_rate, std::string __pyx_v_metacommunity_option, unsigned long __pyx_v_metacommunity_reference); /* proto */
static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_10CCommunity_10reset(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CCommunity *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_10CCommunity_12setup(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CCommunity *__pyx_v_self, PyObject *__pyx_v_database, bool __pyx_v_use_spatial, PyObject *__pyx_v_sample_file, PyObject *__pyx_v_record_fragments, std::vector<double>  __pyx_v_speciation_rates, std::vector<double>  __pyx_v_times); /* proto */
static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_10CCommunity_14add_time(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CCommunity *__pyx_v_self, PyObject *__pyx_v_time); /* proto */
static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_10CCommunity_16apply(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CCommunity *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_10CCommunity_18output(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CCommunity *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_10CCommunity_20speciate_remaining_lineages(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CCommunity *__pyx_v_self, PyObject *__pyx_v_file); /* proto */
static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_10CCommunity_22__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_13pycoalescence_6necsim_6necsim_CCommunity *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_10CCommunity_24__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_13pycoalescence_6necsim_6necsim_CCommunity *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_13pycoalescence_6necsim_6necsim_14CMetacommunity___cinit__(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CMetacommunity *__pyx_v_self, PyObject *__pyx_v_logger); /* proto */
static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_14CMetacommunity_2set_logger(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CMetacommunity *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_14CMetacommunity_4wipe_protracted_parameters(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CMetacommunity *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_14CMetacommunity_6add_protracted_parameters(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CMetacommunity *__pyx_v_self, double __pyx_v_min_speciation_gen, double __pyx_v_max_speciation_gen); /* proto */
static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_14CMetacommunity_8add_metacommunity_parameters(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CMetacommunity *__pyx_v_self, unsigned long __pyx_v_metacommunity_size, double __pyx_v_speciation_rate, std::string __pyx_v_metacommunity_option, unsigned long __pyx_v_metacommunity_reference); /* proto */
static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_14CMetacommunity_10reset(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CMetacommunity *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_14CMetacommunity_12setup(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CMetacommunity *__pyx_v_self, PyObject *__pyx_v_database, bool __pyx_v_use_spatial, PyObject *__pyx_v_sample_file, PyObject *__pyx_v_record_fragments, std::vector<double>  __pyx_v_speciation_rates, std::vector<double>  __pyx_v_times); /* proto */
static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_14CMetacommunity_14add_time(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CMetacommunity *__pyx_v_self, PyObject *__pyx_v_time); /* proto */
static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_14CMetacommunity_16apply(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CMetacommunity *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_14CMetacommunity_18output(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CMetacommunity *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_14CMetacommunity_20speciate_remaining_lineages(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CMetacommunity *__pyx_v_self, PyObject *__pyx_v_file); /* proto */
static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_14CMetacommunity_22__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_13pycoalescence_6necsim_6necsim_CMetacommunity *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_14CMetacommunity_24__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_13pycoalescence_6necsim_6necsim_CMetacommunity *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_13pycoalescence_6necsim_6necsim_27CLandscapeMetricsCalculator___cinit__(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CLandscapeMetricsCalculator *__pyx_v_self, PyObject *__pyx_v_logger); /* proto */
static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_27CLandscapeMetricsCalculator_2set_logger(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CLandscapeMetricsCalculator *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_27CLandscapeMetricsCalculator_4import_map(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CLandscapeMetricsCalculator *__pyx_v_self, PyObject *__pyx_v_file_name); /* proto */
static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_27CLandscapeMetricsCalculator_6calculate_MNN(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CLandscapeMetricsCalculator *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_27CLandscapeMetricsCalculator_8calculate_CLUMPY(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CLandscapeMetricsCalculator *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_27CLandscapeMetricsCalculator_10__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_13pycoalescence_6necsim_6necsim_CLandscapeMetricsCalculator *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_27CLandscapeMetricsCalculator_12__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_13pycoalescence_6necsim_6necsim_CLandscapeMetricsCalculator *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_13pycoalescence_6necsim_6necsim_19CDispersalSimulator___cinit__(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CDispersalSimulator *__pyx_v_self, PyObject *__pyx_v_logger); /* proto */
static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_19CDispersalSimulator_2set_logger(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CDispersalSimulator *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_19CDispersalSimulator_4set_output_database(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CDispersalSimulator *__pyx_v_self, PyObject *__pyx_v_output_database); /* proto */
static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_19CDispersalSimulator_6import_maps_base(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CDispersalSimulator *__pyx_v_self, double __pyx_v_deme, PyObject *__pyx_v_fine_map_file, unsigned long __pyx_v_fine_map_x_size, unsigned long __pyx_v_fine_map_y_size, unsigned long __pyx_v_fine_map_x_offset, unsigned long __pyx_v_fine_map_y_offset, unsigned long __pyx_v_sample_map_x_size, unsigned long __pyx_v_sample_map_y_size, PyObject *__pyx_v_coarse_map_file, unsigned long __pyx_v_coarse_map_x_size, unsigned long __pyx_v_coarse_map_y_size, unsigned long __pyx_v_coarse_map_x_offset, unsigned long __pyx_v_coarse_map_y_offset, unsigned long __pyx_v_coarse_map_scale, PyObject *__pyx_v_landscape_type); /* proto */
static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_19CDispersalSimulator_8import_maps(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CDispersalSimulator *__pyx_v_self, double __pyx_v_deme, PyObject *__pyx_v_fine_map_file, unsigned long __pyx_v_fine_map_x_size, unsigned long __pyx_v_fine_map_y_size, unsigned long __pyx_v_fine_map_x_offset, unsigned long __pyx_v_fine_map_y_offset, unsigned long __pyx_v_sample_map_x_size, unsigned long __pyx_v_sample_map_y_size, PyObject *__pyx_v_coarse_map_file, unsigned long __pyx_v_coarse_map_x_size, unsigned long __pyx_v_coarse_map_y_size, unsigned long __pyx_v_coarse_map_x_offset, unsigned long __pyx_v_coarse_map_y_offset, unsigned long __pyx_v_coarse_map_scale, PyObject *__pyx_v_landscape_type); /* proto */
static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_19CDispersalSimulator_10import_all_maps(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CDispersalSimulator *__pyx_v_self, double __pyx_v_deme, PyObject *__pyx_v_fine_map_file, unsigned long __pyx_v_fine_map_x_size, unsigned long __pyx_v_fine_map_y_size, unsigned long __pyx_v_fine_map_x_offset, unsigned long __pyx_v_fine_map_y_offset, unsigned long __pyx_v_sample_map_x_size, unsigned long __pyx_v_sample_map_y_size, PyObject *__pyx_v_coarse_map_file, unsigned long __pyx_v_coarse_map_x_size, unsigned long __pyx_v_coarse_map_y_size, unsigned long __pyx_v_coarse_map_x_offset, unsigned long __pyx_v_coarse_map_y_offset, unsigned long __pyx_v_coarse_map_scale, PyObject *__pyx_v_landscape_type, std::vector<std::string>  __pyx_v_path_fine, std::vector<unsigned long>  __pyx_v_number_fine, std::vector<double>  __pyx_v_rate_fine, std::vector<double>  __pyx_v_time_fine, std::vector<std::string>  __pyx_v_path_coarse, std::vector<unsigned long>  __pyx_v_number_coarse, std::vector<double>  __pyx_v_rate_coarse, std::vector<double>  __pyx_v_time_coarse); /* proto */
static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_19CDispersalSimulator_12check_completed(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CDispersalSimulator *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_19CDispersalSimulator_14setup(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CDispersalSimulator *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_19CDispersalSimulator_16set_dispersal_parameters(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CDispersalSimulator *__pyx_v_self, PyObject *__pyx_v_dispersal_method, PyObject *__pyx_v_dispersal_file, double __pyx_v_sigma, double __pyx_v_tau, double __pyx_v_m_prob, double __pyx_v_cutoff, double __pyx_v_dispersal_relative_cost, bool __pyx_v_restrict_self); /* proto */
static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_19CDispersalSimulator_18base_setup(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CDispersalSimulator *__pyx_v_self, unsigned long __pyx_v_number_repeats, std::vector<unsigned long>  __pyx_v_number_steps, unsigned long __pyx_v_seed, unsigned long __pyx_v_number_workers); /* proto */
static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_19CDispersalSimulator_20write_database(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CDispersalSimulator *__pyx_v_self, PyObject *__pyx_v_table_name); /* proto */
static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_19CDispersalSimulator_22run_mean_distance_travelled(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CDispersalSimulator *__pyx_v_self, unsigned long __pyx_v_number_repeats, std::vector<unsigned long>  __pyx_v_number_steps, unsigned long __pyx_v_seed, unsigned long __pyx_v_number_workers); /* proto */
static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_19CDispersalSimulator_24run_all_distance_travelled(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CDispersalSimulator *__pyx_v_self, unsigned long __pyx_v_number_repeats, std::vector<unsigned long>  __pyx_v_number_steps, unsigned long __pyx_v_seed, unsigned long __pyx_v_number_workers); /* proto */
static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_19CDispersalSimulator_26run_sample_distance_travelled(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CDispersalSimulator *__pyx_v_self, std::vector<long>  __pyx_v_samples_x, std::vector<long>  __pyx_v_samples_y, unsigned long __pyx_v_number_repeats, std::vector<unsigned long>  __pyx_v_number_steps, unsigned long __pyx_v_seed, unsigned long __pyx_v_number_workers); /* proto */
static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_19CDispersalSimulator_28run_mean_dispersal_distance(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CDispersalSimulator *__pyx_v_self, unsigned long __pyx_v_number_repeats, unsigned long __pyx_v_seed, bool __pyx_v_is_sequential); /* proto */
static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_19CDispersalSimulator_30__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_13pycoalescence_6necsim_6necsim_CDispersalSimulator *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_19CDispersalSimulator_32__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_13pycoalescence_6necsim_6necsim_CDispersalSimulator *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_tp_new_13pycoalescence_6necsim_6necsim_CSpatialSimulation(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_13pycoalescence_6necsim_6necsim_CNSESimulation(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_13pycoalescence_6necsim_6necsim_CPNSESimulation(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_13pycoalescence_6necsim_6necsim_CPSpatialSimulation(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_13pycoalescence_6necsim_6necsim_CCommunity(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_13pycoalescence_6necsim_6necsim_CMetacommunity(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_13pycoalescence_6necsim_6necsim_CLandscapeMetricsCalculator(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_13pycoalescence_6necsim_6necsim_CDispersalSimulator(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tuple_;
static PyObject *__pyx_tuple__2;
static PyObject *__pyx_tuple__3;
static PyObject *__pyx_tuple__4;
static PyObject *__pyx_tuple__5;
static PyObject *__pyx_tuple__6;
static PyObject *__pyx_tuple__7;
static PyObject *__pyx_tuple__8;
static PyObject *__pyx_tuple__9;
static PyObject *__pyx_tuple__10;
static PyObject *__pyx_tuple__11;
static PyObject *__pyx_tuple__12;
static PyObject *__pyx_tuple__13;
static PyObject *__pyx_tuple__14;
static PyObject *__pyx_tuple__15;
static PyObject *__pyx_tuple__16;
static PyObject *__pyx_tuple__17;
static PyObject *__pyx_tuple__18;
static PyObject *__pyx_tuple__19;
static PyObject *__pyx_tuple__20;
/* Late includes */

/* "pycoalescence/necsim/necsim.pyx":23
 *     cdef object logger
 * 
 *     def __cinit__(self, object logger):             # <<<<<<<<<<<<<<
 *         self.logger = logger
 *         self.c_simulation = GenericTree[SpatialTree]()
 */

/* Python wrapper */
static int __pyx_pw_13pycoalescence_6necsim_6necsim_18CSpatialSimulation_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_13pycoalescence_6necsim_6necsim_18CSpatialSimulation_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_logger = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_logger,0};
    PyObject* values[1] = {0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_logger)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__cinit__") < 0)) __PYX_ERR(0, 23, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 1) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
    }
    __pyx_v_logger = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 1, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 23, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pycoalescence.necsim.necsim.CSpatialSimulation.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_13pycoalescence_6necsim_6necsim_18CSpatialSimulation___cinit__(((struct __pyx_obj_13pycoalescence_6necsim_6necsim_CSpatialSimulation *)__pyx_v_self), __pyx_v_logger);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_13pycoalescence_6necsim_6necsim_18CSpatialSimulation___cinit__(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CSpatialSimulation *__pyx_v_self, PyObject *__pyx_v_logger) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  necsim::GenericTree<necsim::SpatialTree>  __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "pycoalescence/necsim/necsim.pyx":24
 * 
 *     def __cinit__(self, object logger):
 *         self.logger = logger             # <<<<<<<<<<<<<<
 *         self.c_simulation = GenericTree[SpatialTree]()
 * 
 */
  __Pyx_INCREF(__pyx_v_logger);
  __Pyx_GIVEREF(__pyx_v_logger);
  __Pyx_GOTREF(__pyx_v_self->logger);
  __Pyx_DECREF(__pyx_v_self->logger);
  __pyx_v_self->logger = __pyx_v_logger;

  /* "pycoalescence/necsim/necsim.pyx":25
 *     def __cinit__(self, object logger):
 *         self.logger = logger
 *         self.c_simulation = GenericTree[SpatialTree]()             # <<<<<<<<<<<<<<
 * 
 *     def set_logger(self):
 */
  try {
    __pyx_t_1 = necsim::GenericTree<necsim::SpatialTree> ();
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 25, __pyx_L1_error)
  }
  __pyx_v_self->c_simulation = __pyx_t_1;

  /* "pycoalescence/necsim/necsim.pyx":23
 *     cdef object logger
 * 
 *     def __cinit__(self, object logger):             # <<<<<<<<<<<<<<
 *         self.logger = logger
 *         self.c_simulation = GenericTree[SpatialTree]()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pycoalescence.necsim.necsim.CSpatialSimulation.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pycoalescence/necsim/necsim.pyx":27
 *         self.c_simulation = GenericTree[SpatialTree]()
 * 
 *     def set_logger(self):             # <<<<<<<<<<<<<<
 *         setGlobalLogger(self.logger, write_to_log)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_18CSpatialSimulation_3set_logger(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_18CSpatialSimulation_3set_logger(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_logger (wrapper)", 0);
  __pyx_r = __pyx_pf_13pycoalescence_6necsim_6necsim_18CSpatialSimulation_2set_logger(((struct __pyx_obj_13pycoalescence_6necsim_6necsim_CSpatialSimulation *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_18CSpatialSimulation_2set_logger(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CSpatialSimulation *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("set_logger", 0);

  /* "pycoalescence/necsim/necsim.pyx":28
 * 
 *     def set_logger(self):
 *         setGlobalLogger(self.logger, write_to_log)             # <<<<<<<<<<<<<<
 * 
 *     def import_from_config(self, config_file):
 */
  __pyx_t_1 = __pyx_v_self->logger;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_write_to_log); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  try {
    necsim::setGlobalLogger(__pyx_t_1, __pyx_t_2);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 28, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pycoalescence/necsim/necsim.pyx":27
 *         self.c_simulation = GenericTree[SpatialTree]()
 * 
 *     def set_logger(self):             # <<<<<<<<<<<<<<
 *         setGlobalLogger(self.logger, write_to_log)
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pycoalescence.necsim.necsim.CSpatialSimulation.set_logger", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pycoalescence/necsim/necsim.pyx":30
 *         setGlobalLogger(self.logger, write_to_log)
 * 
 *     def import_from_config(self, config_file):             # <<<<<<<<<<<<<<
 *         self.set_logger()
 *         self.c_simulation.importSimulationVariables(config_file)
 */

/* Python wrapper */
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_18CSpatialSimulation_5import_from_config(PyObject *__pyx_v_self, PyObject *__pyx_v_config_file); /*proto*/
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_18CSpatialSimulation_5import_from_config(PyObject *__pyx_v_self, PyObject *__pyx_v_config_file) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("import_from_config (wrapper)", 0);
  __pyx_r = __pyx_pf_13pycoalescence_6necsim_6necsim_18CSpatialSimulation_4import_from_config(((struct __pyx_obj_13pycoalescence_6necsim_6necsim_CSpatialSimulation *)__pyx_v_self), ((PyObject *)__pyx_v_config_file));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_18CSpatialSimulation_4import_from_config(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CSpatialSimulation *__pyx_v_self, PyObject *__pyx_v_config_file) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  std::string __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("import_from_config", 0);

  /* "pycoalescence/necsim/necsim.pyx":31
 * 
 *     def import_from_config(self, config_file):
 *         self.set_logger()             # <<<<<<<<<<<<<<
 *         self.c_simulation.importSimulationVariables(config_file)
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_set_logger); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pycoalescence/necsim/necsim.pyx":32
 *     def import_from_config(self, config_file):
 *         self.set_logger()
 *         self.c_simulation.importSimulationVariables(config_file)             # <<<<<<<<<<<<<<
 * 
 *     def import_from_config_string(self, config):
 */
  __pyx_t_4 = __pyx_convert_string_from_py_std__in_string(__pyx_v_config_file); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 32, __pyx_L1_error)
  try {
    __pyx_v_self->c_simulation.importSimulationVariables(__pyx_t_4);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 32, __pyx_L1_error)
  }

  /* "pycoalescence/necsim/necsim.pyx":30
 *         setGlobalLogger(self.logger, write_to_log)
 * 
 *     def import_from_config(self, config_file):             # <<<<<<<<<<<<<<
 *         self.set_logger()
 *         self.c_simulation.importSimulationVariables(config_file)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pycoalescence.necsim.necsim.CSpatialSimulation.import_from_config", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pycoalescence/necsim/necsim.pyx":34
 *         self.c_simulation.importSimulationVariables(config_file)
 * 
 *     def import_from_config_string(self, config):             # <<<<<<<<<<<<<<
 *         self.set_logger()
 *         cdef string config_str = config
 */

/* Python wrapper */
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_18CSpatialSimulation_7import_from_config_string(PyObject *__pyx_v_self, PyObject *__pyx_v_config); /*proto*/
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_18CSpatialSimulation_7import_from_config_string(PyObject *__pyx_v_self, PyObject *__pyx_v_config) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("import_from_config_string (wrapper)", 0);
  __pyx_r = __pyx_pf_13pycoalescence_6necsim_6necsim_18CSpatialSimulation_6import_from_config_string(((struct __pyx_obj_13pycoalescence_6necsim_6necsim_CSpatialSimulation *)__pyx_v_self), ((PyObject *)__pyx_v_config));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_18CSpatialSimulation_6import_from_config_string(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CSpatialSimulation *__pyx_v_self, PyObject *__pyx_v_config) {
  std::string __pyx_v_config_str;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  std::string __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("import_from_config_string", 0);

  /* "pycoalescence/necsim/necsim.pyx":35
 * 
 *     def import_from_config_string(self, config):
 *         self.set_logger()             # <<<<<<<<<<<<<<
 *         cdef string config_str = config
 *         self.c_simulation.importSimulationVariablesFromString(config_str)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_set_logger); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pycoalescence/necsim/necsim.pyx":36
 *     def import_from_config_string(self, config):
 *         self.set_logger()
 *         cdef string config_str = config             # <<<<<<<<<<<<<<
 *         self.c_simulation.importSimulationVariablesFromString(config_str)
 * 
 */
  __pyx_t_4 = __pyx_convert_string_from_py_std__in_string(__pyx_v_config); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 36, __pyx_L1_error)
  __pyx_v_config_str = __pyx_t_4;

  /* "pycoalescence/necsim/necsim.pyx":37
 *         self.set_logger()
 *         cdef string config_str = config
 *         self.c_simulation.importSimulationVariablesFromString(config_str)             # <<<<<<<<<<<<<<
 * 
 *     def setup(self):
 */
  try {
    __pyx_v_self->c_simulation.importSimulationVariablesFromString(__pyx_v_config_str);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 37, __pyx_L1_error)
  }

  /* "pycoalescence/necsim/necsim.pyx":34
 *         self.c_simulation.importSimulationVariables(config_file)
 * 
 *     def import_from_config_string(self, config):             # <<<<<<<<<<<<<<
 *         self.set_logger()
 *         cdef string config_str = config
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pycoalescence.necsim.necsim.CSpatialSimulation.import_from_config_string", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pycoalescence/necsim/necsim.pyx":39
 *         self.c_simulation.importSimulationVariablesFromString(config_str)
 * 
 *     def setup(self):             # <<<<<<<<<<<<<<
 *         self.set_logger()
 *         self.c_simulation.setup()
 */

/* Python wrapper */
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_18CSpatialSimulation_9setup(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_18CSpatialSimulation_9setup(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("setup (wrapper)", 0);
  __pyx_r = __pyx_pf_13pycoalescence_6necsim_6necsim_18CSpatialSimulation_8setup(((struct __pyx_obj_13pycoalescence_6necsim_6necsim_CSpatialSimulation *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_18CSpatialSimulation_8setup(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CSpatialSimulation *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("setup", 0);

  /* "pycoalescence/necsim/necsim.pyx":40
 * 
 *     def setup(self):
 *         self.set_logger()             # <<<<<<<<<<<<<<
 *         self.c_simulation.setup()
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_set_logger); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 40, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 40, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pycoalescence/necsim/necsim.pyx":41
 *     def setup(self):
 *         self.set_logger()
 *         self.c_simulation.setup()             # <<<<<<<<<<<<<<
 * 
 *     def setup_resume(self, pause_directory, out_directory, seed, task, max_time):
 */
  try {
    __pyx_v_self->c_simulation.setup();
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 41, __pyx_L1_error)
  }

  /* "pycoalescence/necsim/necsim.pyx":39
 *         self.c_simulation.importSimulationVariablesFromString(config_str)
 * 
 *     def setup(self):             # <<<<<<<<<<<<<<
 *         self.set_logger()
 *         self.c_simulation.setup()
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pycoalescence.necsim.necsim.CSpatialSimulation.setup", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pycoalescence/necsim/necsim.pyx":43
 *         self.c_simulation.setup()
 * 
 *     def setup_resume(self, pause_directory, out_directory, seed, task, max_time):             # <<<<<<<<<<<<<<
 *         self.set_logger()
 *         self.c_simulation.wipeSimulationVariables()
 */

/* Python wrapper */
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_18CSpatialSimulation_11setup_resume(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_18CSpatialSimulation_11setup_resume(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_pause_directory = 0;
  PyObject *__pyx_v_out_directory = 0;
  PyObject *__pyx_v_seed = 0;
  PyObject *__pyx_v_task = 0;
  PyObject *__pyx_v_max_time = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("setup_resume (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pause_directory,&__pyx_n_s_out_directory,&__pyx_n_s_seed,&__pyx_n_s_task,&__pyx_n_s_max_time,0};
    PyObject* values[5] = {0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pause_directory)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_out_directory)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setup_resume", 1, 5, 5, 1); __PYX_ERR(0, 43, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_seed)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setup_resume", 1, 5, 5, 2); __PYX_ERR(0, 43, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_task)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setup_resume", 1, 5, 5, 3); __PYX_ERR(0, 43, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_max_time)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setup_resume", 1, 5, 5, 4); __PYX_ERR(0, 43, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "setup_resume") < 0)) __PYX_ERR(0, 43, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 5) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
    }
    __pyx_v_pause_directory = values[0];
    __pyx_v_out_directory = values[1];
    __pyx_v_seed = values[2];
    __pyx_v_task = values[3];
    __pyx_v_max_time = values[4];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("setup_resume", 1, 5, 5, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 43, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pycoalescence.necsim.necsim.CSpatialSimulation.setup_resume", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_13pycoalescence_6necsim_6necsim_18CSpatialSimulation_10setup_resume(((struct __pyx_obj_13pycoalescence_6necsim_6necsim_CSpatialSimulation *)__pyx_v_self), __pyx_v_pause_directory, __pyx_v_out_directory, __pyx_v_seed, __pyx_v_task, __pyx_v_max_time);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_18CSpatialSimulation_10setup_resume(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CSpatialSimulation *__pyx_v_self, PyObject *__pyx_v_pause_directory, PyObject *__pyx_v_out_directory, PyObject *__pyx_v_seed, PyObject *__pyx_v_task, PyObject *__pyx_v_max_time) {
  std::string __pyx_v_pause_directory_str;
  std::string __pyx_v_out_directory_str;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  std::string __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  bool __pyx_t_8;
  int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("setup_resume", 0);

  /* "pycoalescence/necsim/necsim.pyx":44
 * 
 *     def setup_resume(self, pause_directory, out_directory, seed, task, max_time):
 *         self.set_logger()             # <<<<<<<<<<<<<<
 *         self.c_simulation.wipeSimulationVariables()
 *         cdef string pause_directory_str = pause_directory
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_set_logger); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 44, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 44, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pycoalescence/necsim/necsim.pyx":45
 *     def setup_resume(self, pause_directory, out_directory, seed, task, max_time):
 *         self.set_logger()
 *         self.c_simulation.wipeSimulationVariables()             # <<<<<<<<<<<<<<
 *         cdef string pause_directory_str = pause_directory
 *         cdef string out_directory_str = out_directory
 */
  try {
    __pyx_v_self->c_simulation.wipeSimulationVariables();
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 45, __pyx_L1_error)
  }

  /* "pycoalescence/necsim/necsim.pyx":46
 *         self.set_logger()
 *         self.c_simulation.wipeSimulationVariables()
 *         cdef string pause_directory_str = pause_directory             # <<<<<<<<<<<<<<
 *         cdef string out_directory_str = out_directory
 *         self.c_simulation.setResumeParameters(pause_directory_str, out_directory_str, seed, task, max_time)
 */
  __pyx_t_4 = __pyx_convert_string_from_py_std__in_string(__pyx_v_pause_directory); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 46, __pyx_L1_error)
  __pyx_v_pause_directory_str = __pyx_t_4;

  /* "pycoalescence/necsim/necsim.pyx":47
 *         self.c_simulation.wipeSimulationVariables()
 *         cdef string pause_directory_str = pause_directory
 *         cdef string out_directory_str = out_directory             # <<<<<<<<<<<<<<
 *         self.c_simulation.setResumeParameters(pause_directory_str, out_directory_str, seed, task, max_time)
 *         self.c_simulation.checkSims(pause_directory_str, seed, task)
 */
  __pyx_t_4 = __pyx_convert_string_from_py_std__in_string(__pyx_v_out_directory); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 47, __pyx_L1_error)
  __pyx_v_out_directory_str = __pyx_t_4;

  /* "pycoalescence/necsim/necsim.pyx":48
 *         cdef string pause_directory_str = pause_directory
 *         cdef string out_directory_str = out_directory
 *         self.c_simulation.setResumeParameters(pause_directory_str, out_directory_str, seed, task, max_time)             # <<<<<<<<<<<<<<
 *         self.c_simulation.checkSims(pause_directory_str, seed, task)
 *         if self.c_simulation.hasPaused():
 */
  __pyx_t_5 = __Pyx_PyInt_As_int(__pyx_v_seed); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 48, __pyx_L1_error)
  __pyx_t_6 = __Pyx_PyInt_As_int(__pyx_v_task); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 48, __pyx_L1_error)
  __pyx_t_7 = __Pyx_PyInt_As_int(__pyx_v_max_time); if (unlikely((__pyx_t_7 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 48, __pyx_L1_error)
  try {
    __pyx_v_self->c_simulation.setResumeParameters(__pyx_v_pause_directory_str, __pyx_v_out_directory_str, __pyx_t_5, __pyx_t_6, __pyx_t_7);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 48, __pyx_L1_error)
  }

  /* "pycoalescence/necsim/necsim.pyx":49
 *         cdef string out_directory_str = out_directory
 *         self.c_simulation.setResumeParameters(pause_directory_str, out_directory_str, seed, task, max_time)
 *         self.c_simulation.checkSims(pause_directory_str, seed, task)             # <<<<<<<<<<<<<<
 *         if self.c_simulation.hasPaused():
 *             self.c_simulation.setup()
 */
  __pyx_t_7 = __Pyx_PyInt_As_int(__pyx_v_seed); if (unlikely((__pyx_t_7 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 49, __pyx_L1_error)
  __pyx_t_6 = __Pyx_PyInt_As_int(__pyx_v_task); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 49, __pyx_L1_error)
  try {
    __pyx_v_self->c_simulation.checkSims(__pyx_v_pause_directory_str, __pyx_t_7, __pyx_t_6);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 49, __pyx_L1_error)
  }

  /* "pycoalescence/necsim/necsim.pyx":50
 *         self.c_simulation.setResumeParameters(pause_directory_str, out_directory_str, seed, task, max_time)
 *         self.c_simulation.checkSims(pause_directory_str, seed, task)
 *         if self.c_simulation.hasPaused():             # <<<<<<<<<<<<<<
 *             self.c_simulation.setup()
 *         else:
 */
  try {
    __pyx_t_8 = __pyx_v_self->c_simulation.hasPaused();
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 50, __pyx_L1_error)
  }
  __pyx_t_9 = (__pyx_t_8 != 0);
  if (likely(__pyx_t_9)) {

    /* "pycoalescence/necsim/necsim.pyx":51
 *         self.c_simulation.checkSims(pause_directory_str, seed, task)
 *         if self.c_simulation.hasPaused():
 *             self.c_simulation.setup()             # <<<<<<<<<<<<<<
 *         else:
 *             raise RuntimeError("Couldn't find paused simulation")
 */
    try {
      __pyx_v_self->c_simulation.setup();
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 51, __pyx_L1_error)
    }

    /* "pycoalescence/necsim/necsim.pyx":50
 *         self.c_simulation.setResumeParameters(pause_directory_str, out_directory_str, seed, task, max_time)
 *         self.c_simulation.checkSims(pause_directory_str, seed, task)
 *         if self.c_simulation.hasPaused():             # <<<<<<<<<<<<<<
 *             self.c_simulation.setup()
 *         else:
 */
    goto __pyx_L3;
  }

  /* "pycoalescence/necsim/necsim.pyx":53
 *             self.c_simulation.setup()
 *         else:
 *             raise RuntimeError("Couldn't find paused simulation")             # <<<<<<<<<<<<<<
 * 
 *     def add_gillespie(self, gillespie_threshold):
 */
  /*else*/ {
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple_, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 53, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 53, __pyx_L1_error)
  }
  __pyx_L3:;

  /* "pycoalescence/necsim/necsim.pyx":43
 *         self.c_simulation.setup()
 * 
 *     def setup_resume(self, pause_directory, out_directory, seed, task, max_time):             # <<<<<<<<<<<<<<
 *         self.set_logger()
 *         self.c_simulation.wipeSimulationVariables()
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pycoalescence.necsim.necsim.CSpatialSimulation.setup_resume", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pycoalescence/necsim/necsim.pyx":55
 *             raise RuntimeError("Couldn't find paused simulation")
 * 
 *     def add_gillespie(self, gillespie_threshold):             # <<<<<<<<<<<<<<
 *         self.set_logger()
 *         self.c_simulation.addGillespie(gillespie_threshold)
 */

/* Python wrapper */
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_18CSpatialSimulation_13add_gillespie(PyObject *__pyx_v_self, PyObject *__pyx_v_gillespie_threshold); /*proto*/
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_18CSpatialSimulation_13add_gillespie(PyObject *__pyx_v_self, PyObject *__pyx_v_gillespie_threshold) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("add_gillespie (wrapper)", 0);
  __pyx_r = __pyx_pf_13pycoalescence_6necsim_6necsim_18CSpatialSimulation_12add_gillespie(((struct __pyx_obj_13pycoalescence_6necsim_6necsim_CSpatialSimulation *)__pyx_v_self), ((PyObject *)__pyx_v_gillespie_threshold));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_18CSpatialSimulation_12add_gillespie(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CSpatialSimulation *__pyx_v_self, PyObject *__pyx_v_gillespie_threshold) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  double __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("add_gillespie", 0);

  /* "pycoalescence/necsim/necsim.pyx":56
 * 
 *     def add_gillespie(self, gillespie_threshold):
 *         self.set_logger()             # <<<<<<<<<<<<<<
 *         self.c_simulation.addGillespie(gillespie_threshold)
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_set_logger); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 56, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 56, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pycoalescence/necsim/necsim.pyx":57
 *     def add_gillespie(self, gillespie_threshold):
 *         self.set_logger()
 *         self.c_simulation.addGillespie(gillespie_threshold)             # <<<<<<<<<<<<<<
 * 
 *     def run(self) -> bool:
 */
  __pyx_t_4 = __pyx_PyFloat_AsDouble(__pyx_v_gillespie_threshold); if (unlikely((__pyx_t_4 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 57, __pyx_L1_error)
  try {
    __pyx_v_self->c_simulation.addGillespie(__pyx_t_4);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 57, __pyx_L1_error)
  }

  /* "pycoalescence/necsim/necsim.pyx":55
 *             raise RuntimeError("Couldn't find paused simulation")
 * 
 *     def add_gillespie(self, gillespie_threshold):             # <<<<<<<<<<<<<<
 *         self.set_logger()
 *         self.c_simulation.addGillespie(gillespie_threshold)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pycoalescence.necsim.necsim.CSpatialSimulation.add_gillespie", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pycoalescence/necsim/necsim.pyx":59
 *         self.c_simulation.addGillespie(gillespie_threshold)
 * 
 *     def run(self) -> bool:             # <<<<<<<<<<<<<<
 *         self.set_logger()
 *         return self.c_simulation.runSimulation()
 */

/* Python wrapper */
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_18CSpatialSimulation_15run(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_18CSpatialSimulation_15run(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("run (wrapper)", 0);
  __pyx_r = __pyx_pf_13pycoalescence_6necsim_6necsim_18CSpatialSimulation_14run(((struct __pyx_obj_13pycoalescence_6necsim_6necsim_CSpatialSimulation *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_18CSpatialSimulation_14run(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CSpatialSimulation *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  bool __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("run", 0);

  /* "pycoalescence/necsim/necsim.pyx":60
 * 
 *     def run(self) -> bool:
 *         self.set_logger()             # <<<<<<<<<<<<<<
 *         return self.c_simulation.runSimulation()
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_set_logger); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 60, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 60, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pycoalescence/necsim/necsim.pyx":61
 *     def run(self) -> bool:
 *         self.set_logger()
 *         return self.c_simulation.runSimulation()             # <<<<<<<<<<<<<<
 * 
 *     def apply_speciation_rates(self, speciation_rates: List[float]):
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_4 = __pyx_v_self->c_simulation.runSimulation();
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 61, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 61, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pycoalescence/necsim/necsim.pyx":59
 *         self.c_simulation.addGillespie(gillespie_threshold)
 * 
 *     def run(self) -> bool:             # <<<<<<<<<<<<<<
 *         self.set_logger()
 *         return self.c_simulation.runSimulation()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pycoalescence.necsim.necsim.CSpatialSimulation.run", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pycoalescence/necsim/necsim.pyx":63
 *         return self.c_simulation.runSimulation()
 * 
 *     def apply_speciation_rates(self, speciation_rates: List[float]):             # <<<<<<<<<<<<<<
 *         self.set_logger()
 *         self.c_simulation.addSpeciationRates(speciation_rates)
 */

/* Python wrapper */
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_18CSpatialSimulation_17apply_speciation_rates(PyObject *__pyx_v_self, PyObject *__pyx_v_speciation_rates); /*proto*/
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_18CSpatialSimulation_17apply_speciation_rates(PyObject *__pyx_v_self, PyObject *__pyx_v_speciation_rates) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("apply_speciation_rates (wrapper)", 0);
  __pyx_r = __pyx_pf_13pycoalescence_6necsim_6necsim_18CSpatialSimulation_16apply_speciation_rates(((struct __pyx_obj_13pycoalescence_6necsim_6necsim_CSpatialSimulation *)__pyx_v_self), ((PyObject *)__pyx_v_speciation_rates));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_18CSpatialSimulation_16apply_speciation_rates(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CSpatialSimulation *__pyx_v_self, PyObject *__pyx_v_speciation_rates) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  std::vector<long double>  __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("apply_speciation_rates", 0);

  /* "pycoalescence/necsim/necsim.pyx":64
 * 
 *     def apply_speciation_rates(self, speciation_rates: List[float]):
 *         self.set_logger()             # <<<<<<<<<<<<<<
 *         self.c_simulation.addSpeciationRates(speciation_rates)
 *         self.c_simulation.applyMultipleRates()
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_set_logger); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 64, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 64, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pycoalescence/necsim/necsim.pyx":65
 *     def apply_speciation_rates(self, speciation_rates: List[float]):
 *         self.set_logger()
 *         self.c_simulation.addSpeciationRates(speciation_rates)             # <<<<<<<<<<<<<<
 *         self.c_simulation.applyMultipleRates()
 * 
 */
  __pyx_t_4 = __pyx_convert_vector_from_py_long__double(__pyx_v_speciation_rates); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 65, __pyx_L1_error)
  try {
    __pyx_v_self->c_simulation.addSpeciationRates(__pyx_t_4);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 65, __pyx_L1_error)
  }

  /* "pycoalescence/necsim/necsim.pyx":66
 *         self.set_logger()
 *         self.c_simulation.addSpeciationRates(speciation_rates)
 *         self.c_simulation.applyMultipleRates()             # <<<<<<<<<<<<<<
 * 
 * cdef class CNSESimulation:
 */
  try {
    __pyx_v_self->c_simulation.applyMultipleRates();
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 66, __pyx_L1_error)
  }

  /* "pycoalescence/necsim/necsim.pyx":63
 *         return self.c_simulation.runSimulation()
 * 
 *     def apply_speciation_rates(self, speciation_rates: List[float]):             # <<<<<<<<<<<<<<
 *         self.set_logger()
 *         self.c_simulation.addSpeciationRates(speciation_rates)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pycoalescence.necsim.necsim.CSpatialSimulation.apply_speciation_rates", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_18CSpatialSimulation_19__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_18CSpatialSimulation_19__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_13pycoalescence_6necsim_6necsim_18CSpatialSimulation_18__reduce_cython__(((struct __pyx_obj_13pycoalescence_6necsim_6necsim_CSpatialSimulation *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_18CSpatialSimulation_18__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_13pycoalescence_6necsim_6necsim_CSpatialSimulation *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pycoalescence.necsim.necsim.CSpatialSimulation.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_18CSpatialSimulation_21__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_18CSpatialSimulation_21__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_13pycoalescence_6necsim_6necsim_18CSpatialSimulation_20__setstate_cython__(((struct __pyx_obj_13pycoalescence_6necsim_6necsim_CSpatialSimulation *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_18CSpatialSimulation_20__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_13pycoalescence_6necsim_6necsim_CSpatialSimulation *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pycoalescence.necsim.necsim.CSpatialSimulation.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pycoalescence/necsim/necsim.pyx":72
 *     cdef object logger
 * 
 *     def __cinit__(self, object logger):             # <<<<<<<<<<<<<<
 *         self.logger = logger
 *         self.c_simulation = GenericTree[Tree]()
 */

/* Python wrapper */
static int __pyx_pw_13pycoalescence_6necsim_6necsim_14CNSESimulation_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_13pycoalescence_6necsim_6necsim_14CNSESimulation_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_logger = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_logger,0};
    PyObject* values[1] = {0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_logger)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__cinit__") < 0)) __PYX_ERR(0, 72, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 1) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
    }
    __pyx_v_logger = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 1, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 72, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pycoalescence.necsim.necsim.CNSESimulation.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_13pycoalescence_6necsim_6necsim_14CNSESimulation___cinit__(((struct __pyx_obj_13pycoalescence_6necsim_6necsim_CNSESimulation *)__pyx_v_self), __pyx_v_logger);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_13pycoalescence_6necsim_6necsim_14CNSESimulation___cinit__(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CNSESimulation *__pyx_v_self, PyObject *__pyx_v_logger) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  necsim::GenericTree<necsim::Tree>  __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "pycoalescence/necsim/necsim.pyx":73
 * 
 *     def __cinit__(self, object logger):
 *         self.logger = logger             # <<<<<<<<<<<<<<
 *         self.c_simulation = GenericTree[Tree]()
 * 
 */
  __Pyx_INCREF(__pyx_v_logger);
  __Pyx_GIVEREF(__pyx_v_logger);
  __Pyx_GOTREF(__pyx_v_self->logger);
  __Pyx_DECREF(__pyx_v_self->logger);
  __pyx_v_self->logger = __pyx_v_logger;

  /* "pycoalescence/necsim/necsim.pyx":74
 *     def __cinit__(self, object logger):
 *         self.logger = logger
 *         self.c_simulation = GenericTree[Tree]()             # <<<<<<<<<<<<<<
 * 
 *     def set_logger(self):
 */
  try {
    __pyx_t_1 = necsim::GenericTree<necsim::Tree> ();
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 74, __pyx_L1_error)
  }
  __pyx_v_self->c_simulation = __pyx_t_1;

  /* "pycoalescence/necsim/necsim.pyx":72
 *     cdef object logger
 * 
 *     def __cinit__(self, object logger):             # <<<<<<<<<<<<<<
 *         self.logger = logger
 *         self.c_simulation = GenericTree[Tree]()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pycoalescence.necsim.necsim.CNSESimulation.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pycoalescence/necsim/necsim.pyx":76
 *         self.c_simulation = GenericTree[Tree]()
 * 
 *     def set_logger(self):             # <<<<<<<<<<<<<<
 *         setGlobalLogger(self.logger, write_to_log)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_14CNSESimulation_3set_logger(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_14CNSESimulation_3set_logger(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_logger (wrapper)", 0);
  __pyx_r = __pyx_pf_13pycoalescence_6necsim_6necsim_14CNSESimulation_2set_logger(((struct __pyx_obj_13pycoalescence_6necsim_6necsim_CNSESimulation *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_14CNSESimulation_2set_logger(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CNSESimulation *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("set_logger", 0);

  /* "pycoalescence/necsim/necsim.pyx":77
 * 
 *     def set_logger(self):
 *         setGlobalLogger(self.logger, write_to_log)             # <<<<<<<<<<<<<<
 * 
 *     def import_from_config(self, config_file):
 */
  __pyx_t_1 = __pyx_v_self->logger;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_write_to_log); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 77, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  try {
    necsim::setGlobalLogger(__pyx_t_1, __pyx_t_2);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 77, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pycoalescence/necsim/necsim.pyx":76
 *         self.c_simulation = GenericTree[Tree]()
 * 
 *     def set_logger(self):             # <<<<<<<<<<<<<<
 *         setGlobalLogger(self.logger, write_to_log)
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pycoalescence.necsim.necsim.CNSESimulation.set_logger", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pycoalescence/necsim/necsim.pyx":79
 *         setGlobalLogger(self.logger, write_to_log)
 * 
 *     def import_from_config(self, config_file):             # <<<<<<<<<<<<<<
 *         self.set_logger()
 *         self.c_simulation.importSimulationVariables(config_file)
 */

/* Python wrapper */
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_14CNSESimulation_5import_from_config(PyObject *__pyx_v_self, PyObject *__pyx_v_config_file); /*proto*/
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_14CNSESimulation_5import_from_config(PyObject *__pyx_v_self, PyObject *__pyx_v_config_file) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("import_from_config (wrapper)", 0);
  __pyx_r = __pyx_pf_13pycoalescence_6necsim_6necsim_14CNSESimulation_4import_from_config(((struct __pyx_obj_13pycoalescence_6necsim_6necsim_CNSESimulation *)__pyx_v_self), ((PyObject *)__pyx_v_config_file));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_14CNSESimulation_4import_from_config(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CNSESimulation *__pyx_v_self, PyObject *__pyx_v_config_file) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  std::string __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("import_from_config", 0);

  /* "pycoalescence/necsim/necsim.pyx":80
 * 
 *     def import_from_config(self, config_file):
 *         self.set_logger()             # <<<<<<<<<<<<<<
 *         self.c_simulation.importSimulationVariables(config_file)
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_set_logger); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 80, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 80, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pycoalescence/necsim/necsim.pyx":81
 *     def import_from_config(self, config_file):
 *         self.set_logger()
 *         self.c_simulation.importSimulationVariables(config_file)             # <<<<<<<<<<<<<<
 * 
 *     def import_from_config_string(self, config):
 */
  __pyx_t_4 = __pyx_convert_string_from_py_std__in_string(__pyx_v_config_file); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 81, __pyx_L1_error)
  try {
    __pyx_v_self->c_simulation.importSimulationVariables(__pyx_t_4);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 81, __pyx_L1_error)
  }

  /* "pycoalescence/necsim/necsim.pyx":79
 *         setGlobalLogger(self.logger, write_to_log)
 * 
 *     def import_from_config(self, config_file):             # <<<<<<<<<<<<<<
 *         self.set_logger()
 *         self.c_simulation.importSimulationVariables(config_file)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pycoalescence.necsim.necsim.CNSESimulation.import_from_config", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pycoalescence/necsim/necsim.pyx":83
 *         self.c_simulation.importSimulationVariables(config_file)
 * 
 *     def import_from_config_string(self, config):             # <<<<<<<<<<<<<<
 *         self.set_logger()
 *         cdef string config_str = config
 */

/* Python wrapper */
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_14CNSESimulation_7import_from_config_string(PyObject *__pyx_v_self, PyObject *__pyx_v_config); /*proto*/
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_14CNSESimulation_7import_from_config_string(PyObject *__pyx_v_self, PyObject *__pyx_v_config) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("import_from_config_string (wrapper)", 0);
  __pyx_r = __pyx_pf_13pycoalescence_6necsim_6necsim_14CNSESimulation_6import_from_config_string(((struct __pyx_obj_13pycoalescence_6necsim_6necsim_CNSESimulation *)__pyx_v_self), ((PyObject *)__pyx_v_config));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_14CNSESimulation_6import_from_config_string(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CNSESimulation *__pyx_v_self, PyObject *__pyx_v_config) {
  std::string __pyx_v_config_str;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  std::string __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("import_from_config_string", 0);

  /* "pycoalescence/necsim/necsim.pyx":84
 * 
 *     def import_from_config_string(self, config):
 *         self.set_logger()             # <<<<<<<<<<<<<<
 *         cdef string config_str = config
 *         self.c_simulation.importSimulationVariablesFromString(config_str)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_set_logger); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pycoalescence/necsim/necsim.pyx":85
 *     def import_from_config_string(self, config):
 *         self.set_logger()
 *         cdef string config_str = config             # <<<<<<<<<<<<<<
 *         self.c_simulation.importSimulationVariablesFromString(config_str)
 * 
 */
  __pyx_t_4 = __pyx_convert_string_from_py_std__in_string(__pyx_v_config); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 85, __pyx_L1_error)
  __pyx_v_config_str = __pyx_t_4;

  /* "pycoalescence/necsim/necsim.pyx":86
 *         self.set_logger()
 *         cdef string config_str = config
 *         self.c_simulation.importSimulationVariablesFromString(config_str)             # <<<<<<<<<<<<<<
 * 
 *     def setup(self):
 */
  try {
    __pyx_v_self->c_simulation.importSimulationVariablesFromString(__pyx_v_config_str);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 86, __pyx_L1_error)
  }

  /* "pycoalescence/necsim/necsim.pyx":83
 *         self.c_simulation.importSimulationVariables(config_file)
 * 
 *     def import_from_config_string(self, config):             # <<<<<<<<<<<<<<
 *         self.set_logger()
 *         cdef string config_str = config
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pycoalescence.necsim.necsim.CNSESimulation.import_from_config_string", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pycoalescence/necsim/necsim.pyx":88
 *         self.c_simulation.importSimulationVariablesFromString(config_str)
 * 
 *     def setup(self):             # <<<<<<<<<<<<<<
 *         self.set_logger()
 *         self.c_simulation.setup()
 */

/* Python wrapper */
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_14CNSESimulation_9setup(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_14CNSESimulation_9setup(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("setup (wrapper)", 0);
  __pyx_r = __pyx_pf_13pycoalescence_6necsim_6necsim_14CNSESimulation_8setup(((struct __pyx_obj_13pycoalescence_6necsim_6necsim_CNSESimulation *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_14CNSESimulation_8setup(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CNSESimulation *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("setup", 0);

  /* "pycoalescence/necsim/necsim.pyx":89
 * 
 *     def setup(self):
 *         self.set_logger()             # <<<<<<<<<<<<<<
 *         self.c_simulation.setup()
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_set_logger); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pycoalescence/necsim/necsim.pyx":90
 *     def setup(self):
 *         self.set_logger()
 *         self.c_simulation.setup()             # <<<<<<<<<<<<<<
 * 
 *     def setup_resume(self, pause_directory, out_directory, seed, task, max_time):
 */
  try {
    __pyx_v_self->c_simulation.setup();
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 90, __pyx_L1_error)
  }

  /* "pycoalescence/necsim/necsim.pyx":88
 *         self.c_simulation.importSimulationVariablesFromString(config_str)
 * 
 *     def setup(self):             # <<<<<<<<<<<<<<
 *         self.set_logger()
 *         self.c_simulation.setup()
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pycoalescence.necsim.necsim.CNSESimulation.setup", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pycoalescence/necsim/necsim.pyx":92
 *         self.c_simulation.setup()
 * 
 *     def setup_resume(self, pause_directory, out_directory, seed, task, max_time):             # <<<<<<<<<<<<<<
 *         self.set_logger()
 *         self.c_simulation.wipeSimulationVariables()
 */

/* Python wrapper */
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_14CNSESimulation_11setup_resume(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_14CNSESimulation_11setup_resume(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_pause_directory = 0;
  PyObject *__pyx_v_out_directory = 0;
  PyObject *__pyx_v_seed = 0;
  PyObject *__pyx_v_task = 0;
  PyObject *__pyx_v_max_time = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("setup_resume (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pause_directory,&__pyx_n_s_out_directory,&__pyx_n_s_seed,&__pyx_n_s_task,&__pyx_n_s_max_time,0};
    PyObject* values[5] = {0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pause_directory)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_out_directory)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setup_resume", 1, 5, 5, 1); __PYX_ERR(0, 92, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_seed)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setup_resume", 1, 5, 5, 2); __PYX_ERR(0, 92, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_task)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setup_resume", 1, 5, 5, 3); __PYX_ERR(0, 92, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_max_time)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setup_resume", 1, 5, 5, 4); __PYX_ERR(0, 92, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "setup_resume") < 0)) __PYX_ERR(0, 92, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 5) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
    }
    __pyx_v_pause_directory = values[0];
    __pyx_v_out_directory = values[1];
    __pyx_v_seed = values[2];
    __pyx_v_task = values[3];
    __pyx_v_max_time = values[4];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("setup_resume", 1, 5, 5, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 92, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pycoalescence.necsim.necsim.CNSESimulation.setup_resume", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_13pycoalescence_6necsim_6necsim_14CNSESimulation_10setup_resume(((struct __pyx_obj_13pycoalescence_6necsim_6necsim_CNSESimulation *)__pyx_v_self), __pyx_v_pause_directory, __pyx_v_out_directory, __pyx_v_seed, __pyx_v_task, __pyx_v_max_time);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_14CNSESimulation_10setup_resume(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CNSESimulation *__pyx_v_self, PyObject *__pyx_v_pause_directory, PyObject *__pyx_v_out_directory, PyObject *__pyx_v_seed, PyObject *__pyx_v_task, PyObject *__pyx_v_max_time) {
  std::string __pyx_v_pause_directory_str;
  std::string __pyx_v_out_directory_str;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  std::string __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  bool __pyx_t_8;
  int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("setup_resume", 0);

  /* "pycoalescence/necsim/necsim.pyx":93
 * 
 *     def setup_resume(self, pause_directory, out_directory, seed, task, max_time):
 *         self.set_logger()             # <<<<<<<<<<<<<<
 *         self.c_simulation.wipeSimulationVariables()
 *         cdef string pause_directory_str = pause_directory
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_set_logger); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 93, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 93, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pycoalescence/necsim/necsim.pyx":94
 *     def setup_resume(self, pause_directory, out_directory, seed, task, max_time):
 *         self.set_logger()
 *         self.c_simulation.wipeSimulationVariables()             # <<<<<<<<<<<<<<
 *         cdef string pause_directory_str = pause_directory
 *         cdef string out_directory_str = out_directory
 */
  try {
    __pyx_v_self->c_simulation.wipeSimulationVariables();
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 94, __pyx_L1_error)
  }

  /* "pycoalescence/necsim/necsim.pyx":95
 *         self.set_logger()
 *         self.c_simulation.wipeSimulationVariables()
 *         cdef string pause_directory_str = pause_directory             # <<<<<<<<<<<<<<
 *         cdef string out_directory_str = out_directory
 *         self.c_simulation.setResumeParameters(pause_directory_str, out_directory_str, seed, task, max_time)
 */
  __pyx_t_4 = __pyx_convert_string_from_py_std__in_string(__pyx_v_pause_directory); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 95, __pyx_L1_error)
  __pyx_v_pause_directory_str = __pyx_t_4;

  /* "pycoalescence/necsim/necsim.pyx":96
 *         self.c_simulation.wipeSimulationVariables()
 *         cdef string pause_directory_str = pause_directory
 *         cdef string out_directory_str = out_directory             # <<<<<<<<<<<<<<
 *         self.c_simulation.setResumeParameters(pause_directory_str, out_directory_str, seed, task, max_time)
 *         self.c_simulation.checkSims(pause_directory_str, seed, task)
 */
  __pyx_t_4 = __pyx_convert_string_from_py_std__in_string(__pyx_v_out_directory); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 96, __pyx_L1_error)
  __pyx_v_out_directory_str = __pyx_t_4;

  /* "pycoalescence/necsim/necsim.pyx":97
 *         cdef string pause_directory_str = pause_directory
 *         cdef string out_directory_str = out_directory
 *         self.c_simulation.setResumeParameters(pause_directory_str, out_directory_str, seed, task, max_time)             # <<<<<<<<<<<<<<
 *         self.c_simulation.checkSims(pause_directory_str, seed, task)
 *         if self.c_simulation.hasPaused():
 */
  __pyx_t_5 = __Pyx_PyInt_As_int(__pyx_v_seed); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 97, __pyx_L1_error)
  __pyx_t_6 = __Pyx_PyInt_As_int(__pyx_v_task); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 97, __pyx_L1_error)
  __pyx_t_7 = __Pyx_PyInt_As_int(__pyx_v_max_time); if (unlikely((__pyx_t_7 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 97, __pyx_L1_error)
  try {
    __pyx_v_self->c_simulation.setResumeParameters(__pyx_v_pause_directory_str, __pyx_v_out_directory_str, __pyx_t_5, __pyx_t_6, __pyx_t_7);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 97, __pyx_L1_error)
  }

  /* "pycoalescence/necsim/necsim.pyx":98
 *         cdef string out_directory_str = out_directory
 *         self.c_simulation.setResumeParameters(pause_directory_str, out_directory_str, seed, task, max_time)
 *         self.c_simulation.checkSims(pause_directory_str, seed, task)             # <<<<<<<<<<<<<<
 *         if self.c_simulation.hasPaused():
 *             self.c_simulation.setup()
 */
  __pyx_t_7 = __Pyx_PyInt_As_int(__pyx_v_seed); if (unlikely((__pyx_t_7 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 98, __pyx_L1_error)
  __pyx_t_6 = __Pyx_PyInt_As_int(__pyx_v_task); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 98, __pyx_L1_error)
  try {
    __pyx_v_self->c_simulation.checkSims(__pyx_v_pause_directory_str, __pyx_t_7, __pyx_t_6);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 98, __pyx_L1_error)
  }

  /* "pycoalescence/necsim/necsim.pyx":99
 *         self.c_simulation.setResumeParameters(pause_directory_str, out_directory_str, seed, task, max_time)
 *         self.c_simulation.checkSims(pause_directory_str, seed, task)
 *         if self.c_simulation.hasPaused():             # <<<<<<<<<<<<<<
 *             self.c_simulation.setup()
 *         else:
 */
  try {
    __pyx_t_8 = __pyx_v_self->c_simulation.hasPaused();
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 99, __pyx_L1_error)
  }
  __pyx_t_9 = (__pyx_t_8 != 0);
  if (likely(__pyx_t_9)) {

    /* "pycoalescence/necsim/necsim.pyx":100
 *         self.c_simulation.checkSims(pause_directory_str, seed, task)
 *         if self.c_simulation.hasPaused():
 *             self.c_simulation.setup()             # <<<<<<<<<<<<<<
 *         else:
 *             raise RuntimeError("Couldn't find paused simulation")
 */
    try {
      __pyx_v_self->c_simulation.setup();
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 100, __pyx_L1_error)
    }

    /* "pycoalescence/necsim/necsim.pyx":99
 *         self.c_simulation.setResumeParameters(pause_directory_str, out_directory_str, seed, task, max_time)
 *         self.c_simulation.checkSims(pause_directory_str, seed, task)
 *         if self.c_simulation.hasPaused():             # <<<<<<<<<<<<<<
 *             self.c_simulation.setup()
 *         else:
 */
    goto __pyx_L3;
  }

  /* "pycoalescence/necsim/necsim.pyx":102
 *             self.c_simulation.setup()
 *         else:
 *             raise RuntimeError("Couldn't find paused simulation")             # <<<<<<<<<<<<<<
 * 
 *     def add_gillespie(self, gillespie_threshold):
 */
  /*else*/ {
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple_, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 102, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 102, __pyx_L1_error)
  }
  __pyx_L3:;

  /* "pycoalescence/necsim/necsim.pyx":92
 *         self.c_simulation.setup()
 * 
 *     def setup_resume(self, pause_directory, out_directory, seed, task, max_time):             # <<<<<<<<<<<<<<
 *         self.set_logger()
 *         self.c_simulation.wipeSimulationVariables()
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pycoalescence.necsim.necsim.CNSESimulation.setup_resume", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pycoalescence/necsim/necsim.pyx":104
 *             raise RuntimeError("Couldn't find paused simulation")
 * 
 *     def add_gillespie(self, gillespie_threshold):             # <<<<<<<<<<<<<<
 *         self.set_logger()
 *         self.c_simulation.addGillespie(gillespie_threshold)
 */

/* Python wrapper */
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_14CNSESimulation_13add_gillespie(PyObject *__pyx_v_self, PyObject *__pyx_v_gillespie_threshold); /*proto*/
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_14CNSESimulation_13add_gillespie(PyObject *__pyx_v_self, PyObject *__pyx_v_gillespie_threshold) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("add_gillespie (wrapper)", 0);
  __pyx_r = __pyx_pf_13pycoalescence_6necsim_6necsim_14CNSESimulation_12add_gillespie(((struct __pyx_obj_13pycoalescence_6necsim_6necsim_CNSESimulation *)__pyx_v_self), ((PyObject *)__pyx_v_gillespie_threshold));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_14CNSESimulation_12add_gillespie(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CNSESimulation *__pyx_v_self, PyObject *__pyx_v_gillespie_threshold) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  double __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("add_gillespie", 0);

  /* "pycoalescence/necsim/necsim.pyx":105
 * 
 *     def add_gillespie(self, gillespie_threshold):
 *         self.set_logger()             # <<<<<<<<<<<<<<
 *         self.c_simulation.addGillespie(gillespie_threshold)
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_set_logger); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 105, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 105, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pycoalescence/necsim/necsim.pyx":106
 *     def add_gillespie(self, gillespie_threshold):
 *         self.set_logger()
 *         self.c_simulation.addGillespie(gillespie_threshold)             # <<<<<<<<<<<<<<
 * 
 *     def run(self) -> bool:
 */
  __pyx_t_4 = __pyx_PyFloat_AsDouble(__pyx_v_gillespie_threshold); if (unlikely((__pyx_t_4 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 106, __pyx_L1_error)
  try {
    __pyx_v_self->c_simulation.addGillespie(__pyx_t_4);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 106, __pyx_L1_error)
  }

  /* "pycoalescence/necsim/necsim.pyx":104
 *             raise RuntimeError("Couldn't find paused simulation")
 * 
 *     def add_gillespie(self, gillespie_threshold):             # <<<<<<<<<<<<<<
 *         self.set_logger()
 *         self.c_simulation.addGillespie(gillespie_threshold)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pycoalescence.necsim.necsim.CNSESimulation.add_gillespie", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pycoalescence/necsim/necsim.pyx":108
 *         self.c_simulation.addGillespie(gillespie_threshold)
 * 
 *     def run(self) -> bool:             # <<<<<<<<<<<<<<
 *         self.set_logger()
 *         return self.c_simulation.runSimulation()
 */

/* Python wrapper */
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_14CNSESimulation_15run(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_14CNSESimulation_15run(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("run (wrapper)", 0);
  __pyx_r = __pyx_pf_13pycoalescence_6necsim_6necsim_14CNSESimulation_14run(((struct __pyx_obj_13pycoalescence_6necsim_6necsim_CNSESimulation *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_14CNSESimulation_14run(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CNSESimulation *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  bool __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("run", 0);

  /* "pycoalescence/necsim/necsim.pyx":109
 * 
 *     def run(self) -> bool:
 *         self.set_logger()             # <<<<<<<<<<<<<<
 *         return self.c_simulation.runSimulation()
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_set_logger); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 109, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 109, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pycoalescence/necsim/necsim.pyx":110
 *     def run(self) -> bool:
 *         self.set_logger()
 *         return self.c_simulation.runSimulation()             # <<<<<<<<<<<<<<
 * 
 *     def apply_speciation_rates(self, speciation_rates: List[float]):
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_4 = __pyx_v_self->c_simulation.runSimulation();
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 110, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 110, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pycoalescence/necsim/necsim.pyx":108
 *         self.c_simulation.addGillespie(gillespie_threshold)
 * 
 *     def run(self) -> bool:             # <<<<<<<<<<<<<<
 *         self.set_logger()
 *         return self.c_simulation.runSimulation()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pycoalescence.necsim.necsim.CNSESimulation.run", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pycoalescence/necsim/necsim.pyx":112
 *         return self.c_simulation.runSimulation()
 * 
 *     def apply_speciation_rates(self, speciation_rates: List[float]):             # <<<<<<<<<<<<<<
 *         self.set_logger()
 *         self.c_simulation.addSpeciationRates(speciation_rates)
 */

/* Python wrapper */
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_14CNSESimulation_17apply_speciation_rates(PyObject *__pyx_v_self, PyObject *__pyx_v_speciation_rates); /*proto*/
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_14CNSESimulation_17apply_speciation_rates(PyObject *__pyx_v_self, PyObject *__pyx_v_speciation_rates) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("apply_speciation_rates (wrapper)", 0);
  __pyx_r = __pyx_pf_13pycoalescence_6necsim_6necsim_14CNSESimulation_16apply_speciation_rates(((struct __pyx_obj_13pycoalescence_6necsim_6necsim_CNSESimulation *)__pyx_v_self), ((PyObject *)__pyx_v_speciation_rates));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_14CNSESimulation_16apply_speciation_rates(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CNSESimulation *__pyx_v_self, PyObject *__pyx_v_speciation_rates) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  std::vector<long double>  __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("apply_speciation_rates", 0);

  /* "pycoalescence/necsim/necsim.pyx":113
 * 
 *     def apply_speciation_rates(self, speciation_rates: List[float]):
 *         self.set_logger()             # <<<<<<<<<<<<<<
 *         self.c_simulation.addSpeciationRates(speciation_rates)
 *         self.c_simulation.applyMultipleRates()
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_set_logger); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 113, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 113, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pycoalescence/necsim/necsim.pyx":114
 *     def apply_speciation_rates(self, speciation_rates: List[float]):
 *         self.set_logger()
 *         self.c_simulation.addSpeciationRates(speciation_rates)             # <<<<<<<<<<<<<<
 *         self.c_simulation.applyMultipleRates()
 * 
 */
  __pyx_t_4 = __pyx_convert_vector_from_py_long__double(__pyx_v_speciation_rates); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 114, __pyx_L1_error)
  try {
    __pyx_v_self->c_simulation.addSpeciationRates(__pyx_t_4);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 114, __pyx_L1_error)
  }

  /* "pycoalescence/necsim/necsim.pyx":115
 *         self.set_logger()
 *         self.c_simulation.addSpeciationRates(speciation_rates)
 *         self.c_simulation.applyMultipleRates()             # <<<<<<<<<<<<<<
 * 
 * cdef class CPNSESimulation:
 */
  try {
    __pyx_v_self->c_simulation.applyMultipleRates();
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 115, __pyx_L1_error)
  }

  /* "pycoalescence/necsim/necsim.pyx":112
 *         return self.c_simulation.runSimulation()
 * 
 *     def apply_speciation_rates(self, speciation_rates: List[float]):             # <<<<<<<<<<<<<<
 *         self.set_logger()
 *         self.c_simulation.addSpeciationRates(speciation_rates)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pycoalescence.necsim.necsim.CNSESimulation.apply_speciation_rates", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_14CNSESimulation_19__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_14CNSESimulation_19__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_13pycoalescence_6necsim_6necsim_14CNSESimulation_18__reduce_cython__(((struct __pyx_obj_13pycoalescence_6necsim_6necsim_CNSESimulation *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_14CNSESimulation_18__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_13pycoalescence_6necsim_6necsim_CNSESimulation *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pycoalescence.necsim.necsim.CNSESimulation.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_14CNSESimulation_21__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_14CNSESimulation_21__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_13pycoalescence_6necsim_6necsim_14CNSESimulation_20__setstate_cython__(((struct __pyx_obj_13pycoalescence_6necsim_6necsim_CNSESimulation *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_14CNSESimulation_20__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_13pycoalescence_6necsim_6necsim_CNSESimulation *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pycoalescence.necsim.necsim.CNSESimulation.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pycoalescence/necsim/necsim.pyx":121
 *     cdef object logger
 * 
 *     def __cinit__(self, object logger):             # <<<<<<<<<<<<<<
 *         self.logger = logger
 *         self.c_simulation = GenericTree[ProtractedTree]()
 */

/* Python wrapper */
static int __pyx_pw_13pycoalescence_6necsim_6necsim_15CPNSESimulation_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_13pycoalescence_6necsim_6necsim_15CPNSESimulation_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_logger = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_logger,0};
    PyObject* values[1] = {0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_logger)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__cinit__") < 0)) __PYX_ERR(0, 121, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 1) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
    }
    __pyx_v_logger = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 1, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 121, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pycoalescence.necsim.necsim.CPNSESimulation.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_13pycoalescence_6necsim_6necsim_15CPNSESimulation___cinit__(((struct __pyx_obj_13pycoalescence_6necsim_6necsim_CPNSESimulation *)__pyx_v_self), __pyx_v_logger);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_13pycoalescence_6necsim_6necsim_15CPNSESimulation___cinit__(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CPNSESimulation *__pyx_v_self, PyObject *__pyx_v_logger) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  necsim::GenericTree<necsim::ProtractedTree>  __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "pycoalescence/necsim/necsim.pyx":122
 * 
 *     def __cinit__(self, object logger):
 *         self.logger = logger             # <<<<<<<<<<<<<<
 *         self.c_simulation = GenericTree[ProtractedTree]()
 * 
 */
  __Pyx_INCREF(__pyx_v_logger);
  __Pyx_GIVEREF(__pyx_v_logger);
  __Pyx_GOTREF(__pyx_v_self->logger);
  __Pyx_DECREF(__pyx_v_self->logger);
  __pyx_v_self->logger = __pyx_v_logger;

  /* "pycoalescence/necsim/necsim.pyx":123
 *     def __cinit__(self, object logger):
 *         self.logger = logger
 *         self.c_simulation = GenericTree[ProtractedTree]()             # <<<<<<<<<<<<<<
 * 
 *     def set_logger(self):
 */
  try {
    __pyx_t_1 = necsim::GenericTree<necsim::ProtractedTree> ();
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 123, __pyx_L1_error)
  }
  __pyx_v_self->c_simulation = __pyx_t_1;

  /* "pycoalescence/necsim/necsim.pyx":121
 *     cdef object logger
 * 
 *     def __cinit__(self, object logger):             # <<<<<<<<<<<<<<
 *         self.logger = logger
 *         self.c_simulation = GenericTree[ProtractedTree]()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pycoalescence.necsim.necsim.CPNSESimulation.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pycoalescence/necsim/necsim.pyx":125
 *         self.c_simulation = GenericTree[ProtractedTree]()
 * 
 *     def set_logger(self):             # <<<<<<<<<<<<<<
 *         setGlobalLogger(self.logger, write_to_log)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_15CPNSESimulation_3set_logger(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_15CPNSESimulation_3set_logger(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_logger (wrapper)", 0);
  __pyx_r = __pyx_pf_13pycoalescence_6necsim_6necsim_15CPNSESimulation_2set_logger(((struct __pyx_obj_13pycoalescence_6necsim_6necsim_CPNSESimulation *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_15CPNSESimulation_2set_logger(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CPNSESimulation *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("set_logger", 0);

  /* "pycoalescence/necsim/necsim.pyx":126
 * 
 *     def set_logger(self):
 *         setGlobalLogger(self.logger, write_to_log)             # <<<<<<<<<<<<<<
 * 
 *     def import_from_config(self, config_file):
 */
  __pyx_t_1 = __pyx_v_self->logger;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_write_to_log); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 126, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  try {
    necsim::setGlobalLogger(__pyx_t_1, __pyx_t_2);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 126, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pycoalescence/necsim/necsim.pyx":125
 *         self.c_simulation = GenericTree[ProtractedTree]()
 * 
 *     def set_logger(self):             # <<<<<<<<<<<<<<
 *         setGlobalLogger(self.logger, write_to_log)
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pycoalescence.necsim.necsim.CPNSESimulation.set_logger", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pycoalescence/necsim/necsim.pyx":128
 *         setGlobalLogger(self.logger, write_to_log)
 * 
 *     def import_from_config(self, config_file):             # <<<<<<<<<<<<<<
 *         self.set_logger()
 *         self.c_simulation.importSimulationVariables(config_file)
 */

/* Python wrapper */
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_15CPNSESimulation_5import_from_config(PyObject *__pyx_v_self, PyObject *__pyx_v_config_file); /*proto*/
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_15CPNSESimulation_5import_from_config(PyObject *__pyx_v_self, PyObject *__pyx_v_config_file) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("import_from_config (wrapper)", 0);
  __pyx_r = __pyx_pf_13pycoalescence_6necsim_6necsim_15CPNSESimulation_4import_from_config(((struct __pyx_obj_13pycoalescence_6necsim_6necsim_CPNSESimulation *)__pyx_v_self), ((PyObject *)__pyx_v_config_file));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_15CPNSESimulation_4import_from_config(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CPNSESimulation *__pyx_v_self, PyObject *__pyx_v_config_file) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  std::string __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("import_from_config", 0);

  /* "pycoalescence/necsim/necsim.pyx":129
 * 
 *     def import_from_config(self, config_file):
 *         self.set_logger()             # <<<<<<<<<<<<<<
 *         self.c_simulation.importSimulationVariables(config_file)
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_set_logger); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 129, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 129, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pycoalescence/necsim/necsim.pyx":130
 *     def import_from_config(self, config_file):
 *         self.set_logger()
 *         self.c_simulation.importSimulationVariables(config_file)             # <<<<<<<<<<<<<<
 * 
 *     def import_from_config_string(self, config):
 */
  __pyx_t_4 = __pyx_convert_string_from_py_std__in_string(__pyx_v_config_file); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 130, __pyx_L1_error)
  try {
    __pyx_v_self->c_simulation.importSimulationVariables(__pyx_t_4);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 130, __pyx_L1_error)
  }

  /* "pycoalescence/necsim/necsim.pyx":128
 *         setGlobalLogger(self.logger, write_to_log)
 * 
 *     def import_from_config(self, config_file):             # <<<<<<<<<<<<<<
 *         self.set_logger()
 *         self.c_simulation.importSimulationVariables(config_file)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pycoalescence.necsim.necsim.CPNSESimulation.import_from_config", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pycoalescence/necsim/necsim.pyx":132
 *         self.c_simulation.importSimulationVariables(config_file)
 * 
 *     def import_from_config_string(self, config):             # <<<<<<<<<<<<<<
 *         self.set_logger()
 *         cdef string config_str = config
 */

/* Python wrapper */
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_15CPNSESimulation_7import_from_config_string(PyObject *__pyx_v_self, PyObject *__pyx_v_config); /*proto*/
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_15CPNSESimulation_7import_from_config_string(PyObject *__pyx_v_self, PyObject *__pyx_v_config) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("import_from_config_string (wrapper)", 0);
  __pyx_r = __pyx_pf_13pycoalescence_6necsim_6necsim_15CPNSESimulation_6import_from_config_string(((struct __pyx_obj_13pycoalescence_6necsim_6necsim_CPNSESimulation *)__pyx_v_self), ((PyObject *)__pyx_v_config));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_15CPNSESimulation_6import_from_config_string(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CPNSESimulation *__pyx_v_self, PyObject *__pyx_v_config) {
  std::string __pyx_v_config_str;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  std::string __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("import_from_config_string", 0);

  /* "pycoalescence/necsim/necsim.pyx":133
 * 
 *     def import_from_config_string(self, config):
 *         self.set_logger()             # <<<<<<<<<<<<<<
 *         cdef string config_str = config
 *         self.c_simulation.importSimulationVariablesFromString(config_str)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_set_logger); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 133, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 133, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pycoalescence/necsim/necsim.pyx":134
 *     def import_from_config_string(self, config):
 *         self.set_logger()
 *         cdef string config_str = config             # <<<<<<<<<<<<<<
 *         self.c_simulation.importSimulationVariablesFromString(config_str)
 * 
 */
  __pyx_t_4 = __pyx_convert_string_from_py_std__in_string(__pyx_v_config); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 134, __pyx_L1_error)
  __pyx_v_config_str = __pyx_t_4;

  /* "pycoalescence/necsim/necsim.pyx":135
 *         self.set_logger()
 *         cdef string config_str = config
 *         self.c_simulation.importSimulationVariablesFromString(config_str)             # <<<<<<<<<<<<<<
 * 
 *     def setup(self):
 */
  try {
    __pyx_v_self->c_simulation.importSimulationVariablesFromString(__pyx_v_config_str);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 135, __pyx_L1_error)
  }

  /* "pycoalescence/necsim/necsim.pyx":132
 *         self.c_simulation.importSimulationVariables(config_file)
 * 
 *     def import_from_config_string(self, config):             # <<<<<<<<<<<<<<
 *         self.set_logger()
 *         cdef string config_str = config
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pycoalescence.necsim.necsim.CPNSESimulation.import_from_config_string", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pycoalescence/necsim/necsim.pyx":137
 *         self.c_simulation.importSimulationVariablesFromString(config_str)
 * 
 *     def setup(self):             # <<<<<<<<<<<<<<
 *         self.set_logger()
 *         self.c_simulation.setup()
 */

/* Python wrapper */
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_15CPNSESimulation_9setup(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_15CPNSESimulation_9setup(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("setup (wrapper)", 0);
  __pyx_r = __pyx_pf_13pycoalescence_6necsim_6necsim_15CPNSESimulation_8setup(((struct __pyx_obj_13pycoalescence_6necsim_6necsim_CPNSESimulation *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_15CPNSESimulation_8setup(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CPNSESimulation *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("setup", 0);

  /* "pycoalescence/necsim/necsim.pyx":138
 * 
 *     def setup(self):
 *         self.set_logger()             # <<<<<<<<<<<<<<
 *         self.c_simulation.setup()
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_set_logger); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 138, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 138, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pycoalescence/necsim/necsim.pyx":139
 *     def setup(self):
 *         self.set_logger()
 *         self.c_simulation.setup()             # <<<<<<<<<<<<<<
 * 
 *     def setup_resume(self, pause_directory, out_directory, seed, task, max_time):
 */
  try {
    __pyx_v_self->c_simulation.setup();
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 139, __pyx_L1_error)
  }

  /* "pycoalescence/necsim/necsim.pyx":137
 *         self.c_simulation.importSimulationVariablesFromString(config_str)
 * 
 *     def setup(self):             # <<<<<<<<<<<<<<
 *         self.set_logger()
 *         self.c_simulation.setup()
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pycoalescence.necsim.necsim.CPNSESimulation.setup", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pycoalescence/necsim/necsim.pyx":141
 *         self.c_simulation.setup()
 * 
 *     def setup_resume(self, pause_directory, out_directory, seed, task, max_time):             # <<<<<<<<<<<<<<
 *         self.set_logger()
 *         self.c_simulation.wipeSimulationVariables()
 */

/* Python wrapper */
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_15CPNSESimulation_11setup_resume(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_15CPNSESimulation_11setup_resume(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_pause_directory = 0;
  PyObject *__pyx_v_out_directory = 0;
  PyObject *__pyx_v_seed = 0;
  PyObject *__pyx_v_task = 0;
  PyObject *__pyx_v_max_time = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("setup_resume (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pause_directory,&__pyx_n_s_out_directory,&__pyx_n_s_seed,&__pyx_n_s_task,&__pyx_n_s_max_time,0};
    PyObject* values[5] = {0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pause_directory)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_out_directory)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setup_resume", 1, 5, 5, 1); __PYX_ERR(0, 141, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_seed)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setup_resume", 1, 5, 5, 2); __PYX_ERR(0, 141, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_task)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setup_resume", 1, 5, 5, 3); __PYX_ERR(0, 141, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_max_time)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setup_resume", 1, 5, 5, 4); __PYX_ERR(0, 141, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "setup_resume") < 0)) __PYX_ERR(0, 141, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 5) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
    }
    __pyx_v_pause_directory = values[0];
    __pyx_v_out_directory = values[1];
    __pyx_v_seed = values[2];
    __pyx_v_task = values[3];
    __pyx_v_max_time = values[4];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("setup_resume", 1, 5, 5, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 141, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pycoalescence.necsim.necsim.CPNSESimulation.setup_resume", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_13pycoalescence_6necsim_6necsim_15CPNSESimulation_10setup_resume(((struct __pyx_obj_13pycoalescence_6necsim_6necsim_CPNSESimulation *)__pyx_v_self), __pyx_v_pause_directory, __pyx_v_out_directory, __pyx_v_seed, __pyx_v_task, __pyx_v_max_time);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_15CPNSESimulation_10setup_resume(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CPNSESimulation *__pyx_v_self, PyObject *__pyx_v_pause_directory, PyObject *__pyx_v_out_directory, PyObject *__pyx_v_seed, PyObject *__pyx_v_task, PyObject *__pyx_v_max_time) {
  std::string __pyx_v_pause_directory_str;
  std::string __pyx_v_out_directory_str;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  std::string __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  bool __pyx_t_8;
  int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("setup_resume", 0);

  /* "pycoalescence/necsim/necsim.pyx":142
 * 
 *     def setup_resume(self, pause_directory, out_directory, seed, task, max_time):
 *         self.set_logger()             # <<<<<<<<<<<<<<
 *         self.c_simulation.wipeSimulationVariables()
 *         cdef string pause_directory_str = pause_directory
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_set_logger); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 142, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 142, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pycoalescence/necsim/necsim.pyx":143
 *     def setup_resume(self, pause_directory, out_directory, seed, task, max_time):
 *         self.set_logger()
 *         self.c_simulation.wipeSimulationVariables()             # <<<<<<<<<<<<<<
 *         cdef string pause_directory_str = pause_directory
 *         cdef string out_directory_str = out_directory
 */
  try {
    __pyx_v_self->c_simulation.wipeSimulationVariables();
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 143, __pyx_L1_error)
  }

  /* "pycoalescence/necsim/necsim.pyx":144
 *         self.set_logger()
 *         self.c_simulation.wipeSimulationVariables()
 *         cdef string pause_directory_str = pause_directory             # <<<<<<<<<<<<<<
 *         cdef string out_directory_str = out_directory
 *         self.c_simulation.setResumeParameters(pause_directory_str, out_directory_str, seed, task, max_time)
 */
  __pyx_t_4 = __pyx_convert_string_from_py_std__in_string(__pyx_v_pause_directory); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 144, __pyx_L1_error)
  __pyx_v_pause_directory_str = __pyx_t_4;

  /* "pycoalescence/necsim/necsim.pyx":145
 *         self.c_simulation.wipeSimulationVariables()
 *         cdef string pause_directory_str = pause_directory
 *         cdef string out_directory_str = out_directory             # <<<<<<<<<<<<<<
 *         self.c_simulation.setResumeParameters(pause_directory_str, out_directory_str, seed, task, max_time)
 *         self.c_simulation.checkSims(pause_directory_str, seed, task)
 */
  __pyx_t_4 = __pyx_convert_string_from_py_std__in_string(__pyx_v_out_directory); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 145, __pyx_L1_error)
  __pyx_v_out_directory_str = __pyx_t_4;

  /* "pycoalescence/necsim/necsim.pyx":146
 *         cdef string pause_directory_str = pause_directory
 *         cdef string out_directory_str = out_directory
 *         self.c_simulation.setResumeParameters(pause_directory_str, out_directory_str, seed, task, max_time)             # <<<<<<<<<<<<<<
 *         self.c_simulation.checkSims(pause_directory_str, seed, task)
 *         if self.c_simulation.hasPaused():
 */
  __pyx_t_5 = __Pyx_PyInt_As_int(__pyx_v_seed); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 146, __pyx_L1_error)
  __pyx_t_6 = __Pyx_PyInt_As_int(__pyx_v_task); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 146, __pyx_L1_error)
  __pyx_t_7 = __Pyx_PyInt_As_int(__pyx_v_max_time); if (unlikely((__pyx_t_7 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 146, __pyx_L1_error)
  try {
    __pyx_v_self->c_simulation.setResumeParameters(__pyx_v_pause_directory_str, __pyx_v_out_directory_str, __pyx_t_5, __pyx_t_6, __pyx_t_7);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 146, __pyx_L1_error)
  }

  /* "pycoalescence/necsim/necsim.pyx":147
 *         cdef string out_directory_str = out_directory
 *         self.c_simulation.setResumeParameters(pause_directory_str, out_directory_str, seed, task, max_time)
 *         self.c_simulation.checkSims(pause_directory_str, seed, task)             # <<<<<<<<<<<<<<
 *         if self.c_simulation.hasPaused():
 *             self.c_simulation.setup()
 */
  __pyx_t_7 = __Pyx_PyInt_As_int(__pyx_v_seed); if (unlikely((__pyx_t_7 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 147, __pyx_L1_error)
  __pyx_t_6 = __Pyx_PyInt_As_int(__pyx_v_task); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 147, __pyx_L1_error)
  try {
    __pyx_v_self->c_simulation.checkSims(__pyx_v_pause_directory_str, __pyx_t_7, __pyx_t_6);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 147, __pyx_L1_error)
  }

  /* "pycoalescence/necsim/necsim.pyx":148
 *         self.c_simulation.setResumeParameters(pause_directory_str, out_directory_str, seed, task, max_time)
 *         self.c_simulation.checkSims(pause_directory_str, seed, task)
 *         if self.c_simulation.hasPaused():             # <<<<<<<<<<<<<<
 *             self.c_simulation.setup()
 *         else:
 */
  try {
    __pyx_t_8 = __pyx_v_self->c_simulation.hasPaused();
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 148, __pyx_L1_error)
  }
  __pyx_t_9 = (__pyx_t_8 != 0);
  if (likely(__pyx_t_9)) {

    /* "pycoalescence/necsim/necsim.pyx":149
 *         self.c_simulation.checkSims(pause_directory_str, seed, task)
 *         if self.c_simulation.hasPaused():
 *             self.c_simulation.setup()             # <<<<<<<<<<<<<<
 *         else:
 *             raise RuntimeError("Couldn't find paused simulation")
 */
    try {
      __pyx_v_self->c_simulation.setup();
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 149, __pyx_L1_error)
    }

    /* "pycoalescence/necsim/necsim.pyx":148
 *         self.c_simulation.setResumeParameters(pause_directory_str, out_directory_str, seed, task, max_time)
 *         self.c_simulation.checkSims(pause_directory_str, seed, task)
 *         if self.c_simulation.hasPaused():             # <<<<<<<<<<<<<<
 *             self.c_simulation.setup()
 *         else:
 */
    goto __pyx_L3;
  }

  /* "pycoalescence/necsim/necsim.pyx":151
 *             self.c_simulation.setup()
 *         else:
 *             raise RuntimeError("Couldn't find paused simulation")             # <<<<<<<<<<<<<<
 * 
 *     def add_gillespie(self, gillespie_threshold):
 */
  /*else*/ {
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple_, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 151, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 151, __pyx_L1_error)
  }
  __pyx_L3:;

  /* "pycoalescence/necsim/necsim.pyx":141
 *         self.c_simulation.setup()
 * 
 *     def setup_resume(self, pause_directory, out_directory, seed, task, max_time):             # <<<<<<<<<<<<<<
 *         self.set_logger()
 *         self.c_simulation.wipeSimulationVariables()
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pycoalescence.necsim.necsim.CPNSESimulation.setup_resume", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pycoalescence/necsim/necsim.pyx":153
 *             raise RuntimeError("Couldn't find paused simulation")
 * 
 *     def add_gillespie(self, gillespie_threshold):             # <<<<<<<<<<<<<<
 *         self.set_logger()
 *         self.c_simulation.addGillespie(gillespie_threshold)
 */

/* Python wrapper */
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_15CPNSESimulation_13add_gillespie(PyObject *__pyx_v_self, PyObject *__pyx_v_gillespie_threshold); /*proto*/
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_15CPNSESimulation_13add_gillespie(PyObject *__pyx_v_self, PyObject *__pyx_v_gillespie_threshold) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("add_gillespie (wrapper)", 0);
  __pyx_r = __pyx_pf_13pycoalescence_6necsim_6necsim_15CPNSESimulation_12add_gillespie(((struct __pyx_obj_13pycoalescence_6necsim_6necsim_CPNSESimulation *)__pyx_v_self), ((PyObject *)__pyx_v_gillespie_threshold));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_15CPNSESimulation_12add_gillespie(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CPNSESimulation *__pyx_v_self, PyObject *__pyx_v_gillespie_threshold) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  double __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("add_gillespie", 0);

  /* "pycoalescence/necsim/necsim.pyx":154
 * 
 *     def add_gillespie(self, gillespie_threshold):
 *         self.set_logger()             # <<<<<<<<<<<<<<
 *         self.c_simulation.addGillespie(gillespie_threshold)
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_set_logger); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 154, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 154, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pycoalescence/necsim/necsim.pyx":155
 *     def add_gillespie(self, gillespie_threshold):
 *         self.set_logger()
 *         self.c_simulation.addGillespie(gillespie_threshold)             # <<<<<<<<<<<<<<
 * 
 *     def run(self) -> bool:
 */
  __pyx_t_4 = __pyx_PyFloat_AsDouble(__pyx_v_gillespie_threshold); if (unlikely((__pyx_t_4 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 155, __pyx_L1_error)
  try {
    __pyx_v_self->c_simulation.addGillespie(__pyx_t_4);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 155, __pyx_L1_error)
  }

  /* "pycoalescence/necsim/necsim.pyx":153
 *             raise RuntimeError("Couldn't find paused simulation")
 * 
 *     def add_gillespie(self, gillespie_threshold):             # <<<<<<<<<<<<<<
 *         self.set_logger()
 *         self.c_simulation.addGillespie(gillespie_threshold)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pycoalescence.necsim.necsim.CPNSESimulation.add_gillespie", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pycoalescence/necsim/necsim.pyx":157
 *         self.c_simulation.addGillespie(gillespie_threshold)
 * 
 *     def run(self) -> bool:             # <<<<<<<<<<<<<<
 *         self.set_logger()
 *         return self.c_simulation.runSimulation()
 */

/* Python wrapper */
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_15CPNSESimulation_15run(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_15CPNSESimulation_15run(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("run (wrapper)", 0);
  __pyx_r = __pyx_pf_13pycoalescence_6necsim_6necsim_15CPNSESimulation_14run(((struct __pyx_obj_13pycoalescence_6necsim_6necsim_CPNSESimulation *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_15CPNSESimulation_14run(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CPNSESimulation *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  bool __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("run", 0);

  /* "pycoalescence/necsim/necsim.pyx":158
 * 
 *     def run(self) -> bool:
 *         self.set_logger()             # <<<<<<<<<<<<<<
 *         return self.c_simulation.runSimulation()
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_set_logger); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 158, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 158, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pycoalescence/necsim/necsim.pyx":159
 *     def run(self) -> bool:
 *         self.set_logger()
 *         return self.c_simulation.runSimulation()             # <<<<<<<<<<<<<<
 * 
 *     def apply_speciation_rates(self, speciation_rates: List[float]):
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_4 = __pyx_v_self->c_simulation.runSimulation();
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 159, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 159, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pycoalescence/necsim/necsim.pyx":157
 *         self.c_simulation.addGillespie(gillespie_threshold)
 * 
 *     def run(self) -> bool:             # <<<<<<<<<<<<<<
 *         self.set_logger()
 *         return self.c_simulation.runSimulation()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pycoalescence.necsim.necsim.CPNSESimulation.run", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pycoalescence/necsim/necsim.pyx":161
 *         return self.c_simulation.runSimulation()
 * 
 *     def apply_speciation_rates(self, speciation_rates: List[float]):             # <<<<<<<<<<<<<<
 *         self.set_logger()
 *         self.c_simulation.addSpeciationRates(speciation_rates)
 */

/* Python wrapper */
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_15CPNSESimulation_17apply_speciation_rates(PyObject *__pyx_v_self, PyObject *__pyx_v_speciation_rates); /*proto*/
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_15CPNSESimulation_17apply_speciation_rates(PyObject *__pyx_v_self, PyObject *__pyx_v_speciation_rates) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("apply_speciation_rates (wrapper)", 0);
  __pyx_r = __pyx_pf_13pycoalescence_6necsim_6necsim_15CPNSESimulation_16apply_speciation_rates(((struct __pyx_obj_13pycoalescence_6necsim_6necsim_CPNSESimulation *)__pyx_v_self), ((PyObject *)__pyx_v_speciation_rates));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_15CPNSESimulation_16apply_speciation_rates(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CPNSESimulation *__pyx_v_self, PyObject *__pyx_v_speciation_rates) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  std::vector<long double>  __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("apply_speciation_rates", 0);

  /* "pycoalescence/necsim/necsim.pyx":162
 * 
 *     def apply_speciation_rates(self, speciation_rates: List[float]):
 *         self.set_logger()             # <<<<<<<<<<<<<<
 *         self.c_simulation.addSpeciationRates(speciation_rates)
 *         self.c_simulation.applyMultipleRates()
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_set_logger); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 162, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 162, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pycoalescence/necsim/necsim.pyx":163
 *     def apply_speciation_rates(self, speciation_rates: List[float]):
 *         self.set_logger()
 *         self.c_simulation.addSpeciationRates(speciation_rates)             # <<<<<<<<<<<<<<
 *         self.c_simulation.applyMultipleRates()
 * 
 */
  __pyx_t_4 = __pyx_convert_vector_from_py_long__double(__pyx_v_speciation_rates); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 163, __pyx_L1_error)
  try {
    __pyx_v_self->c_simulation.addSpeciationRates(__pyx_t_4);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 163, __pyx_L1_error)
  }

  /* "pycoalescence/necsim/necsim.pyx":164
 *         self.set_logger()
 *         self.c_simulation.addSpeciationRates(speciation_rates)
 *         self.c_simulation.applyMultipleRates()             # <<<<<<<<<<<<<<
 * 
 * 
 */
  try {
    __pyx_v_self->c_simulation.applyMultipleRates();
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 164, __pyx_L1_error)
  }

  /* "pycoalescence/necsim/necsim.pyx":161
 *         return self.c_simulation.runSimulation()
 * 
 *     def apply_speciation_rates(self, speciation_rates: List[float]):             # <<<<<<<<<<<<<<
 *         self.set_logger()
 *         self.c_simulation.addSpeciationRates(speciation_rates)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pycoalescence.necsim.necsim.CPNSESimulation.apply_speciation_rates", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_15CPNSESimulation_19__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_15CPNSESimulation_19__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_13pycoalescence_6necsim_6necsim_15CPNSESimulation_18__reduce_cython__(((struct __pyx_obj_13pycoalescence_6necsim_6necsim_CPNSESimulation *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_15CPNSESimulation_18__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_13pycoalescence_6necsim_6necsim_CPNSESimulation *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pycoalescence.necsim.necsim.CPNSESimulation.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_15CPNSESimulation_21__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_15CPNSESimulation_21__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_13pycoalescence_6necsim_6necsim_15CPNSESimulation_20__setstate_cython__(((struct __pyx_obj_13pycoalescence_6necsim_6necsim_CPNSESimulation *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_15CPNSESimulation_20__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_13pycoalescence_6necsim_6necsim_CPNSESimulation *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pycoalescence.necsim.necsim.CPNSESimulation.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pycoalescence/necsim/necsim.pyx":171
 *     cdef object logger
 * 
 *     def __cinit__(self, object logger):             # <<<<<<<<<<<<<<
 *         self.logger = logger
 *         self.c_simulation = GenericTree[ProtractedSpatialTree]()
 */

/* Python wrapper */
static int __pyx_pw_13pycoalescence_6necsim_6necsim_19CPSpatialSimulation_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_13pycoalescence_6necsim_6necsim_19CPSpatialSimulation_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_logger = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_logger,0};
    PyObject* values[1] = {0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_logger)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__cinit__") < 0)) __PYX_ERR(0, 171, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 1) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
    }
    __pyx_v_logger = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 1, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 171, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pycoalescence.necsim.necsim.CPSpatialSimulation.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_13pycoalescence_6necsim_6necsim_19CPSpatialSimulation___cinit__(((struct __pyx_obj_13pycoalescence_6necsim_6necsim_CPSpatialSimulation *)__pyx_v_self), __pyx_v_logger);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_13pycoalescence_6necsim_6necsim_19CPSpatialSimulation___cinit__(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CPSpatialSimulation *__pyx_v_self, PyObject *__pyx_v_logger) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  necsim::GenericTree<necsim::ProtractedSpatialTree>  __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "pycoalescence/necsim/necsim.pyx":172
 * 
 *     def __cinit__(self, object logger):
 *         self.logger = logger             # <<<<<<<<<<<<<<
 *         self.c_simulation = GenericTree[ProtractedSpatialTree]()
 * 
 */
  __Pyx_INCREF(__pyx_v_logger);
  __Pyx_GIVEREF(__pyx_v_logger);
  __Pyx_GOTREF(__pyx_v_self->logger);
  __Pyx_DECREF(__pyx_v_self->logger);
  __pyx_v_self->logger = __pyx_v_logger;

  /* "pycoalescence/necsim/necsim.pyx":173
 *     def __cinit__(self, object logger):
 *         self.logger = logger
 *         self.c_simulation = GenericTree[ProtractedSpatialTree]()             # <<<<<<<<<<<<<<
 * 
 *     def set_logger(self):
 */
  try {
    __pyx_t_1 = necsim::GenericTree<necsim::ProtractedSpatialTree> ();
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 173, __pyx_L1_error)
  }
  __pyx_v_self->c_simulation = __pyx_t_1;

  /* "pycoalescence/necsim/necsim.pyx":171
 *     cdef object logger
 * 
 *     def __cinit__(self, object logger):             # <<<<<<<<<<<<<<
 *         self.logger = logger
 *         self.c_simulation = GenericTree[ProtractedSpatialTree]()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pycoalescence.necsim.necsim.CPSpatialSimulation.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pycoalescence/necsim/necsim.pyx":175
 *         self.c_simulation = GenericTree[ProtractedSpatialTree]()
 * 
 *     def set_logger(self):             # <<<<<<<<<<<<<<
 *         setGlobalLogger(self.logger, write_to_log)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_19CPSpatialSimulation_3set_logger(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_19CPSpatialSimulation_3set_logger(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_logger (wrapper)", 0);
  __pyx_r = __pyx_pf_13pycoalescence_6necsim_6necsim_19CPSpatialSimulation_2set_logger(((struct __pyx_obj_13pycoalescence_6necsim_6necsim_CPSpatialSimulation *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_19CPSpatialSimulation_2set_logger(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CPSpatialSimulation *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("set_logger", 0);

  /* "pycoalescence/necsim/necsim.pyx":176
 * 
 *     def set_logger(self):
 *         setGlobalLogger(self.logger, write_to_log)             # <<<<<<<<<<<<<<
 * 
 *     def import_from_config(self, config_file):
 */
  __pyx_t_1 = __pyx_v_self->logger;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_write_to_log); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 176, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  try {
    necsim::setGlobalLogger(__pyx_t_1, __pyx_t_2);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 176, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pycoalescence/necsim/necsim.pyx":175
 *         self.c_simulation = GenericTree[ProtractedSpatialTree]()
 * 
 *     def set_logger(self):             # <<<<<<<<<<<<<<
 *         setGlobalLogger(self.logger, write_to_log)
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pycoalescence.necsim.necsim.CPSpatialSimulation.set_logger", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pycoalescence/necsim/necsim.pyx":178
 *         setGlobalLogger(self.logger, write_to_log)
 * 
 *     def import_from_config(self, config_file):             # <<<<<<<<<<<<<<
 *         self.set_logger()
 *         self.c_simulation.importSimulationVariables(config_file)
 */

/* Python wrapper */
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_19CPSpatialSimulation_5import_from_config(PyObject *__pyx_v_self, PyObject *__pyx_v_config_file); /*proto*/
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_19CPSpatialSimulation_5import_from_config(PyObject *__pyx_v_self, PyObject *__pyx_v_config_file) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("import_from_config (wrapper)", 0);
  __pyx_r = __pyx_pf_13pycoalescence_6necsim_6necsim_19CPSpatialSimulation_4import_from_config(((struct __pyx_obj_13pycoalescence_6necsim_6necsim_CPSpatialSimulation *)__pyx_v_self), ((PyObject *)__pyx_v_config_file));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_19CPSpatialSimulation_4import_from_config(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CPSpatialSimulation *__pyx_v_self, PyObject *__pyx_v_config_file) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  std::string __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("import_from_config", 0);

  /* "pycoalescence/necsim/necsim.pyx":179
 * 
 *     def import_from_config(self, config_file):
 *         self.set_logger()             # <<<<<<<<<<<<<<
 *         self.c_simulation.importSimulationVariables(config_file)
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_set_logger); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 179, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 179, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pycoalescence/necsim/necsim.pyx":180
 *     def import_from_config(self, config_file):
 *         self.set_logger()
 *         self.c_simulation.importSimulationVariables(config_file)             # <<<<<<<<<<<<<<
 * 
 *     def import_from_config_string(self, config):
 */
  __pyx_t_4 = __pyx_convert_string_from_py_std__in_string(__pyx_v_config_file); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 180, __pyx_L1_error)
  try {
    __pyx_v_self->c_simulation.importSimulationVariables(__pyx_t_4);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 180, __pyx_L1_error)
  }

  /* "pycoalescence/necsim/necsim.pyx":178
 *         setGlobalLogger(self.logger, write_to_log)
 * 
 *     def import_from_config(self, config_file):             # <<<<<<<<<<<<<<
 *         self.set_logger()
 *         self.c_simulation.importSimulationVariables(config_file)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pycoalescence.necsim.necsim.CPSpatialSimulation.import_from_config", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pycoalescence/necsim/necsim.pyx":182
 *         self.c_simulation.importSimulationVariables(config_file)
 * 
 *     def import_from_config_string(self, config):             # <<<<<<<<<<<<<<
 *         self.set_logger()
 *         cdef string config_str = config
 */

/* Python wrapper */
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_19CPSpatialSimulation_7import_from_config_string(PyObject *__pyx_v_self, PyObject *__pyx_v_config); /*proto*/
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_19CPSpatialSimulation_7import_from_config_string(PyObject *__pyx_v_self, PyObject *__pyx_v_config) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("import_from_config_string (wrapper)", 0);
  __pyx_r = __pyx_pf_13pycoalescence_6necsim_6necsim_19CPSpatialSimulation_6import_from_config_string(((struct __pyx_obj_13pycoalescence_6necsim_6necsim_CPSpatialSimulation *)__pyx_v_self), ((PyObject *)__pyx_v_config));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_19CPSpatialSimulation_6import_from_config_string(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CPSpatialSimulation *__pyx_v_self, PyObject *__pyx_v_config) {
  std::string __pyx_v_config_str;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  std::string __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("import_from_config_string", 0);

  /* "pycoalescence/necsim/necsim.pyx":183
 * 
 *     def import_from_config_string(self, config):
 *         self.set_logger()             # <<<<<<<<<<<<<<
 *         cdef string config_str = config
 *         self.c_simulation.importSimulationVariablesFromString(config_str)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_set_logger); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 183, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 183, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pycoalescence/necsim/necsim.pyx":184
 *     def import_from_config_string(self, config):
 *         self.set_logger()
 *         cdef string config_str = config             # <<<<<<<<<<<<<<
 *         self.c_simulation.importSimulationVariablesFromString(config_str)
 * 
 */
  __pyx_t_4 = __pyx_convert_string_from_py_std__in_string(__pyx_v_config); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 184, __pyx_L1_error)
  __pyx_v_config_str = __pyx_t_4;

  /* "pycoalescence/necsim/necsim.pyx":185
 *         self.set_logger()
 *         cdef string config_str = config
 *         self.c_simulation.importSimulationVariablesFromString(config_str)             # <<<<<<<<<<<<<<
 * 
 *     def setup(self):
 */
  try {
    __pyx_v_self->c_simulation.importSimulationVariablesFromString(__pyx_v_config_str);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 185, __pyx_L1_error)
  }

  /* "pycoalescence/necsim/necsim.pyx":182
 *         self.c_simulation.importSimulationVariables(config_file)
 * 
 *     def import_from_config_string(self, config):             # <<<<<<<<<<<<<<
 *         self.set_logger()
 *         cdef string config_str = config
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pycoalescence.necsim.necsim.CPSpatialSimulation.import_from_config_string", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pycoalescence/necsim/necsim.pyx":187
 *         self.c_simulation.importSimulationVariablesFromString(config_str)
 * 
 *     def setup(self):             # <<<<<<<<<<<<<<
 *         self.set_logger()
 *         self.c_simulation.setup()
 */

/* Python wrapper */
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_19CPSpatialSimulation_9setup(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_19CPSpatialSimulation_9setup(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("setup (wrapper)", 0);
  __pyx_r = __pyx_pf_13pycoalescence_6necsim_6necsim_19CPSpatialSimulation_8setup(((struct __pyx_obj_13pycoalescence_6necsim_6necsim_CPSpatialSimulation *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_19CPSpatialSimulation_8setup(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CPSpatialSimulation *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("setup", 0);

  /* "pycoalescence/necsim/necsim.pyx":188
 * 
 *     def setup(self):
 *         self.set_logger()             # <<<<<<<<<<<<<<
 *         self.c_simulation.setup()
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_set_logger); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 188, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 188, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pycoalescence/necsim/necsim.pyx":189
 *     def setup(self):
 *         self.set_logger()
 *         self.c_simulation.setup()             # <<<<<<<<<<<<<<
 * 
 *     def setup_resume(self, pause_directory, out_directory, seed, task, max_time):
 */
  try {
    __pyx_v_self->c_simulation.setup();
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 189, __pyx_L1_error)
  }

  /* "pycoalescence/necsim/necsim.pyx":187
 *         self.c_simulation.importSimulationVariablesFromString(config_str)
 * 
 *     def setup(self):             # <<<<<<<<<<<<<<
 *         self.set_logger()
 *         self.c_simulation.setup()
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pycoalescence.necsim.necsim.CPSpatialSimulation.setup", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pycoalescence/necsim/necsim.pyx":191
 *         self.c_simulation.setup()
 * 
 *     def setup_resume(self, pause_directory, out_directory, seed, task, max_time):             # <<<<<<<<<<<<<<
 *         self.set_logger()
 *         self.c_simulation.wipeSimulationVariables()
 */

/* Python wrapper */
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_19CPSpatialSimulation_11setup_resume(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_19CPSpatialSimulation_11setup_resume(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_pause_directory = 0;
  PyObject *__pyx_v_out_directory = 0;
  PyObject *__pyx_v_seed = 0;
  PyObject *__pyx_v_task = 0;
  PyObject *__pyx_v_max_time = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("setup_resume (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pause_directory,&__pyx_n_s_out_directory,&__pyx_n_s_seed,&__pyx_n_s_task,&__pyx_n_s_max_time,0};
    PyObject* values[5] = {0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pause_directory)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_out_directory)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setup_resume", 1, 5, 5, 1); __PYX_ERR(0, 191, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_seed)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setup_resume", 1, 5, 5, 2); __PYX_ERR(0, 191, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_task)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setup_resume", 1, 5, 5, 3); __PYX_ERR(0, 191, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_max_time)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setup_resume", 1, 5, 5, 4); __PYX_ERR(0, 191, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "setup_resume") < 0)) __PYX_ERR(0, 191, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 5) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
    }
    __pyx_v_pause_directory = values[0];
    __pyx_v_out_directory = values[1];
    __pyx_v_seed = values[2];
    __pyx_v_task = values[3];
    __pyx_v_max_time = values[4];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("setup_resume", 1, 5, 5, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 191, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pycoalescence.necsim.necsim.CPSpatialSimulation.setup_resume", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_13pycoalescence_6necsim_6necsim_19CPSpatialSimulation_10setup_resume(((struct __pyx_obj_13pycoalescence_6necsim_6necsim_CPSpatialSimulation *)__pyx_v_self), __pyx_v_pause_directory, __pyx_v_out_directory, __pyx_v_seed, __pyx_v_task, __pyx_v_max_time);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_19CPSpatialSimulation_10setup_resume(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CPSpatialSimulation *__pyx_v_self, PyObject *__pyx_v_pause_directory, PyObject *__pyx_v_out_directory, PyObject *__pyx_v_seed, PyObject *__pyx_v_task, PyObject *__pyx_v_max_time) {
  std::string __pyx_v_pause_directory_str;
  std::string __pyx_v_out_directory_str;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  std::string __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  bool __pyx_t_8;
  int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("setup_resume", 0);

  /* "pycoalescence/necsim/necsim.pyx":192
 * 
 *     def setup_resume(self, pause_directory, out_directory, seed, task, max_time):
 *         self.set_logger()             # <<<<<<<<<<<<<<
 *         self.c_simulation.wipeSimulationVariables()
 *         cdef string pause_directory_str = pause_directory
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_set_logger); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 192, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 192, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pycoalescence/necsim/necsim.pyx":193
 *     def setup_resume(self, pause_directory, out_directory, seed, task, max_time):
 *         self.set_logger()
 *         self.c_simulation.wipeSimulationVariables()             # <<<<<<<<<<<<<<
 *         cdef string pause_directory_str = pause_directory
 *         cdef string out_directory_str = out_directory
 */
  try {
    __pyx_v_self->c_simulation.wipeSimulationVariables();
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 193, __pyx_L1_error)
  }

  /* "pycoalescence/necsim/necsim.pyx":194
 *         self.set_logger()
 *         self.c_simulation.wipeSimulationVariables()
 *         cdef string pause_directory_str = pause_directory             # <<<<<<<<<<<<<<
 *         cdef string out_directory_str = out_directory
 *         self.c_simulation.setResumeParameters(pause_directory_str, out_directory_str, seed, task, max_time)
 */
  __pyx_t_4 = __pyx_convert_string_from_py_std__in_string(__pyx_v_pause_directory); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 194, __pyx_L1_error)
  __pyx_v_pause_directory_str = __pyx_t_4;

  /* "pycoalescence/necsim/necsim.pyx":195
 *         self.c_simulation.wipeSimulationVariables()
 *         cdef string pause_directory_str = pause_directory
 *         cdef string out_directory_str = out_directory             # <<<<<<<<<<<<<<
 *         self.c_simulation.setResumeParameters(pause_directory_str, out_directory_str, seed, task, max_time)
 *         self.c_simulation.checkSims(pause_directory_str, seed, task)
 */
  __pyx_t_4 = __pyx_convert_string_from_py_std__in_string(__pyx_v_out_directory); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 195, __pyx_L1_error)
  __pyx_v_out_directory_str = __pyx_t_4;

  /* "pycoalescence/necsim/necsim.pyx":196
 *         cdef string pause_directory_str = pause_directory
 *         cdef string out_directory_str = out_directory
 *         self.c_simulation.setResumeParameters(pause_directory_str, out_directory_str, seed, task, max_time)             # <<<<<<<<<<<<<<
 *         self.c_simulation.checkSims(pause_directory_str, seed, task)
 *         if self.c_simulation.hasPaused():
 */
  __pyx_t_5 = __Pyx_PyInt_As_int(__pyx_v_seed); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 196, __pyx_L1_error)
  __pyx_t_6 = __Pyx_PyInt_As_int(__pyx_v_task); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 196, __pyx_L1_error)
  __pyx_t_7 = __Pyx_PyInt_As_int(__pyx_v_max_time); if (unlikely((__pyx_t_7 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 196, __pyx_L1_error)
  try {
    __pyx_v_self->c_simulation.setResumeParameters(__pyx_v_pause_directory_str, __pyx_v_out_directory_str, __pyx_t_5, __pyx_t_6, __pyx_t_7);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 196, __pyx_L1_error)
  }

  /* "pycoalescence/necsim/necsim.pyx":197
 *         cdef string out_directory_str = out_directory
 *         self.c_simulation.setResumeParameters(pause_directory_str, out_directory_str, seed, task, max_time)
 *         self.c_simulation.checkSims(pause_directory_str, seed, task)             # <<<<<<<<<<<<<<
 *         if self.c_simulation.hasPaused():
 *             self.c_simulation.setup()
 */
  __pyx_t_7 = __Pyx_PyInt_As_int(__pyx_v_seed); if (unlikely((__pyx_t_7 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 197, __pyx_L1_error)
  __pyx_t_6 = __Pyx_PyInt_As_int(__pyx_v_task); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 197, __pyx_L1_error)
  try {
    __pyx_v_self->c_simulation.checkSims(__pyx_v_pause_directory_str, __pyx_t_7, __pyx_t_6);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 197, __pyx_L1_error)
  }

  /* "pycoalescence/necsim/necsim.pyx":198
 *         self.c_simulation.setResumeParameters(pause_directory_str, out_directory_str, seed, task, max_time)
 *         self.c_simulation.checkSims(pause_directory_str, seed, task)
 *         if self.c_simulation.hasPaused():             # <<<<<<<<<<<<<<
 *             self.c_simulation.setup()
 *         else:
 */
  try {
    __pyx_t_8 = __pyx_v_self->c_simulation.hasPaused();
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 198, __pyx_L1_error)
  }
  __pyx_t_9 = (__pyx_t_8 != 0);
  if (likely(__pyx_t_9)) {

    /* "pycoalescence/necsim/necsim.pyx":199
 *         self.c_simulation.checkSims(pause_directory_str, seed, task)
 *         if self.c_simulation.hasPaused():
 *             self.c_simulation.setup()             # <<<<<<<<<<<<<<
 *         else:
 *             raise RuntimeError("Couldn't find paused simulation")
 */
    try {
      __pyx_v_self->c_simulation.setup();
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 199, __pyx_L1_error)
    }

    /* "pycoalescence/necsim/necsim.pyx":198
 *         self.c_simulation.setResumeParameters(pause_directory_str, out_directory_str, seed, task, max_time)
 *         self.c_simulation.checkSims(pause_directory_str, seed, task)
 *         if self.c_simulation.hasPaused():             # <<<<<<<<<<<<<<
 *             self.c_simulation.setup()
 *         else:
 */
    goto __pyx_L3;
  }

  /* "pycoalescence/necsim/necsim.pyx":201
 *             self.c_simulation.setup()
 *         else:
 *             raise RuntimeError("Couldn't find paused simulation")             # <<<<<<<<<<<<<<
 * 
 *     def add_gillespie(self, gillespie_threshold):
 */
  /*else*/ {
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple_, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 201, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 201, __pyx_L1_error)
  }
  __pyx_L3:;

  /* "pycoalescence/necsim/necsim.pyx":191
 *         self.c_simulation.setup()
 * 
 *     def setup_resume(self, pause_directory, out_directory, seed, task, max_time):             # <<<<<<<<<<<<<<
 *         self.set_logger()
 *         self.c_simulation.wipeSimulationVariables()
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pycoalescence.necsim.necsim.CPSpatialSimulation.setup_resume", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pycoalescence/necsim/necsim.pyx":203
 *             raise RuntimeError("Couldn't find paused simulation")
 * 
 *     def add_gillespie(self, gillespie_threshold):             # <<<<<<<<<<<<<<
 *         self.set_logger()
 *         self.c_simulation.addGillespie(gillespie_threshold)
 */

/* Python wrapper */
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_19CPSpatialSimulation_13add_gillespie(PyObject *__pyx_v_self, PyObject *__pyx_v_gillespie_threshold); /*proto*/
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_19CPSpatialSimulation_13add_gillespie(PyObject *__pyx_v_self, PyObject *__pyx_v_gillespie_threshold) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("add_gillespie (wrapper)", 0);
  __pyx_r = __pyx_pf_13pycoalescence_6necsim_6necsim_19CPSpatialSimulation_12add_gillespie(((struct __pyx_obj_13pycoalescence_6necsim_6necsim_CPSpatialSimulation *)__pyx_v_self), ((PyObject *)__pyx_v_gillespie_threshold));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_19CPSpatialSimulation_12add_gillespie(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CPSpatialSimulation *__pyx_v_self, PyObject *__pyx_v_gillespie_threshold) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  double __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("add_gillespie", 0);

  /* "pycoalescence/necsim/necsim.pyx":204
 * 
 *     def add_gillespie(self, gillespie_threshold):
 *         self.set_logger()             # <<<<<<<<<<<<<<
 *         self.c_simulation.addGillespie(gillespie_threshold)
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_set_logger); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 204, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 204, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pycoalescence/necsim/necsim.pyx":205
 *     def add_gillespie(self, gillespie_threshold):
 *         self.set_logger()
 *         self.c_simulation.addGillespie(gillespie_threshold)             # <<<<<<<<<<<<<<
 * 
 *     def run(self) -> bool:
 */
  __pyx_t_4 = __pyx_PyFloat_AsDouble(__pyx_v_gillespie_threshold); if (unlikely((__pyx_t_4 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 205, __pyx_L1_error)
  try {
    __pyx_v_self->c_simulation.addGillespie(__pyx_t_4);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 205, __pyx_L1_error)
  }

  /* "pycoalescence/necsim/necsim.pyx":203
 *             raise RuntimeError("Couldn't find paused simulation")
 * 
 *     def add_gillespie(self, gillespie_threshold):             # <<<<<<<<<<<<<<
 *         self.set_logger()
 *         self.c_simulation.addGillespie(gillespie_threshold)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pycoalescence.necsim.necsim.CPSpatialSimulation.add_gillespie", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pycoalescence/necsim/necsim.pyx":207
 *         self.c_simulation.addGillespie(gillespie_threshold)
 * 
 *     def run(self) -> bool:             # <<<<<<<<<<<<<<
 *         self.set_logger()
 *         return self.c_simulation.runSimulation()
 */

/* Python wrapper */
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_19CPSpatialSimulation_15run(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_19CPSpatialSimulation_15run(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("run (wrapper)", 0);
  __pyx_r = __pyx_pf_13pycoalescence_6necsim_6necsim_19CPSpatialSimulation_14run(((struct __pyx_obj_13pycoalescence_6necsim_6necsim_CPSpatialSimulation *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_19CPSpatialSimulation_14run(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CPSpatialSimulation *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  bool __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("run", 0);

  /* "pycoalescence/necsim/necsim.pyx":208
 * 
 *     def run(self) -> bool:
 *         self.set_logger()             # <<<<<<<<<<<<<<
 *         return self.c_simulation.runSimulation()
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_set_logger); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 208, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 208, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pycoalescence/necsim/necsim.pyx":209
 *     def run(self) -> bool:
 *         self.set_logger()
 *         return self.c_simulation.runSimulation()             # <<<<<<<<<<<<<<
 * 
 *     def apply_speciation_rates(self, speciation_rates: List[float]):
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_4 = __pyx_v_self->c_simulation.runSimulation();
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 209, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 209, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pycoalescence/necsim/necsim.pyx":207
 *         self.c_simulation.addGillespie(gillespie_threshold)
 * 
 *     def run(self) -> bool:             # <<<<<<<<<<<<<<
 *         self.set_logger()
 *         return self.c_simulation.runSimulation()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pycoalescence.necsim.necsim.CPSpatialSimulation.run", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pycoalescence/necsim/necsim.pyx":211
 *         return self.c_simulation.runSimulation()
 * 
 *     def apply_speciation_rates(self, speciation_rates: List[float]):             # <<<<<<<<<<<<<<
 *         self.set_logger()
 *         self.c_simulation.addSpeciationRates(speciation_rates)
 */

/* Python wrapper */
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_19CPSpatialSimulation_17apply_speciation_rates(PyObject *__pyx_v_self, PyObject *__pyx_v_speciation_rates); /*proto*/
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_19CPSpatialSimulation_17apply_speciation_rates(PyObject *__pyx_v_self, PyObject *__pyx_v_speciation_rates) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("apply_speciation_rates (wrapper)", 0);
  __pyx_r = __pyx_pf_13pycoalescence_6necsim_6necsim_19CPSpatialSimulation_16apply_speciation_rates(((struct __pyx_obj_13pycoalescence_6necsim_6necsim_CPSpatialSimulation *)__pyx_v_self), ((PyObject *)__pyx_v_speciation_rates));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_19CPSpatialSimulation_16apply_speciation_rates(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CPSpatialSimulation *__pyx_v_self, PyObject *__pyx_v_speciation_rates) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  std::vector<long double>  __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("apply_speciation_rates", 0);

  /* "pycoalescence/necsim/necsim.pyx":212
 * 
 *     def apply_speciation_rates(self, speciation_rates: List[float]):
 *         self.set_logger()             # <<<<<<<<<<<<<<
 *         self.c_simulation.addSpeciationRates(speciation_rates)
 *         self.c_simulation.applyMultipleRates()
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_set_logger); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 212, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 212, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pycoalescence/necsim/necsim.pyx":213
 *     def apply_speciation_rates(self, speciation_rates: List[float]):
 *         self.set_logger()
 *         self.c_simulation.addSpeciationRates(speciation_rates)             # <<<<<<<<<<<<<<
 *         self.c_simulation.applyMultipleRates()
 * 
 */
  __pyx_t_4 = __pyx_convert_vector_from_py_long__double(__pyx_v_speciation_rates); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 213, __pyx_L1_error)
  try {
    __pyx_v_self->c_simulation.addSpeciationRates(__pyx_t_4);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 213, __pyx_L1_error)
  }

  /* "pycoalescence/necsim/necsim.pyx":214
 *         self.set_logger()
 *         self.c_simulation.addSpeciationRates(speciation_rates)
 *         self.c_simulation.applyMultipleRates()             # <<<<<<<<<<<<<<
 * 
 * cdef class CCommunity:
 */
  try {
    __pyx_v_self->c_simulation.applyMultipleRates();
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 214, __pyx_L1_error)
  }

  /* "pycoalescence/necsim/necsim.pyx":211
 *         return self.c_simulation.runSimulation()
 * 
 *     def apply_speciation_rates(self, speciation_rates: List[float]):             # <<<<<<<<<<<<<<
 *         self.set_logger()
 *         self.c_simulation.addSpeciationRates(speciation_rates)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pycoalescence.necsim.necsim.CPSpatialSimulation.apply_speciation_rates", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_19CPSpatialSimulation_19__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_19CPSpatialSimulation_19__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_13pycoalescence_6necsim_6necsim_19CPSpatialSimulation_18__reduce_cython__(((struct __pyx_obj_13pycoalescence_6necsim_6necsim_CPSpatialSimulation *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_19CPSpatialSimulation_18__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_13pycoalescence_6necsim_6necsim_CPSpatialSimulation *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__8, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pycoalescence.necsim.necsim.CPSpatialSimulation.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_19CPSpatialSimulation_21__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_19CPSpatialSimulation_21__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_13pycoalescence_6necsim_6necsim_19CPSpatialSimulation_20__setstate_cython__(((struct __pyx_obj_13pycoalescence_6necsim_6necsim_CPSpatialSimulation *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_19CPSpatialSimulation_20__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_13pycoalescence_6necsim_6necsim_CPSpatialSimulation *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__9, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pycoalescence.necsim.necsim.CPSpatialSimulation.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pycoalescence/necsim/necsim.pyx":221
 *     cdef object logger
 * 
 *     def __cinit__(self, object logger):             # <<<<<<<<<<<<<<
 *         self.logger = logger
 *         self.c_community = Community()
 */

/* Python wrapper */
static int __pyx_pw_13pycoalescence_6necsim_6necsim_10CCommunity_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_13pycoalescence_6necsim_6necsim_10CCommunity_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_logger = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_logger,0};
    PyObject* values[1] = {0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_logger)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__cinit__") < 0)) __PYX_ERR(0, 221, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 1) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
    }
    __pyx_v_logger = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 1, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 221, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pycoalescence.necsim.necsim.CCommunity.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_13pycoalescence_6necsim_6necsim_10CCommunity___cinit__(((struct __pyx_obj_13pycoalescence_6necsim_6necsim_CCommunity *)__pyx_v_self), __pyx_v_logger);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_13pycoalescence_6necsim_6necsim_10CCommunity___cinit__(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CCommunity *__pyx_v_self, PyObject *__pyx_v_logger) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  necsim::Community __pyx_t_1;
  std::shared_ptr<struct necsim::SpecSimParameters>  __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "pycoalescence/necsim/necsim.pyx":222
 * 
 *     def __cinit__(self, object logger):
 *         self.logger = logger             # <<<<<<<<<<<<<<
 *         self.c_community = Community()
 *         self.c_spec_parameters = make_shared[SpecSimParameters]()
 */
  __Pyx_INCREF(__pyx_v_logger);
  __Pyx_GIVEREF(__pyx_v_logger);
  __Pyx_GOTREF(__pyx_v_self->logger);
  __Pyx_DECREF(__pyx_v_self->logger);
  __pyx_v_self->logger = __pyx_v_logger;

  /* "pycoalescence/necsim/necsim.pyx":223
 *     def __cinit__(self, object logger):
 *         self.logger = logger
 *         self.c_community = Community()             # <<<<<<<<<<<<<<
 *         self.c_spec_parameters = make_shared[SpecSimParameters]()
 * 
 */
  try {
    __pyx_t_1 = necsim::Community();
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 223, __pyx_L1_error)
  }
  __pyx_v_self->c_community = __pyx_t_1;

  /* "pycoalescence/necsim/necsim.pyx":224
 *         self.logger = logger
 *         self.c_community = Community()
 *         self.c_spec_parameters = make_shared[SpecSimParameters]()             # <<<<<<<<<<<<<<
 * 
 *     def set_logger(self):
 */
  try {
    __pyx_t_2 = std::make_shared<struct necsim::SpecSimParameters>();
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 224, __pyx_L1_error)
  }
  __pyx_v_self->c_spec_parameters = __pyx_t_2;

  /* "pycoalescence/necsim/necsim.pyx":221
 *     cdef object logger
 * 
 *     def __cinit__(self, object logger):             # <<<<<<<<<<<<<<
 *         self.logger = logger
 *         self.c_community = Community()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pycoalescence.necsim.necsim.CCommunity.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pycoalescence/necsim/necsim.pyx":226
 *         self.c_spec_parameters = make_shared[SpecSimParameters]()
 * 
 *     def set_logger(self):             # <<<<<<<<<<<<<<
 *         setGlobalLogger(self.logger, write_to_log)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_10CCommunity_3set_logger(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_10CCommunity_3set_logger(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_logger (wrapper)", 0);
  __pyx_r = __pyx_pf_13pycoalescence_6necsim_6necsim_10CCommunity_2set_logger(((struct __pyx_obj_13pycoalescence_6necsim_6necsim_CCommunity *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_10CCommunity_2set_logger(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CCommunity *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("set_logger", 0);

  /* "pycoalescence/necsim/necsim.pyx":227
 * 
 *     def set_logger(self):
 *         setGlobalLogger(self.logger, write_to_log)             # <<<<<<<<<<<<<<
 * 
 *     def wipe_protracted_parameters(self):
 */
  __pyx_t_1 = __pyx_v_self->logger;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_write_to_log); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 227, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  try {
    necsim::setGlobalLogger(__pyx_t_1, __pyx_t_2);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 227, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pycoalescence/necsim/necsim.pyx":226
 *         self.c_spec_parameters = make_shared[SpecSimParameters]()
 * 
 *     def set_logger(self):             # <<<<<<<<<<<<<<
 *         setGlobalLogger(self.logger, write_to_log)
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pycoalescence.necsim.necsim.CCommunity.set_logger", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pycoalescence/necsim/necsim.pyx":229
 *         setGlobalLogger(self.logger, write_to_log)
 * 
 *     def wipe_protracted_parameters(self):             # <<<<<<<<<<<<<<
 *         self.set_logger()
 *         deref(self.c_spec_parameters).wipeProtractedParameters()
 */

/* Python wrapper */
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_10CCommunity_5wipe_protracted_parameters(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_10CCommunity_5wipe_protracted_parameters(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("wipe_protracted_parameters (wrapper)", 0);
  __pyx_r = __pyx_pf_13pycoalescence_6necsim_6necsim_10CCommunity_4wipe_protracted_parameters(((struct __pyx_obj_13pycoalescence_6necsim_6necsim_CCommunity *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_10CCommunity_4wipe_protracted_parameters(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CCommunity *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("wipe_protracted_parameters", 0);

  /* "pycoalescence/necsim/necsim.pyx":230
 * 
 *     def wipe_protracted_parameters(self):
 *         self.set_logger()             # <<<<<<<<<<<<<<
 *         deref(self.c_spec_parameters).wipeProtractedParameters()
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_set_logger); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 230, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 230, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pycoalescence/necsim/necsim.pyx":231
 *     def wipe_protracted_parameters(self):
 *         self.set_logger()
 *         deref(self.c_spec_parameters).wipeProtractedParameters()             # <<<<<<<<<<<<<<
 * 
 *     def add_protracted_parameters(self, double min_speciation_gen, double max_speciation_gen):
 */
  try {
    (*__pyx_v_self->c_spec_parameters).wipeProtractedParameters();
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 231, __pyx_L1_error)
  }

  /* "pycoalescence/necsim/necsim.pyx":229
 *         setGlobalLogger(self.logger, write_to_log)
 * 
 *     def wipe_protracted_parameters(self):             # <<<<<<<<<<<<<<
 *         self.set_logger()
 *         deref(self.c_spec_parameters).wipeProtractedParameters()
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pycoalescence.necsim.necsim.CCommunity.wipe_protracted_parameters", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pycoalescence/necsim/necsim.pyx":233
 *         deref(self.c_spec_parameters).wipeProtractedParameters()
 * 
 *     def add_protracted_parameters(self, double min_speciation_gen, double max_speciation_gen):             # <<<<<<<<<<<<<<
 *         self.set_logger()
 *         deref(self.c_spec_parameters).addProtractedParameters(min_speciation_gen, max_speciation_gen)
 */

/* Python wrapper */
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_10CCommunity_7add_protracted_parameters(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_10CCommunity_7add_protracted_parameters(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  double __pyx_v_min_speciation_gen;
  double __pyx_v_max_speciation_gen;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("add_protracted_parameters (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_min_speciation_gen,&__pyx_n_s_max_speciation_gen,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_min_speciation_gen)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_max_speciation_gen)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("add_protracted_parameters", 1, 2, 2, 1); __PYX_ERR(0, 233, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "add_protracted_parameters") < 0)) __PYX_ERR(0, 233, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_min_speciation_gen = __pyx_PyFloat_AsDouble(values[0]); if (unlikely((__pyx_v_min_speciation_gen == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 233, __pyx_L3_error)
    __pyx_v_max_speciation_gen = __pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_max_speciation_gen == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 233, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("add_protracted_parameters", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 233, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pycoalescence.necsim.necsim.CCommunity.add_protracted_parameters", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_13pycoalescence_6necsim_6necsim_10CCommunity_6add_protracted_parameters(((struct __pyx_obj_13pycoalescence_6necsim_6necsim_CCommunity *)__pyx_v_self), __pyx_v_min_speciation_gen, __pyx_v_max_speciation_gen);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_10CCommunity_6add_protracted_parameters(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CCommunity *__pyx_v_self, double __pyx_v_min_speciation_gen, double __pyx_v_max_speciation_gen) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("add_protracted_parameters", 0);

  /* "pycoalescence/necsim/necsim.pyx":234
 * 
 *     def add_protracted_parameters(self, double min_speciation_gen, double max_speciation_gen):
 *         self.set_logger()             # <<<<<<<<<<<<<<
 *         deref(self.c_spec_parameters).addProtractedParameters(min_speciation_gen, max_speciation_gen)
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_set_logger); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 234, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 234, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pycoalescence/necsim/necsim.pyx":235
 *     def add_protracted_parameters(self, double min_speciation_gen, double max_speciation_gen):
 *         self.set_logger()
 *         deref(self.c_spec_parameters).addProtractedParameters(min_speciation_gen, max_speciation_gen)             # <<<<<<<<<<<<<<
 * 
 *     def add_metacommunity_parameters(self, const unsigned long & metacommunity_size,
 */
  try {
    (*__pyx_v_self->c_spec_parameters).addProtractedParameters(__pyx_v_min_speciation_gen, __pyx_v_max_speciation_gen);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 235, __pyx_L1_error)
  }

  /* "pycoalescence/necsim/necsim.pyx":233
 *         deref(self.c_spec_parameters).wipeProtractedParameters()
 * 
 *     def add_protracted_parameters(self, double min_speciation_gen, double max_speciation_gen):             # <<<<<<<<<<<<<<
 *         self.set_logger()
 *         deref(self.c_spec_parameters).addProtractedParameters(min_speciation_gen, max_speciation_gen)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pycoalescence.necsim.necsim.CCommunity.add_protracted_parameters", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pycoalescence/necsim/necsim.pyx":237
 *         deref(self.c_spec_parameters).addProtractedParameters(min_speciation_gen, max_speciation_gen)
 * 
 *     def add_metacommunity_parameters(self, const unsigned long & metacommunity_size,             # <<<<<<<<<<<<<<
 *                                      const double & speciation_rate,
 *                                      const string & metacommunity_option,
 */

/* Python wrapper */
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_10CCommunity_9add_metacommunity_parameters(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_10CCommunity_9add_metacommunity_parameters(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  unsigned long __pyx_v_metacommunity_size;
  double __pyx_v_speciation_rate;
  std::string __pyx_v_metacommunity_option;
  unsigned long __pyx_v_metacommunity_reference;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("add_metacommunity_parameters (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_metacommunity_size,&__pyx_n_s_speciation_rate,&__pyx_n_s_metacommunity_option,&__pyx_n_s_metacommunity_reference,0};
    PyObject* values[4] = {0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_metacommunity_size)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_speciation_rate)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("add_metacommunity_parameters", 1, 4, 4, 1); __PYX_ERR(0, 237, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_metacommunity_option)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("add_metacommunity_parameters", 1, 4, 4, 2); __PYX_ERR(0, 237, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_metacommunity_reference)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("add_metacommunity_parameters", 1, 4, 4, 3); __PYX_ERR(0, 237, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "add_metacommunity_parameters") < 0)) __PYX_ERR(0, 237, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 4) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
    }
    __pyx_v_metacommunity_size = __Pyx_PyInt_As_unsigned_long(values[0]); if (unlikely((__pyx_v_metacommunity_size == (unsigned long)-1) && PyErr_Occurred())) __PYX_ERR(0, 237, __pyx_L3_error)
    __pyx_v_speciation_rate = __pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_speciation_rate == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 238, __pyx_L3_error)
    __pyx_v_metacommunity_option = __pyx_convert_string_from_py_std__in_string(values[2]); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 239, __pyx_L3_error)
    __pyx_v_metacommunity_reference = __Pyx_PyInt_As_unsigned_long(values[3]); if (unlikely((__pyx_v_metacommunity_reference == (unsigned long)-1) && PyErr_Occurred())) __PYX_ERR(0, 240, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("add_metacommunity_parameters", 1, 4, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 237, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pycoalescence.necsim.necsim.CCommunity.add_metacommunity_parameters", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_13pycoalescence_6necsim_6necsim_10CCommunity_8add_metacommunity_parameters(((struct __pyx_obj_13pycoalescence_6necsim_6necsim_CCommunity *)__pyx_v_self), __pyx_v_metacommunity_size, __pyx_v_speciation_rate, __pyx_v_metacommunity_option, __pyx_v_metacommunity_reference);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_10CCommunity_8add_metacommunity_parameters(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CCommunity *__pyx_v_self, unsigned long __pyx_v_metacommunity_size, double __pyx_v_speciation_rate, std::string __pyx_v_metacommunity_option, unsigned long __pyx_v_metacommunity_reference) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("add_metacommunity_parameters", 0);

  /* "pycoalescence/necsim/necsim.pyx":241
 *                                      const string & metacommunity_option,
 *                                      const unsigned long & metacommunity_reference):
 *         self.set_logger()             # <<<<<<<<<<<<<<
 *         deref(self.c_spec_parameters).addMetacommunityParameters(metacommunity_size, speciation_rate,
 *                                                                  metacommunity_option,
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_set_logger); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 241, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 241, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pycoalescence/necsim/necsim.pyx":242
 *                                      const unsigned long & metacommunity_reference):
 *         self.set_logger()
 *         deref(self.c_spec_parameters).addMetacommunityParameters(metacommunity_size, speciation_rate,             # <<<<<<<<<<<<<<
 *                                                                  metacommunity_option,
 *                                                                  metacommunity_reference)
 */
  try {
    (*__pyx_v_self->c_spec_parameters).addMetacommunityParameters(__pyx_v_metacommunity_size, __pyx_v_speciation_rate, __pyx_v_metacommunity_option, __pyx_v_metacommunity_reference);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 242, __pyx_L1_error)
  }

  /* "pycoalescence/necsim/necsim.pyx":237
 *         deref(self.c_spec_parameters).addProtractedParameters(min_speciation_gen, max_speciation_gen)
 * 
 *     def add_metacommunity_parameters(self, const unsigned long & metacommunity_size,             # <<<<<<<<<<<<<<
 *                                      const double & speciation_rate,
 *                                      const string & metacommunity_option,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pycoalescence.necsim.necsim.CCommunity.add_metacommunity_parameters", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pycoalescence/necsim/necsim.pyx":246
 *                                                                  metacommunity_reference)
 * 
 *     def reset(self):             # <<<<<<<<<<<<<<
 *         self.set_logger()
 *         deref(self.c_spec_parameters).wipe()
 */

/* Python wrapper */
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_10CCommunity_11reset(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_10CCommunity_11reset(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("reset (wrapper)", 0);
  __pyx_r = __pyx_pf_13pycoalescence_6necsim_6necsim_10CCommunity_10reset(((struct __pyx_obj_13pycoalescence_6necsim_6necsim_CCommunity *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_10CCommunity_10reset(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CCommunity *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("reset", 0);

  /* "pycoalescence/necsim/necsim.pyx":247
 * 
 *     def reset(self):
 *         self.set_logger()             # <<<<<<<<<<<<<<
 *         deref(self.c_spec_parameters).wipe()
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_set_logger); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 247, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 247, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pycoalescence/necsim/necsim.pyx":248
 *     def reset(self):
 *         self.set_logger()
 *         deref(self.c_spec_parameters).wipe()             # <<<<<<<<<<<<<<
 * 
 *     def setup(self,
 */
  try {
    (*__pyx_v_self->c_spec_parameters).wipe();
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 248, __pyx_L1_error)
  }

  /* "pycoalescence/necsim/necsim.pyx":246
 *                                                                  metacommunity_reference)
 * 
 *     def reset(self):             # <<<<<<<<<<<<<<
 *         self.set_logger()
 *         deref(self.c_spec_parameters).wipe()
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pycoalescence.necsim.necsim.CCommunity.reset", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pycoalescence/necsim/necsim.pyx":250
 *         deref(self.c_spec_parameters).wipe()
 * 
 *     def setup(self,             # <<<<<<<<<<<<<<
 *               database,
 *               bool use_spatial,
 */

/* Python wrapper */
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_10CCommunity_13setup(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_10CCommunity_13setup(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_database = 0;
  bool __pyx_v_use_spatial;
  PyObject *__pyx_v_sample_file = 0;
  PyObject *__pyx_v_record_fragments = 0;
  std::vector<double>  __pyx_v_speciation_rates;
  std::vector<double>  __pyx_v_times;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("setup (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_database,&__pyx_n_s_use_spatial,&__pyx_n_s_sample_file,&__pyx_n_s_record_fragments,&__pyx_n_s_speciation_rates,&__pyx_n_s_times,0};
    PyObject* values[6] = {0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_database)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_use_spatial)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setup", 1, 6, 6, 1); __PYX_ERR(0, 250, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_sample_file)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setup", 1, 6, 6, 2); __PYX_ERR(0, 250, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_record_fragments)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setup", 1, 6, 6, 3); __PYX_ERR(0, 250, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_speciation_rates)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setup", 1, 6, 6, 4); __PYX_ERR(0, 250, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_times)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setup", 1, 6, 6, 5); __PYX_ERR(0, 250, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "setup") < 0)) __PYX_ERR(0, 250, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 6) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
    }
    __pyx_v_database = values[0];
    __pyx_v_use_spatial = __Pyx_PyObject_IsTrue(values[1]); if (unlikely((__pyx_v_use_spatial == ((bool)-1)) && PyErr_Occurred())) __PYX_ERR(0, 252, __pyx_L3_error)
    __pyx_v_sample_file = values[2];
    __pyx_v_record_fragments = values[3];
    __pyx_v_speciation_rates = __pyx_convert_vector_from_py_double(values[4]); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 255, __pyx_L3_error)
    __pyx_v_times = __pyx_convert_vector_from_py_double(values[5]); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 256, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("setup", 1, 6, 6, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 250, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pycoalescence.necsim.necsim.CCommunity.setup", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_13pycoalescence_6necsim_6necsim_10CCommunity_12setup(((struct __pyx_obj_13pycoalescence_6necsim_6necsim_CCommunity *)__pyx_v_self), __pyx_v_database, __pyx_v_use_spatial, __pyx_v_sample_file, __pyx_v_record_fragments, __pyx_v_speciation_rates, __pyx_v_times);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_10CCommunity_12setup(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CCommunity *__pyx_v_self, PyObject *__pyx_v_database, bool __pyx_v_use_spatial, PyObject *__pyx_v_sample_file, PyObject *__pyx_v_record_fragments, std::vector<double>  __pyx_v_speciation_rates, std::vector<double>  __pyx_v_times) {
  std::string __pyx_v_database_str;
  std::string __pyx_v_sample_file_str;
  std::string __pyx_v_fragment_file_str;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  std::string __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("setup", 0);

  /* "pycoalescence/necsim/necsim.pyx":257
 *               vector[double] speciation_rates,
 *               vector[double] times):
 *         self.set_logger()             # <<<<<<<<<<<<<<
 *         cdef string database_str = database
 *         cdef string sample_file_str = sample_file
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_set_logger); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 257, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 257, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pycoalescence/necsim/necsim.pyx":258
 *               vector[double] times):
 *         self.set_logger()
 *         cdef string database_str = database             # <<<<<<<<<<<<<<
 *         cdef string sample_file_str = sample_file
 *         cdef string fragment_file_str = record_fragments
 */
  __pyx_t_4 = __pyx_convert_string_from_py_std__in_string(__pyx_v_database); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 258, __pyx_L1_error)
  __pyx_v_database_str = __pyx_t_4;

  /* "pycoalescence/necsim/necsim.pyx":259
 *         self.set_logger()
 *         cdef string database_str = database
 *         cdef string sample_file_str = sample_file             # <<<<<<<<<<<<<<
 *         cdef string fragment_file_str = record_fragments
 *         deref(self.c_spec_parameters).setup(database_str, use_spatial, sample_file_str, times, fragment_file_str,
 */
  __pyx_t_4 = __pyx_convert_string_from_py_std__in_string(__pyx_v_sample_file); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 259, __pyx_L1_error)
  __pyx_v_sample_file_str = __pyx_t_4;

  /* "pycoalescence/necsim/necsim.pyx":260
 *         cdef string database_str = database
 *         cdef string sample_file_str = sample_file
 *         cdef string fragment_file_str = record_fragments             # <<<<<<<<<<<<<<
 *         deref(self.c_spec_parameters).setup(database_str, use_spatial, sample_file_str, times, fragment_file_str,
 *                                             speciation_rates)
 */
  __pyx_t_4 = __pyx_convert_string_from_py_std__in_string(__pyx_v_record_fragments); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 260, __pyx_L1_error)
  __pyx_v_fragment_file_str = __pyx_t_4;

  /* "pycoalescence/necsim/necsim.pyx":261
 *         cdef string sample_file_str = sample_file
 *         cdef string fragment_file_str = record_fragments
 *         deref(self.c_spec_parameters).setup(database_str, use_spatial, sample_file_str, times, fragment_file_str,             # <<<<<<<<<<<<<<
 *                                             speciation_rates)
 * 
 */
  try {
    (*__pyx_v_self->c_spec_parameters).setup(__pyx_v_database_str, __pyx_v_use_spatial, __pyx_v_sample_file_str, __pyx_v_times, __pyx_v_fragment_file_str, __pyx_v_speciation_rates);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 261, __pyx_L1_error)
  }

  /* "pycoalescence/necsim/necsim.pyx":250
 *         deref(self.c_spec_parameters).wipe()
 * 
 *     def setup(self,             # <<<<<<<<<<<<<<
 *               database,
 *               bool use_spatial,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pycoalescence.necsim.necsim.CCommunity.setup", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pycoalescence/necsim/necsim.pyx":264
 *                                             speciation_rates)
 * 
 *     def add_time(self, time):             # <<<<<<<<<<<<<<
 *         self.set_logger()
 *         deref(self.c_spec_parameters).addTime(time)
 */

/* Python wrapper */
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_10CCommunity_15add_time(PyObject *__pyx_v_self, PyObject *__pyx_v_time); /*proto*/
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_10CCommunity_15add_time(PyObject *__pyx_v_self, PyObject *__pyx_v_time) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("add_time (wrapper)", 0);
  __pyx_r = __pyx_pf_13pycoalescence_6necsim_6necsim_10CCommunity_14add_time(((struct __pyx_obj_13pycoalescence_6necsim_6necsim_CCommunity *)__pyx_v_self), ((PyObject *)__pyx_v_time));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_10CCommunity_14add_time(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CCommunity *__pyx_v_self, PyObject *__pyx_v_time) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  double __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("add_time", 0);

  /* "pycoalescence/necsim/necsim.pyx":265
 * 
 *     def add_time(self, time):
 *         self.set_logger()             # <<<<<<<<<<<<<<
 *         deref(self.c_spec_parameters).addTime(time)
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_set_logger); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 265, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 265, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pycoalescence/necsim/necsim.pyx":266
 *     def add_time(self, time):
 *         self.set_logger()
 *         deref(self.c_spec_parameters).addTime(time)             # <<<<<<<<<<<<<<
 * 
 *     def apply(self):
 */
  __pyx_t_4 = __pyx_PyFloat_AsDouble(__pyx_v_time); if (unlikely((__pyx_t_4 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 266, __pyx_L1_error)
  try {
    (*__pyx_v_self->c_spec_parameters).addTime(__pyx_t_4);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 266, __pyx_L1_error)
  }

  /* "pycoalescence/necsim/necsim.pyx":264
 *                                             speciation_rates)
 * 
 *     def add_time(self, time):             # <<<<<<<<<<<<<<
 *         self.set_logger()
 *         deref(self.c_spec_parameters).addTime(time)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pycoalescence.necsim.necsim.CCommunity.add_time", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pycoalescence/necsim/necsim.pyx":268
 *         deref(self.c_spec_parameters).addTime(time)
 * 
 *     def apply(self):             # <<<<<<<<<<<<<<
 *         self.set_logger()
 *         self.c_community.applyNoOutput(self.c_spec_parameters)
 */

/* Python wrapper */
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_10CCommunity_17apply(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_10CCommunity_17apply(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("apply (wrapper)", 0);
  __pyx_r = __pyx_pf_13pycoalescence_6necsim_6necsim_10CCommunity_16apply(((struct __pyx_obj_13pycoalescence_6necsim_6necsim_CCommunity *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_10CCommunity_16apply(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CCommunity *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("apply", 0);

  /* "pycoalescence/necsim/necsim.pyx":269
 * 
 *     def apply(self):
 *         self.set_logger()             # <<<<<<<<<<<<<<
 *         self.c_community.applyNoOutput(self.c_spec_parameters)
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_set_logger); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 269, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 269, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pycoalescence/necsim/necsim.pyx":270
 *     def apply(self):
 *         self.set_logger()
 *         self.c_community.applyNoOutput(self.c_spec_parameters)             # <<<<<<<<<<<<<<
 * 
 *     def output(self):
 */
  try {
    __pyx_v_self->c_community.applyNoOutput(__pyx_v_self->c_spec_parameters);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 270, __pyx_L1_error)
  }

  /* "pycoalescence/necsim/necsim.pyx":268
 *         deref(self.c_spec_parameters).addTime(time)
 * 
 *     def apply(self):             # <<<<<<<<<<<<<<
 *         self.set_logger()
 *         self.c_community.applyNoOutput(self.c_spec_parameters)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pycoalescence.necsim.necsim.CCommunity.apply", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pycoalescence/necsim/necsim.pyx":272
 *         self.c_community.applyNoOutput(self.c_spec_parameters)
 * 
 *     def output(self):             # <<<<<<<<<<<<<<
 *         self.set_logger()
 *         self.c_community.output()
 */

/* Python wrapper */
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_10CCommunity_19output(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_10CCommunity_19output(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("output (wrapper)", 0);
  __pyx_r = __pyx_pf_13pycoalescence_6necsim_6necsim_10CCommunity_18output(((struct __pyx_obj_13pycoalescence_6necsim_6necsim_CCommunity *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_10CCommunity_18output(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CCommunity *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("output", 0);

  /* "pycoalescence/necsim/necsim.pyx":273
 * 
 *     def output(self):
 *         self.set_logger()             # <<<<<<<<<<<<<<
 *         self.c_community.output()
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_set_logger); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 273, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 273, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pycoalescence/necsim/necsim.pyx":274
 *     def output(self):
 *         self.set_logger()
 *         self.c_community.output()             # <<<<<<<<<<<<<<
 * 
 *     def speciate_remaining_lineages(self, file):
 */
  try {
    __pyx_v_self->c_community.output();
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 274, __pyx_L1_error)
  }

  /* "pycoalescence/necsim/necsim.pyx":272
 *         self.c_community.applyNoOutput(self.c_spec_parameters)
 * 
 *     def output(self):             # <<<<<<<<<<<<<<
 *         self.set_logger()
 *         self.c_community.output()
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pycoalescence.necsim.necsim.CCommunity.output", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pycoalescence/necsim/necsim.pyx":276
 *         self.c_community.output()
 * 
 *     def speciate_remaining_lineages(self, file):             # <<<<<<<<<<<<<<
 *         self.set_logger()
 *         cdef string file_str = file
 */

/* Python wrapper */
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_10CCommunity_21speciate_remaining_lineages(PyObject *__pyx_v_self, PyObject *__pyx_v_file); /*proto*/
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_10CCommunity_21speciate_remaining_lineages(PyObject *__pyx_v_self, PyObject *__pyx_v_file) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("speciate_remaining_lineages (wrapper)", 0);
  __pyx_r = __pyx_pf_13pycoalescence_6necsim_6necsim_10CCommunity_20speciate_remaining_lineages(((struct __pyx_obj_13pycoalescence_6necsim_6necsim_CCommunity *)__pyx_v_self), ((PyObject *)__pyx_v_file));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_10CCommunity_20speciate_remaining_lineages(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CCommunity *__pyx_v_self, PyObject *__pyx_v_file) {
  std::string __pyx_v_file_str;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  std::string __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("speciate_remaining_lineages", 0);

  /* "pycoalescence/necsim/necsim.pyx":277
 * 
 *     def speciate_remaining_lineages(self, file):
 *         self.set_logger()             # <<<<<<<<<<<<<<
 *         cdef string file_str = file
 *         self.c_community.speciateRemainingLineages(file_str)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_set_logger); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 277, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 277, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pycoalescence/necsim/necsim.pyx":278
 *     def speciate_remaining_lineages(self, file):
 *         self.set_logger()
 *         cdef string file_str = file             # <<<<<<<<<<<<<<
 *         self.c_community.speciateRemainingLineages(file_str)
 * 
 */
  __pyx_t_4 = __pyx_convert_string_from_py_std__in_string(__pyx_v_file); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 278, __pyx_L1_error)
  __pyx_v_file_str = __pyx_t_4;

  /* "pycoalescence/necsim/necsim.pyx":279
 *         self.set_logger()
 *         cdef string file_str = file
 *         self.c_community.speciateRemainingLineages(file_str)             # <<<<<<<<<<<<<<
 * 
 * cdef class CMetacommunity:
 */
  try {
    __pyx_v_self->c_community.speciateRemainingLineages(__pyx_v_file_str);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 279, __pyx_L1_error)
  }

  /* "pycoalescence/necsim/necsim.pyx":276
 *         self.c_community.output()
 * 
 *     def speciate_remaining_lineages(self, file):             # <<<<<<<<<<<<<<
 *         self.set_logger()
 *         cdef string file_str = file
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pycoalescence.necsim.necsim.CCommunity.speciate_remaining_lineages", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_10CCommunity_23__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_10CCommunity_23__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_13pycoalescence_6necsim_6necsim_10CCommunity_22__reduce_cython__(((struct __pyx_obj_13pycoalescence_6necsim_6necsim_CCommunity *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_10CCommunity_22__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_13pycoalescence_6necsim_6necsim_CCommunity *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__10, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pycoalescence.necsim.necsim.CCommunity.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_10CCommunity_25__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_10CCommunity_25__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_13pycoalescence_6necsim_6necsim_10CCommunity_24__setstate_cython__(((struct __pyx_obj_13pycoalescence_6necsim_6necsim_CCommunity *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_10CCommunity_24__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_13pycoalescence_6necsim_6necsim_CCommunity *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__11, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pycoalescence.necsim.necsim.CCommunity.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pycoalescence/necsim/necsim.pyx":286
 *     cdef object logger
 * 
 *     def __cinit__(self, object logger):             # <<<<<<<<<<<<<<
 *         self.logger = logger
 *         self.c_community = Metacommunity()
 */

/* Python wrapper */
static int __pyx_pw_13pycoalescence_6necsim_6necsim_14CMetacommunity_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_13pycoalescence_6necsim_6necsim_14CMetacommunity_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_logger = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_logger,0};
    PyObject* values[1] = {0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_logger)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__cinit__") < 0)) __PYX_ERR(0, 286, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 1) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
    }
    __pyx_v_logger = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 1, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 286, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pycoalescence.necsim.necsim.CMetacommunity.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_13pycoalescence_6necsim_6necsim_14CMetacommunity___cinit__(((struct __pyx_obj_13pycoalescence_6necsim_6necsim_CMetacommunity *)__pyx_v_self), __pyx_v_logger);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_13pycoalescence_6necsim_6necsim_14CMetacommunity___cinit__(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CMetacommunity *__pyx_v_self, PyObject *__pyx_v_logger) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  necsim::Metacommunity __pyx_t_1;
  std::shared_ptr<struct necsim::SpecSimParameters>  __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "pycoalescence/necsim/necsim.pyx":287
 * 
 *     def __cinit__(self, object logger):
 *         self.logger = logger             # <<<<<<<<<<<<<<
 *         self.c_community = Metacommunity()
 *         self.c_spec_parameters = make_shared[SpecSimParameters]()
 */
  __Pyx_INCREF(__pyx_v_logger);
  __Pyx_GIVEREF(__pyx_v_logger);
  __Pyx_GOTREF(__pyx_v_self->logger);
  __Pyx_DECREF(__pyx_v_self->logger);
  __pyx_v_self->logger = __pyx_v_logger;

  /* "pycoalescence/necsim/necsim.pyx":288
 *     def __cinit__(self, object logger):
 *         self.logger = logger
 *         self.c_community = Metacommunity()             # <<<<<<<<<<<<<<
 *         self.c_spec_parameters = make_shared[SpecSimParameters]()
 * 
 */
  try {
    __pyx_t_1 = necsim::Metacommunity();
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 288, __pyx_L1_error)
  }
  __pyx_v_self->c_community = __pyx_t_1;

  /* "pycoalescence/necsim/necsim.pyx":289
 *         self.logger = logger
 *         self.c_community = Metacommunity()
 *         self.c_spec_parameters = make_shared[SpecSimParameters]()             # <<<<<<<<<<<<<<
 * 
 *     def set_logger(self):
 */
  try {
    __pyx_t_2 = std::make_shared<struct necsim::SpecSimParameters>();
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 289, __pyx_L1_error)
  }
  __pyx_v_self->c_spec_parameters = __pyx_t_2;

  /* "pycoalescence/necsim/necsim.pyx":286
 *     cdef object logger
 * 
 *     def __cinit__(self, object logger):             # <<<<<<<<<<<<<<
 *         self.logger = logger
 *         self.c_community = Metacommunity()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pycoalescence.necsim.necsim.CMetacommunity.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pycoalescence/necsim/necsim.pyx":291
 *         self.c_spec_parameters = make_shared[SpecSimParameters]()
 * 
 *     def set_logger(self):             # <<<<<<<<<<<<<<
 *         setGlobalLogger(self.logger, write_to_log)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_14CMetacommunity_3set_logger(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_14CMetacommunity_3set_logger(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_logger (wrapper)", 0);
  __pyx_r = __pyx_pf_13pycoalescence_6necsim_6necsim_14CMetacommunity_2set_logger(((struct __pyx_obj_13pycoalescence_6necsim_6necsim_CMetacommunity *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_14CMetacommunity_2set_logger(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CMetacommunity *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("set_logger", 0);

  /* "pycoalescence/necsim/necsim.pyx":292
 * 
 *     def set_logger(self):
 *         setGlobalLogger(self.logger, write_to_log)             # <<<<<<<<<<<<<<
 * 
 *     def wipe_protracted_parameters(self):
 */
  __pyx_t_1 = __pyx_v_self->logger;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_write_to_log); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 292, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  try {
    necsim::setGlobalLogger(__pyx_t_1, __pyx_t_2);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 292, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pycoalescence/necsim/necsim.pyx":291
 *         self.c_spec_parameters = make_shared[SpecSimParameters]()
 * 
 *     def set_logger(self):             # <<<<<<<<<<<<<<
 *         setGlobalLogger(self.logger, write_to_log)
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pycoalescence.necsim.necsim.CMetacommunity.set_logger", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pycoalescence/necsim/necsim.pyx":294
 *         setGlobalLogger(self.logger, write_to_log)
 * 
 *     def wipe_protracted_parameters(self):             # <<<<<<<<<<<<<<
 *         self.set_logger()
 *         deref(self.c_spec_parameters).wipeProtractedParameters()
 */

/* Python wrapper */
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_14CMetacommunity_5wipe_protracted_parameters(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_14CMetacommunity_5wipe_protracted_parameters(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("wipe_protracted_parameters (wrapper)", 0);
  __pyx_r = __pyx_pf_13pycoalescence_6necsim_6necsim_14CMetacommunity_4wipe_protracted_parameters(((struct __pyx_obj_13pycoalescence_6necsim_6necsim_CMetacommunity *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_14CMetacommunity_4wipe_protracted_parameters(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CMetacommunity *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("wipe_protracted_parameters", 0);

  /* "pycoalescence/necsim/necsim.pyx":295
 * 
 *     def wipe_protracted_parameters(self):
 *         self.set_logger()             # <<<<<<<<<<<<<<
 *         deref(self.c_spec_parameters).wipeProtractedParameters()
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_set_logger); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 295, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 295, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pycoalescence/necsim/necsim.pyx":296
 *     def wipe_protracted_parameters(self):
 *         self.set_logger()
 *         deref(self.c_spec_parameters).wipeProtractedParameters()             # <<<<<<<<<<<<<<
 * 
 *     def add_protracted_parameters(self, double min_speciation_gen, double max_speciation_gen):
 */
  try {
    (*__pyx_v_self->c_spec_parameters).wipeProtractedParameters();
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 296, __pyx_L1_error)
  }

  /* "pycoalescence/necsim/necsim.pyx":294
 *         setGlobalLogger(self.logger, write_to_log)
 * 
 *     def wipe_protracted_parameters(self):             # <<<<<<<<<<<<<<
 *         self.set_logger()
 *         deref(self.c_spec_parameters).wipeProtractedParameters()
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pycoalescence.necsim.necsim.CMetacommunity.wipe_protracted_parameters", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pycoalescence/necsim/necsim.pyx":298
 *         deref(self.c_spec_parameters).wipeProtractedParameters()
 * 
 *     def add_protracted_parameters(self, double min_speciation_gen, double max_speciation_gen):             # <<<<<<<<<<<<<<
 *         self.set_logger()
 *         deref(self.c_spec_parameters).addProtractedParameters(min_speciation_gen, max_speciation_gen)
 */

/* Python wrapper */
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_14CMetacommunity_7add_protracted_parameters(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_14CMetacommunity_7add_protracted_parameters(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  double __pyx_v_min_speciation_gen;
  double __pyx_v_max_speciation_gen;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("add_protracted_parameters (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_min_speciation_gen,&__pyx_n_s_max_speciation_gen,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_min_speciation_gen)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_max_speciation_gen)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("add_protracted_parameters", 1, 2, 2, 1); __PYX_ERR(0, 298, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "add_protracted_parameters") < 0)) __PYX_ERR(0, 298, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_min_speciation_gen = __pyx_PyFloat_AsDouble(values[0]); if (unlikely((__pyx_v_min_speciation_gen == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 298, __pyx_L3_error)
    __pyx_v_max_speciation_gen = __pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_max_speciation_gen == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 298, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("add_protracted_parameters", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 298, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pycoalescence.necsim.necsim.CMetacommunity.add_protracted_parameters", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_13pycoalescence_6necsim_6necsim_14CMetacommunity_6add_protracted_parameters(((struct __pyx_obj_13pycoalescence_6necsim_6necsim_CMetacommunity *)__pyx_v_self), __pyx_v_min_speciation_gen, __pyx_v_max_speciation_gen);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_14CMetacommunity_6add_protracted_parameters(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CMetacommunity *__pyx_v_self, double __pyx_v_min_speciation_gen, double __pyx_v_max_speciation_gen) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("add_protracted_parameters", 0);

  /* "pycoalescence/necsim/necsim.pyx":299
 * 
 *     def add_protracted_parameters(self, double min_speciation_gen, double max_speciation_gen):
 *         self.set_logger()             # <<<<<<<<<<<<<<
 *         deref(self.c_spec_parameters).addProtractedParameters(min_speciation_gen, max_speciation_gen)
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_set_logger); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 299, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 299, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pycoalescence/necsim/necsim.pyx":300
 *     def add_protracted_parameters(self, double min_speciation_gen, double max_speciation_gen):
 *         self.set_logger()
 *         deref(self.c_spec_parameters).addProtractedParameters(min_speciation_gen, max_speciation_gen)             # <<<<<<<<<<<<<<
 * 
 *     def add_metacommunity_parameters(self, const unsigned long & metacommunity_size,
 */
  try {
    (*__pyx_v_self->c_spec_parameters).addProtractedParameters(__pyx_v_min_speciation_gen, __pyx_v_max_speciation_gen);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 300, __pyx_L1_error)
  }

  /* "pycoalescence/necsim/necsim.pyx":298
 *         deref(self.c_spec_parameters).wipeProtractedParameters()
 * 
 *     def add_protracted_parameters(self, double min_speciation_gen, double max_speciation_gen):             # <<<<<<<<<<<<<<
 *         self.set_logger()
 *         deref(self.c_spec_parameters).addProtractedParameters(min_speciation_gen, max_speciation_gen)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pycoalescence.necsim.necsim.CMetacommunity.add_protracted_parameters", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pycoalescence/necsim/necsim.pyx":302
 *         deref(self.c_spec_parameters).addProtractedParameters(min_speciation_gen, max_speciation_gen)
 * 
 *     def add_metacommunity_parameters(self, const unsigned long & metacommunity_size,             # <<<<<<<<<<<<<<
 *                                      const double & speciation_rate,
 *                                      const string & metacommunity_option,
 */

/* Python wrapper */
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_14CMetacommunity_9add_metacommunity_parameters(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_14CMetacommunity_9add_metacommunity_parameters(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  unsigned long __pyx_v_metacommunity_size;
  double __pyx_v_speciation_rate;
  std::string __pyx_v_metacommunity_option;
  unsigned long __pyx_v_metacommunity_reference;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("add_metacommunity_parameters (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_metacommunity_size,&__pyx_n_s_speciation_rate,&__pyx_n_s_metacommunity_option,&__pyx_n_s_metacommunity_reference,0};
    PyObject* values[4] = {0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_metacommunity_size)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_speciation_rate)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("add_metacommunity_parameters", 1, 4, 4, 1); __PYX_ERR(0, 302, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_metacommunity_option)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("add_metacommunity_parameters", 1, 4, 4, 2); __PYX_ERR(0, 302, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_metacommunity_reference)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("add_metacommunity_parameters", 1, 4, 4, 3); __PYX_ERR(0, 302, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "add_metacommunity_parameters") < 0)) __PYX_ERR(0, 302, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 4) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
    }
    __pyx_v_metacommunity_size = __Pyx_PyInt_As_unsigned_long(values[0]); if (unlikely((__pyx_v_metacommunity_size == (unsigned long)-1) && PyErr_Occurred())) __PYX_ERR(0, 302, __pyx_L3_error)
    __pyx_v_speciation_rate = __pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_speciation_rate == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 303, __pyx_L3_error)
    __pyx_v_metacommunity_option = __pyx_convert_string_from_py_std__in_string(values[2]); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 304, __pyx_L3_error)
    __pyx_v_metacommunity_reference = __Pyx_PyInt_As_unsigned_long(values[3]); if (unlikely((__pyx_v_metacommunity_reference == (unsigned long)-1) && PyErr_Occurred())) __PYX_ERR(0, 305, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("add_metacommunity_parameters", 1, 4, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 302, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pycoalescence.necsim.necsim.CMetacommunity.add_metacommunity_parameters", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_13pycoalescence_6necsim_6necsim_14CMetacommunity_8add_metacommunity_parameters(((struct __pyx_obj_13pycoalescence_6necsim_6necsim_CMetacommunity *)__pyx_v_self), __pyx_v_metacommunity_size, __pyx_v_speciation_rate, __pyx_v_metacommunity_option, __pyx_v_metacommunity_reference);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_14CMetacommunity_8add_metacommunity_parameters(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CMetacommunity *__pyx_v_self, unsigned long __pyx_v_metacommunity_size, double __pyx_v_speciation_rate, std::string __pyx_v_metacommunity_option, unsigned long __pyx_v_metacommunity_reference) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("add_metacommunity_parameters", 0);

  /* "pycoalescence/necsim/necsim.pyx":306
 *                                      const string & metacommunity_option,
 *                                      const unsigned long & metacommunity_reference):
 *         self.set_logger()             # <<<<<<<<<<<<<<
 *         deref(self.c_spec_parameters).addMetacommunityParameters(metacommunity_size, speciation_rate,
 *                                                                  metacommunity_option,
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_set_logger); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 306, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 306, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pycoalescence/necsim/necsim.pyx":307
 *                                      const unsigned long & metacommunity_reference):
 *         self.set_logger()
 *         deref(self.c_spec_parameters).addMetacommunityParameters(metacommunity_size, speciation_rate,             # <<<<<<<<<<<<<<
 *                                                                  metacommunity_option,
 *                                                                  metacommunity_reference)
 */
  try {
    (*__pyx_v_self->c_spec_parameters).addMetacommunityParameters(__pyx_v_metacommunity_size, __pyx_v_speciation_rate, __pyx_v_metacommunity_option, __pyx_v_metacommunity_reference);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 307, __pyx_L1_error)
  }

  /* "pycoalescence/necsim/necsim.pyx":302
 *         deref(self.c_spec_parameters).addProtractedParameters(min_speciation_gen, max_speciation_gen)
 * 
 *     def add_metacommunity_parameters(self, const unsigned long & metacommunity_size,             # <<<<<<<<<<<<<<
 *                                      const double & speciation_rate,
 *                                      const string & metacommunity_option,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pycoalescence.necsim.necsim.CMetacommunity.add_metacommunity_parameters", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pycoalescence/necsim/necsim.pyx":311
 *                                                                  metacommunity_reference)
 * 
 *     def reset(self):             # <<<<<<<<<<<<<<
 *         self.set_logger()
 *         deref(self.c_spec_parameters).wipe()
 */

/* Python wrapper */
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_14CMetacommunity_11reset(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_14CMetacommunity_11reset(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("reset (wrapper)", 0);
  __pyx_r = __pyx_pf_13pycoalescence_6necsim_6necsim_14CMetacommunity_10reset(((struct __pyx_obj_13pycoalescence_6necsim_6necsim_CMetacommunity *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_14CMetacommunity_10reset(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CMetacommunity *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("reset", 0);

  /* "pycoalescence/necsim/necsim.pyx":312
 * 
 *     def reset(self):
 *         self.set_logger()             # <<<<<<<<<<<<<<
 *         deref(self.c_spec_parameters).wipe()
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_set_logger); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 312, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 312, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pycoalescence/necsim/necsim.pyx":313
 *     def reset(self):
 *         self.set_logger()
 *         deref(self.c_spec_parameters).wipe()             # <<<<<<<<<<<<<<
 * 
 *     def setup(self,
 */
  try {
    (*__pyx_v_self->c_spec_parameters).wipe();
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 313, __pyx_L1_error)
  }

  /* "pycoalescence/necsim/necsim.pyx":311
 *                                                                  metacommunity_reference)
 * 
 *     def reset(self):             # <<<<<<<<<<<<<<
 *         self.set_logger()
 *         deref(self.c_spec_parameters).wipe()
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pycoalescence.necsim.necsim.CMetacommunity.reset", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pycoalescence/necsim/necsim.pyx":315
 *         deref(self.c_spec_parameters).wipe()
 * 
 *     def setup(self,             # <<<<<<<<<<<<<<
 *               database,
 *               bool use_spatial,
 */

/* Python wrapper */
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_14CMetacommunity_13setup(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_14CMetacommunity_13setup(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_database = 0;
  bool __pyx_v_use_spatial;
  PyObject *__pyx_v_sample_file = 0;
  PyObject *__pyx_v_record_fragments = 0;
  std::vector<double>  __pyx_v_speciation_rates;
  std::vector<double>  __pyx_v_times;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("setup (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_database,&__pyx_n_s_use_spatial,&__pyx_n_s_sample_file,&__pyx_n_s_record_fragments,&__pyx_n_s_speciation_rates,&__pyx_n_s_times,0};
    PyObject* values[6] = {0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_database)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_use_spatial)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setup", 1, 6, 6, 1); __PYX_ERR(0, 315, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_sample_file)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setup", 1, 6, 6, 2); __PYX_ERR(0, 315, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_record_fragments)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setup", 1, 6, 6, 3); __PYX_ERR(0, 315, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_speciation_rates)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setup", 1, 6, 6, 4); __PYX_ERR(0, 315, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_times)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setup", 1, 6, 6, 5); __PYX_ERR(0, 315, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "setup") < 0)) __PYX_ERR(0, 315, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 6) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
    }
    __pyx_v_database = values[0];
    __pyx_v_use_spatial = __Pyx_PyObject_IsTrue(values[1]); if (unlikely((__pyx_v_use_spatial == ((bool)-1)) && PyErr_Occurred())) __PYX_ERR(0, 317, __pyx_L3_error)
    __pyx_v_sample_file = values[2];
    __pyx_v_record_fragments = values[3];
    __pyx_v_speciation_rates = __pyx_convert_vector_from_py_double(values[4]); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 320, __pyx_L3_error)
    __pyx_v_times = __pyx_convert_vector_from_py_double(values[5]); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 321, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("setup", 1, 6, 6, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 315, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pycoalescence.necsim.necsim.CMetacommunity.setup", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_13pycoalescence_6necsim_6necsim_14CMetacommunity_12setup(((struct __pyx_obj_13pycoalescence_6necsim_6necsim_CMetacommunity *)__pyx_v_self), __pyx_v_database, __pyx_v_use_spatial, __pyx_v_sample_file, __pyx_v_record_fragments, __pyx_v_speciation_rates, __pyx_v_times);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_14CMetacommunity_12setup(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CMetacommunity *__pyx_v_self, PyObject *__pyx_v_database, bool __pyx_v_use_spatial, PyObject *__pyx_v_sample_file, PyObject *__pyx_v_record_fragments, std::vector<double>  __pyx_v_speciation_rates, std::vector<double>  __pyx_v_times) {
  std::string __pyx_v_database_str;
  std::string __pyx_v_sample_file_str;
  std::string __pyx_v_fragment_file_str;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  std::string __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("setup", 0);

  /* "pycoalescence/necsim/necsim.pyx":322
 *               vector[double] speciation_rates,
 *               vector[double] times):
 *         self.set_logger()             # <<<<<<<<<<<<<<
 *         cdef string database_str = database
 *         cdef string sample_file_str = sample_file
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_set_logger); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 322, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 322, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pycoalescence/necsim/necsim.pyx":323
 *               vector[double] times):
 *         self.set_logger()
 *         cdef string database_str = database             # <<<<<<<<<<<<<<
 *         cdef string sample_file_str = sample_file
 *         cdef string fragment_file_str = record_fragments
 */
  __pyx_t_4 = __pyx_convert_string_from_py_std__in_string(__pyx_v_database); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 323, __pyx_L1_error)
  __pyx_v_database_str = __pyx_t_4;

  /* "pycoalescence/necsim/necsim.pyx":324
 *         self.set_logger()
 *         cdef string database_str = database
 *         cdef string sample_file_str = sample_file             # <<<<<<<<<<<<<<
 *         cdef string fragment_file_str = record_fragments
 *         deref(self.c_spec_parameters).setup(database_str, use_spatial, sample_file_str, times, fragment_file_str,
 */
  __pyx_t_4 = __pyx_convert_string_from_py_std__in_string(__pyx_v_sample_file); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 324, __pyx_L1_error)
  __pyx_v_sample_file_str = __pyx_t_4;

  /* "pycoalescence/necsim/necsim.pyx":325
 *         cdef string database_str = database
 *         cdef string sample_file_str = sample_file
 *         cdef string fragment_file_str = record_fragments             # <<<<<<<<<<<<<<
 *         deref(self.c_spec_parameters).setup(database_str, use_spatial, sample_file_str, times, fragment_file_str,
 *                                             speciation_rates)
 */
  __pyx_t_4 = __pyx_convert_string_from_py_std__in_string(__pyx_v_record_fragments); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 325, __pyx_L1_error)
  __pyx_v_fragment_file_str = __pyx_t_4;

  /* "pycoalescence/necsim/necsim.pyx":326
 *         cdef string sample_file_str = sample_file
 *         cdef string fragment_file_str = record_fragments
 *         deref(self.c_spec_parameters).setup(database_str, use_spatial, sample_file_str, times, fragment_file_str,             # <<<<<<<<<<<<<<
 *                                             speciation_rates)
 * 
 */
  try {
    (*__pyx_v_self->c_spec_parameters).setup(__pyx_v_database_str, __pyx_v_use_spatial, __pyx_v_sample_file_str, __pyx_v_times, __pyx_v_fragment_file_str, __pyx_v_speciation_rates);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 326, __pyx_L1_error)
  }

  /* "pycoalescence/necsim/necsim.pyx":315
 *         deref(self.c_spec_parameters).wipe()
 * 
 *     def setup(self,             # <<<<<<<<<<<<<<
 *               database,
 *               bool use_spatial,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pycoalescence.necsim.necsim.CMetacommunity.setup", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pycoalescence/necsim/necsim.pyx":329
 *                                             speciation_rates)
 * 
 *     def add_time(self, time):             # <<<<<<<<<<<<<<
 *         self.set_logger()
 *         deref(self.c_spec_parameters).addTime(time)
 */

/* Python wrapper */
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_14CMetacommunity_15add_time(PyObject *__pyx_v_self, PyObject *__pyx_v_time); /*proto*/
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_14CMetacommunity_15add_time(PyObject *__pyx_v_self, PyObject *__pyx_v_time) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("add_time (wrapper)", 0);
  __pyx_r = __pyx_pf_13pycoalescence_6necsim_6necsim_14CMetacommunity_14add_time(((struct __pyx_obj_13pycoalescence_6necsim_6necsim_CMetacommunity *)__pyx_v_self), ((PyObject *)__pyx_v_time));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_14CMetacommunity_14add_time(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CMetacommunity *__pyx_v_self, PyObject *__pyx_v_time) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  double __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("add_time", 0);

  /* "pycoalescence/necsim/necsim.pyx":330
 * 
 *     def add_time(self, time):
 *         self.set_logger()             # <<<<<<<<<<<<<<
 *         deref(self.c_spec_parameters).addTime(time)
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_set_logger); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 330, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 330, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pycoalescence/necsim/necsim.pyx":331
 *     def add_time(self, time):
 *         self.set_logger()
 *         deref(self.c_spec_parameters).addTime(time)             # <<<<<<<<<<<<<<
 * 
 *     def apply(self):
 */
  __pyx_t_4 = __pyx_PyFloat_AsDouble(__pyx_v_time); if (unlikely((__pyx_t_4 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 331, __pyx_L1_error)
  try {
    (*__pyx_v_self->c_spec_parameters).addTime(__pyx_t_4);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 331, __pyx_L1_error)
  }

  /* "pycoalescence/necsim/necsim.pyx":329
 *                                             speciation_rates)
 * 
 *     def add_time(self, time):             # <<<<<<<<<<<<<<
 *         self.set_logger()
 *         deref(self.c_spec_parameters).addTime(time)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pycoalescence.necsim.necsim.CMetacommunity.add_time", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pycoalescence/necsim/necsim.pyx":333
 *         deref(self.c_spec_parameters).addTime(time)
 * 
 *     def apply(self):             # <<<<<<<<<<<<<<
 *         self.set_logger()
 *         self.c_community.applyNoOutput(self.c_spec_parameters)
 */

/* Python wrapper */
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_14CMetacommunity_17apply(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_14CMetacommunity_17apply(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("apply (wrapper)", 0);
  __pyx_r = __pyx_pf_13pycoalescence_6necsim_6necsim_14CMetacommunity_16apply(((struct __pyx_obj_13pycoalescence_6necsim_6necsim_CMetacommunity *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_14CMetacommunity_16apply(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CMetacommunity *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("apply", 0);

  /* "pycoalescence/necsim/necsim.pyx":334
 * 
 *     def apply(self):
 *         self.set_logger()             # <<<<<<<<<<<<<<
 *         self.c_community.applyNoOutput(self.c_spec_parameters)
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_set_logger); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 334, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 334, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pycoalescence/necsim/necsim.pyx":335
 *     def apply(self):
 *         self.set_logger()
 *         self.c_community.applyNoOutput(self.c_spec_parameters)             # <<<<<<<<<<<<<<
 * 
 *     def output(self):
 */
  try {
    __pyx_v_self->c_community.applyNoOutput(__pyx_v_self->c_spec_parameters);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 335, __pyx_L1_error)
  }

  /* "pycoalescence/necsim/necsim.pyx":333
 *         deref(self.c_spec_parameters).addTime(time)
 * 
 *     def apply(self):             # <<<<<<<<<<<<<<
 *         self.set_logger()
 *         self.c_community.applyNoOutput(self.c_spec_parameters)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pycoalescence.necsim.necsim.CMetacommunity.apply", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pycoalescence/necsim/necsim.pyx":337
 *         self.c_community.applyNoOutput(self.c_spec_parameters)
 * 
 *     def output(self):             # <<<<<<<<<<<<<<
 *         self.set_logger()
 *         self.c_community.output()
 */

/* Python wrapper */
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_14CMetacommunity_19output(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_14CMetacommunity_19output(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("output (wrapper)", 0);
  __pyx_r = __pyx_pf_13pycoalescence_6necsim_6necsim_14CMetacommunity_18output(((struct __pyx_obj_13pycoalescence_6necsim_6necsim_CMetacommunity *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_14CMetacommunity_18output(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CMetacommunity *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("output", 0);

  /* "pycoalescence/necsim/necsim.pyx":338
 * 
 *     def output(self):
 *         self.set_logger()             # <<<<<<<<<<<<<<
 *         self.c_community.output()
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_set_logger); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 338, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 338, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pycoalescence/necsim/necsim.pyx":339
 *     def output(self):
 *         self.set_logger()
 *         self.c_community.output()             # <<<<<<<<<<<<<<
 * 
 *     def speciate_remaining_lineages(self, file):
 */
  try {
    __pyx_v_self->c_community.output();
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 339, __pyx_L1_error)
  }

  /* "pycoalescence/necsim/necsim.pyx":337
 *         self.c_community.applyNoOutput(self.c_spec_parameters)
 * 
 *     def output(self):             # <<<<<<<<<<<<<<
 *         self.set_logger()
 *         self.c_community.output()
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pycoalescence.necsim.necsim.CMetacommunity.output", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pycoalescence/necsim/necsim.pyx":341
 *         self.c_community.output()
 * 
 *     def speciate_remaining_lineages(self, file):             # <<<<<<<<<<<<<<
 *         self.set_logger()
 *         cdef string file_str = file
 */

/* Python wrapper */
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_14CMetacommunity_21speciate_remaining_lineages(PyObject *__pyx_v_self, PyObject *__pyx_v_file); /*proto*/
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_14CMetacommunity_21speciate_remaining_lineages(PyObject *__pyx_v_self, PyObject *__pyx_v_file) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("speciate_remaining_lineages (wrapper)", 0);
  __pyx_r = __pyx_pf_13pycoalescence_6necsim_6necsim_14CMetacommunity_20speciate_remaining_lineages(((struct __pyx_obj_13pycoalescence_6necsim_6necsim_CMetacommunity *)__pyx_v_self), ((PyObject *)__pyx_v_file));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_14CMetacommunity_20speciate_remaining_lineages(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CMetacommunity *__pyx_v_self, PyObject *__pyx_v_file) {
  std::string __pyx_v_file_str;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  std::string __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("speciate_remaining_lineages", 0);

  /* "pycoalescence/necsim/necsim.pyx":342
 * 
 *     def speciate_remaining_lineages(self, file):
 *         self.set_logger()             # <<<<<<<<<<<<<<
 *         cdef string file_str = file
 *         self.c_community.speciateRemainingLineages(file_str)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_set_logger); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 342, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 342, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pycoalescence/necsim/necsim.pyx":343
 *     def speciate_remaining_lineages(self, file):
 *         self.set_logger()
 *         cdef string file_str = file             # <<<<<<<<<<<<<<
 *         self.c_community.speciateRemainingLineages(file_str)
 * 
 */
  __pyx_t_4 = __pyx_convert_string_from_py_std__in_string(__pyx_v_file); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 343, __pyx_L1_error)
  __pyx_v_file_str = __pyx_t_4;

  /* "pycoalescence/necsim/necsim.pyx":344
 *         self.set_logger()
 *         cdef string file_str = file
 *         self.c_community.speciateRemainingLineages(file_str)             # <<<<<<<<<<<<<<
 * 
 * cdef class CLandscapeMetricsCalculator:
 */
  try {
    __pyx_v_self->c_community.speciateRemainingLineages(__pyx_v_file_str);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 344, __pyx_L1_error)
  }

  /* "pycoalescence/necsim/necsim.pyx":341
 *         self.c_community.output()
 * 
 *     def speciate_remaining_lineages(self, file):             # <<<<<<<<<<<<<<
 *         self.set_logger()
 *         cdef string file_str = file
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pycoalescence.necsim.necsim.CMetacommunity.speciate_remaining_lineages", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_14CMetacommunity_23__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_14CMetacommunity_23__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_13pycoalescence_6necsim_6necsim_14CMetacommunity_22__reduce_cython__(((struct __pyx_obj_13pycoalescence_6necsim_6necsim_CMetacommunity *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_14CMetacommunity_22__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_13pycoalescence_6necsim_6necsim_CMetacommunity *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__12, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pycoalescence.necsim.necsim.CMetacommunity.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_14CMetacommunity_25__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_14CMetacommunity_25__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_13pycoalescence_6necsim_6necsim_14CMetacommunity_24__setstate_cython__(((struct __pyx_obj_13pycoalescence_6necsim_6necsim_CMetacommunity *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_14CMetacommunity_24__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_13pycoalescence_6necsim_6necsim_CMetacommunity *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__13, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pycoalescence.necsim.necsim.CMetacommunity.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pycoalescence/necsim/necsim.pyx":350
 *     cdef object logger
 * 
 *     def __cinit__(self, object logger):             # <<<<<<<<<<<<<<
 *         self.logger = logger
 *         self.c_landscapes = LandscapeMetricsCalculator()
 */

/* Python wrapper */
static int __pyx_pw_13pycoalescence_6necsim_6necsim_27CLandscapeMetricsCalculator_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_13pycoalescence_6necsim_6necsim_27CLandscapeMetricsCalculator_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_logger = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_logger,0};
    PyObject* values[1] = {0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_logger)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__cinit__") < 0)) __PYX_ERR(0, 350, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 1) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
    }
    __pyx_v_logger = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 1, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 350, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pycoalescence.necsim.necsim.CLandscapeMetricsCalculator.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_13pycoalescence_6necsim_6necsim_27CLandscapeMetricsCalculator___cinit__(((struct __pyx_obj_13pycoalescence_6necsim_6necsim_CLandscapeMetricsCalculator *)__pyx_v_self), __pyx_v_logger);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_13pycoalescence_6necsim_6necsim_27CLandscapeMetricsCalculator___cinit__(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CLandscapeMetricsCalculator *__pyx_v_self, PyObject *__pyx_v_logger) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  LandscapeMetricsCalculator __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "pycoalescence/necsim/necsim.pyx":351
 * 
 *     def __cinit__(self, object logger):
 *         self.logger = logger             # <<<<<<<<<<<<<<
 *         self.c_landscapes = LandscapeMetricsCalculator()
 * 
 */
  __Pyx_INCREF(__pyx_v_logger);
  __Pyx_GIVEREF(__pyx_v_logger);
  __Pyx_GOTREF(__pyx_v_self->logger);
  __Pyx_DECREF(__pyx_v_self->logger);
  __pyx_v_self->logger = __pyx_v_logger;

  /* "pycoalescence/necsim/necsim.pyx":352
 *     def __cinit__(self, object logger):
 *         self.logger = logger
 *         self.c_landscapes = LandscapeMetricsCalculator()             # <<<<<<<<<<<<<<
 * 
 *     def set_logger(self):
 */
  try {
    __pyx_t_1 = LandscapeMetricsCalculator();
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 352, __pyx_L1_error)
  }
  __pyx_v_self->c_landscapes = __pyx_t_1;

  /* "pycoalescence/necsim/necsim.pyx":350
 *     cdef object logger
 * 
 *     def __cinit__(self, object logger):             # <<<<<<<<<<<<<<
 *         self.logger = logger
 *         self.c_landscapes = LandscapeMetricsCalculator()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pycoalescence.necsim.necsim.CLandscapeMetricsCalculator.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pycoalescence/necsim/necsim.pyx":354
 *         self.c_landscapes = LandscapeMetricsCalculator()
 * 
 *     def set_logger(self):             # <<<<<<<<<<<<<<
 *         setGlobalLogger(self.logger, write_to_log)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_27CLandscapeMetricsCalculator_3set_logger(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_27CLandscapeMetricsCalculator_3set_logger(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_logger (wrapper)", 0);
  __pyx_r = __pyx_pf_13pycoalescence_6necsim_6necsim_27CLandscapeMetricsCalculator_2set_logger(((struct __pyx_obj_13pycoalescence_6necsim_6necsim_CLandscapeMetricsCalculator *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_27CLandscapeMetricsCalculator_2set_logger(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CLandscapeMetricsCalculator *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("set_logger", 0);

  /* "pycoalescence/necsim/necsim.pyx":355
 * 
 *     def set_logger(self):
 *         setGlobalLogger(self.logger, write_to_log)             # <<<<<<<<<<<<<<
 * 
 *     def import_map(self, file_name):
 */
  __pyx_t_1 = __pyx_v_self->logger;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_write_to_log); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 355, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  try {
    necsim::setGlobalLogger(__pyx_t_1, __pyx_t_2);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 355, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pycoalescence/necsim/necsim.pyx":354
 *         self.c_landscapes = LandscapeMetricsCalculator()
 * 
 *     def set_logger(self):             # <<<<<<<<<<<<<<
 *         setGlobalLogger(self.logger, write_to_log)
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pycoalescence.necsim.necsim.CLandscapeMetricsCalculator.set_logger", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pycoalescence/necsim/necsim.pyx":357
 *         setGlobalLogger(self.logger, write_to_log)
 * 
 *     def import_map(self, file_name):             # <<<<<<<<<<<<<<
 *         self.set_logger()
 *         cdef string file_name_str = file_name
 */

/* Python wrapper */
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_27CLandscapeMetricsCalculator_5import_map(PyObject *__pyx_v_self, PyObject *__pyx_v_file_name); /*proto*/
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_27CLandscapeMetricsCalculator_5import_map(PyObject *__pyx_v_self, PyObject *__pyx_v_file_name) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("import_map (wrapper)", 0);
  __pyx_r = __pyx_pf_13pycoalescence_6necsim_6necsim_27CLandscapeMetricsCalculator_4import_map(((struct __pyx_obj_13pycoalescence_6necsim_6necsim_CLandscapeMetricsCalculator *)__pyx_v_self), ((PyObject *)__pyx_v_file_name));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_27CLandscapeMetricsCalculator_4import_map(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CLandscapeMetricsCalculator *__pyx_v_self, PyObject *__pyx_v_file_name) {
  std::string __pyx_v_file_name_str;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  std::string __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("import_map", 0);

  /* "pycoalescence/necsim/necsim.pyx":358
 * 
 *     def import_map(self, file_name):
 *         self.set_logger()             # <<<<<<<<<<<<<<
 *         cdef string file_name_str = file_name
 *         self.c_landscapes.importMap(file_name_str)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_set_logger); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 358, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 358, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pycoalescence/necsim/necsim.pyx":359
 *     def import_map(self, file_name):
 *         self.set_logger()
 *         cdef string file_name_str = file_name             # <<<<<<<<<<<<<<
 *         self.c_landscapes.importMap(file_name_str)
 * 
 */
  __pyx_t_4 = __pyx_convert_string_from_py_std__in_string(__pyx_v_file_name); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 359, __pyx_L1_error)
  __pyx_v_file_name_str = __pyx_t_4;

  /* "pycoalescence/necsim/necsim.pyx":360
 *         self.set_logger()
 *         cdef string file_name_str = file_name
 *         self.c_landscapes.importMap(file_name_str)             # <<<<<<<<<<<<<<
 * 
 *     def calculate_MNN(self):
 */
  try {
    __pyx_v_self->c_landscapes.importMap(__pyx_v_file_name_str);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 360, __pyx_L1_error)
  }

  /* "pycoalescence/necsim/necsim.pyx":357
 *         setGlobalLogger(self.logger, write_to_log)
 * 
 *     def import_map(self, file_name):             # <<<<<<<<<<<<<<
 *         self.set_logger()
 *         cdef string file_name_str = file_name
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pycoalescence.necsim.necsim.CLandscapeMetricsCalculator.import_map", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pycoalescence/necsim/necsim.pyx":362
 *         self.c_landscapes.importMap(file_name_str)
 * 
 *     def calculate_MNN(self):             # <<<<<<<<<<<<<<
 *         self.set_logger()
 *         return self.c_landscapes.calculateMNN()
 */

/* Python wrapper */
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_27CLandscapeMetricsCalculator_7calculate_MNN(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_27CLandscapeMetricsCalculator_7calculate_MNN(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("calculate_MNN (wrapper)", 0);
  __pyx_r = __pyx_pf_13pycoalescence_6necsim_6necsim_27CLandscapeMetricsCalculator_6calculate_MNN(((struct __pyx_obj_13pycoalescence_6necsim_6necsim_CLandscapeMetricsCalculator *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_27CLandscapeMetricsCalculator_6calculate_MNN(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CLandscapeMetricsCalculator *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  double __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("calculate_MNN", 0);

  /* "pycoalescence/necsim/necsim.pyx":363
 * 
 *     def calculate_MNN(self):
 *         self.set_logger()             # <<<<<<<<<<<<<<
 *         return self.c_landscapes.calculateMNN()
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_set_logger); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 363, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 363, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pycoalescence/necsim/necsim.pyx":364
 *     def calculate_MNN(self):
 *         self.set_logger()
 *         return self.c_landscapes.calculateMNN()             # <<<<<<<<<<<<<<
 * 
 *     def calculate_CLUMPY(self):
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_4 = __pyx_v_self->c_landscapes.calculateMNN();
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 364, __pyx_L1_error)
  }
  __pyx_t_1 = PyFloat_FromDouble(__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 364, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pycoalescence/necsim/necsim.pyx":362
 *         self.c_landscapes.importMap(file_name_str)
 * 
 *     def calculate_MNN(self):             # <<<<<<<<<<<<<<
 *         self.set_logger()
 *         return self.c_landscapes.calculateMNN()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pycoalescence.necsim.necsim.CLandscapeMetricsCalculator.calculate_MNN", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pycoalescence/necsim/necsim.pyx":366
 *         return self.c_landscapes.calculateMNN()
 * 
 *     def calculate_CLUMPY(self):             # <<<<<<<<<<<<<<
 *         self.set_logger()
 *         return self.c_landscapes.calculateClumpiness()
 */

/* Python wrapper */
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_27CLandscapeMetricsCalculator_9calculate_CLUMPY(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_27CLandscapeMetricsCalculator_9calculate_CLUMPY(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("calculate_CLUMPY (wrapper)", 0);
  __pyx_r = __pyx_pf_13pycoalescence_6necsim_6necsim_27CLandscapeMetricsCalculator_8calculate_CLUMPY(((struct __pyx_obj_13pycoalescence_6necsim_6necsim_CLandscapeMetricsCalculator *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_27CLandscapeMetricsCalculator_8calculate_CLUMPY(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CLandscapeMetricsCalculator *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  double __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("calculate_CLUMPY", 0);

  /* "pycoalescence/necsim/necsim.pyx":367
 * 
 *     def calculate_CLUMPY(self):
 *         self.set_logger()             # <<<<<<<<<<<<<<
 *         return self.c_landscapes.calculateClumpiness()
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_set_logger); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 367, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 367, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pycoalescence/necsim/necsim.pyx":368
 *     def calculate_CLUMPY(self):
 *         self.set_logger()
 *         return self.c_landscapes.calculateClumpiness()             # <<<<<<<<<<<<<<
 * 
 * cdef class CDispersalSimulator:
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_4 = __pyx_v_self->c_landscapes.calculateClumpiness();
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 368, __pyx_L1_error)
  }
  __pyx_t_1 = PyFloat_FromDouble(__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 368, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pycoalescence/necsim/necsim.pyx":366
 *         return self.c_landscapes.calculateMNN()
 * 
 *     def calculate_CLUMPY(self):             # <<<<<<<<<<<<<<
 *         self.set_logger()
 *         return self.c_landscapes.calculateClumpiness()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pycoalescence.necsim.necsim.CLandscapeMetricsCalculator.calculate_CLUMPY", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_27CLandscapeMetricsCalculator_11__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_27CLandscapeMetricsCalculator_11__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_13pycoalescence_6necsim_6necsim_27CLandscapeMetricsCalculator_10__reduce_cython__(((struct __pyx_obj_13pycoalescence_6necsim_6necsim_CLandscapeMetricsCalculator *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_27CLandscapeMetricsCalculator_10__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_13pycoalescence_6necsim_6necsim_CLandscapeMetricsCalculator *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__14, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pycoalescence.necsim.necsim.CLandscapeMetricsCalculator.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_27CLandscapeMetricsCalculator_13__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_27CLandscapeMetricsCalculator_13__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_13pycoalescence_6necsim_6necsim_27CLandscapeMetricsCalculator_12__setstate_cython__(((struct __pyx_obj_13pycoalescence_6necsim_6necsim_CLandscapeMetricsCalculator *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_27CLandscapeMetricsCalculator_12__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_13pycoalescence_6necsim_6necsim_CLandscapeMetricsCalculator *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__15, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pycoalescence.necsim.necsim.CLandscapeMetricsCalculator.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pycoalescence/necsim/necsim.pyx":379
 *     cdef bool printing
 * 
 *     def __cinit__(self, logger):             # <<<<<<<<<<<<<<
 *         self.logger = logger
 *         self.c_simulator = SimulateDispersal()
 */

/* Python wrapper */
static int __pyx_pw_13pycoalescence_6necsim_6necsim_19CDispersalSimulator_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_13pycoalescence_6necsim_6necsim_19CDispersalSimulator_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_logger = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_logger,0};
    PyObject* values[1] = {0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_logger)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__cinit__") < 0)) __PYX_ERR(0, 379, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 1) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
    }
    __pyx_v_logger = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 1, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 379, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pycoalescence.necsim.necsim.CDispersalSimulator.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_13pycoalescence_6necsim_6necsim_19CDispersalSimulator___cinit__(((struct __pyx_obj_13pycoalescence_6necsim_6necsim_CDispersalSimulator *)__pyx_v_self), __pyx_v_logger);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_13pycoalescence_6necsim_6necsim_19CDispersalSimulator___cinit__(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CDispersalSimulator *__pyx_v_self, PyObject *__pyx_v_logger) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  necsim::SimulateDispersal __pyx_t_1;
  std::shared_ptr<necsim::SimParameters>  __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "pycoalescence/necsim/necsim.pyx":380
 * 
 *     def __cinit__(self, logger):
 *         self.logger = logger             # <<<<<<<<<<<<<<
 *         self.c_simulator = SimulateDispersal()
 *         self.c_sim_parameters = make_shared[SimParameters]()
 */
  __Pyx_INCREF(__pyx_v_logger);
  __Pyx_GIVEREF(__pyx_v_logger);
  __Pyx_GOTREF(__pyx_v_self->logger);
  __Pyx_DECREF(__pyx_v_self->logger);
  __pyx_v_self->logger = __pyx_v_logger;

  /* "pycoalescence/necsim/necsim.pyx":381
 *     def __cinit__(self, logger):
 *         self.logger = logger
 *         self.c_simulator = SimulateDispersal()             # <<<<<<<<<<<<<<
 *         self.c_sim_parameters = make_shared[SimParameters]()
 *         self.output_database = "none"
 */
  try {
    __pyx_t_1 = necsim::SimulateDispersal();
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 381, __pyx_L1_error)
  }
  __pyx_v_self->c_simulator = __pyx_t_1;

  /* "pycoalescence/necsim/necsim.pyx":382
 *         self.logger = logger
 *         self.c_simulator = SimulateDispersal()
 *         self.c_sim_parameters = make_shared[SimParameters]()             # <<<<<<<<<<<<<<
 *         self.output_database = "none"
 *         self.has_imported_maps = False
 */
  try {
    __pyx_t_2 = std::make_shared<necsim::SimParameters>();
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 382, __pyx_L1_error)
  }
  __pyx_v_self->c_sim_parameters = __pyx_t_2;

  /* "pycoalescence/necsim/necsim.pyx":383
 *         self.c_simulator = SimulateDispersal()
 *         self.c_sim_parameters = make_shared[SimParameters]()
 *         self.output_database = "none"             # <<<<<<<<<<<<<<
 *         self.has_imported_maps = False
 *         self.needs_update = True
 */
  __Pyx_INCREF(__pyx_n_u_none);
  __Pyx_GIVEREF(__pyx_n_u_none);
  __Pyx_GOTREF(__pyx_v_self->output_database);
  __Pyx_DECREF(__pyx_v_self->output_database);
  __pyx_v_self->output_database = __pyx_n_u_none;

  /* "pycoalescence/necsim/necsim.pyx":384
 *         self.c_sim_parameters = make_shared[SimParameters]()
 *         self.output_database = "none"
 *         self.has_imported_maps = False             # <<<<<<<<<<<<<<
 *         self.needs_update = True
 *         self.printing = True
 */
  __pyx_v_self->has_imported_maps = 0;

  /* "pycoalescence/necsim/necsim.pyx":385
 *         self.output_database = "none"
 *         self.has_imported_maps = False
 *         self.needs_update = True             # <<<<<<<<<<<<<<
 *         self.printing = True
 * 
 */
  __pyx_v_self->needs_update = 1;

  /* "pycoalescence/necsim/necsim.pyx":386
 *         self.has_imported_maps = False
 *         self.needs_update = True
 *         self.printing = True             # <<<<<<<<<<<<<<
 * 
 *     def set_logger(self):
 */
  __pyx_v_self->printing = 1;

  /* "pycoalescence/necsim/necsim.pyx":379
 *     cdef bool printing
 * 
 *     def __cinit__(self, logger):             # <<<<<<<<<<<<<<
 *         self.logger = logger
 *         self.c_simulator = SimulateDispersal()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pycoalescence.necsim.necsim.CDispersalSimulator.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pycoalescence/necsim/necsim.pyx":388
 *         self.printing = True
 * 
 *     def set_logger(self):             # <<<<<<<<<<<<<<
 *         setGlobalLogger(self.logger, write_to_log)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_19CDispersalSimulator_3set_logger(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_19CDispersalSimulator_3set_logger(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_logger (wrapper)", 0);
  __pyx_r = __pyx_pf_13pycoalescence_6necsim_6necsim_19CDispersalSimulator_2set_logger(((struct __pyx_obj_13pycoalescence_6necsim_6necsim_CDispersalSimulator *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_19CDispersalSimulator_2set_logger(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CDispersalSimulator *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("set_logger", 0);

  /* "pycoalescence/necsim/necsim.pyx":389
 * 
 *     def set_logger(self):
 *         setGlobalLogger(self.logger, write_to_log)             # <<<<<<<<<<<<<<
 * 
 *     def set_output_database(self, output_database):
 */
  __pyx_t_1 = __pyx_v_self->logger;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_write_to_log); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 389, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  try {
    necsim::setGlobalLogger(__pyx_t_1, __pyx_t_2);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 389, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pycoalescence/necsim/necsim.pyx":388
 *         self.printing = True
 * 
 *     def set_logger(self):             # <<<<<<<<<<<<<<
 *         setGlobalLogger(self.logger, write_to_log)
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pycoalescence.necsim.necsim.CDispersalSimulator.set_logger", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pycoalescence/necsim/necsim.pyx":391
 *         setGlobalLogger(self.logger, write_to_log)
 * 
 *     def set_output_database(self, output_database):             # <<<<<<<<<<<<<<
 *         self.set_logger()
 *         cdef string output_f = output_database
 */

/* Python wrapper */
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_19CDispersalSimulator_5set_output_database(PyObject *__pyx_v_self, PyObject *__pyx_v_output_database); /*proto*/
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_19CDispersalSimulator_5set_output_database(PyObject *__pyx_v_self, PyObject *__pyx_v_output_database) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_output_database (wrapper)", 0);
  __pyx_r = __pyx_pf_13pycoalescence_6necsim_6necsim_19CDispersalSimulator_4set_output_database(((struct __pyx_obj_13pycoalescence_6necsim_6necsim_CDispersalSimulator *)__pyx_v_self), ((PyObject *)__pyx_v_output_database));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_19CDispersalSimulator_4set_output_database(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CDispersalSimulator *__pyx_v_self, PyObject *__pyx_v_output_database) {
  std::string __pyx_v_output_f;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  std::string __pyx_t_4;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("set_output_database", 0);

  /* "pycoalescence/necsim/necsim.pyx":392
 * 
 *     def set_output_database(self, output_database):
 *         self.set_logger()             # <<<<<<<<<<<<<<
 *         cdef string output_f = output_database
 *         if self.output_database == "none":
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_set_logger); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 392, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 392, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pycoalescence/necsim/necsim.pyx":393
 *     def set_output_database(self, output_database):
 *         self.set_logger()
 *         cdef string output_f = output_database             # <<<<<<<<<<<<<<
 *         if self.output_database == "none":
 *             self.output_database = output_database
 */
  __pyx_t_4 = __pyx_convert_string_from_py_std__in_string(__pyx_v_output_database); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 393, __pyx_L1_error)
  __pyx_v_output_f = __pyx_t_4;

  /* "pycoalescence/necsim/necsim.pyx":394
 *         self.set_logger()
 *         cdef string output_f = output_database
 *         if self.output_database == "none":             # <<<<<<<<<<<<<<
 *             self.output_database = output_database
 *             self.c_simulator.setOutputDatabase(output_f)
 */
  __pyx_t_5 = (__Pyx_PyUnicode_Equals(__pyx_v_self->output_database, __pyx_n_u_none, Py_EQ)); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 394, __pyx_L1_error)
  if (__pyx_t_5) {

    /* "pycoalescence/necsim/necsim.pyx":395
 *         cdef string output_f = output_database
 *         if self.output_database == "none":
 *             self.output_database = output_database             # <<<<<<<<<<<<<<
 *             self.c_simulator.setOutputDatabase(output_f)
 * 
 */
    __Pyx_INCREF(__pyx_v_output_database);
    __Pyx_GIVEREF(__pyx_v_output_database);
    __Pyx_GOTREF(__pyx_v_self->output_database);
    __Pyx_DECREF(__pyx_v_self->output_database);
    __pyx_v_self->output_database = __pyx_v_output_database;

    /* "pycoalescence/necsim/necsim.pyx":396
 *         if self.output_database == "none":
 *             self.output_database = output_database
 *             self.c_simulator.setOutputDatabase(output_f)             # <<<<<<<<<<<<<<
 * 
 *     def import_maps_base(self,
 */
    try {
      __pyx_v_self->c_simulator.setOutputDatabase(__pyx_v_output_f);
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 396, __pyx_L1_error)
    }

    /* "pycoalescence/necsim/necsim.pyx":394
 *         self.set_logger()
 *         cdef string output_f = output_database
 *         if self.output_database == "none":             # <<<<<<<<<<<<<<
 *             self.output_database = output_database
 *             self.c_simulator.setOutputDatabase(output_f)
 */
  }

  /* "pycoalescence/necsim/necsim.pyx":391
 *         setGlobalLogger(self.logger, write_to_log)
 * 
 *     def set_output_database(self, output_database):             # <<<<<<<<<<<<<<
 *         self.set_logger()
 *         cdef string output_f = output_database
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pycoalescence.necsim.necsim.CDispersalSimulator.set_output_database", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pycoalescence/necsim/necsim.pyx":398
 *             self.c_simulator.setOutputDatabase(output_f)
 * 
 *     def import_maps_base(self,             # <<<<<<<<<<<<<<
 *                     double deme,
 *                     fine_map_file,
 */

/* Python wrapper */
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_19CDispersalSimulator_7import_maps_base(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_19CDispersalSimulator_7import_maps_base(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  double __pyx_v_deme;
  PyObject *__pyx_v_fine_map_file = 0;
  unsigned long __pyx_v_fine_map_x_size;
  unsigned long __pyx_v_fine_map_y_size;
  unsigned long __pyx_v_fine_map_x_offset;
  unsigned long __pyx_v_fine_map_y_offset;
  unsigned long __pyx_v_sample_map_x_size;
  unsigned long __pyx_v_sample_map_y_size;
  PyObject *__pyx_v_coarse_map_file = 0;
  unsigned long __pyx_v_coarse_map_x_size;
  unsigned long __pyx_v_coarse_map_y_size;
  unsigned long __pyx_v_coarse_map_x_offset;
  unsigned long __pyx_v_coarse_map_y_offset;
  unsigned long __pyx_v_coarse_map_scale;
  PyObject *__pyx_v_landscape_type = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("import_maps_base (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_deme,&__pyx_n_s_fine_map_file,&__pyx_n_s_fine_map_x_size,&__pyx_n_s_fine_map_y_size,&__pyx_n_s_fine_map_x_offset,&__pyx_n_s_fine_map_y_offset,&__pyx_n_s_sample_map_x_size,&__pyx_n_s_sample_map_y_size,&__pyx_n_s_coarse_map_file,&__pyx_n_s_coarse_map_x_size,&__pyx_n_s_coarse_map_y_size,&__pyx_n_s_coarse_map_x_offset,&__pyx_n_s_coarse_map_y_offset,&__pyx_n_s_coarse_map_scale,&__pyx_n_s_landscape_type,0};
    PyObject* values[15] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 15: values[14] = PyTuple_GET_ITEM(__pyx_args, 14);
        CYTHON_FALLTHROUGH;
        case 14: values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
        CYTHON_FALLTHROUGH;
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_deme)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_fine_map_file)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("import_maps_base", 1, 15, 15, 1); __PYX_ERR(0, 398, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_fine_map_x_size)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("import_maps_base", 1, 15, 15, 2); __PYX_ERR(0, 398, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_fine_map_y_size)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("import_maps_base", 1, 15, 15, 3); __PYX_ERR(0, 398, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_fine_map_x_offset)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("import_maps_base", 1, 15, 15, 4); __PYX_ERR(0, 398, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_fine_map_y_offset)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("import_maps_base", 1, 15, 15, 5); __PYX_ERR(0, 398, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_sample_map_x_size)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("import_maps_base", 1, 15, 15, 6); __PYX_ERR(0, 398, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_sample_map_y_size)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("import_maps_base", 1, 15, 15, 7); __PYX_ERR(0, 398, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_coarse_map_file)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("import_maps_base", 1, 15, 15, 8); __PYX_ERR(0, 398, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_coarse_map_x_size)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("import_maps_base", 1, 15, 15, 9); __PYX_ERR(0, 398, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_coarse_map_y_size)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("import_maps_base", 1, 15, 15, 10); __PYX_ERR(0, 398, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (likely((values[11] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_coarse_map_x_offset)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("import_maps_base", 1, 15, 15, 11); __PYX_ERR(0, 398, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 12:
        if (likely((values[12] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_coarse_map_y_offset)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("import_maps_base", 1, 15, 15, 12); __PYX_ERR(0, 398, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 13:
        if (likely((values[13] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_coarse_map_scale)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("import_maps_base", 1, 15, 15, 13); __PYX_ERR(0, 398, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 14:
        if (likely((values[14] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_landscape_type)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("import_maps_base", 1, 15, 15, 14); __PYX_ERR(0, 398, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "import_maps_base") < 0)) __PYX_ERR(0, 398, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 15) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
      values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
      values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
      values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
      values[14] = PyTuple_GET_ITEM(__pyx_args, 14);
    }
    __pyx_v_deme = __pyx_PyFloat_AsDouble(values[0]); if (unlikely((__pyx_v_deme == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 399, __pyx_L3_error)
    __pyx_v_fine_map_file = values[1];
    __pyx_v_fine_map_x_size = __Pyx_PyInt_As_unsigned_long(values[2]); if (unlikely((__pyx_v_fine_map_x_size == (unsigned long)-1) && PyErr_Occurred())) __PYX_ERR(0, 401, __pyx_L3_error)
    __pyx_v_fine_map_y_size = __Pyx_PyInt_As_unsigned_long(values[3]); if (unlikely((__pyx_v_fine_map_y_size == (unsigned long)-1) && PyErr_Occurred())) __PYX_ERR(0, 402, __pyx_L3_error)
    __pyx_v_fine_map_x_offset = __Pyx_PyInt_As_unsigned_long(values[4]); if (unlikely((__pyx_v_fine_map_x_offset == (unsigned long)-1) && PyErr_Occurred())) __PYX_ERR(0, 403, __pyx_L3_error)
    __pyx_v_fine_map_y_offset = __Pyx_PyInt_As_unsigned_long(values[5]); if (unlikely((__pyx_v_fine_map_y_offset == (unsigned long)-1) && PyErr_Occurred())) __PYX_ERR(0, 404, __pyx_L3_error)
    __pyx_v_sample_map_x_size = __Pyx_PyInt_As_unsigned_long(values[6]); if (unlikely((__pyx_v_sample_map_x_size == (unsigned long)-1) && PyErr_Occurred())) __PYX_ERR(0, 405, __pyx_L3_error)
    __pyx_v_sample_map_y_size = __Pyx_PyInt_As_unsigned_long(values[7]); if (unlikely((__pyx_v_sample_map_y_size == (unsigned long)-1) && PyErr_Occurred())) __PYX_ERR(0, 406, __pyx_L3_error)
    __pyx_v_coarse_map_file = values[8];
    __pyx_v_coarse_map_x_size = __Pyx_PyInt_As_unsigned_long(values[9]); if (unlikely((__pyx_v_coarse_map_x_size == (unsigned long)-1) && PyErr_Occurred())) __PYX_ERR(0, 408, __pyx_L3_error)
    __pyx_v_coarse_map_y_size = __Pyx_PyInt_As_unsigned_long(values[10]); if (unlikely((__pyx_v_coarse_map_y_size == (unsigned long)-1) && PyErr_Occurred())) __PYX_ERR(0, 409, __pyx_L3_error)
    __pyx_v_coarse_map_x_offset = __Pyx_PyInt_As_unsigned_long(values[11]); if (unlikely((__pyx_v_coarse_map_x_offset == (unsigned long)-1) && PyErr_Occurred())) __PYX_ERR(0, 410, __pyx_L3_error)
    __pyx_v_coarse_map_y_offset = __Pyx_PyInt_As_unsigned_long(values[12]); if (unlikely((__pyx_v_coarse_map_y_offset == (unsigned long)-1) && PyErr_Occurred())) __PYX_ERR(0, 411, __pyx_L3_error)
    __pyx_v_coarse_map_scale = __Pyx_PyInt_As_unsigned_long(values[13]); if (unlikely((__pyx_v_coarse_map_scale == (unsigned long)-1) && PyErr_Occurred())) __PYX_ERR(0, 412, __pyx_L3_error)
    __pyx_v_landscape_type = values[14];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("import_maps_base", 1, 15, 15, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 398, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pycoalescence.necsim.necsim.CDispersalSimulator.import_maps_base", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_13pycoalescence_6necsim_6necsim_19CDispersalSimulator_6import_maps_base(((struct __pyx_obj_13pycoalescence_6necsim_6necsim_CDispersalSimulator *)__pyx_v_self), __pyx_v_deme, __pyx_v_fine_map_file, __pyx_v_fine_map_x_size, __pyx_v_fine_map_y_size, __pyx_v_fine_map_x_offset, __pyx_v_fine_map_y_offset, __pyx_v_sample_map_x_size, __pyx_v_sample_map_y_size, __pyx_v_coarse_map_file, __pyx_v_coarse_map_x_size, __pyx_v_coarse_map_y_size, __pyx_v_coarse_map_x_offset, __pyx_v_coarse_map_y_offset, __pyx_v_coarse_map_scale, __pyx_v_landscape_type);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_19CDispersalSimulator_6import_maps_base(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CDispersalSimulator *__pyx_v_self, double __pyx_v_deme, PyObject *__pyx_v_fine_map_file, unsigned long __pyx_v_fine_map_x_size, unsigned long __pyx_v_fine_map_y_size, unsigned long __pyx_v_fine_map_x_offset, unsigned long __pyx_v_fine_map_y_offset, unsigned long __pyx_v_sample_map_x_size, unsigned long __pyx_v_sample_map_y_size, PyObject *__pyx_v_coarse_map_file, unsigned long __pyx_v_coarse_map_x_size, unsigned long __pyx_v_coarse_map_y_size, unsigned long __pyx_v_coarse_map_x_offset, unsigned long __pyx_v_coarse_map_y_offset, unsigned long __pyx_v_coarse_map_scale, PyObject *__pyx_v_landscape_type) {
  std::string __pyx_v_fine_map_file_str;
  std::string __pyx_v_coarse_map_file_str;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  std::string __pyx_t_4;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("import_maps_base", 0);

  /* "pycoalescence/necsim/necsim.pyx":415
 *                     landscape_type,
 *                     ):
 *         self.set_logger()             # <<<<<<<<<<<<<<
 *         cdef string fine_map_file_str = fine_map_file
 *         cdef string coarse_map_file_str = coarse_map_file
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_set_logger); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 415, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 415, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pycoalescence/necsim/necsim.pyx":416
 *                     ):
 *         self.set_logger()
 *         cdef string fine_map_file_str = fine_map_file             # <<<<<<<<<<<<<<
 *         cdef string coarse_map_file_str = coarse_map_file
 *         if self.has_imported_maps:
 */
  __pyx_t_4 = __pyx_convert_string_from_py_std__in_string(__pyx_v_fine_map_file); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 416, __pyx_L1_error)
  __pyx_v_fine_map_file_str = __pyx_t_4;

  /* "pycoalescence/necsim/necsim.pyx":417
 *         self.set_logger()
 *         cdef string fine_map_file_str = fine_map_file
 *         cdef string coarse_map_file_str = coarse_map_file             # <<<<<<<<<<<<<<
 *         if self.has_imported_maps:
 *             raise RuntimeError("Maps have already been imported")
 */
  __pyx_t_4 = __pyx_convert_string_from_py_std__in_string(__pyx_v_coarse_map_file); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 417, __pyx_L1_error)
  __pyx_v_coarse_map_file_str = __pyx_t_4;

  /* "pycoalescence/necsim/necsim.pyx":418
 *         cdef string fine_map_file_str = fine_map_file
 *         cdef string coarse_map_file_str = coarse_map_file
 *         if self.has_imported_maps:             # <<<<<<<<<<<<<<
 *             raise RuntimeError("Maps have already been imported")
 *         deref(self.c_sim_parameters).setMapParameters(fine_map_file_str, coarse_map_file_str,
 */
  __pyx_t_5 = (__pyx_v_self->has_imported_maps != 0);
  if (unlikely(__pyx_t_5)) {

    /* "pycoalescence/necsim/necsim.pyx":419
 *         cdef string coarse_map_file_str = coarse_map_file
 *         if self.has_imported_maps:
 *             raise RuntimeError("Maps have already been imported")             # <<<<<<<<<<<<<<
 *         deref(self.c_sim_parameters).setMapParameters(fine_map_file_str, coarse_map_file_str,
 *                                                       sample_map_x_size,
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple__16, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 419, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 419, __pyx_L1_error)

    /* "pycoalescence/necsim/necsim.pyx":418
 *         cdef string fine_map_file_str = fine_map_file
 *         cdef string coarse_map_file_str = coarse_map_file
 *         if self.has_imported_maps:             # <<<<<<<<<<<<<<
 *             raise RuntimeError("Maps have already been imported")
 *         deref(self.c_sim_parameters).setMapParameters(fine_map_file_str, coarse_map_file_str,
 */
  }

  /* "pycoalescence/necsim/necsim.pyx":430
 *                                                       coarse_map_y_size, coarse_map_x_offset,
 *                                                       coarse_map_y_offset, coarse_map_scale,
 *                                                       deme, landscape_type)             # <<<<<<<<<<<<<<
 *     def import_maps(self, double deme,
 *                     fine_map_file,
 */
  __pyx_t_4 = __pyx_convert_string_from_py_std__in_string(__pyx_v_landscape_type); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 430, __pyx_L1_error)

  /* "pycoalescence/necsim/necsim.pyx":420
 *         if self.has_imported_maps:
 *             raise RuntimeError("Maps have already been imported")
 *         deref(self.c_sim_parameters).setMapParameters(fine_map_file_str, coarse_map_file_str,             # <<<<<<<<<<<<<<
 *                                                       sample_map_x_size,
 *                                                       sample_map_y_size,
 */
  try {
    (*__pyx_v_self->c_sim_parameters).setMapParameters(__pyx_v_fine_map_file_str, __pyx_v_coarse_map_file_str, __pyx_v_sample_map_x_size, __pyx_v_sample_map_y_size, __pyx_v_fine_map_x_size, __pyx_v_fine_map_y_size, __pyx_v_fine_map_x_offset, __pyx_v_fine_map_y_offset, __pyx_v_coarse_map_x_size, __pyx_v_coarse_map_y_size, __pyx_v_coarse_map_x_offset, __pyx_v_coarse_map_y_offset, __pyx_v_coarse_map_scale, __pyx_v_deme, __pyx_t_4);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 420, __pyx_L1_error)
  }

  /* "pycoalescence/necsim/necsim.pyx":398
 *             self.c_simulator.setOutputDatabase(output_f)
 * 
 *     def import_maps_base(self,             # <<<<<<<<<<<<<<
 *                     double deme,
 *                     fine_map_file,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pycoalescence.necsim.necsim.CDispersalSimulator.import_maps_base", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pycoalescence/necsim/necsim.pyx":431
 *                                                       coarse_map_y_offset, coarse_map_scale,
 *                                                       deme, landscape_type)
 *     def import_maps(self, double deme,             # <<<<<<<<<<<<<<
 *                     fine_map_file,
 *                     unsigned long fine_map_x_size,
 */

/* Python wrapper */
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_19CDispersalSimulator_9import_maps(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_19CDispersalSimulator_9import_maps(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  double __pyx_v_deme;
  PyObject *__pyx_v_fine_map_file = 0;
  unsigned long __pyx_v_fine_map_x_size;
  unsigned long __pyx_v_fine_map_y_size;
  unsigned long __pyx_v_fine_map_x_offset;
  unsigned long __pyx_v_fine_map_y_offset;
  unsigned long __pyx_v_sample_map_x_size;
  unsigned long __pyx_v_sample_map_y_size;
  PyObject *__pyx_v_coarse_map_file = 0;
  unsigned long __pyx_v_coarse_map_x_size;
  unsigned long __pyx_v_coarse_map_y_size;
  unsigned long __pyx_v_coarse_map_x_offset;
  unsigned long __pyx_v_coarse_map_y_offset;
  unsigned long __pyx_v_coarse_map_scale;
  PyObject *__pyx_v_landscape_type = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("import_maps (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_deme,&__pyx_n_s_fine_map_file,&__pyx_n_s_fine_map_x_size,&__pyx_n_s_fine_map_y_size,&__pyx_n_s_fine_map_x_offset,&__pyx_n_s_fine_map_y_offset,&__pyx_n_s_sample_map_x_size,&__pyx_n_s_sample_map_y_size,&__pyx_n_s_coarse_map_file,&__pyx_n_s_coarse_map_x_size,&__pyx_n_s_coarse_map_y_size,&__pyx_n_s_coarse_map_x_offset,&__pyx_n_s_coarse_map_y_offset,&__pyx_n_s_coarse_map_scale,&__pyx_n_s_landscape_type,0};
    PyObject* values[15] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 15: values[14] = PyTuple_GET_ITEM(__pyx_args, 14);
        CYTHON_FALLTHROUGH;
        case 14: values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
        CYTHON_FALLTHROUGH;
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_deme)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_fine_map_file)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("import_maps", 1, 15, 15, 1); __PYX_ERR(0, 431, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_fine_map_x_size)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("import_maps", 1, 15, 15, 2); __PYX_ERR(0, 431, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_fine_map_y_size)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("import_maps", 1, 15, 15, 3); __PYX_ERR(0, 431, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_fine_map_x_offset)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("import_maps", 1, 15, 15, 4); __PYX_ERR(0, 431, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_fine_map_y_offset)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("import_maps", 1, 15, 15, 5); __PYX_ERR(0, 431, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_sample_map_x_size)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("import_maps", 1, 15, 15, 6); __PYX_ERR(0, 431, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_sample_map_y_size)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("import_maps", 1, 15, 15, 7); __PYX_ERR(0, 431, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_coarse_map_file)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("import_maps", 1, 15, 15, 8); __PYX_ERR(0, 431, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_coarse_map_x_size)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("import_maps", 1, 15, 15, 9); __PYX_ERR(0, 431, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_coarse_map_y_size)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("import_maps", 1, 15, 15, 10); __PYX_ERR(0, 431, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (likely((values[11] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_coarse_map_x_offset)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("import_maps", 1, 15, 15, 11); __PYX_ERR(0, 431, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 12:
        if (likely((values[12] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_coarse_map_y_offset)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("import_maps", 1, 15, 15, 12); __PYX_ERR(0, 431, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 13:
        if (likely((values[13] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_coarse_map_scale)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("import_maps", 1, 15, 15, 13); __PYX_ERR(0, 431, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 14:
        if (likely((values[14] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_landscape_type)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("import_maps", 1, 15, 15, 14); __PYX_ERR(0, 431, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "import_maps") < 0)) __PYX_ERR(0, 431, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 15) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
      values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
      values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
      values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
      values[14] = PyTuple_GET_ITEM(__pyx_args, 14);
    }
    __pyx_v_deme = __pyx_PyFloat_AsDouble(values[0]); if (unlikely((__pyx_v_deme == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 431, __pyx_L3_error)
    __pyx_v_fine_map_file = values[1];
    __pyx_v_fine_map_x_size = __Pyx_PyInt_As_unsigned_long(values[2]); if (unlikely((__pyx_v_fine_map_x_size == (unsigned long)-1) && PyErr_Occurred())) __PYX_ERR(0, 433, __pyx_L3_error)
    __pyx_v_fine_map_y_size = __Pyx_PyInt_As_unsigned_long(values[3]); if (unlikely((__pyx_v_fine_map_y_size == (unsigned long)-1) && PyErr_Occurred())) __PYX_ERR(0, 434, __pyx_L3_error)
    __pyx_v_fine_map_x_offset = __Pyx_PyInt_As_unsigned_long(values[4]); if (unlikely((__pyx_v_fine_map_x_offset == (unsigned long)-1) && PyErr_Occurred())) __PYX_ERR(0, 435, __pyx_L3_error)
    __pyx_v_fine_map_y_offset = __Pyx_PyInt_As_unsigned_long(values[5]); if (unlikely((__pyx_v_fine_map_y_offset == (unsigned long)-1) && PyErr_Occurred())) __PYX_ERR(0, 436, __pyx_L3_error)
    __pyx_v_sample_map_x_size = __Pyx_PyInt_As_unsigned_long(values[6]); if (unlikely((__pyx_v_sample_map_x_size == (unsigned long)-1) && PyErr_Occurred())) __PYX_ERR(0, 437, __pyx_L3_error)
    __pyx_v_sample_map_y_size = __Pyx_PyInt_As_unsigned_long(values[7]); if (unlikely((__pyx_v_sample_map_y_size == (unsigned long)-1) && PyErr_Occurred())) __PYX_ERR(0, 438, __pyx_L3_error)
    __pyx_v_coarse_map_file = values[8];
    __pyx_v_coarse_map_x_size = __Pyx_PyInt_As_unsigned_long(values[9]); if (unlikely((__pyx_v_coarse_map_x_size == (unsigned long)-1) && PyErr_Occurred())) __PYX_ERR(0, 440, __pyx_L3_error)
    __pyx_v_coarse_map_y_size = __Pyx_PyInt_As_unsigned_long(values[10]); if (unlikely((__pyx_v_coarse_map_y_size == (unsigned long)-1) && PyErr_Occurred())) __PYX_ERR(0, 441, __pyx_L3_error)
    __pyx_v_coarse_map_x_offset = __Pyx_PyInt_As_unsigned_long(values[11]); if (unlikely((__pyx_v_coarse_map_x_offset == (unsigned long)-1) && PyErr_Occurred())) __PYX_ERR(0, 442, __pyx_L3_error)
    __pyx_v_coarse_map_y_offset = __Pyx_PyInt_As_unsigned_long(values[12]); if (unlikely((__pyx_v_coarse_map_y_offset == (unsigned long)-1) && PyErr_Occurred())) __PYX_ERR(0, 443, __pyx_L3_error)
    __pyx_v_coarse_map_scale = __Pyx_PyInt_As_unsigned_long(values[13]); if (unlikely((__pyx_v_coarse_map_scale == (unsigned long)-1) && PyErr_Occurred())) __PYX_ERR(0, 444, __pyx_L3_error)
    __pyx_v_landscape_type = values[14];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("import_maps", 1, 15, 15, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 431, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pycoalescence.necsim.necsim.CDispersalSimulator.import_maps", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_13pycoalescence_6necsim_6necsim_19CDispersalSimulator_8import_maps(((struct __pyx_obj_13pycoalescence_6necsim_6necsim_CDispersalSimulator *)__pyx_v_self), __pyx_v_deme, __pyx_v_fine_map_file, __pyx_v_fine_map_x_size, __pyx_v_fine_map_y_size, __pyx_v_fine_map_x_offset, __pyx_v_fine_map_y_offset, __pyx_v_sample_map_x_size, __pyx_v_sample_map_y_size, __pyx_v_coarse_map_file, __pyx_v_coarse_map_x_size, __pyx_v_coarse_map_y_size, __pyx_v_coarse_map_x_offset, __pyx_v_coarse_map_y_offset, __pyx_v_coarse_map_scale, __pyx_v_landscape_type);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_19CDispersalSimulator_8import_maps(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CDispersalSimulator *__pyx_v_self, double __pyx_v_deme, PyObject *__pyx_v_fine_map_file, unsigned long __pyx_v_fine_map_x_size, unsigned long __pyx_v_fine_map_y_size, unsigned long __pyx_v_fine_map_x_offset, unsigned long __pyx_v_fine_map_y_offset, unsigned long __pyx_v_sample_map_x_size, unsigned long __pyx_v_sample_map_y_size, PyObject *__pyx_v_coarse_map_file, unsigned long __pyx_v_coarse_map_x_size, unsigned long __pyx_v_coarse_map_y_size, unsigned long __pyx_v_coarse_map_x_offset, unsigned long __pyx_v_coarse_map_y_offset, unsigned long __pyx_v_coarse_map_scale, PyObject *__pyx_v_landscape_type) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  int __pyx_t_16;
  PyObject *__pyx_t_17 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("import_maps", 0);

  /* "pycoalescence/necsim/necsim.pyx":446
 *                     unsigned long coarse_map_scale,
 *                     landscape_type):
 *         self.import_maps_base(deme,             # <<<<<<<<<<<<<<
 *                          fine_map_file,
 *                          fine_map_x_size,
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_import_maps_base); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 446, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyFloat_FromDouble(__pyx_v_deme); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 446, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);

  /* "pycoalescence/necsim/necsim.pyx":448
 *         self.import_maps_base(deme,
 *                          fine_map_file,
 *                          fine_map_x_size,             # <<<<<<<<<<<<<<
 *                          fine_map_y_size,
 *                          fine_map_x_offset,
 */
  __pyx_t_4 = __Pyx_PyInt_From_unsigned_long(__pyx_v_fine_map_x_size); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 448, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);

  /* "pycoalescence/necsim/necsim.pyx":449
 *                          fine_map_file,
 *                          fine_map_x_size,
 *                          fine_map_y_size,             # <<<<<<<<<<<<<<
 *                          fine_map_x_offset,
 *                          fine_map_y_offset,
 */
  __pyx_t_5 = __Pyx_PyInt_From_unsigned_long(__pyx_v_fine_map_y_size); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 449, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);

  /* "pycoalescence/necsim/necsim.pyx":450
 *                          fine_map_x_size,
 *                          fine_map_y_size,
 *                          fine_map_x_offset,             # <<<<<<<<<<<<<<
 *                          fine_map_y_offset,
 *                          sample_map_x_size,
 */
  __pyx_t_6 = __Pyx_PyInt_From_unsigned_long(__pyx_v_fine_map_x_offset); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 450, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);

  /* "pycoalescence/necsim/necsim.pyx":451
 *                          fine_map_y_size,
 *                          fine_map_x_offset,
 *                          fine_map_y_offset,             # <<<<<<<<<<<<<<
 *                          sample_map_x_size,
 *                          sample_map_y_size,
 */
  __pyx_t_7 = __Pyx_PyInt_From_unsigned_long(__pyx_v_fine_map_y_offset); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 451, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);

  /* "pycoalescence/necsim/necsim.pyx":452
 *                          fine_map_x_offset,
 *                          fine_map_y_offset,
 *                          sample_map_x_size,             # <<<<<<<<<<<<<<
 *                          sample_map_y_size,
 *                          coarse_map_file,
 */
  __pyx_t_8 = __Pyx_PyInt_From_unsigned_long(__pyx_v_sample_map_x_size); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 452, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);

  /* "pycoalescence/necsim/necsim.pyx":453
 *                          fine_map_y_offset,
 *                          sample_map_x_size,
 *                          sample_map_y_size,             # <<<<<<<<<<<<<<
 *                          coarse_map_file,
 *                          coarse_map_x_size,
 */
  __pyx_t_9 = __Pyx_PyInt_From_unsigned_long(__pyx_v_sample_map_y_size); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 453, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);

  /* "pycoalescence/necsim/necsim.pyx":455
 *                          sample_map_y_size,
 *                          coarse_map_file,
 *                          coarse_map_x_size,             # <<<<<<<<<<<<<<
 *                          coarse_map_y_size,
 *                          coarse_map_x_offset,
 */
  __pyx_t_10 = __Pyx_PyInt_From_unsigned_long(__pyx_v_coarse_map_x_size); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 455, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);

  /* "pycoalescence/necsim/necsim.pyx":456
 *                          coarse_map_file,
 *                          coarse_map_x_size,
 *                          coarse_map_y_size,             # <<<<<<<<<<<<<<
 *                          coarse_map_x_offset,
 *                          coarse_map_y_offset,
 */
  __pyx_t_11 = __Pyx_PyInt_From_unsigned_long(__pyx_v_coarse_map_y_size); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 456, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);

  /* "pycoalescence/necsim/necsim.pyx":457
 *                          coarse_map_x_size,
 *                          coarse_map_y_size,
 *                          coarse_map_x_offset,             # <<<<<<<<<<<<<<
 *                          coarse_map_y_offset,
 *                          coarse_map_scale,
 */
  __pyx_t_12 = __Pyx_PyInt_From_unsigned_long(__pyx_v_coarse_map_x_offset); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 457, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);

  /* "pycoalescence/necsim/necsim.pyx":458
 *                          coarse_map_y_size,
 *                          coarse_map_x_offset,
 *                          coarse_map_y_offset,             # <<<<<<<<<<<<<<
 *                          coarse_map_scale,
 *                          landscape_type)
 */
  __pyx_t_13 = __Pyx_PyInt_From_unsigned_long(__pyx_v_coarse_map_y_offset); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 458, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);

  /* "pycoalescence/necsim/necsim.pyx":459
 *                          coarse_map_x_offset,
 *                          coarse_map_y_offset,
 *                          coarse_map_scale,             # <<<<<<<<<<<<<<
 *                          landscape_type)
 *         self.setup()
 */
  __pyx_t_14 = __Pyx_PyInt_From_unsigned_long(__pyx_v_coarse_map_scale); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 459, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);

  /* "pycoalescence/necsim/necsim.pyx":460
 *                          coarse_map_y_offset,
 *                          coarse_map_scale,
 *                          landscape_type)             # <<<<<<<<<<<<<<
 *         self.setup()
 *         self.c_simulator.importMaps()
 */
  __pyx_t_15 = NULL;
  __pyx_t_16 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_15 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_15)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_15);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_16 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[16] = {__pyx_t_15, __pyx_t_3, __pyx_v_fine_map_file, __pyx_t_4, __pyx_t_5, __pyx_t_6, __pyx_t_7, __pyx_t_8, __pyx_t_9, __pyx_v_coarse_map_file, __pyx_t_10, __pyx_t_11, __pyx_t_12, __pyx_t_13, __pyx_t_14, __pyx_v_landscape_type};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_16, 15+__pyx_t_16); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 446, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[16] = {__pyx_t_15, __pyx_t_3, __pyx_v_fine_map_file, __pyx_t_4, __pyx_t_5, __pyx_t_6, __pyx_t_7, __pyx_t_8, __pyx_t_9, __pyx_v_coarse_map_file, __pyx_t_10, __pyx_t_11, __pyx_t_12, __pyx_t_13, __pyx_t_14, __pyx_v_landscape_type};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_16, 15+__pyx_t_16); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 446, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
  } else
  #endif
  {
    __pyx_t_17 = PyTuple_New(15+__pyx_t_16); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 446, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_17);
    if (__pyx_t_15) {
      __Pyx_GIVEREF(__pyx_t_15); PyTuple_SET_ITEM(__pyx_t_17, 0, __pyx_t_15); __pyx_t_15 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_17, 0+__pyx_t_16, __pyx_t_3);
    __Pyx_INCREF(__pyx_v_fine_map_file);
    __Pyx_GIVEREF(__pyx_v_fine_map_file);
    PyTuple_SET_ITEM(__pyx_t_17, 1+__pyx_t_16, __pyx_v_fine_map_file);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_17, 2+__pyx_t_16, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_17, 3+__pyx_t_16, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_17, 4+__pyx_t_16, __pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_17, 5+__pyx_t_16, __pyx_t_7);
    __Pyx_GIVEREF(__pyx_t_8);
    PyTuple_SET_ITEM(__pyx_t_17, 6+__pyx_t_16, __pyx_t_8);
    __Pyx_GIVEREF(__pyx_t_9);
    PyTuple_SET_ITEM(__pyx_t_17, 7+__pyx_t_16, __pyx_t_9);
    __Pyx_INCREF(__pyx_v_coarse_map_file);
    __Pyx_GIVEREF(__pyx_v_coarse_map_file);
    PyTuple_SET_ITEM(__pyx_t_17, 8+__pyx_t_16, __pyx_v_coarse_map_file);
    __Pyx_GIVEREF(__pyx_t_10);
    PyTuple_SET_ITEM(__pyx_t_17, 9+__pyx_t_16, __pyx_t_10);
    __Pyx_GIVEREF(__pyx_t_11);
    PyTuple_SET_ITEM(__pyx_t_17, 10+__pyx_t_16, __pyx_t_11);
    __Pyx_GIVEREF(__pyx_t_12);
    PyTuple_SET_ITEM(__pyx_t_17, 11+__pyx_t_16, __pyx_t_12);
    __Pyx_GIVEREF(__pyx_t_13);
    PyTuple_SET_ITEM(__pyx_t_17, 12+__pyx_t_16, __pyx_t_13);
    __Pyx_GIVEREF(__pyx_t_14);
    PyTuple_SET_ITEM(__pyx_t_17, 13+__pyx_t_16, __pyx_t_14);
    __Pyx_INCREF(__pyx_v_landscape_type);
    __Pyx_GIVEREF(__pyx_v_landscape_type);
    PyTuple_SET_ITEM(__pyx_t_17, 14+__pyx_t_16, __pyx_v_landscape_type);
    __pyx_t_3 = 0;
    __pyx_t_4 = 0;
    __pyx_t_5 = 0;
    __pyx_t_6 = 0;
    __pyx_t_7 = 0;
    __pyx_t_8 = 0;
    __pyx_t_9 = 0;
    __pyx_t_10 = 0;
    __pyx_t_11 = 0;
    __pyx_t_12 = 0;
    __pyx_t_13 = 0;
    __pyx_t_14 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_17, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 446, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pycoalescence/necsim/necsim.pyx":461
 *                          coarse_map_scale,
 *                          landscape_type)
 *         self.setup()             # <<<<<<<<<<<<<<
 *         self.c_simulator.importMaps()
 *         self.has_imported_maps = True
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_setup); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 461, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_17 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_17 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_17)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_17);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_17) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_17) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 461, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pycoalescence/necsim/necsim.pyx":462
 *                          landscape_type)
 *         self.setup()
 *         self.c_simulator.importMaps()             # <<<<<<<<<<<<<<
 *         self.has_imported_maps = True
 * 
 */
  try {
    __pyx_v_self->c_simulator.importMaps();
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 462, __pyx_L1_error)
  }

  /* "pycoalescence/necsim/necsim.pyx":463
 *         self.setup()
 *         self.c_simulator.importMaps()
 *         self.has_imported_maps = True             # <<<<<<<<<<<<<<
 * 
 *     def import_all_maps(self, double deme,
 */
  __pyx_v_self->has_imported_maps = 1;

  /* "pycoalescence/necsim/necsim.pyx":431
 *                                                       coarse_map_y_offset, coarse_map_scale,
 *                                                       deme, landscape_type)
 *     def import_maps(self, double deme,             # <<<<<<<<<<<<<<
 *                     fine_map_file,
 *                     unsigned long fine_map_x_size,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_XDECREF(__pyx_t_17);
  __Pyx_AddTraceback("pycoalescence.necsim.necsim.CDispersalSimulator.import_maps", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pycoalescence/necsim/necsim.pyx":465
 *         self.has_imported_maps = True
 * 
 *     def import_all_maps(self, double deme,             # <<<<<<<<<<<<<<
 *                         fine_map_file,
 *                         unsigned long int fine_map_x_size,
 */

/* Python wrapper */
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_19CDispersalSimulator_11import_all_maps(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_19CDispersalSimulator_11import_all_maps(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  double __pyx_v_deme;
  PyObject *__pyx_v_fine_map_file = 0;
  unsigned long __pyx_v_fine_map_x_size;
  unsigned long __pyx_v_fine_map_y_size;
  unsigned long __pyx_v_fine_map_x_offset;
  unsigned long __pyx_v_fine_map_y_offset;
  unsigned long __pyx_v_sample_map_x_size;
  unsigned long __pyx_v_sample_map_y_size;
  PyObject *__pyx_v_coarse_map_file = 0;
  unsigned long __pyx_v_coarse_map_x_size;
  unsigned long __pyx_v_coarse_map_y_size;
  unsigned long __pyx_v_coarse_map_x_offset;
  unsigned long __pyx_v_coarse_map_y_offset;
  unsigned long __pyx_v_coarse_map_scale;
  PyObject *__pyx_v_landscape_type = 0;
  std::vector<std::string>  __pyx_v_path_fine;
  std::vector<unsigned long>  __pyx_v_number_fine;
  std::vector<double>  __pyx_v_rate_fine;
  std::vector<double>  __pyx_v_time_fine;
  std::vector<std::string>  __pyx_v_path_coarse;
  std::vector<unsigned long>  __pyx_v_number_coarse;
  std::vector<double>  __pyx_v_rate_coarse;
  std::vector<double>  __pyx_v_time_coarse;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("import_all_maps (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_deme,&__pyx_n_s_fine_map_file,&__pyx_n_s_fine_map_x_size,&__pyx_n_s_fine_map_y_size,&__pyx_n_s_fine_map_x_offset,&__pyx_n_s_fine_map_y_offset,&__pyx_n_s_sample_map_x_size,&__pyx_n_s_sample_map_y_size,&__pyx_n_s_coarse_map_file,&__pyx_n_s_coarse_map_x_size,&__pyx_n_s_coarse_map_y_size,&__pyx_n_s_coarse_map_x_offset,&__pyx_n_s_coarse_map_y_offset,&__pyx_n_s_coarse_map_scale,&__pyx_n_s_landscape_type,&__pyx_n_s_path_fine,&__pyx_n_s_number_fine,&__pyx_n_s_rate_fine,&__pyx_n_s_time_fine,&__pyx_n_s_path_coarse,&__pyx_n_s_number_coarse,&__pyx_n_s_rate_coarse,&__pyx_n_s_time_coarse,0};
    PyObject* values[23] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 23: values[22] = PyTuple_GET_ITEM(__pyx_args, 22);
        CYTHON_FALLTHROUGH;
        case 22: values[21] = PyTuple_GET_ITEM(__pyx_args, 21);
        CYTHON_FALLTHROUGH;
        case 21: values[20] = PyTuple_GET_ITEM(__pyx_args, 20);
        CYTHON_FALLTHROUGH;
        case 20: values[19] = PyTuple_GET_ITEM(__pyx_args, 19);
        CYTHON_FALLTHROUGH;
        case 19: values[18] = PyTuple_GET_ITEM(__pyx_args, 18);
        CYTHON_FALLTHROUGH;
        case 18: values[17] = PyTuple_GET_ITEM(__pyx_args, 17);
        CYTHON_FALLTHROUGH;
        case 17: values[16] = PyTuple_GET_ITEM(__pyx_args, 16);
        CYTHON_FALLTHROUGH;
        case 16: values[15] = PyTuple_GET_ITEM(__pyx_args, 15);
        CYTHON_FALLTHROUGH;
        case 15: values[14] = PyTuple_GET_ITEM(__pyx_args, 14);
        CYTHON_FALLTHROUGH;
        case 14: values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
        CYTHON_FALLTHROUGH;
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_deme)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_fine_map_file)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("import_all_maps", 1, 23, 23, 1); __PYX_ERR(0, 465, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_fine_map_x_size)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("import_all_maps", 1, 23, 23, 2); __PYX_ERR(0, 465, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_fine_map_y_size)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("import_all_maps", 1, 23, 23, 3); __PYX_ERR(0, 465, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_fine_map_x_offset)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("import_all_maps", 1, 23, 23, 4); __PYX_ERR(0, 465, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_fine_map_y_offset)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("import_all_maps", 1, 23, 23, 5); __PYX_ERR(0, 465, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_sample_map_x_size)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("import_all_maps", 1, 23, 23, 6); __PYX_ERR(0, 465, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_sample_map_y_size)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("import_all_maps", 1, 23, 23, 7); __PYX_ERR(0, 465, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_coarse_map_file)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("import_all_maps", 1, 23, 23, 8); __PYX_ERR(0, 465, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_coarse_map_x_size)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("import_all_maps", 1, 23, 23, 9); __PYX_ERR(0, 465, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_coarse_map_y_size)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("import_all_maps", 1, 23, 23, 10); __PYX_ERR(0, 465, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (likely((values[11] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_coarse_map_x_offset)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("import_all_maps", 1, 23, 23, 11); __PYX_ERR(0, 465, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 12:
        if (likely((values[12] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_coarse_map_y_offset)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("import_all_maps", 1, 23, 23, 12); __PYX_ERR(0, 465, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 13:
        if (likely((values[13] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_coarse_map_scale)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("import_all_maps", 1, 23, 23, 13); __PYX_ERR(0, 465, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 14:
        if (likely((values[14] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_landscape_type)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("import_all_maps", 1, 23, 23, 14); __PYX_ERR(0, 465, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 15:
        if (likely((values[15] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_path_fine)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("import_all_maps", 1, 23, 23, 15); __PYX_ERR(0, 465, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 16:
        if (likely((values[16] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_number_fine)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("import_all_maps", 1, 23, 23, 16); __PYX_ERR(0, 465, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 17:
        if (likely((values[17] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_rate_fine)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("import_all_maps", 1, 23, 23, 17); __PYX_ERR(0, 465, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 18:
        if (likely((values[18] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_time_fine)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("import_all_maps", 1, 23, 23, 18); __PYX_ERR(0, 465, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 19:
        if (likely((values[19] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_path_coarse)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("import_all_maps", 1, 23, 23, 19); __PYX_ERR(0, 465, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 20:
        if (likely((values[20] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_number_coarse)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("import_all_maps", 1, 23, 23, 20); __PYX_ERR(0, 465, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 21:
        if (likely((values[21] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_rate_coarse)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("import_all_maps", 1, 23, 23, 21); __PYX_ERR(0, 465, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 22:
        if (likely((values[22] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_time_coarse)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("import_all_maps", 1, 23, 23, 22); __PYX_ERR(0, 465, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "import_all_maps") < 0)) __PYX_ERR(0, 465, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 23) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
      values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
      values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
      values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
      values[14] = PyTuple_GET_ITEM(__pyx_args, 14);
      values[15] = PyTuple_GET_ITEM(__pyx_args, 15);
      values[16] = PyTuple_GET_ITEM(__pyx_args, 16);
      values[17] = PyTuple_GET_ITEM(__pyx_args, 17);
      values[18] = PyTuple_GET_ITEM(__pyx_args, 18);
      values[19] = PyTuple_GET_ITEM(__pyx_args, 19);
      values[20] = PyTuple_GET_ITEM(__pyx_args, 20);
      values[21] = PyTuple_GET_ITEM(__pyx_args, 21);
      values[22] = PyTuple_GET_ITEM(__pyx_args, 22);
    }
    __pyx_v_deme = __pyx_PyFloat_AsDouble(values[0]); if (unlikely((__pyx_v_deme == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 465, __pyx_L3_error)
    __pyx_v_fine_map_file = values[1];
    __pyx_v_fine_map_x_size = __Pyx_PyInt_As_unsigned_long(values[2]); if (unlikely((__pyx_v_fine_map_x_size == (unsigned long)-1) && PyErr_Occurred())) __PYX_ERR(0, 467, __pyx_L3_error)
    __pyx_v_fine_map_y_size = __Pyx_PyInt_As_unsigned_long(values[3]); if (unlikely((__pyx_v_fine_map_y_size == (unsigned long)-1) && PyErr_Occurred())) __PYX_ERR(0, 468, __pyx_L3_error)
    __pyx_v_fine_map_x_offset = __Pyx_PyInt_As_unsigned_long(values[4]); if (unlikely((__pyx_v_fine_map_x_offset == (unsigned long)-1) && PyErr_Occurred())) __PYX_ERR(0, 469, __pyx_L3_error)
    __pyx_v_fine_map_y_offset = __Pyx_PyInt_As_unsigned_long(values[5]); if (unlikely((__pyx_v_fine_map_y_offset == (unsigned long)-1) && PyErr_Occurred())) __PYX_ERR(0, 470, __pyx_L3_error)
    __pyx_v_sample_map_x_size = __Pyx_PyInt_As_unsigned_long(values[6]); if (unlikely((__pyx_v_sample_map_x_size == (unsigned long)-1) && PyErr_Occurred())) __PYX_ERR(0, 471, __pyx_L3_error)
    __pyx_v_sample_map_y_size = __Pyx_PyInt_As_unsigned_long(values[7]); if (unlikely((__pyx_v_sample_map_y_size == (unsigned long)-1) && PyErr_Occurred())) __PYX_ERR(0, 472, __pyx_L3_error)
    __pyx_v_coarse_map_file = values[8];
    __pyx_v_coarse_map_x_size = __Pyx_PyInt_As_unsigned_long(values[9]); if (unlikely((__pyx_v_coarse_map_x_size == (unsigned long)-1) && PyErr_Occurred())) __PYX_ERR(0, 474, __pyx_L3_error)
    __pyx_v_coarse_map_y_size = __Pyx_PyInt_As_unsigned_long(values[10]); if (unlikely((__pyx_v_coarse_map_y_size == (unsigned long)-1) && PyErr_Occurred())) __PYX_ERR(0, 475, __pyx_L3_error)
    __pyx_v_coarse_map_x_offset = __Pyx_PyInt_As_unsigned_long(values[11]); if (unlikely((__pyx_v_coarse_map_x_offset == (unsigned long)-1) && PyErr_Occurred())) __PYX_ERR(0, 476, __pyx_L3_error)
    __pyx_v_coarse_map_y_offset = __Pyx_PyInt_As_unsigned_long(values[12]); if (unlikely((__pyx_v_coarse_map_y_offset == (unsigned long)-1) && PyErr_Occurred())) __PYX_ERR(0, 477, __pyx_L3_error)
    __pyx_v_coarse_map_scale = __Pyx_PyInt_As_unsigned_long(values[13]); if (unlikely((__pyx_v_coarse_map_scale == (unsigned long)-1) && PyErr_Occurred())) __PYX_ERR(0, 478, __pyx_L3_error)
    __pyx_v_landscape_type = values[14];
    __pyx_v_path_fine = __pyx_convert_vector_from_py_std_3a__3a_string(values[15]); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 480, __pyx_L3_error)
    __pyx_v_number_fine = __pyx_convert_vector_from_py_unsigned_long(values[16]); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 481, __pyx_L3_error)
    __pyx_v_rate_fine = __pyx_convert_vector_from_py_double(values[17]); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 482, __pyx_L3_error)
    __pyx_v_time_fine = __pyx_convert_vector_from_py_double(values[18]); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 483, __pyx_L3_error)
    __pyx_v_path_coarse = __pyx_convert_vector_from_py_std_3a__3a_string(values[19]); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 484, __pyx_L3_error)
    __pyx_v_number_coarse = __pyx_convert_vector_from_py_unsigned_long(values[20]); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 485, __pyx_L3_error)
    __pyx_v_rate_coarse = __pyx_convert_vector_from_py_double(values[21]); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 486, __pyx_L3_error)
    __pyx_v_time_coarse = __pyx_convert_vector_from_py_double(values[22]); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 487, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("import_all_maps", 1, 23, 23, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 465, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pycoalescence.necsim.necsim.CDispersalSimulator.import_all_maps", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_13pycoalescence_6necsim_6necsim_19CDispersalSimulator_10import_all_maps(((struct __pyx_obj_13pycoalescence_6necsim_6necsim_CDispersalSimulator *)__pyx_v_self), __pyx_v_deme, __pyx_v_fine_map_file, __pyx_v_fine_map_x_size, __pyx_v_fine_map_y_size, __pyx_v_fine_map_x_offset, __pyx_v_fine_map_y_offset, __pyx_v_sample_map_x_size, __pyx_v_sample_map_y_size, __pyx_v_coarse_map_file, __pyx_v_coarse_map_x_size, __pyx_v_coarse_map_y_size, __pyx_v_coarse_map_x_offset, __pyx_v_coarse_map_y_offset, __pyx_v_coarse_map_scale, __pyx_v_landscape_type, __pyx_v_path_fine, __pyx_v_number_fine, __pyx_v_rate_fine, __pyx_v_time_fine, __pyx_v_path_coarse, __pyx_v_number_coarse, __pyx_v_rate_coarse, __pyx_v_time_coarse);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_19CDispersalSimulator_10import_all_maps(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CDispersalSimulator *__pyx_v_self, double __pyx_v_deme, PyObject *__pyx_v_fine_map_file, unsigned long __pyx_v_fine_map_x_size, unsigned long __pyx_v_fine_map_y_size, unsigned long __pyx_v_fine_map_x_offset, unsigned long __pyx_v_fine_map_y_offset, unsigned long __pyx_v_sample_map_x_size, unsigned long __pyx_v_sample_map_y_size, PyObject *__pyx_v_coarse_map_file, unsigned long __pyx_v_coarse_map_x_size, unsigned long __pyx_v_coarse_map_y_size, unsigned long __pyx_v_coarse_map_x_offset, unsigned long __pyx_v_coarse_map_y_offset, unsigned long __pyx_v_coarse_map_scale, PyObject *__pyx_v_landscape_type, std::vector<std::string>  __pyx_v_path_fine, std::vector<unsigned long>  __pyx_v_number_fine, std::vector<double>  __pyx_v_rate_fine, std::vector<double>  __pyx_v_time_fine, std::vector<std::string>  __pyx_v_path_coarse, std::vector<unsigned long>  __pyx_v_number_coarse, std::vector<double>  __pyx_v_rate_coarse, std::vector<double>  __pyx_v_time_coarse) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  int __pyx_t_16;
  PyObject *__pyx_t_17 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("import_all_maps", 0);

  /* "pycoalescence/necsim/necsim.pyx":488
 *                         vector[double] rate_coarse,
 *                         vector[double] time_coarse):
 *         self.set_logger()             # <<<<<<<<<<<<<<
 *         self.import_maps_base(deme,
 *                          fine_map_file,
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_set_logger); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 488, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 488, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pycoalescence/necsim/necsim.pyx":489
 *                         vector[double] time_coarse):
 *         self.set_logger()
 *         self.import_maps_base(deme,             # <<<<<<<<<<<<<<
 *                          fine_map_file,
 *                          fine_map_x_size,
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_import_maps_base); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 489, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyFloat_FromDouble(__pyx_v_deme); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 489, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);

  /* "pycoalescence/necsim/necsim.pyx":491
 *         self.import_maps_base(deme,
 *                          fine_map_file,
 *                          fine_map_x_size,             # <<<<<<<<<<<<<<
 *                          fine_map_y_size,
 *                          fine_map_x_offset,
 */
  __pyx_t_4 = __Pyx_PyInt_From_unsigned_long(__pyx_v_fine_map_x_size); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 491, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);

  /* "pycoalescence/necsim/necsim.pyx":492
 *                          fine_map_file,
 *                          fine_map_x_size,
 *                          fine_map_y_size,             # <<<<<<<<<<<<<<
 *                          fine_map_x_offset,
 *                          fine_map_y_offset,
 */
  __pyx_t_5 = __Pyx_PyInt_From_unsigned_long(__pyx_v_fine_map_y_size); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 492, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);

  /* "pycoalescence/necsim/necsim.pyx":493
 *                          fine_map_x_size,
 *                          fine_map_y_size,
 *                          fine_map_x_offset,             # <<<<<<<<<<<<<<
 *                          fine_map_y_offset,
 *                          sample_map_x_size,
 */
  __pyx_t_6 = __Pyx_PyInt_From_unsigned_long(__pyx_v_fine_map_x_offset); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 493, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);

  /* "pycoalescence/necsim/necsim.pyx":494
 *                          fine_map_y_size,
 *                          fine_map_x_offset,
 *                          fine_map_y_offset,             # <<<<<<<<<<<<<<
 *                          sample_map_x_size,
 *                          sample_map_y_size,
 */
  __pyx_t_7 = __Pyx_PyInt_From_unsigned_long(__pyx_v_fine_map_y_offset); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 494, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);

  /* "pycoalescence/necsim/necsim.pyx":495
 *                          fine_map_x_offset,
 *                          fine_map_y_offset,
 *                          sample_map_x_size,             # <<<<<<<<<<<<<<
 *                          sample_map_y_size,
 *                          coarse_map_file,
 */
  __pyx_t_8 = __Pyx_PyInt_From_unsigned_long(__pyx_v_sample_map_x_size); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 495, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);

  /* "pycoalescence/necsim/necsim.pyx":496
 *                          fine_map_y_offset,
 *                          sample_map_x_size,
 *                          sample_map_y_size,             # <<<<<<<<<<<<<<
 *                          coarse_map_file,
 *                          coarse_map_x_size,
 */
  __pyx_t_9 = __Pyx_PyInt_From_unsigned_long(__pyx_v_sample_map_y_size); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 496, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);

  /* "pycoalescence/necsim/necsim.pyx":498
 *                          sample_map_y_size,
 *                          coarse_map_file,
 *                          coarse_map_x_size,             # <<<<<<<<<<<<<<
 *                          coarse_map_y_size,
 *                          coarse_map_x_offset,
 */
  __pyx_t_10 = __Pyx_PyInt_From_unsigned_long(__pyx_v_coarse_map_x_size); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 498, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);

  /* "pycoalescence/necsim/necsim.pyx":499
 *                          coarse_map_file,
 *                          coarse_map_x_size,
 *                          coarse_map_y_size,             # <<<<<<<<<<<<<<
 *                          coarse_map_x_offset,
 *                          coarse_map_y_offset,
 */
  __pyx_t_11 = __Pyx_PyInt_From_unsigned_long(__pyx_v_coarse_map_y_size); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 499, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);

  /* "pycoalescence/necsim/necsim.pyx":500
 *                          coarse_map_x_size,
 *                          coarse_map_y_size,
 *                          coarse_map_x_offset,             # <<<<<<<<<<<<<<
 *                          coarse_map_y_offset,
 *                          coarse_map_scale,
 */
  __pyx_t_12 = __Pyx_PyInt_From_unsigned_long(__pyx_v_coarse_map_x_offset); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 500, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);

  /* "pycoalescence/necsim/necsim.pyx":501
 *                          coarse_map_y_size,
 *                          coarse_map_x_offset,
 *                          coarse_map_y_offset,             # <<<<<<<<<<<<<<
 *                          coarse_map_scale,
 *                          landscape_type)
 */
  __pyx_t_13 = __Pyx_PyInt_From_unsigned_long(__pyx_v_coarse_map_y_offset); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 501, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);

  /* "pycoalescence/necsim/necsim.pyx":502
 *                          coarse_map_x_offset,
 *                          coarse_map_y_offset,
 *                          coarse_map_scale,             # <<<<<<<<<<<<<<
 *                          landscape_type)
 *         deref(self.c_sim_parameters).setHistoricalMapParameters(path_fine,
 */
  __pyx_t_14 = __Pyx_PyInt_From_unsigned_long(__pyx_v_coarse_map_scale); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 502, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);

  /* "pycoalescence/necsim/necsim.pyx":503
 *                          coarse_map_y_offset,
 *                          coarse_map_scale,
 *                          landscape_type)             # <<<<<<<<<<<<<<
 *         deref(self.c_sim_parameters).setHistoricalMapParameters(path_fine,
 *                                                                 number_fine,
 */
  __pyx_t_15 = NULL;
  __pyx_t_16 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_15 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_15)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_15);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_16 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[16] = {__pyx_t_15, __pyx_t_3, __pyx_v_fine_map_file, __pyx_t_4, __pyx_t_5, __pyx_t_6, __pyx_t_7, __pyx_t_8, __pyx_t_9, __pyx_v_coarse_map_file, __pyx_t_10, __pyx_t_11, __pyx_t_12, __pyx_t_13, __pyx_t_14, __pyx_v_landscape_type};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_16, 15+__pyx_t_16); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 489, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[16] = {__pyx_t_15, __pyx_t_3, __pyx_v_fine_map_file, __pyx_t_4, __pyx_t_5, __pyx_t_6, __pyx_t_7, __pyx_t_8, __pyx_t_9, __pyx_v_coarse_map_file, __pyx_t_10, __pyx_t_11, __pyx_t_12, __pyx_t_13, __pyx_t_14, __pyx_v_landscape_type};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_16, 15+__pyx_t_16); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 489, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
  } else
  #endif
  {
    __pyx_t_17 = PyTuple_New(15+__pyx_t_16); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 489, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_17);
    if (__pyx_t_15) {
      __Pyx_GIVEREF(__pyx_t_15); PyTuple_SET_ITEM(__pyx_t_17, 0, __pyx_t_15); __pyx_t_15 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_17, 0+__pyx_t_16, __pyx_t_3);
    __Pyx_INCREF(__pyx_v_fine_map_file);
    __Pyx_GIVEREF(__pyx_v_fine_map_file);
    PyTuple_SET_ITEM(__pyx_t_17, 1+__pyx_t_16, __pyx_v_fine_map_file);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_17, 2+__pyx_t_16, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_17, 3+__pyx_t_16, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_17, 4+__pyx_t_16, __pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_17, 5+__pyx_t_16, __pyx_t_7);
    __Pyx_GIVEREF(__pyx_t_8);
    PyTuple_SET_ITEM(__pyx_t_17, 6+__pyx_t_16, __pyx_t_8);
    __Pyx_GIVEREF(__pyx_t_9);
    PyTuple_SET_ITEM(__pyx_t_17, 7+__pyx_t_16, __pyx_t_9);
    __Pyx_INCREF(__pyx_v_coarse_map_file);
    __Pyx_GIVEREF(__pyx_v_coarse_map_file);
    PyTuple_SET_ITEM(__pyx_t_17, 8+__pyx_t_16, __pyx_v_coarse_map_file);
    __Pyx_GIVEREF(__pyx_t_10);
    PyTuple_SET_ITEM(__pyx_t_17, 9+__pyx_t_16, __pyx_t_10);
    __Pyx_GIVEREF(__pyx_t_11);
    PyTuple_SET_ITEM(__pyx_t_17, 10+__pyx_t_16, __pyx_t_11);
    __Pyx_GIVEREF(__pyx_t_12);
    PyTuple_SET_ITEM(__pyx_t_17, 11+__pyx_t_16, __pyx_t_12);
    __Pyx_GIVEREF(__pyx_t_13);
    PyTuple_SET_ITEM(__pyx_t_17, 12+__pyx_t_16, __pyx_t_13);
    __Pyx_GIVEREF(__pyx_t_14);
    PyTuple_SET_ITEM(__pyx_t_17, 13+__pyx_t_16, __pyx_t_14);
    __Pyx_INCREF(__pyx_v_landscape_type);
    __Pyx_GIVEREF(__pyx_v_landscape_type);
    PyTuple_SET_ITEM(__pyx_t_17, 14+__pyx_t_16, __pyx_v_landscape_type);
    __pyx_t_3 = 0;
    __pyx_t_4 = 0;
    __pyx_t_5 = 0;
    __pyx_t_6 = 0;
    __pyx_t_7 = 0;
    __pyx_t_8 = 0;
    __pyx_t_9 = 0;
    __pyx_t_10 = 0;
    __pyx_t_11 = 0;
    __pyx_t_12 = 0;
    __pyx_t_13 = 0;
    __pyx_t_14 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_17, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 489, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pycoalescence/necsim/necsim.pyx":504
 *                          coarse_map_scale,
 *                          landscape_type)
 *         deref(self.c_sim_parameters).setHistoricalMapParameters(path_fine,             # <<<<<<<<<<<<<<
 *                                                                 number_fine,
 *                                                                 rate_fine,
 */
  try {
    (*__pyx_v_self->c_sim_parameters).setHistoricalMapParameters(__pyx_v_path_fine, __pyx_v_number_fine, __pyx_v_rate_fine, __pyx_v_time_fine, __pyx_v_path_coarse, __pyx_v_number_coarse, __pyx_v_rate_coarse, __pyx_v_time_coarse);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 504, __pyx_L1_error)
  }

  /* "pycoalescence/necsim/necsim.pyx":512
 *                                                                 rate_coarse,
 *                                                                 time_coarse)
 *         self.setup()             # <<<<<<<<<<<<<<
 *         self.c_simulator.importMaps()
 *         self.has_imported_maps = True
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_setup); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 512, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_17 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_17 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_17)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_17);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_17) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_17) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 512, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pycoalescence/necsim/necsim.pyx":513
 *                                                                 time_coarse)
 *         self.setup()
 *         self.c_simulator.importMaps()             # <<<<<<<<<<<<<<
 *         self.has_imported_maps = True
 * 
 */
  try {
    __pyx_v_self->c_simulator.importMaps();
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 513, __pyx_L1_error)
  }

  /* "pycoalescence/necsim/necsim.pyx":514
 *         self.setup()
 *         self.c_simulator.importMaps()
 *         self.has_imported_maps = True             # <<<<<<<<<<<<<<
 * 
 *     def check_completed(self):
 */
  __pyx_v_self->has_imported_maps = 1;

  /* "pycoalescence/necsim/necsim.pyx":465
 *         self.has_imported_maps = True
 * 
 *     def import_all_maps(self, double deme,             # <<<<<<<<<<<<<<
 *                         fine_map_file,
 *                         unsigned long int fine_map_x_size,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_XDECREF(__pyx_t_17);
  __Pyx_AddTraceback("pycoalescence.necsim.necsim.CDispersalSimulator.import_all_maps", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pycoalescence/necsim/necsim.pyx":516
 *         self.has_imported_maps = True
 * 
 *     def check_completed(self):             # <<<<<<<<<<<<<<
 *         if self.output_database == "none":
 *             raise RuntimeError("Output database has not been set.")
 */

/* Python wrapper */
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_19CDispersalSimulator_13check_completed(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_19CDispersalSimulator_13check_completed(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("check_completed (wrapper)", 0);
  __pyx_r = __pyx_pf_13pycoalescence_6necsim_6necsim_19CDispersalSimulator_12check_completed(((struct __pyx_obj_13pycoalescence_6necsim_6necsim_CDispersalSimulator *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_19CDispersalSimulator_12check_completed(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CDispersalSimulator *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("check_completed", 0);

  /* "pycoalescence/necsim/necsim.pyx":517
 * 
 *     def check_completed(self):
 *         if self.output_database == "none":             # <<<<<<<<<<<<<<
 *             raise RuntimeError("Output database has not been set.")
 *         if not self.has_imported_maps:
 */
  __pyx_t_1 = (__Pyx_PyUnicode_Equals(__pyx_v_self->output_database, __pyx_n_u_none, Py_EQ)); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 517, __pyx_L1_error)
  if (unlikely(__pyx_t_1)) {

    /* "pycoalescence/necsim/necsim.pyx":518
 *     def check_completed(self):
 *         if self.output_database == "none":
 *             raise RuntimeError("Output database has not been set.")             # <<<<<<<<<<<<<<
 *         if not self.has_imported_maps:
 *             raise RuntimeError("Maps have not been imported - cannot start simulations.")
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple__17, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 518, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 518, __pyx_L1_error)

    /* "pycoalescence/necsim/necsim.pyx":517
 * 
 *     def check_completed(self):
 *         if self.output_database == "none":             # <<<<<<<<<<<<<<
 *             raise RuntimeError("Output database has not been set.")
 *         if not self.has_imported_maps:
 */
  }

  /* "pycoalescence/necsim/necsim.pyx":519
 *         if self.output_database == "none":
 *             raise RuntimeError("Output database has not been set.")
 *         if not self.has_imported_maps:             # <<<<<<<<<<<<<<
 *             raise RuntimeError("Maps have not been imported - cannot start simulations.")
 * 
 */
  __pyx_t_1 = ((!(__pyx_v_self->has_imported_maps != 0)) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "pycoalescence/necsim/necsim.pyx":520
 *             raise RuntimeError("Output database has not been set.")
 *         if not self.has_imported_maps:
 *             raise RuntimeError("Maps have not been imported - cannot start simulations.")             # <<<<<<<<<<<<<<
 * 
 *     def setup(self):
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple__18, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 520, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 520, __pyx_L1_error)

    /* "pycoalescence/necsim/necsim.pyx":519
 *         if self.output_database == "none":
 *             raise RuntimeError("Output database has not been set.")
 *         if not self.has_imported_maps:             # <<<<<<<<<<<<<<
 *             raise RuntimeError("Maps have not been imported - cannot start simulations.")
 * 
 */
  }

  /* "pycoalescence/necsim/necsim.pyx":516
 *         self.has_imported_maps = True
 * 
 *     def check_completed(self):             # <<<<<<<<<<<<<<
 *         if self.output_database == "none":
 *             raise RuntimeError("Output database has not been set.")
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pycoalescence.necsim.necsim.CDispersalSimulator.check_completed", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pycoalescence/necsim/necsim.pyx":522
 *             raise RuntimeError("Maps have not been imported - cannot start simulations.")
 * 
 *     def setup(self):             # <<<<<<<<<<<<<<
 *         if self.needs_update:
 *             self.set_logger()
 */

/* Python wrapper */
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_19CDispersalSimulator_15setup(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_19CDispersalSimulator_15setup(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("setup (wrapper)", 0);
  __pyx_r = __pyx_pf_13pycoalescence_6necsim_6necsim_19CDispersalSimulator_14setup(((struct __pyx_obj_13pycoalescence_6necsim_6necsim_CDispersalSimulator *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_19CDispersalSimulator_14setup(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CDispersalSimulator *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("setup", 0);

  /* "pycoalescence/necsim/necsim.pyx":523
 * 
 *     def setup(self):
 *         if self.needs_update:             # <<<<<<<<<<<<<<
 *             self.set_logger()
 *             self.c_simulator.setSimulationParameters(self.c_sim_parameters, self.printing)
 */
  __pyx_t_1 = (__pyx_v_self->needs_update != 0);
  if (__pyx_t_1) {

    /* "pycoalescence/necsim/necsim.pyx":524
 *     def setup(self):
 *         if self.needs_update:
 *             self.set_logger()             # <<<<<<<<<<<<<<
 *             self.c_simulator.setSimulationParameters(self.c_sim_parameters, self.printing)
 *             self.c_simulator.setDispersalParameters()
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_set_logger); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 524, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 524, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "pycoalescence/necsim/necsim.pyx":525
 *         if self.needs_update:
 *             self.set_logger()
 *             self.c_simulator.setSimulationParameters(self.c_sim_parameters, self.printing)             # <<<<<<<<<<<<<<
 *             self.c_simulator.setDispersalParameters()
 *             self.printing = False
 */
    try {
      __pyx_v_self->c_simulator.setSimulationParameters(__pyx_v_self->c_sim_parameters, __pyx_v_self->printing);
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 525, __pyx_L1_error)
    }

    /* "pycoalescence/necsim/necsim.pyx":526
 *             self.set_logger()
 *             self.c_simulator.setSimulationParameters(self.c_sim_parameters, self.printing)
 *             self.c_simulator.setDispersalParameters()             # <<<<<<<<<<<<<<
 *             self.printing = False
 *             self.needs_update = False
 */
    try {
      __pyx_v_self->c_simulator.setDispersalParameters();
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 526, __pyx_L1_error)
    }

    /* "pycoalescence/necsim/necsim.pyx":527
 *             self.c_simulator.setSimulationParameters(self.c_sim_parameters, self.printing)
 *             self.c_simulator.setDispersalParameters()
 *             self.printing = False             # <<<<<<<<<<<<<<
 *             self.needs_update = False
 * 
 */
    __pyx_v_self->printing = 0;

    /* "pycoalescence/necsim/necsim.pyx":528
 *             self.c_simulator.setDispersalParameters()
 *             self.printing = False
 *             self.needs_update = False             # <<<<<<<<<<<<<<
 * 
 *     def set_dispersal_parameters(self,
 */
    __pyx_v_self->needs_update = 0;

    /* "pycoalescence/necsim/necsim.pyx":523
 * 
 *     def setup(self):
 *         if self.needs_update:             # <<<<<<<<<<<<<<
 *             self.set_logger()
 *             self.c_simulator.setSimulationParameters(self.c_sim_parameters, self.printing)
 */
  }

  /* "pycoalescence/necsim/necsim.pyx":522
 *             raise RuntimeError("Maps have not been imported - cannot start simulations.")
 * 
 *     def setup(self):             # <<<<<<<<<<<<<<
 *         if self.needs_update:
 *             self.set_logger()
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pycoalescence.necsim.necsim.CDispersalSimulator.setup", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pycoalescence/necsim/necsim.pyx":530
 *             self.needs_update = False
 * 
 *     def set_dispersal_parameters(self,             # <<<<<<<<<<<<<<
 *                                  dispersal_method,
 *                                  dispersal_file,
 */

/* Python wrapper */
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_19CDispersalSimulator_17set_dispersal_parameters(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_19CDispersalSimulator_17set_dispersal_parameters(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_dispersal_method = 0;
  PyObject *__pyx_v_dispersal_file = 0;
  double __pyx_v_sigma;
  double __pyx_v_tau;
  double __pyx_v_m_prob;
  double __pyx_v_cutoff;
  double __pyx_v_dispersal_relative_cost;
  bool __pyx_v_restrict_self;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_dispersal_parameters (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_dispersal_method,&__pyx_n_s_dispersal_file,&__pyx_n_s_sigma,&__pyx_n_s_tau,&__pyx_n_s_m_prob,&__pyx_n_s_cutoff,&__pyx_n_s_dispersal_relative_cost,&__pyx_n_s_restrict_self,0};
    PyObject* values[8] = {0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_dispersal_method)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_dispersal_file)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("set_dispersal_parameters", 1, 8, 8, 1); __PYX_ERR(0, 530, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_sigma)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("set_dispersal_parameters", 1, 8, 8, 2); __PYX_ERR(0, 530, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_tau)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("set_dispersal_parameters", 1, 8, 8, 3); __PYX_ERR(0, 530, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_m_prob)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("set_dispersal_parameters", 1, 8, 8, 4); __PYX_ERR(0, 530, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_cutoff)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("set_dispersal_parameters", 1, 8, 8, 5); __PYX_ERR(0, 530, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_dispersal_relative_cost)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("set_dispersal_parameters", 1, 8, 8, 6); __PYX_ERR(0, 530, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_restrict_self)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("set_dispersal_parameters", 1, 8, 8, 7); __PYX_ERR(0, 530, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "set_dispersal_parameters") < 0)) __PYX_ERR(0, 530, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 8) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
    }
    __pyx_v_dispersal_method = values[0];
    __pyx_v_dispersal_file = values[1];
    __pyx_v_sigma = __pyx_PyFloat_AsDouble(values[2]); if (unlikely((__pyx_v_sigma == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 533, __pyx_L3_error)
    __pyx_v_tau = __pyx_PyFloat_AsDouble(values[3]); if (unlikely((__pyx_v_tau == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 534, __pyx_L3_error)
    __pyx_v_m_prob = __pyx_PyFloat_AsDouble(values[4]); if (unlikely((__pyx_v_m_prob == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 535, __pyx_L3_error)
    __pyx_v_cutoff = __pyx_PyFloat_AsDouble(values[5]); if (unlikely((__pyx_v_cutoff == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 536, __pyx_L3_error)
    __pyx_v_dispersal_relative_cost = __pyx_PyFloat_AsDouble(values[6]); if (unlikely((__pyx_v_dispersal_relative_cost == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 537, __pyx_L3_error)
    __pyx_v_restrict_self = __Pyx_PyObject_IsTrue(values[7]); if (unlikely((__pyx_v_restrict_self == ((bool)-1)) && PyErr_Occurred())) __PYX_ERR(0, 538, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("set_dispersal_parameters", 1, 8, 8, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 530, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pycoalescence.necsim.necsim.CDispersalSimulator.set_dispersal_parameters", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_13pycoalescence_6necsim_6necsim_19CDispersalSimulator_16set_dispersal_parameters(((struct __pyx_obj_13pycoalescence_6necsim_6necsim_CDispersalSimulator *)__pyx_v_self), __pyx_v_dispersal_method, __pyx_v_dispersal_file, __pyx_v_sigma, __pyx_v_tau, __pyx_v_m_prob, __pyx_v_cutoff, __pyx_v_dispersal_relative_cost, __pyx_v_restrict_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_19CDispersalSimulator_16set_dispersal_parameters(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CDispersalSimulator *__pyx_v_self, PyObject *__pyx_v_dispersal_method, PyObject *__pyx_v_dispersal_file, double __pyx_v_sigma, double __pyx_v_tau, double __pyx_v_m_prob, double __pyx_v_cutoff, double __pyx_v_dispersal_relative_cost, bool __pyx_v_restrict_self) {
  CYTHON_UNUSED std::string __pyx_v_dispersal_method_str;
  std::string __pyx_v_dispersal_file_str;
  std::string __pyx_v_closed_str;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  std::string __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("set_dispersal_parameters", 0);

  /* "pycoalescence/necsim/necsim.pyx":539
 *                                  double dispersal_relative_cost,
 *                                  bool restrict_self):
 *         cdef string dispersal_method_str = dispersal_method             # <<<<<<<<<<<<<<
 *         cdef string dispersal_file_str = dispersal_file
 *         cdef string closed_str = b"closed"
 */
  __pyx_t_1 = __pyx_convert_string_from_py_std__in_string(__pyx_v_dispersal_method); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 539, __pyx_L1_error)
  __pyx_v_dispersal_method_str = __pyx_t_1;

  /* "pycoalescence/necsim/necsim.pyx":540
 *                                  bool restrict_self):
 *         cdef string dispersal_method_str = dispersal_method
 *         cdef string dispersal_file_str = dispersal_file             # <<<<<<<<<<<<<<
 *         cdef string closed_str = b"closed"
 *         self.set_logger()
 */
  __pyx_t_1 = __pyx_convert_string_from_py_std__in_string(__pyx_v_dispersal_file); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 540, __pyx_L1_error)
  __pyx_v_dispersal_file_str = __pyx_t_1;

  /* "pycoalescence/necsim/necsim.pyx":541
 *         cdef string dispersal_method_str = dispersal_method
 *         cdef string dispersal_file_str = dispersal_file
 *         cdef string closed_str = b"closed"             # <<<<<<<<<<<<<<
 *         self.set_logger()
 *         deref(self.c_sim_parameters).setDispersalParameters(dispersal_method,
 */
  __pyx_t_1 = __pyx_convert_string_from_py_std__in_string(__pyx_n_b_closed); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 541, __pyx_L1_error)
  __pyx_v_closed_str = __pyx_t_1;

  /* "pycoalescence/necsim/necsim.pyx":542
 *         cdef string dispersal_file_str = dispersal_file
 *         cdef string closed_str = b"closed"
 *         self.set_logger()             # <<<<<<<<<<<<<<
 *         deref(self.c_sim_parameters).setDispersalParameters(dispersal_method,
 *                                                             sigma,
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_set_logger); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 542, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 542, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pycoalescence/necsim/necsim.pyx":543
 *         cdef string closed_str = b"closed"
 *         self.set_logger()
 *         deref(self.c_sim_parameters).setDispersalParameters(dispersal_method,             # <<<<<<<<<<<<<<
 *                                                             sigma,
 *                                                             tau,
 */
  __pyx_t_1 = __pyx_convert_string_from_py_std__in_string(__pyx_v_dispersal_method); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 543, __pyx_L1_error)

  /* "pycoalescence/necsim/necsim.pyx":551
 *                                                             restrict_self,
 *                                                             closed_str,
 *                                                             dispersal_file_str)             # <<<<<<<<<<<<<<
 *         self.needs_update = True
 * 
 */
  try {
    (*__pyx_v_self->c_sim_parameters).setDispersalParameters(__pyx_t_1, __pyx_v_sigma, __pyx_v_tau, __pyx_v_m_prob, __pyx_v_cutoff, __pyx_v_dispersal_relative_cost, __pyx_v_restrict_self, __pyx_v_closed_str, __pyx_v_dispersal_file_str);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 543, __pyx_L1_error)
  }

  /* "pycoalescence/necsim/necsim.pyx":552
 *                                                             closed_str,
 *                                                             dispersal_file_str)
 *         self.needs_update = True             # <<<<<<<<<<<<<<
 * 
 *     def base_setup(self, unsigned long number_repeats,
 */
  __pyx_v_self->needs_update = 1;

  /* "pycoalescence/necsim/necsim.pyx":530
 *             self.needs_update = False
 * 
 *     def set_dispersal_parameters(self,             # <<<<<<<<<<<<<<
 *                                  dispersal_method,
 *                                  dispersal_file,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pycoalescence.necsim.necsim.CDispersalSimulator.set_dispersal_parameters", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pycoalescence/necsim/necsim.pyx":554
 *         self.needs_update = True
 * 
 *     def base_setup(self, unsigned long number_repeats,             # <<<<<<<<<<<<<<
 *                    vector[unsigned long] number_steps, unsigned long seed, unsigned long number_workers):
 *         self.set_logger()
 */

/* Python wrapper */
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_19CDispersalSimulator_19base_setup(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_19CDispersalSimulator_19base_setup(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  unsigned long __pyx_v_number_repeats;
  std::vector<unsigned long>  __pyx_v_number_steps;
  unsigned long __pyx_v_seed;
  unsigned long __pyx_v_number_workers;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("base_setup (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_number_repeats,&__pyx_n_s_number_steps,&__pyx_n_s_seed,&__pyx_n_s_number_workers,0};
    PyObject* values[4] = {0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_number_repeats)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_number_steps)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("base_setup", 1, 4, 4, 1); __PYX_ERR(0, 554, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_seed)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("base_setup", 1, 4, 4, 2); __PYX_ERR(0, 554, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_number_workers)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("base_setup", 1, 4, 4, 3); __PYX_ERR(0, 554, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "base_setup") < 0)) __PYX_ERR(0, 554, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 4) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
    }
    __pyx_v_number_repeats = __Pyx_PyInt_As_unsigned_long(values[0]); if (unlikely((__pyx_v_number_repeats == (unsigned long)-1) && PyErr_Occurred())) __PYX_ERR(0, 554, __pyx_L3_error)
    __pyx_v_number_steps = __pyx_convert_vector_from_py_unsigned_long(values[1]); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 555, __pyx_L3_error)
    __pyx_v_seed = __Pyx_PyInt_As_unsigned_long(values[2]); if (unlikely((__pyx_v_seed == (unsigned long)-1) && PyErr_Occurred())) __PYX_ERR(0, 555, __pyx_L3_error)
    __pyx_v_number_workers = __Pyx_PyInt_As_unsigned_long(values[3]); if (unlikely((__pyx_v_number_workers == (unsigned long)-1) && PyErr_Occurred())) __PYX_ERR(0, 555, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("base_setup", 1, 4, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 554, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pycoalescence.necsim.necsim.CDispersalSimulator.base_setup", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_13pycoalescence_6necsim_6necsim_19CDispersalSimulator_18base_setup(((struct __pyx_obj_13pycoalescence_6necsim_6necsim_CDispersalSimulator *)__pyx_v_self), __pyx_v_number_repeats, __pyx_v_number_steps, __pyx_v_seed, __pyx_v_number_workers);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_19CDispersalSimulator_18base_setup(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CDispersalSimulator *__pyx_v_self, unsigned long __pyx_v_number_repeats, std::vector<unsigned long>  __pyx_v_number_steps, unsigned long __pyx_v_seed, unsigned long __pyx_v_number_workers) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("base_setup", 0);

  /* "pycoalescence/necsim/necsim.pyx":556
 *     def base_setup(self, unsigned long number_repeats,
 *                    vector[unsigned long] number_steps, unsigned long seed, unsigned long number_workers):
 *         self.set_logger()             # <<<<<<<<<<<<<<
 *         self.setup()
 *         self.c_simulator.setSeed(seed)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_set_logger); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 556, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 556, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pycoalescence/necsim/necsim.pyx":557
 *                    vector[unsigned long] number_steps, unsigned long seed, unsigned long number_workers):
 *         self.set_logger()
 *         self.setup()             # <<<<<<<<<<<<<<
 *         self.c_simulator.setSeed(seed)
 *         self.c_simulator.setNumberRepeats(number_repeats)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_setup); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 557, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 557, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pycoalescence/necsim/necsim.pyx":558
 *         self.set_logger()
 *         self.setup()
 *         self.c_simulator.setSeed(seed)             # <<<<<<<<<<<<<<
 *         self.c_simulator.setNumberRepeats(number_repeats)
 *         self.c_simulator.setNumberSteps(number_steps)
 */
  try {
    __pyx_v_self->c_simulator.setSeed(__pyx_v_seed);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 558, __pyx_L1_error)
  }

  /* "pycoalescence/necsim/necsim.pyx":559
 *         self.setup()
 *         self.c_simulator.setSeed(seed)
 *         self.c_simulator.setNumberRepeats(number_repeats)             # <<<<<<<<<<<<<<
 *         self.c_simulator.setNumberSteps(number_steps)
 *         self.c_simulator.setNumberWorkers(number_workers)
 */
  try {
    __pyx_v_self->c_simulator.setNumberRepeats(__pyx_v_number_repeats);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 559, __pyx_L1_error)
  }

  /* "pycoalescence/necsim/necsim.pyx":560
 *         self.c_simulator.setSeed(seed)
 *         self.c_simulator.setNumberRepeats(number_repeats)
 *         self.c_simulator.setNumberSteps(number_steps)             # <<<<<<<<<<<<<<
 *         self.c_simulator.setNumberWorkers(number_workers)
 *         if not self.has_imported_maps:
 */
  try {
    __pyx_v_self->c_simulator.setNumberSteps(__pyx_v_number_steps);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 560, __pyx_L1_error)
  }

  /* "pycoalescence/necsim/necsim.pyx":561
 *         self.c_simulator.setNumberRepeats(number_repeats)
 *         self.c_simulator.setNumberSteps(number_steps)
 *         self.c_simulator.setNumberWorkers(number_workers)             # <<<<<<<<<<<<<<
 *         if not self.has_imported_maps:
 *             self.c_simulator.importMaps()
 */
  try {
    __pyx_v_self->c_simulator.setNumberWorkers(__pyx_v_number_workers);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 561, __pyx_L1_error)
  }

  /* "pycoalescence/necsim/necsim.pyx":562
 *         self.c_simulator.setNumberSteps(number_steps)
 *         self.c_simulator.setNumberWorkers(number_workers)
 *         if not self.has_imported_maps:             # <<<<<<<<<<<<<<
 *             self.c_simulator.importMaps()
 *         self.check_completed()
 */
  __pyx_t_4 = ((!(__pyx_v_self->has_imported_maps != 0)) != 0);
  if (__pyx_t_4) {

    /* "pycoalescence/necsim/necsim.pyx":563
 *         self.c_simulator.setNumberWorkers(number_workers)
 *         if not self.has_imported_maps:
 *             self.c_simulator.importMaps()             # <<<<<<<<<<<<<<
 *         self.check_completed()
 * 
 */
    try {
      __pyx_v_self->c_simulator.importMaps();
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 563, __pyx_L1_error)
    }

    /* "pycoalescence/necsim/necsim.pyx":562
 *         self.c_simulator.setNumberSteps(number_steps)
 *         self.c_simulator.setNumberWorkers(number_workers)
 *         if not self.has_imported_maps:             # <<<<<<<<<<<<<<
 *             self.c_simulator.importMaps()
 *         self.check_completed()
 */
  }

  /* "pycoalescence/necsim/necsim.pyx":564
 *         if not self.has_imported_maps:
 *             self.c_simulator.importMaps()
 *         self.check_completed()             # <<<<<<<<<<<<<<
 * 
 *     def write_database(self, table_name):
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_check_completed); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 564, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 564, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pycoalescence/necsim/necsim.pyx":554
 *         self.needs_update = True
 * 
 *     def base_setup(self, unsigned long number_repeats,             # <<<<<<<<<<<<<<
 *                    vector[unsigned long] number_steps, unsigned long seed, unsigned long number_workers):
 *         self.set_logger()
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pycoalescence.necsim.necsim.CDispersalSimulator.base_setup", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pycoalescence/necsim/necsim.pyx":566
 *         self.check_completed()
 * 
 *     def write_database(self, table_name):             # <<<<<<<<<<<<<<
 *         cdef string table_name_str = table_name
 *         self.c_simulator.writeDatabase(table_name_str)
 */

/* Python wrapper */
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_19CDispersalSimulator_21write_database(PyObject *__pyx_v_self, PyObject *__pyx_v_table_name); /*proto*/
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_19CDispersalSimulator_21write_database(PyObject *__pyx_v_self, PyObject *__pyx_v_table_name) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("write_database (wrapper)", 0);
  __pyx_r = __pyx_pf_13pycoalescence_6necsim_6necsim_19CDispersalSimulator_20write_database(((struct __pyx_obj_13pycoalescence_6necsim_6necsim_CDispersalSimulator *)__pyx_v_self), ((PyObject *)__pyx_v_table_name));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_19CDispersalSimulator_20write_database(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CDispersalSimulator *__pyx_v_self, PyObject *__pyx_v_table_name) {
  std::string __pyx_v_table_name_str;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  std::string __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("write_database", 0);

  /* "pycoalescence/necsim/necsim.pyx":567
 * 
 *     def write_database(self, table_name):
 *         cdef string table_name_str = table_name             # <<<<<<<<<<<<<<
 *         self.c_simulator.writeDatabase(table_name_str)
 * 
 */
  __pyx_t_1 = __pyx_convert_string_from_py_std__in_string(__pyx_v_table_name); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 567, __pyx_L1_error)
  __pyx_v_table_name_str = __pyx_t_1;

  /* "pycoalescence/necsim/necsim.pyx":568
 *     def write_database(self, table_name):
 *         cdef string table_name_str = table_name
 *         self.c_simulator.writeDatabase(table_name_str)             # <<<<<<<<<<<<<<
 * 
 *     def run_mean_distance_travelled(self, unsigned long number_repeats,
 */
  try {
    __pyx_v_self->c_simulator.writeDatabase(__pyx_v_table_name_str);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 568, __pyx_L1_error)
  }

  /* "pycoalescence/necsim/necsim.pyx":566
 *         self.check_completed()
 * 
 *     def write_database(self, table_name):             # <<<<<<<<<<<<<<
 *         cdef string table_name_str = table_name
 *         self.c_simulator.writeDatabase(table_name_str)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pycoalescence.necsim.necsim.CDispersalSimulator.write_database", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pycoalescence/necsim/necsim.pyx":570
 *         self.c_simulator.writeDatabase(table_name_str)
 * 
 *     def run_mean_distance_travelled(self, unsigned long number_repeats,             # <<<<<<<<<<<<<<
 *                                     vector[unsigned long] number_steps, unsigned long seed,
 *                                     unsigned long number_workers):
 */

/* Python wrapper */
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_19CDispersalSimulator_23run_mean_distance_travelled(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_19CDispersalSimulator_23run_mean_distance_travelled(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  unsigned long __pyx_v_number_repeats;
  std::vector<unsigned long>  __pyx_v_number_steps;
  unsigned long __pyx_v_seed;
  unsigned long __pyx_v_number_workers;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("run_mean_distance_travelled (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_number_repeats,&__pyx_n_s_number_steps,&__pyx_n_s_seed,&__pyx_n_s_number_workers,0};
    PyObject* values[4] = {0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_number_repeats)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_number_steps)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("run_mean_distance_travelled", 1, 4, 4, 1); __PYX_ERR(0, 570, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_seed)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("run_mean_distance_travelled", 1, 4, 4, 2); __PYX_ERR(0, 570, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_number_workers)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("run_mean_distance_travelled", 1, 4, 4, 3); __PYX_ERR(0, 570, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "run_mean_distance_travelled") < 0)) __PYX_ERR(0, 570, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 4) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
    }
    __pyx_v_number_repeats = __Pyx_PyInt_As_unsigned_long(values[0]); if (unlikely((__pyx_v_number_repeats == (unsigned long)-1) && PyErr_Occurred())) __PYX_ERR(0, 570, __pyx_L3_error)
    __pyx_v_number_steps = __pyx_convert_vector_from_py_unsigned_long(values[1]); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 571, __pyx_L3_error)
    __pyx_v_seed = __Pyx_PyInt_As_unsigned_long(values[2]); if (unlikely((__pyx_v_seed == (unsigned long)-1) && PyErr_Occurred())) __PYX_ERR(0, 571, __pyx_L3_error)
    __pyx_v_number_workers = __Pyx_PyInt_As_unsigned_long(values[3]); if (unlikely((__pyx_v_number_workers == (unsigned long)-1) && PyErr_Occurred())) __PYX_ERR(0, 572, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("run_mean_distance_travelled", 1, 4, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 570, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pycoalescence.necsim.necsim.CDispersalSimulator.run_mean_distance_travelled", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_13pycoalescence_6necsim_6necsim_19CDispersalSimulator_22run_mean_distance_travelled(((struct __pyx_obj_13pycoalescence_6necsim_6necsim_CDispersalSimulator *)__pyx_v_self), __pyx_v_number_repeats, __pyx_v_number_steps, __pyx_v_seed, __pyx_v_number_workers);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_19CDispersalSimulator_22run_mean_distance_travelled(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CDispersalSimulator *__pyx_v_self, unsigned long __pyx_v_number_repeats, std::vector<unsigned long>  __pyx_v_number_steps, unsigned long __pyx_v_seed, unsigned long __pyx_v_number_workers) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("run_mean_distance_travelled", 0);

  /* "pycoalescence/necsim/necsim.pyx":573
 *                                     vector[unsigned long] number_steps, unsigned long seed,
 *                                     unsigned long number_workers):
 *         self.base_setup(number_repeats, number_steps, seed, number_workers)             # <<<<<<<<<<<<<<
 *         self.c_simulator.runMeanDistanceTravelled()
 *         self.write_database("DISTANCES_TRAVELLED")
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_base_setup); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 573, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_From_unsigned_long(__pyx_v_number_repeats); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 573, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __pyx_convert_vector_to_py_unsigned_long(__pyx_v_number_steps); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 573, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyInt_From_unsigned_long(__pyx_v_seed); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 573, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyInt_From_unsigned_long(__pyx_v_number_workers); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 573, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = NULL;
  __pyx_t_8 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_8 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[5] = {__pyx_t_7, __pyx_t_3, __pyx_t_4, __pyx_t_5, __pyx_t_6};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_8, 4+__pyx_t_8); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 573, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[5] = {__pyx_t_7, __pyx_t_3, __pyx_t_4, __pyx_t_5, __pyx_t_6};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_8, 4+__pyx_t_8); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 573, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  } else
  #endif
  {
    __pyx_t_9 = PyTuple_New(4+__pyx_t_8); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 573, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    if (__pyx_t_7) {
      __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_7); __pyx_t_7 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_9, 0+__pyx_t_8, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_9, 1+__pyx_t_8, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_9, 2+__pyx_t_8, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_9, 3+__pyx_t_8, __pyx_t_6);
    __pyx_t_3 = 0;
    __pyx_t_4 = 0;
    __pyx_t_5 = 0;
    __pyx_t_6 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_9, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 573, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pycoalescence/necsim/necsim.pyx":574
 *                                     unsigned long number_workers):
 *         self.base_setup(number_repeats, number_steps, seed, number_workers)
 *         self.c_simulator.runMeanDistanceTravelled()             # <<<<<<<<<<<<<<
 *         self.write_database("DISTANCES_TRAVELLED")
 * 
 */
  try {
    __pyx_v_self->c_simulator.runMeanDistanceTravelled();
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 574, __pyx_L1_error)
  }

  /* "pycoalescence/necsim/necsim.pyx":575
 *         self.base_setup(number_repeats, number_steps, seed, number_workers)
 *         self.c_simulator.runMeanDistanceTravelled()
 *         self.write_database("DISTANCES_TRAVELLED")             # <<<<<<<<<<<<<<
 * 
 *     def run_all_distance_travelled(self, unsigned long number_repeats, vector[unsigned long] number_steps,
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_write_database); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 575, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_9 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_9)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_9);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_9, __pyx_n_u_DISTANCES_TRAVELLED) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_n_u_DISTANCES_TRAVELLED);
  __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 575, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pycoalescence/necsim/necsim.pyx":570
 *         self.c_simulator.writeDatabase(table_name_str)
 * 
 *     def run_mean_distance_travelled(self, unsigned long number_repeats,             # <<<<<<<<<<<<<<
 *                                     vector[unsigned long] number_steps, unsigned long seed,
 *                                     unsigned long number_workers):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("pycoalescence.necsim.necsim.CDispersalSimulator.run_mean_distance_travelled", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pycoalescence/necsim/necsim.pyx":577
 *         self.write_database("DISTANCES_TRAVELLED")
 * 
 *     def run_all_distance_travelled(self, unsigned long number_repeats, vector[unsigned long] number_steps,             # <<<<<<<<<<<<<<
 *                                    unsigned long seed, unsigned long number_workers):
 *         self.base_setup(number_repeats, number_steps, seed, number_workers)
 */

/* Python wrapper */
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_19CDispersalSimulator_25run_all_distance_travelled(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_19CDispersalSimulator_25run_all_distance_travelled(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  unsigned long __pyx_v_number_repeats;
  std::vector<unsigned long>  __pyx_v_number_steps;
  unsigned long __pyx_v_seed;
  unsigned long __pyx_v_number_workers;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("run_all_distance_travelled (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_number_repeats,&__pyx_n_s_number_steps,&__pyx_n_s_seed,&__pyx_n_s_number_workers,0};
    PyObject* values[4] = {0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_number_repeats)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_number_steps)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("run_all_distance_travelled", 1, 4, 4, 1); __PYX_ERR(0, 577, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_seed)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("run_all_distance_travelled", 1, 4, 4, 2); __PYX_ERR(0, 577, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_number_workers)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("run_all_distance_travelled", 1, 4, 4, 3); __PYX_ERR(0, 577, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "run_all_distance_travelled") < 0)) __PYX_ERR(0, 577, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 4) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
    }
    __pyx_v_number_repeats = __Pyx_PyInt_As_unsigned_long(values[0]); if (unlikely((__pyx_v_number_repeats == (unsigned long)-1) && PyErr_Occurred())) __PYX_ERR(0, 577, __pyx_L3_error)
    __pyx_v_number_steps = __pyx_convert_vector_from_py_unsigned_long(values[1]); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 577, __pyx_L3_error)
    __pyx_v_seed = __Pyx_PyInt_As_unsigned_long(values[2]); if (unlikely((__pyx_v_seed == (unsigned long)-1) && PyErr_Occurred())) __PYX_ERR(0, 578, __pyx_L3_error)
    __pyx_v_number_workers = __Pyx_PyInt_As_unsigned_long(values[3]); if (unlikely((__pyx_v_number_workers == (unsigned long)-1) && PyErr_Occurred())) __PYX_ERR(0, 578, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("run_all_distance_travelled", 1, 4, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 577, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pycoalescence.necsim.necsim.CDispersalSimulator.run_all_distance_travelled", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_13pycoalescence_6necsim_6necsim_19CDispersalSimulator_24run_all_distance_travelled(((struct __pyx_obj_13pycoalescence_6necsim_6necsim_CDispersalSimulator *)__pyx_v_self), __pyx_v_number_repeats, __pyx_v_number_steps, __pyx_v_seed, __pyx_v_number_workers);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_19CDispersalSimulator_24run_all_distance_travelled(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CDispersalSimulator *__pyx_v_self, unsigned long __pyx_v_number_repeats, std::vector<unsigned long>  __pyx_v_number_steps, unsigned long __pyx_v_seed, unsigned long __pyx_v_number_workers) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("run_all_distance_travelled", 0);

  /* "pycoalescence/necsim/necsim.pyx":579
 *     def run_all_distance_travelled(self, unsigned long number_repeats, vector[unsigned long] number_steps,
 *                                    unsigned long seed, unsigned long number_workers):
 *         self.base_setup(number_repeats, number_steps, seed, number_workers)             # <<<<<<<<<<<<<<
 *         self.c_simulator.runAllDistanceTravelled()
 *         self.write_database("DISTANCES_TRAVELLED")
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_base_setup); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 579, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_From_unsigned_long(__pyx_v_number_repeats); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 579, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __pyx_convert_vector_to_py_unsigned_long(__pyx_v_number_steps); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 579, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyInt_From_unsigned_long(__pyx_v_seed); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 579, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyInt_From_unsigned_long(__pyx_v_number_workers); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 579, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = NULL;
  __pyx_t_8 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_8 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[5] = {__pyx_t_7, __pyx_t_3, __pyx_t_4, __pyx_t_5, __pyx_t_6};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_8, 4+__pyx_t_8); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 579, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[5] = {__pyx_t_7, __pyx_t_3, __pyx_t_4, __pyx_t_5, __pyx_t_6};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_8, 4+__pyx_t_8); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 579, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  } else
  #endif
  {
    __pyx_t_9 = PyTuple_New(4+__pyx_t_8); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 579, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    if (__pyx_t_7) {
      __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_7); __pyx_t_7 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_9, 0+__pyx_t_8, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_9, 1+__pyx_t_8, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_9, 2+__pyx_t_8, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_9, 3+__pyx_t_8, __pyx_t_6);
    __pyx_t_3 = 0;
    __pyx_t_4 = 0;
    __pyx_t_5 = 0;
    __pyx_t_6 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_9, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 579, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pycoalescence/necsim/necsim.pyx":580
 *                                    unsigned long seed, unsigned long number_workers):
 *         self.base_setup(number_repeats, number_steps, seed, number_workers)
 *         self.c_simulator.runAllDistanceTravelled()             # <<<<<<<<<<<<<<
 *         self.write_database("DISTANCES_TRAVELLED")
 * 
 */
  try {
    __pyx_v_self->c_simulator.runAllDistanceTravelled();
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 580, __pyx_L1_error)
  }

  /* "pycoalescence/necsim/necsim.pyx":581
 *         self.base_setup(number_repeats, number_steps, seed, number_workers)
 *         self.c_simulator.runAllDistanceTravelled()
 *         self.write_database("DISTANCES_TRAVELLED")             # <<<<<<<<<<<<<<
 * 
 *     def run_sample_distance_travelled(self, vector[long] samples_x,
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_write_database); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 581, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_9 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_9)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_9);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_9, __pyx_n_u_DISTANCES_TRAVELLED) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_n_u_DISTANCES_TRAVELLED);
  __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 581, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pycoalescence/necsim/necsim.pyx":577
 *         self.write_database("DISTANCES_TRAVELLED")
 * 
 *     def run_all_distance_travelled(self, unsigned long number_repeats, vector[unsigned long] number_steps,             # <<<<<<<<<<<<<<
 *                                    unsigned long seed, unsigned long number_workers):
 *         self.base_setup(number_repeats, number_steps, seed, number_workers)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("pycoalescence.necsim.necsim.CDispersalSimulator.run_all_distance_travelled", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pycoalescence/necsim/necsim.pyx":583
 *         self.write_database("DISTANCES_TRAVELLED")
 * 
 *     def run_sample_distance_travelled(self, vector[long] samples_x,             # <<<<<<<<<<<<<<
 *                                       vector[long] samples_y, unsigned long number_repeats,
 *                                       vector[unsigned long] number_steps, unsigned long seed,
 */

/* Python wrapper */
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_19CDispersalSimulator_27run_sample_distance_travelled(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_19CDispersalSimulator_27run_sample_distance_travelled(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  std::vector<long>  __pyx_v_samples_x;
  std::vector<long>  __pyx_v_samples_y;
  unsigned long __pyx_v_number_repeats;
  std::vector<unsigned long>  __pyx_v_number_steps;
  unsigned long __pyx_v_seed;
  unsigned long __pyx_v_number_workers;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("run_sample_distance_travelled (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_samples_x,&__pyx_n_s_samples_y,&__pyx_n_s_number_repeats,&__pyx_n_s_number_steps,&__pyx_n_s_seed,&__pyx_n_s_number_workers,0};
    PyObject* values[6] = {0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_samples_x)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_samples_y)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("run_sample_distance_travelled", 1, 6, 6, 1); __PYX_ERR(0, 583, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_number_repeats)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("run_sample_distance_travelled", 1, 6, 6, 2); __PYX_ERR(0, 583, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_number_steps)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("run_sample_distance_travelled", 1, 6, 6, 3); __PYX_ERR(0, 583, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_seed)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("run_sample_distance_travelled", 1, 6, 6, 4); __PYX_ERR(0, 583, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_number_workers)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("run_sample_distance_travelled", 1, 6, 6, 5); __PYX_ERR(0, 583, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "run_sample_distance_travelled") < 0)) __PYX_ERR(0, 583, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 6) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
    }
    __pyx_v_samples_x = __pyx_convert_vector_from_py_long(values[0]); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 583, __pyx_L3_error)
    __pyx_v_samples_y = __pyx_convert_vector_from_py_long(values[1]); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 584, __pyx_L3_error)
    __pyx_v_number_repeats = __Pyx_PyInt_As_unsigned_long(values[2]); if (unlikely((__pyx_v_number_repeats == (unsigned long)-1) && PyErr_Occurred())) __PYX_ERR(0, 584, __pyx_L3_error)
    __pyx_v_number_steps = __pyx_convert_vector_from_py_unsigned_long(values[3]); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 585, __pyx_L3_error)
    __pyx_v_seed = __Pyx_PyInt_As_unsigned_long(values[4]); if (unlikely((__pyx_v_seed == (unsigned long)-1) && PyErr_Occurred())) __PYX_ERR(0, 585, __pyx_L3_error)
    __pyx_v_number_workers = __Pyx_PyInt_As_unsigned_long(values[5]); if (unlikely((__pyx_v_number_workers == (unsigned long)-1) && PyErr_Occurred())) __PYX_ERR(0, 586, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("run_sample_distance_travelled", 1, 6, 6, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 583, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pycoalescence.necsim.necsim.CDispersalSimulator.run_sample_distance_travelled", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_13pycoalescence_6necsim_6necsim_19CDispersalSimulator_26run_sample_distance_travelled(((struct __pyx_obj_13pycoalescence_6necsim_6necsim_CDispersalSimulator *)__pyx_v_self), __pyx_v_samples_x, __pyx_v_samples_y, __pyx_v_number_repeats, __pyx_v_number_steps, __pyx_v_seed, __pyx_v_number_workers);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_19CDispersalSimulator_26run_sample_distance_travelled(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CDispersalSimulator *__pyx_v_self, std::vector<long>  __pyx_v_samples_x, std::vector<long>  __pyx_v_samples_y, unsigned long __pyx_v_number_repeats, std::vector<unsigned long>  __pyx_v_number_steps, unsigned long __pyx_v_seed, unsigned long __pyx_v_number_workers) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("run_sample_distance_travelled", 0);

  /* "pycoalescence/necsim/necsim.pyx":587
 *                                       vector[unsigned long] number_steps, unsigned long seed,
 *                                       unsigned long number_workers):
 *         self.base_setup(number_repeats, number_steps, seed, number_workers)             # <<<<<<<<<<<<<<
 *         self.c_simulator.runSampleDistanceTravelled(samples_x, samples_y)
 *         self.write_database("DISTANCES_TRAVELLED")
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_base_setup); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 587, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_From_unsigned_long(__pyx_v_number_repeats); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 587, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __pyx_convert_vector_to_py_unsigned_long(__pyx_v_number_steps); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 587, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyInt_From_unsigned_long(__pyx_v_seed); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 587, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyInt_From_unsigned_long(__pyx_v_number_workers); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 587, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = NULL;
  __pyx_t_8 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_8 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[5] = {__pyx_t_7, __pyx_t_3, __pyx_t_4, __pyx_t_5, __pyx_t_6};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_8, 4+__pyx_t_8); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 587, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[5] = {__pyx_t_7, __pyx_t_3, __pyx_t_4, __pyx_t_5, __pyx_t_6};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_8, 4+__pyx_t_8); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 587, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  } else
  #endif
  {
    __pyx_t_9 = PyTuple_New(4+__pyx_t_8); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 587, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    if (__pyx_t_7) {
      __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_7); __pyx_t_7 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_9, 0+__pyx_t_8, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_9, 1+__pyx_t_8, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_9, 2+__pyx_t_8, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_9, 3+__pyx_t_8, __pyx_t_6);
    __pyx_t_3 = 0;
    __pyx_t_4 = 0;
    __pyx_t_5 = 0;
    __pyx_t_6 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_9, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 587, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pycoalescence/necsim/necsim.pyx":588
 *                                       unsigned long number_workers):
 *         self.base_setup(number_repeats, number_steps, seed, number_workers)
 *         self.c_simulator.runSampleDistanceTravelled(samples_x, samples_y)             # <<<<<<<<<<<<<<
 *         self.write_database("DISTANCES_TRAVELLED")
 * 
 */
  try {
    __pyx_v_self->c_simulator.runSampleDistanceTravelled(__pyx_v_samples_x, __pyx_v_samples_y);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 588, __pyx_L1_error)
  }

  /* "pycoalescence/necsim/necsim.pyx":589
 *         self.base_setup(number_repeats, number_steps, seed, number_workers)
 *         self.c_simulator.runSampleDistanceTravelled(samples_x, samples_y)
 *         self.write_database("DISTANCES_TRAVELLED")             # <<<<<<<<<<<<<<
 * 
 *     def run_mean_dispersal_distance(self, unsigned long number_repeats, unsigned long seed, bool is_sequential):
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_write_database); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 589, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_9 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_9)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_9);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_9, __pyx_n_u_DISTANCES_TRAVELLED) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_n_u_DISTANCES_TRAVELLED);
  __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 589, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pycoalescence/necsim/necsim.pyx":583
 *         self.write_database("DISTANCES_TRAVELLED")
 * 
 *     def run_sample_distance_travelled(self, vector[long] samples_x,             # <<<<<<<<<<<<<<
 *                                       vector[long] samples_y, unsigned long number_repeats,
 *                                       vector[unsigned long] number_steps, unsigned long seed,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("pycoalescence.necsim.necsim.CDispersalSimulator.run_sample_distance_travelled", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pycoalescence/necsim/necsim.pyx":591
 *         self.write_database("DISTANCES_TRAVELLED")
 * 
 *     def run_mean_dispersal_distance(self, unsigned long number_repeats, unsigned long seed, bool is_sequential):             # <<<<<<<<<<<<<<
 *         self.set_logger()
 *         self.setup()
 */

/* Python wrapper */
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_19CDispersalSimulator_29run_mean_dispersal_distance(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_19CDispersalSimulator_29run_mean_dispersal_distance(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  unsigned long __pyx_v_number_repeats;
  unsigned long __pyx_v_seed;
  bool __pyx_v_is_sequential;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("run_mean_dispersal_distance (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_number_repeats,&__pyx_n_s_seed,&__pyx_n_s_is_sequential,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_number_repeats)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_seed)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("run_mean_dispersal_distance", 1, 3, 3, 1); __PYX_ERR(0, 591, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_is_sequential)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("run_mean_dispersal_distance", 1, 3, 3, 2); __PYX_ERR(0, 591, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "run_mean_dispersal_distance") < 0)) __PYX_ERR(0, 591, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_number_repeats = __Pyx_PyInt_As_unsigned_long(values[0]); if (unlikely((__pyx_v_number_repeats == (unsigned long)-1) && PyErr_Occurred())) __PYX_ERR(0, 591, __pyx_L3_error)
    __pyx_v_seed = __Pyx_PyInt_As_unsigned_long(values[1]); if (unlikely((__pyx_v_seed == (unsigned long)-1) && PyErr_Occurred())) __PYX_ERR(0, 591, __pyx_L3_error)
    __pyx_v_is_sequential = __Pyx_PyObject_IsTrue(values[2]); if (unlikely((__pyx_v_is_sequential == ((bool)-1)) && PyErr_Occurred())) __PYX_ERR(0, 591, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("run_mean_dispersal_distance", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 591, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pycoalescence.necsim.necsim.CDispersalSimulator.run_mean_dispersal_distance", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_13pycoalescence_6necsim_6necsim_19CDispersalSimulator_28run_mean_dispersal_distance(((struct __pyx_obj_13pycoalescence_6necsim_6necsim_CDispersalSimulator *)__pyx_v_self), __pyx_v_number_repeats, __pyx_v_seed, __pyx_v_is_sequential);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_19CDispersalSimulator_28run_mean_dispersal_distance(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CDispersalSimulator *__pyx_v_self, unsigned long __pyx_v_number_repeats, unsigned long __pyx_v_seed, bool __pyx_v_is_sequential) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("run_mean_dispersal_distance", 0);

  /* "pycoalescence/necsim/necsim.pyx":592
 * 
 *     def run_mean_dispersal_distance(self, unsigned long number_repeats, unsigned long seed, bool is_sequential):
 *         self.set_logger()             # <<<<<<<<<<<<<<
 *         self.setup()
 *         self.c_simulator.setSeed(seed)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_set_logger); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 592, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 592, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pycoalescence/necsim/necsim.pyx":593
 *     def run_mean_dispersal_distance(self, unsigned long number_repeats, unsigned long seed, bool is_sequential):
 *         self.set_logger()
 *         self.setup()             # <<<<<<<<<<<<<<
 *         self.c_simulator.setSeed(seed)
 *         self.c_simulator.setNumberRepeats(number_repeats)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_setup); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 593, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 593, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pycoalescence/necsim/necsim.pyx":594
 *         self.set_logger()
 *         self.setup()
 *         self.c_simulator.setSeed(seed)             # <<<<<<<<<<<<<<
 *         self.c_simulator.setNumberRepeats(number_repeats)
 *         self.c_simulator.setSequential(is_sequential)
 */
  try {
    __pyx_v_self->c_simulator.setSeed(__pyx_v_seed);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 594, __pyx_L1_error)
  }

  /* "pycoalescence/necsim/necsim.pyx":595
 *         self.setup()
 *         self.c_simulator.setSeed(seed)
 *         self.c_simulator.setNumberRepeats(number_repeats)             # <<<<<<<<<<<<<<
 *         self.c_simulator.setSequential(is_sequential)
 *         if not self.has_imported_maps:
 */
  try {
    __pyx_v_self->c_simulator.setNumberRepeats(__pyx_v_number_repeats);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 595, __pyx_L1_error)
  }

  /* "pycoalescence/necsim/necsim.pyx":596
 *         self.c_simulator.setSeed(seed)
 *         self.c_simulator.setNumberRepeats(number_repeats)
 *         self.c_simulator.setSequential(is_sequential)             # <<<<<<<<<<<<<<
 *         if not self.has_imported_maps:
 *             self.c_simulator.importMaps()
 */
  try {
    __pyx_v_self->c_simulator.setSequential(__pyx_v_is_sequential);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 596, __pyx_L1_error)
  }

  /* "pycoalescence/necsim/necsim.pyx":597
 *         self.c_simulator.setNumberRepeats(number_repeats)
 *         self.c_simulator.setSequential(is_sequential)
 *         if not self.has_imported_maps:             # <<<<<<<<<<<<<<
 *             self.c_simulator.importMaps()
 *         self.check_completed()
 */
  __pyx_t_4 = ((!(__pyx_v_self->has_imported_maps != 0)) != 0);
  if (__pyx_t_4) {

    /* "pycoalescence/necsim/necsim.pyx":598
 *         self.c_simulator.setSequential(is_sequential)
 *         if not self.has_imported_maps:
 *             self.c_simulator.importMaps()             # <<<<<<<<<<<<<<
 *         self.check_completed()
 *         self.c_simulator.runMeanDispersalDistance()
 */
    try {
      __pyx_v_self->c_simulator.importMaps();
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 598, __pyx_L1_error)
    }

    /* "pycoalescence/necsim/necsim.pyx":597
 *         self.c_simulator.setNumberRepeats(number_repeats)
 *         self.c_simulator.setSequential(is_sequential)
 *         if not self.has_imported_maps:             # <<<<<<<<<<<<<<
 *             self.c_simulator.importMaps()
 *         self.check_completed()
 */
  }

  /* "pycoalescence/necsim/necsim.pyx":599
 *         if not self.has_imported_maps:
 *             self.c_simulator.importMaps()
 *         self.check_completed()             # <<<<<<<<<<<<<<
 *         self.c_simulator.runMeanDispersalDistance()
 *         self.write_database("DISPERSAL_DISTANCES")
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_check_completed); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 599, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 599, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pycoalescence/necsim/necsim.pyx":600
 *             self.c_simulator.importMaps()
 *         self.check_completed()
 *         self.c_simulator.runMeanDispersalDistance()             # <<<<<<<<<<<<<<
 *         self.write_database("DISPERSAL_DISTANCES")
 */
  try {
    __pyx_v_self->c_simulator.runMeanDispersalDistance();
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 600, __pyx_L1_error)
  }

  /* "pycoalescence/necsim/necsim.pyx":601
 *         self.check_completed()
 *         self.c_simulator.runMeanDispersalDistance()
 *         self.write_database("DISPERSAL_DISTANCES")             # <<<<<<<<<<<<<<
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_write_database); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 601, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_n_u_DISPERSAL_DISTANCES) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_n_u_DISPERSAL_DISTANCES);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 601, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pycoalescence/necsim/necsim.pyx":591
 *         self.write_database("DISTANCES_TRAVELLED")
 * 
 *     def run_mean_dispersal_distance(self, unsigned long number_repeats, unsigned long seed, bool is_sequential):             # <<<<<<<<<<<<<<
 *         self.set_logger()
 *         self.setup()
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pycoalescence.necsim.necsim.CDispersalSimulator.run_mean_dispersal_distance", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_19CDispersalSimulator_31__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_19CDispersalSimulator_31__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_13pycoalescence_6necsim_6necsim_19CDispersalSimulator_30__reduce_cython__(((struct __pyx_obj_13pycoalescence_6necsim_6necsim_CDispersalSimulator *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_19CDispersalSimulator_30__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_13pycoalescence_6necsim_6necsim_CDispersalSimulator *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__19, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pycoalescence.necsim.necsim.CDispersalSimulator.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_19CDispersalSimulator_33__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_13pycoalescence_6necsim_6necsim_19CDispersalSimulator_33__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_13pycoalescence_6necsim_6necsim_19CDispersalSimulator_32__setstate_cython__(((struct __pyx_obj_13pycoalescence_6necsim_6necsim_CDispersalSimulator *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_13pycoalescence_6necsim_6necsim_19CDispersalSimulator_32__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_13pycoalescence_6necsim_6necsim_CDispersalSimulator *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__20, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pycoalescence.necsim.necsim.CDispersalSimulator.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "string.from_py":13
 * 
 * @cname("__pyx_convert_string_from_py_std__in_string")
 * cdef string __pyx_convert_string_from_py_std__in_string(object o) except *:             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t length = 0
 *     cdef const char* data = __Pyx_PyObject_AsStringAndSize(o, &length)
 */

static std::string __pyx_convert_string_from_py_std__in_string(PyObject *__pyx_v_o) {
  Py_ssize_t __pyx_v_length;
  char const *__pyx_v_data;
  std::string __pyx_r;
  __Pyx_RefNannyDeclarations
  char const *__pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_convert_string_from_py_std__in_string", 0);

  /* "string.from_py":14
 * @cname("__pyx_convert_string_from_py_std__in_string")
 * cdef string __pyx_convert_string_from_py_std__in_string(object o) except *:
 *     cdef Py_ssize_t length = 0             # <<<<<<<<<<<<<<
 *     cdef const char* data = __Pyx_PyObject_AsStringAndSize(o, &length)
 *     return string(data, length)
 */
  __pyx_v_length = 0;

  /* "string.from_py":15
 * cdef string __pyx_convert_string_from_py_std__in_string(object o) except *:
 *     cdef Py_ssize_t length = 0
 *     cdef const char* data = __Pyx_PyObject_AsStringAndSize(o, &length)             # <<<<<<<<<<<<<<
 *     return string(data, length)
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_AsStringAndSize(__pyx_v_o, (&__pyx_v_length)); if (unlikely(__pyx_t_1 == ((char const *)NULL))) __PYX_ERR(1, 15, __pyx_L1_error)
  __pyx_v_data = __pyx_t_1;

  /* "string.from_py":16
 *     cdef Py_ssize_t length = 0
 *     cdef const char* data = __Pyx_PyObject_AsStringAndSize(o, &length)
 *     return string(data, length)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = std::string(__pyx_v_data, __pyx_v_length);
  goto __pyx_L0;

  /* "string.from_py":13
 * 
 * @cname("__pyx_convert_string_from_py_std__in_string")
 * cdef string __pyx_convert_string_from_py_std__in_string(object o) except *:             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t length = 0
 *     cdef const char* data = __Pyx_PyObject_AsStringAndSize(o, &length)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("string.from_py.__pyx_convert_string_from_py_std__in_string", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_pretend_to_initialize(&__pyx_r);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "vector.from_py":45
 * 
 * @cname("__pyx_convert_vector_from_py_double")
 * cdef vector[X] __pyx_convert_vector_from_py_double(object o) except *:             # <<<<<<<<<<<<<<
 *     cdef vector[X] v
 *     for item in o:
 */

static std::vector<double>  __pyx_convert_vector_from_py_double(PyObject *__pyx_v_o) {
  std::vector<double>  __pyx_v_v;
  PyObject *__pyx_v_item = NULL;
  std::vector<double>  __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *(*__pyx_t_3)(PyObject *);
  PyObject *__pyx_t_4 = NULL;
  double __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_convert_vector_from_py_double", 0);

  /* "vector.from_py":47
 * cdef vector[X] __pyx_convert_vector_from_py_double(object o) except *:
 *     cdef vector[X] v
 *     for item in o:             # <<<<<<<<<<<<<<
 *         v.push_back(<X>item)
 *     return v
 */
  if (likely(PyList_CheckExact(__pyx_v_o)) || PyTuple_CheckExact(__pyx_v_o)) {
    __pyx_t_1 = __pyx_v_o; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
    __pyx_t_3 = NULL;
  } else {
    __pyx_t_2 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_v_o); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 47, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 47, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_3)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(1, 47, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 47, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      } else {
        if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(1, 47, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 47, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      }
    } else {
      __pyx_t_4 = __pyx_t_3(__pyx_t_1);
      if (unlikely(!__pyx_t_4)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(1, 47, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_4);
    }
    __Pyx_XDECREF_SET(__pyx_v_item, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "vector.from_py":48
 *     cdef vector[X] v
 *     for item in o:
 *         v.push_back(<X>item)             # <<<<<<<<<<<<<<
 *     return v
 * 
 */
    __pyx_t_5 = __pyx_PyFloat_AsDouble(__pyx_v_item); if (unlikely((__pyx_t_5 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 48, __pyx_L1_error)
    __pyx_v_v.push_back(((double)__pyx_t_5));

    /* "vector.from_py":47
 * cdef vector[X] __pyx_convert_vector_from_py_double(object o) except *:
 *     cdef vector[X] v
 *     for item in o:             # <<<<<<<<<<<<<<
 *         v.push_back(<X>item)
 *     return v
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "vector.from_py":49
 *     for item in o:
 *         v.push_back(<X>item)
 *     return v             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_v;
  goto __pyx_L0;

  /* "vector.from_py":45
 * 
 * @cname("__pyx_convert_vector_from_py_double")
 * cdef vector[X] __pyx_convert_vector_from_py_double(object o) except *:             # <<<<<<<<<<<<<<
 *     cdef vector[X] v
 *     for item in o:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("vector.from_py.__pyx_convert_vector_from_py_double", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_pretend_to_initialize(&__pyx_r);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_item);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static std::vector<std::string>  __pyx_convert_vector_from_py_std_3a__3a_string(PyObject *__pyx_v_o) {
  std::vector<std::string>  __pyx_v_v;
  PyObject *__pyx_v_item = NULL;
  std::vector<std::string>  __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *(*__pyx_t_3)(PyObject *);
  PyObject *__pyx_t_4 = NULL;
  std::string __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_convert_vector_from_py_std_3a__3a_string", 0);

  /* "vector.from_py":47
 * cdef vector[X] __pyx_convert_vector_from_py_std_3a__3a_string(object o) except *:
 *     cdef vector[X] v
 *     for item in o:             # <<<<<<<<<<<<<<
 *         v.push_back(<X>item)
 *     return v
 */
  if (likely(PyList_CheckExact(__pyx_v_o)) || PyTuple_CheckExact(__pyx_v_o)) {
    __pyx_t_1 = __pyx_v_o; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
    __pyx_t_3 = NULL;
  } else {
    __pyx_t_2 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_v_o); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 47, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 47, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_3)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(1, 47, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 47, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      } else {
        if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(1, 47, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 47, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      }
    } else {
      __pyx_t_4 = __pyx_t_3(__pyx_t_1);
      if (unlikely(!__pyx_t_4)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(1, 47, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_4);
    }
    __Pyx_XDECREF_SET(__pyx_v_item, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "vector.from_py":48
 *     cdef vector[X] v
 *     for item in o:
 *         v.push_back(<X>item)             # <<<<<<<<<<<<<<
 *     return v
 * 
 */
    __pyx_t_5 = __pyx_convert_string_from_py_std__in_string(__pyx_v_item); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 48, __pyx_L1_error)
    __pyx_v_v.push_back(((std::string)__pyx_t_5));

    /* "vector.from_py":47
 * cdef vector[X] __pyx_convert_vector_from_py_std_3a__3a_string(object o) except *:
 *     cdef vector[X] v
 *     for item in o:             # <<<<<<<<<<<<<<
 *         v.push_back(<X>item)
 *     return v
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "vector.from_py":49
 *     for item in o:
 *         v.push_back(<X>item)
 *     return v             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_v;
  goto __pyx_L0;

  /* "vector.from_py":45
 * 
 * @cname("__pyx_convert_vector_from_py_std_3a__3a_string")
 * cdef vector[X] __pyx_convert_vector_from_py_std_3a__3a_string(object o) except *:             # <<<<<<<<<<<<<<
 *     cdef vector[X] v
 *     for item in o:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("vector.from_py.__pyx_convert_vector_from_py_std_3a__3a_string", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_pretend_to_initialize(&__pyx_r);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_item);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static std::vector<unsigned long>  __pyx_convert_vector_from_py_unsigned_long(PyObject *__pyx_v_o) {
  std::vector<unsigned long>  __pyx_v_v;
  PyObject *__pyx_v_item = NULL;
  std::vector<unsigned long>  __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *(*__pyx_t_3)(PyObject *);
  PyObject *__pyx_t_4 = NULL;
  unsigned long __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_convert_vector_from_py_unsigned_long", 0);

  /* "vector.from_py":47
 * cdef vector[X] __pyx_convert_vector_from_py_unsigned_long(object o) except *:
 *     cdef vector[X] v
 *     for item in o:             # <<<<<<<<<<<<<<
 *         v.push_back(<X>item)
 *     return v
 */
  if (likely(PyList_CheckExact(__pyx_v_o)) || PyTuple_CheckExact(__pyx_v_o)) {
    __pyx_t_1 = __pyx_v_o; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
    __pyx_t_3 = NULL;
  } else {
    __pyx_t_2 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_v_o); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 47, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 47, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_3)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(1, 47, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 47, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      } else {
        if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(1, 47, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 47, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      }
    } else {
      __pyx_t_4 = __pyx_t_3(__pyx_t_1);
      if (unlikely(!__pyx_t_4)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(1, 47, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_4);
    }
    __Pyx_XDECREF_SET(__pyx_v_item, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "vector.from_py":48
 *     cdef vector[X] v
 *     for item in o:
 *         v.push_back(<X>item)             # <<<<<<<<<<<<<<
 *     return v
 * 
 */
    __pyx_t_5 = __Pyx_PyInt_As_unsigned_long(__pyx_v_item); if (unlikely((__pyx_t_5 == (unsigned long)-1) && PyErr_Occurred())) __PYX_ERR(1, 48, __pyx_L1_error)
    __pyx_v_v.push_back(((unsigned long)__pyx_t_5));

    /* "vector.from_py":47
 * cdef vector[X] __pyx_convert_vector_from_py_unsigned_long(object o) except *:
 *     cdef vector[X] v
 *     for item in o:             # <<<<<<<<<<<<<<
 *         v.push_back(<X>item)
 *     return v
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "vector.from_py":49
 *     for item in o:
 *         v.push_back(<X>item)
 *     return v             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_v;
  goto __pyx_L0;

  /* "vector.from_py":45
 * 
 * @cname("__pyx_convert_vector_from_py_unsigned_long")
 * cdef vector[X] __pyx_convert_vector_from_py_unsigned_long(object o) except *:             # <<<<<<<<<<<<<<
 *     cdef vector[X] v
 *     for item in o:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("vector.from_py.__pyx_convert_vector_from_py_unsigned_long", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_pretend_to_initialize(&__pyx_r);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_item);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static std::vector<long>  __pyx_convert_vector_from_py_long(PyObject *__pyx_v_o) {
  std::vector<long>  __pyx_v_v;
  PyObject *__pyx_v_item = NULL;
  std::vector<long>  __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *(*__pyx_t_3)(PyObject *);
  PyObject *__pyx_t_4 = NULL;
  long __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_convert_vector_from_py_long", 0);

  /* "vector.from_py":47
 * cdef vector[X] __pyx_convert_vector_from_py_long(object o) except *:
 *     cdef vector[X] v
 *     for item in o:             # <<<<<<<<<<<<<<
 *         v.push_back(<X>item)
 *     return v
 */
  if (likely(PyList_CheckExact(__pyx_v_o)) || PyTuple_CheckExact(__pyx_v_o)) {
    __pyx_t_1 = __pyx_v_o; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
    __pyx_t_3 = NULL;
  } else {
    __pyx_t_2 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_v_o); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 47, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 47, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_3)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(1, 47, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 47, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      } else {
        if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(1, 47, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 47, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      }
    } else {
      __pyx_t_4 = __pyx_t_3(__pyx_t_1);
      if (unlikely(!__pyx_t_4)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(1, 47, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_4);
    }
    __Pyx_XDECREF_SET(__pyx_v_item, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "vector.from_py":48
 *     cdef vector[X] v
 *     for item in o:
 *         v.push_back(<X>item)             # <<<<<<<<<<<<<<
 *     return v
 * 
 */
    __pyx_t_5 = __Pyx_PyInt_As_long(__pyx_v_item); if (unlikely((__pyx_t_5 == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 48, __pyx_L1_error)
    __pyx_v_v.push_back(((long)__pyx_t_5));

    /* "vector.from_py":47
 * cdef vector[X] __pyx_convert_vector_from_py_long(object o) except *:
 *     cdef vector[X] v
 *     for item in o:             # <<<<<<<<<<<<<<
 *         v.push_back(<X>item)
 *     return v
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "vector.from_py":49
 *     for item in o:
 *         v.push_back(<X>item)
 *     return v             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_v;
  goto __pyx_L0;

  /* "vector.from_py":45
 * 
 * @cname("__pyx_convert_vector_from_py_long")
 * cdef vector[X] __pyx_convert_vector_from_py_long(object o) except *:             # <<<<<<<<<<<<<<
 *     cdef vector[X] v
 *     for item in o:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("vector.from_py.__pyx_convert_vector_from_py_long", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_pretend_to_initialize(&__pyx_r);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_item);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static std::vector<long double>  __pyx_convert_vector_from_py_long__double(PyObject *__pyx_v_o) {
  std::vector<long double>  __pyx_v_v;
  PyObject *__pyx_v_item = NULL;
  std::vector<long double>  __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *(*__pyx_t_3)(PyObject *);
  PyObject *__pyx_t_4 = NULL;
  long double __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_convert_vector_from_py_long__double", 0);

  /* "vector.from_py":47
 * cdef vector[X] __pyx_convert_vector_from_py_long__double(object o) except *:
 *     cdef vector[X] v
 *     for item in o:             # <<<<<<<<<<<<<<
 *         v.push_back(<X>item)
 *     return v
 */
  if (likely(PyList_CheckExact(__pyx_v_o)) || PyTuple_CheckExact(__pyx_v_o)) {
    __pyx_t_1 = __pyx_v_o; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
    __pyx_t_3 = NULL;
  } else {
    __pyx_t_2 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_v_o); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 47, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 47, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_3)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(1, 47, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 47, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      } else {
        if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(1, 47, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 47, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      }
    } else {
      __pyx_t_4 = __pyx_t_3(__pyx_t_1);
      if (unlikely(!__pyx_t_4)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(1, 47, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_4);
    }
    __Pyx_XDECREF_SET(__pyx_v_item, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "vector.from_py":48
 *     cdef vector[X] v
 *     for item in o:
 *         v.push_back(<X>item)             # <<<<<<<<<<<<<<
 *     return v
 * 
 */
    __pyx_t_5 = __pyx_PyFloat_AsDouble(__pyx_v_item); if (unlikely((__pyx_t_5 == (long double)-1) && PyErr_Occurred())) __PYX_ERR(1, 48, __pyx_L1_error)
    __pyx_v_v.push_back(((long double)__pyx_t_5));

    /* "vector.from_py":47
 * cdef vector[X] __pyx_convert_vector_from_py_long__double(object o) except *:
 *     cdef vector[X] v
 *     for item in o:             # <<<<<<<<<<<<<<
 *         v.push_back(<X>item)
 *     return v
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "vector.from_py":49
 *     for item in o:
 *         v.push_back(<X>item)
 *     return v             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_v;
  goto __pyx_L0;

  /* "vector.from_py":45
 * 
 * @cname("__pyx_convert_vector_from_py_long__double")
 * cdef vector[X] __pyx_convert_vector_from_py_long__double(object o) except *:             # <<<<<<<<<<<<<<
 *     cdef vector[X] v
 *     for item in o:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("vector.from_py.__pyx_convert_vector_from_py_long__double", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_pretend_to_initialize(&__pyx_r);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_item);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "vector.to_py":60
 * 
 * @cname("__pyx_convert_vector_to_py_unsigned_long")
 * cdef object __pyx_convert_vector_to_py_unsigned_long(vector[X]& v):             # <<<<<<<<<<<<<<
 *     return [v[i] for i in range(v.size())]
 * 
 */

static PyObject *__pyx_convert_vector_to_py_unsigned_long(const std::vector<unsigned long>  &__pyx_v_v) {
  size_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  size_t __pyx_t_2;
  size_t __pyx_t_3;
  size_t __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_convert_vector_to_py_unsigned_long", 0);

  /* "vector.to_py":61
 * @cname("__pyx_convert_vector_to_py_unsigned_long")
 * cdef object __pyx_convert_vector_to_py_unsigned_long(vector[X]& v):
 *     return [v[i] for i in range(v.size())]             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 61, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __pyx_v_v.size();
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;
    __pyx_t_5 = __Pyx_PyInt_From_unsigned_long((__pyx_v_v[__pyx_v_i])); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 61, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_5))) __PYX_ERR(1, 61, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "vector.to_py":60
 * 
 * @cname("__pyx_convert_vector_to_py_unsigned_long")
 * cdef object __pyx_convert_vector_to_py_unsigned_long(vector[X]& v):             # <<<<<<<<<<<<<<
 *     return [v[i] for i in range(v.size())]
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("vector.to_py.__pyx_convert_vector_to_py_unsigned_long", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_tp_new_13pycoalescence_6necsim_6necsim_CSpatialSimulation(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_13pycoalescence_6necsim_6necsim_CSpatialSimulation *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_13pycoalescence_6necsim_6necsim_CSpatialSimulation *)o);
  new((void*)&(p->c_simulation)) necsim::GenericTree<necsim::SpatialTree> ();
  p->logger = Py_None; Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_13pycoalescence_6necsim_6necsim_18CSpatialSimulation_1__cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_13pycoalescence_6necsim_6necsim_CSpatialSimulation(PyObject *o) {
  struct __pyx_obj_13pycoalescence_6necsim_6necsim_CSpatialSimulation *p = (struct __pyx_obj_13pycoalescence_6necsim_6necsim_CSpatialSimulation *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  __Pyx_call_destructor(p->c_simulation);
  Py_CLEAR(p->logger);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_13pycoalescence_6necsim_6necsim_CSpatialSimulation(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_13pycoalescence_6necsim_6necsim_CSpatialSimulation *p = (struct __pyx_obj_13pycoalescence_6necsim_6necsim_CSpatialSimulation *)o;
  if (p->logger) {
    e = (*v)(p->logger, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_13pycoalescence_6necsim_6necsim_CSpatialSimulation(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_13pycoalescence_6necsim_6necsim_CSpatialSimulation *p = (struct __pyx_obj_13pycoalescence_6necsim_6necsim_CSpatialSimulation *)o;
  tmp = ((PyObject*)p->logger);
  p->logger = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyMethodDef __pyx_methods_13pycoalescence_6necsim_6necsim_CSpatialSimulation[] = {
  {"set_logger", (PyCFunction)__pyx_pw_13pycoalescence_6necsim_6necsim_18CSpatialSimulation_3set_logger, METH_NOARGS, 0},
  {"import_from_config", (PyCFunction)__pyx_pw_13pycoalescence_6necsim_6necsim_18CSpatialSimulation_5import_from_config, METH_O, 0},
  {"import_from_config_string", (PyCFunction)__pyx_pw_13pycoalescence_6necsim_6necsim_18CSpatialSimulation_7import_from_config_string, METH_O, 0},
  {"setup", (PyCFunction)__pyx_pw_13pycoalescence_6necsim_6necsim_18CSpatialSimulation_9setup, METH_NOARGS, 0},
  {"setup_resume", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_13pycoalescence_6necsim_6necsim_18CSpatialSimulation_11setup_resume, METH_VARARGS|METH_KEYWORDS, 0},
  {"add_gillespie", (PyCFunction)__pyx_pw_13pycoalescence_6necsim_6necsim_18CSpatialSimulation_13add_gillespie, METH_O, 0},
  {"run", (PyCFunction)__pyx_pw_13pycoalescence_6necsim_6necsim_18CSpatialSimulation_15run, METH_NOARGS, 0},
  {"apply_speciation_rates", (PyCFunction)__pyx_pw_13pycoalescence_6necsim_6necsim_18CSpatialSimulation_17apply_speciation_rates, METH_O, 0},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_13pycoalescence_6necsim_6necsim_18CSpatialSimulation_19__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_13pycoalescence_6necsim_6necsim_18CSpatialSimulation_21__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_13pycoalescence_6necsim_6necsim_CSpatialSimulation = {
  PyVarObject_HEAD_INIT(0, 0)
  "pycoalescence.necsim.necsim.CSpatialSimulation", /*tp_name*/
  sizeof(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CSpatialSimulation), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_13pycoalescence_6necsim_6necsim_CSpatialSimulation, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_13pycoalescence_6necsim_6necsim_CSpatialSimulation, /*tp_traverse*/
  __pyx_tp_clear_13pycoalescence_6necsim_6necsim_CSpatialSimulation, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_13pycoalescence_6necsim_6necsim_CSpatialSimulation, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_13pycoalescence_6necsim_6necsim_CSpatialSimulation, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static PyObject *__pyx_tp_new_13pycoalescence_6necsim_6necsim_CNSESimulation(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_13pycoalescence_6necsim_6necsim_CNSESimulation *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_13pycoalescence_6necsim_6necsim_CNSESimulation *)o);
  new((void*)&(p->c_simulation)) necsim::GenericTree<necsim::Tree> ();
  p->logger = Py_None; Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_13pycoalescence_6necsim_6necsim_14CNSESimulation_1__cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_13pycoalescence_6necsim_6necsim_CNSESimulation(PyObject *o) {
  struct __pyx_obj_13pycoalescence_6necsim_6necsim_CNSESimulation *p = (struct __pyx_obj_13pycoalescence_6necsim_6necsim_CNSESimulation *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  __Pyx_call_destructor(p->c_simulation);
  Py_CLEAR(p->logger);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_13pycoalescence_6necsim_6necsim_CNSESimulation(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_13pycoalescence_6necsim_6necsim_CNSESimulation *p = (struct __pyx_obj_13pycoalescence_6necsim_6necsim_CNSESimulation *)o;
  if (p->logger) {
    e = (*v)(p->logger, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_13pycoalescence_6necsim_6necsim_CNSESimulation(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_13pycoalescence_6necsim_6necsim_CNSESimulation *p = (struct __pyx_obj_13pycoalescence_6necsim_6necsim_CNSESimulation *)o;
  tmp = ((PyObject*)p->logger);
  p->logger = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyMethodDef __pyx_methods_13pycoalescence_6necsim_6necsim_CNSESimulation[] = {
  {"set_logger", (PyCFunction)__pyx_pw_13pycoalescence_6necsim_6necsim_14CNSESimulation_3set_logger, METH_NOARGS, 0},
  {"import_from_config", (PyCFunction)__pyx_pw_13pycoalescence_6necsim_6necsim_14CNSESimulation_5import_from_config, METH_O, 0},
  {"import_from_config_string", (PyCFunction)__pyx_pw_13pycoalescence_6necsim_6necsim_14CNSESimulation_7import_from_config_string, METH_O, 0},
  {"setup", (PyCFunction)__pyx_pw_13pycoalescence_6necsim_6necsim_14CNSESimulation_9setup, METH_NOARGS, 0},
  {"setup_resume", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_13pycoalescence_6necsim_6necsim_14CNSESimulation_11setup_resume, METH_VARARGS|METH_KEYWORDS, 0},
  {"add_gillespie", (PyCFunction)__pyx_pw_13pycoalescence_6necsim_6necsim_14CNSESimulation_13add_gillespie, METH_O, 0},
  {"run", (PyCFunction)__pyx_pw_13pycoalescence_6necsim_6necsim_14CNSESimulation_15run, METH_NOARGS, 0},
  {"apply_speciation_rates", (PyCFunction)__pyx_pw_13pycoalescence_6necsim_6necsim_14CNSESimulation_17apply_speciation_rates, METH_O, 0},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_13pycoalescence_6necsim_6necsim_14CNSESimulation_19__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_13pycoalescence_6necsim_6necsim_14CNSESimulation_21__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_13pycoalescence_6necsim_6necsim_CNSESimulation = {
  PyVarObject_HEAD_INIT(0, 0)
  "pycoalescence.necsim.necsim.CNSESimulation", /*tp_name*/
  sizeof(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CNSESimulation), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_13pycoalescence_6necsim_6necsim_CNSESimulation, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_13pycoalescence_6necsim_6necsim_CNSESimulation, /*tp_traverse*/
  __pyx_tp_clear_13pycoalescence_6necsim_6necsim_CNSESimulation, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_13pycoalescence_6necsim_6necsim_CNSESimulation, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_13pycoalescence_6necsim_6necsim_CNSESimulation, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static PyObject *__pyx_tp_new_13pycoalescence_6necsim_6necsim_CPNSESimulation(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_13pycoalescence_6necsim_6necsim_CPNSESimulation *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_13pycoalescence_6necsim_6necsim_CPNSESimulation *)o);
  new((void*)&(p->c_simulation)) necsim::GenericTree<necsim::ProtractedTree> ();
  p->logger = Py_None; Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_13pycoalescence_6necsim_6necsim_15CPNSESimulation_1__cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_13pycoalescence_6necsim_6necsim_CPNSESimulation(PyObject *o) {
  struct __pyx_obj_13pycoalescence_6necsim_6necsim_CPNSESimulation *p = (struct __pyx_obj_13pycoalescence_6necsim_6necsim_CPNSESimulation *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  __Pyx_call_destructor(p->c_simulation);
  Py_CLEAR(p->logger);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_13pycoalescence_6necsim_6necsim_CPNSESimulation(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_13pycoalescence_6necsim_6necsim_CPNSESimulation *p = (struct __pyx_obj_13pycoalescence_6necsim_6necsim_CPNSESimulation *)o;
  if (p->logger) {
    e = (*v)(p->logger, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_13pycoalescence_6necsim_6necsim_CPNSESimulation(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_13pycoalescence_6necsim_6necsim_CPNSESimulation *p = (struct __pyx_obj_13pycoalescence_6necsim_6necsim_CPNSESimulation *)o;
  tmp = ((PyObject*)p->logger);
  p->logger = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyMethodDef __pyx_methods_13pycoalescence_6necsim_6necsim_CPNSESimulation[] = {
  {"set_logger", (PyCFunction)__pyx_pw_13pycoalescence_6necsim_6necsim_15CPNSESimulation_3set_logger, METH_NOARGS, 0},
  {"import_from_config", (PyCFunction)__pyx_pw_13pycoalescence_6necsim_6necsim_15CPNSESimulation_5import_from_config, METH_O, 0},
  {"import_from_config_string", (PyCFunction)__pyx_pw_13pycoalescence_6necsim_6necsim_15CPNSESimulation_7import_from_config_string, METH_O, 0},
  {"setup", (PyCFunction)__pyx_pw_13pycoalescence_6necsim_6necsim_15CPNSESimulation_9setup, METH_NOARGS, 0},
  {"setup_resume", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_13pycoalescence_6necsim_6necsim_15CPNSESimulation_11setup_resume, METH_VARARGS|METH_KEYWORDS, 0},
  {"add_gillespie", (PyCFunction)__pyx_pw_13pycoalescence_6necsim_6necsim_15CPNSESimulation_13add_gillespie, METH_O, 0},
  {"run", (PyCFunction)__pyx_pw_13pycoalescence_6necsim_6necsim_15CPNSESimulation_15run, METH_NOARGS, 0},
  {"apply_speciation_rates", (PyCFunction)__pyx_pw_13pycoalescence_6necsim_6necsim_15CPNSESimulation_17apply_speciation_rates, METH_O, 0},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_13pycoalescence_6necsim_6necsim_15CPNSESimulation_19__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_13pycoalescence_6necsim_6necsim_15CPNSESimulation_21__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_13pycoalescence_6necsim_6necsim_CPNSESimulation = {
  PyVarObject_HEAD_INIT(0, 0)
  "pycoalescence.necsim.necsim.CPNSESimulation", /*tp_name*/
  sizeof(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CPNSESimulation), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_13pycoalescence_6necsim_6necsim_CPNSESimulation, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_13pycoalescence_6necsim_6necsim_CPNSESimulation, /*tp_traverse*/
  __pyx_tp_clear_13pycoalescence_6necsim_6necsim_CPNSESimulation, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_13pycoalescence_6necsim_6necsim_CPNSESimulation, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_13pycoalescence_6necsim_6necsim_CPNSESimulation, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static PyObject *__pyx_tp_new_13pycoalescence_6necsim_6necsim_CPSpatialSimulation(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_13pycoalescence_6necsim_6necsim_CPSpatialSimulation *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_13pycoalescence_6necsim_6necsim_CPSpatialSimulation *)o);
  new((void*)&(p->c_simulation)) necsim::GenericTree<necsim::ProtractedSpatialTree> ();
  p->logger = Py_None; Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_13pycoalescence_6necsim_6necsim_19CPSpatialSimulation_1__cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_13pycoalescence_6necsim_6necsim_CPSpatialSimulation(PyObject *o) {
  struct __pyx_obj_13pycoalescence_6necsim_6necsim_CPSpatialSimulation *p = (struct __pyx_obj_13pycoalescence_6necsim_6necsim_CPSpatialSimulation *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  __Pyx_call_destructor(p->c_simulation);
  Py_CLEAR(p->logger);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_13pycoalescence_6necsim_6necsim_CPSpatialSimulation(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_13pycoalescence_6necsim_6necsim_CPSpatialSimulation *p = (struct __pyx_obj_13pycoalescence_6necsim_6necsim_CPSpatialSimulation *)o;
  if (p->logger) {
    e = (*v)(p->logger, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_13pycoalescence_6necsim_6necsim_CPSpatialSimulation(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_13pycoalescence_6necsim_6necsim_CPSpatialSimulation *p = (struct __pyx_obj_13pycoalescence_6necsim_6necsim_CPSpatialSimulation *)o;
  tmp = ((PyObject*)p->logger);
  p->logger = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyMethodDef __pyx_methods_13pycoalescence_6necsim_6necsim_CPSpatialSimulation[] = {
  {"set_logger", (PyCFunction)__pyx_pw_13pycoalescence_6necsim_6necsim_19CPSpatialSimulation_3set_logger, METH_NOARGS, 0},
  {"import_from_config", (PyCFunction)__pyx_pw_13pycoalescence_6necsim_6necsim_19CPSpatialSimulation_5import_from_config, METH_O, 0},
  {"import_from_config_string", (PyCFunction)__pyx_pw_13pycoalescence_6necsim_6necsim_19CPSpatialSimulation_7import_from_config_string, METH_O, 0},
  {"setup", (PyCFunction)__pyx_pw_13pycoalescence_6necsim_6necsim_19CPSpatialSimulation_9setup, METH_NOARGS, 0},
  {"setup_resume", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_13pycoalescence_6necsim_6necsim_19CPSpatialSimulation_11setup_resume, METH_VARARGS|METH_KEYWORDS, 0},
  {"add_gillespie", (PyCFunction)__pyx_pw_13pycoalescence_6necsim_6necsim_19CPSpatialSimulation_13add_gillespie, METH_O, 0},
  {"run", (PyCFunction)__pyx_pw_13pycoalescence_6necsim_6necsim_19CPSpatialSimulation_15run, METH_NOARGS, 0},
  {"apply_speciation_rates", (PyCFunction)__pyx_pw_13pycoalescence_6necsim_6necsim_19CPSpatialSimulation_17apply_speciation_rates, METH_O, 0},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_13pycoalescence_6necsim_6necsim_19CPSpatialSimulation_19__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_13pycoalescence_6necsim_6necsim_19CPSpatialSimulation_21__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_13pycoalescence_6necsim_6necsim_CPSpatialSimulation = {
  PyVarObject_HEAD_INIT(0, 0)
  "pycoalescence.necsim.necsim.CPSpatialSimulation", /*tp_name*/
  sizeof(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CPSpatialSimulation), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_13pycoalescence_6necsim_6necsim_CPSpatialSimulation, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_13pycoalescence_6necsim_6necsim_CPSpatialSimulation, /*tp_traverse*/
  __pyx_tp_clear_13pycoalescence_6necsim_6necsim_CPSpatialSimulation, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_13pycoalescence_6necsim_6necsim_CPSpatialSimulation, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_13pycoalescence_6necsim_6necsim_CPSpatialSimulation, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static PyObject *__pyx_tp_new_13pycoalescence_6necsim_6necsim_CCommunity(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_13pycoalescence_6necsim_6necsim_CCommunity *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_13pycoalescence_6necsim_6necsim_CCommunity *)o);
  new((void*)&(p->c_community)) necsim::Community();
  new((void*)&(p->c_spec_parameters)) std::shared_ptr<struct necsim::SpecSimParameters> ();
  p->logger = Py_None; Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_13pycoalescence_6necsim_6necsim_10CCommunity_1__cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_13pycoalescence_6necsim_6necsim_CCommunity(PyObject *o) {
  struct __pyx_obj_13pycoalescence_6necsim_6necsim_CCommunity *p = (struct __pyx_obj_13pycoalescence_6necsim_6necsim_CCommunity *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  __Pyx_call_destructor(p->c_community);
  __Pyx_call_destructor(p->c_spec_parameters);
  Py_CLEAR(p->logger);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_13pycoalescence_6necsim_6necsim_CCommunity(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_13pycoalescence_6necsim_6necsim_CCommunity *p = (struct __pyx_obj_13pycoalescence_6necsim_6necsim_CCommunity *)o;
  if (p->logger) {
    e = (*v)(p->logger, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_13pycoalescence_6necsim_6necsim_CCommunity(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_13pycoalescence_6necsim_6necsim_CCommunity *p = (struct __pyx_obj_13pycoalescence_6necsim_6necsim_CCommunity *)o;
  tmp = ((PyObject*)p->logger);
  p->logger = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyMethodDef __pyx_methods_13pycoalescence_6necsim_6necsim_CCommunity[] = {
  {"set_logger", (PyCFunction)__pyx_pw_13pycoalescence_6necsim_6necsim_10CCommunity_3set_logger, METH_NOARGS, 0},
  {"wipe_protracted_parameters", (PyCFunction)__pyx_pw_13pycoalescence_6necsim_6necsim_10CCommunity_5wipe_protracted_parameters, METH_NOARGS, 0},
  {"add_protracted_parameters", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_13pycoalescence_6necsim_6necsim_10CCommunity_7add_protracted_parameters, METH_VARARGS|METH_KEYWORDS, 0},
  {"add_metacommunity_parameters", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_13pycoalescence_6necsim_6necsim_10CCommunity_9add_metacommunity_parameters, METH_VARARGS|METH_KEYWORDS, 0},
  {"reset", (PyCFunction)__pyx_pw_13pycoalescence_6necsim_6necsim_10CCommunity_11reset, METH_NOARGS, 0},
  {"setup", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_13pycoalescence_6necsim_6necsim_10CCommunity_13setup, METH_VARARGS|METH_KEYWORDS, 0},
  {"add_time", (PyCFunction)__pyx_pw_13pycoalescence_6necsim_6necsim_10CCommunity_15add_time, METH_O, 0},
  {"apply", (PyCFunction)__pyx_pw_13pycoalescence_6necsim_6necsim_10CCommunity_17apply, METH_NOARGS, 0},
  {"output", (PyCFunction)__pyx_pw_13pycoalescence_6necsim_6necsim_10CCommunity_19output, METH_NOARGS, 0},
  {"speciate_remaining_lineages", (PyCFunction)__pyx_pw_13pycoalescence_6necsim_6necsim_10CCommunity_21speciate_remaining_lineages, METH_O, 0},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_13pycoalescence_6necsim_6necsim_10CCommunity_23__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_13pycoalescence_6necsim_6necsim_10CCommunity_25__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_13pycoalescence_6necsim_6necsim_CCommunity = {
  PyVarObject_HEAD_INIT(0, 0)
  "pycoalescence.necsim.necsim.CCommunity", /*tp_name*/
  sizeof(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CCommunity), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_13pycoalescence_6necsim_6necsim_CCommunity, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_13pycoalescence_6necsim_6necsim_CCommunity, /*tp_traverse*/
  __pyx_tp_clear_13pycoalescence_6necsim_6necsim_CCommunity, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_13pycoalescence_6necsim_6necsim_CCommunity, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_13pycoalescence_6necsim_6necsim_CCommunity, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static PyObject *__pyx_tp_new_13pycoalescence_6necsim_6necsim_CMetacommunity(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_13pycoalescence_6necsim_6necsim_CMetacommunity *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_13pycoalescence_6necsim_6necsim_CMetacommunity *)o);
  new((void*)&(p->c_community)) necsim::Metacommunity();
  new((void*)&(p->c_spec_parameters)) std::shared_ptr<struct necsim::SpecSimParameters> ();
  p->logger = Py_None; Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_13pycoalescence_6necsim_6necsim_14CMetacommunity_1__cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_13pycoalescence_6necsim_6necsim_CMetacommunity(PyObject *o) {
  struct __pyx_obj_13pycoalescence_6necsim_6necsim_CMetacommunity *p = (struct __pyx_obj_13pycoalescence_6necsim_6necsim_CMetacommunity *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  __Pyx_call_destructor(p->c_community);
  __Pyx_call_destructor(p->c_spec_parameters);
  Py_CLEAR(p->logger);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_13pycoalescence_6necsim_6necsim_CMetacommunity(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_13pycoalescence_6necsim_6necsim_CMetacommunity *p = (struct __pyx_obj_13pycoalescence_6necsim_6necsim_CMetacommunity *)o;
  if (p->logger) {
    e = (*v)(p->logger, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_13pycoalescence_6necsim_6necsim_CMetacommunity(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_13pycoalescence_6necsim_6necsim_CMetacommunity *p = (struct __pyx_obj_13pycoalescence_6necsim_6necsim_CMetacommunity *)o;
  tmp = ((PyObject*)p->logger);
  p->logger = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyMethodDef __pyx_methods_13pycoalescence_6necsim_6necsim_CMetacommunity[] = {
  {"set_logger", (PyCFunction)__pyx_pw_13pycoalescence_6necsim_6necsim_14CMetacommunity_3set_logger, METH_NOARGS, 0},
  {"wipe_protracted_parameters", (PyCFunction)__pyx_pw_13pycoalescence_6necsim_6necsim_14CMetacommunity_5wipe_protracted_parameters, METH_NOARGS, 0},
  {"add_protracted_parameters", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_13pycoalescence_6necsim_6necsim_14CMetacommunity_7add_protracted_parameters, METH_VARARGS|METH_KEYWORDS, 0},
  {"add_metacommunity_parameters", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_13pycoalescence_6necsim_6necsim_14CMetacommunity_9add_metacommunity_parameters, METH_VARARGS|METH_KEYWORDS, 0},
  {"reset", (PyCFunction)__pyx_pw_13pycoalescence_6necsim_6necsim_14CMetacommunity_11reset, METH_NOARGS, 0},
  {"setup", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_13pycoalescence_6necsim_6necsim_14CMetacommunity_13setup, METH_VARARGS|METH_KEYWORDS, 0},
  {"add_time", (PyCFunction)__pyx_pw_13pycoalescence_6necsim_6necsim_14CMetacommunity_15add_time, METH_O, 0},
  {"apply", (PyCFunction)__pyx_pw_13pycoalescence_6necsim_6necsim_14CMetacommunity_17apply, METH_NOARGS, 0},
  {"output", (PyCFunction)__pyx_pw_13pycoalescence_6necsim_6necsim_14CMetacommunity_19output, METH_NOARGS, 0},
  {"speciate_remaining_lineages", (PyCFunction)__pyx_pw_13pycoalescence_6necsim_6necsim_14CMetacommunity_21speciate_remaining_lineages, METH_O, 0},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_13pycoalescence_6necsim_6necsim_14CMetacommunity_23__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_13pycoalescence_6necsim_6necsim_14CMetacommunity_25__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_13pycoalescence_6necsim_6necsim_CMetacommunity = {
  PyVarObject_HEAD_INIT(0, 0)
  "pycoalescence.necsim.necsim.CMetacommunity", /*tp_name*/
  sizeof(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CMetacommunity), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_13pycoalescence_6necsim_6necsim_CMetacommunity, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_13pycoalescence_6necsim_6necsim_CMetacommunity, /*tp_traverse*/
  __pyx_tp_clear_13pycoalescence_6necsim_6necsim_CMetacommunity, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_13pycoalescence_6necsim_6necsim_CMetacommunity, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_13pycoalescence_6necsim_6necsim_CMetacommunity, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static PyObject *__pyx_tp_new_13pycoalescence_6necsim_6necsim_CLandscapeMetricsCalculator(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_13pycoalescence_6necsim_6necsim_CLandscapeMetricsCalculator *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_13pycoalescence_6necsim_6necsim_CLandscapeMetricsCalculator *)o);
  new((void*)&(p->c_landscapes)) LandscapeMetricsCalculator();
  p->logger = Py_None; Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_13pycoalescence_6necsim_6necsim_27CLandscapeMetricsCalculator_1__cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_13pycoalescence_6necsim_6necsim_CLandscapeMetricsCalculator(PyObject *o) {
  struct __pyx_obj_13pycoalescence_6necsim_6necsim_CLandscapeMetricsCalculator *p = (struct __pyx_obj_13pycoalescence_6necsim_6necsim_CLandscapeMetricsCalculator *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  __Pyx_call_destructor(p->c_landscapes);
  Py_CLEAR(p->logger);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_13pycoalescence_6necsim_6necsim_CLandscapeMetricsCalculator(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_13pycoalescence_6necsim_6necsim_CLandscapeMetricsCalculator *p = (struct __pyx_obj_13pycoalescence_6necsim_6necsim_CLandscapeMetricsCalculator *)o;
  if (p->logger) {
    e = (*v)(p->logger, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_13pycoalescence_6necsim_6necsim_CLandscapeMetricsCalculator(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_13pycoalescence_6necsim_6necsim_CLandscapeMetricsCalculator *p = (struct __pyx_obj_13pycoalescence_6necsim_6necsim_CLandscapeMetricsCalculator *)o;
  tmp = ((PyObject*)p->logger);
  p->logger = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyMethodDef __pyx_methods_13pycoalescence_6necsim_6necsim_CLandscapeMetricsCalculator[] = {
  {"set_logger", (PyCFunction)__pyx_pw_13pycoalescence_6necsim_6necsim_27CLandscapeMetricsCalculator_3set_logger, METH_NOARGS, 0},
  {"import_map", (PyCFunction)__pyx_pw_13pycoalescence_6necsim_6necsim_27CLandscapeMetricsCalculator_5import_map, METH_O, 0},
  {"calculate_MNN", (PyCFunction)__pyx_pw_13pycoalescence_6necsim_6necsim_27CLandscapeMetricsCalculator_7calculate_MNN, METH_NOARGS, 0},
  {"calculate_CLUMPY", (PyCFunction)__pyx_pw_13pycoalescence_6necsim_6necsim_27CLandscapeMetricsCalculator_9calculate_CLUMPY, METH_NOARGS, 0},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_13pycoalescence_6necsim_6necsim_27CLandscapeMetricsCalculator_11__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_13pycoalescence_6necsim_6necsim_27CLandscapeMetricsCalculator_13__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_13pycoalescence_6necsim_6necsim_CLandscapeMetricsCalculator = {
  PyVarObject_HEAD_INIT(0, 0)
  "pycoalescence.necsim.necsim.CLandscapeMetricsCalculator", /*tp_name*/
  sizeof(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CLandscapeMetricsCalculator), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_13pycoalescence_6necsim_6necsim_CLandscapeMetricsCalculator, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_13pycoalescence_6necsim_6necsim_CLandscapeMetricsCalculator, /*tp_traverse*/
  __pyx_tp_clear_13pycoalescence_6necsim_6necsim_CLandscapeMetricsCalculator, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_13pycoalescence_6necsim_6necsim_CLandscapeMetricsCalculator, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_13pycoalescence_6necsim_6necsim_CLandscapeMetricsCalculator, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static PyObject *__pyx_tp_new_13pycoalescence_6necsim_6necsim_CDispersalSimulator(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_13pycoalescence_6necsim_6necsim_CDispersalSimulator *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_13pycoalescence_6necsim_6necsim_CDispersalSimulator *)o);
  new((void*)&(p->c_simulator)) necsim::SimulateDispersal();
  new((void*)&(p->c_sim_parameters)) std::shared_ptr<necsim::SimParameters> ();
  p->logger = Py_None; Py_INCREF(Py_None);
  p->output_database = Py_None; Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_13pycoalescence_6necsim_6necsim_19CDispersalSimulator_1__cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_13pycoalescence_6necsim_6necsim_CDispersalSimulator(PyObject *o) {
  struct __pyx_obj_13pycoalescence_6necsim_6necsim_CDispersalSimulator *p = (struct __pyx_obj_13pycoalescence_6necsim_6necsim_CDispersalSimulator *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  __Pyx_call_destructor(p->c_simulator);
  __Pyx_call_destructor(p->c_sim_parameters);
  Py_CLEAR(p->logger);
  Py_CLEAR(p->output_database);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_13pycoalescence_6necsim_6necsim_CDispersalSimulator(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_13pycoalescence_6necsim_6necsim_CDispersalSimulator *p = (struct __pyx_obj_13pycoalescence_6necsim_6necsim_CDispersalSimulator *)o;
  if (p->logger) {
    e = (*v)(p->logger, a); if (e) return e;
  }
  if (p->output_database) {
    e = (*v)(p->output_database, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_13pycoalescence_6necsim_6necsim_CDispersalSimulator(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_13pycoalescence_6necsim_6necsim_CDispersalSimulator *p = (struct __pyx_obj_13pycoalescence_6necsim_6necsim_CDispersalSimulator *)o;
  tmp = ((PyObject*)p->logger);
  p->logger = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->output_database);
  p->output_database = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyMethodDef __pyx_methods_13pycoalescence_6necsim_6necsim_CDispersalSimulator[] = {
  {"set_logger", (PyCFunction)__pyx_pw_13pycoalescence_6necsim_6necsim_19CDispersalSimulator_3set_logger, METH_NOARGS, 0},
  {"set_output_database", (PyCFunction)__pyx_pw_13pycoalescence_6necsim_6necsim_19CDispersalSimulator_5set_output_database, METH_O, 0},
  {"import_maps_base", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_13pycoalescence_6necsim_6necsim_19CDispersalSimulator_7import_maps_base, METH_VARARGS|METH_KEYWORDS, 0},
  {"import_maps", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_13pycoalescence_6necsim_6necsim_19CDispersalSimulator_9import_maps, METH_VARARGS|METH_KEYWORDS, 0},
  {"import_all_maps", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_13pycoalescence_6necsim_6necsim_19CDispersalSimulator_11import_all_maps, METH_VARARGS|METH_KEYWORDS, 0},
  {"check_completed", (PyCFunction)__pyx_pw_13pycoalescence_6necsim_6necsim_19CDispersalSimulator_13check_completed, METH_NOARGS, 0},
  {"setup", (PyCFunction)__pyx_pw_13pycoalescence_6necsim_6necsim_19CDispersalSimulator_15setup, METH_NOARGS, 0},
  {"set_dispersal_parameters", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_13pycoalescence_6necsim_6necsim_19CDispersalSimulator_17set_dispersal_parameters, METH_VARARGS|METH_KEYWORDS, 0},
  {"base_setup", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_13pycoalescence_6necsim_6necsim_19CDispersalSimulator_19base_setup, METH_VARARGS|METH_KEYWORDS, 0},
  {"write_database", (PyCFunction)__pyx_pw_13pycoalescence_6necsim_6necsim_19CDispersalSimulator_21write_database, METH_O, 0},
  {"run_mean_distance_travelled", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_13pycoalescence_6necsim_6necsim_19CDispersalSimulator_23run_mean_distance_travelled, METH_VARARGS|METH_KEYWORDS, 0},
  {"run_all_distance_travelled", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_13pycoalescence_6necsim_6necsim_19CDispersalSimulator_25run_all_distance_travelled, METH_VARARGS|METH_KEYWORDS, 0},
  {"run_sample_distance_travelled", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_13pycoalescence_6necsim_6necsim_19CDispersalSimulator_27run_sample_distance_travelled, METH_VARARGS|METH_KEYWORDS, 0},
  {"run_mean_dispersal_distance", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_13pycoalescence_6necsim_6necsim_19CDispersalSimulator_29run_mean_dispersal_distance, METH_VARARGS|METH_KEYWORDS, 0},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_13pycoalescence_6necsim_6necsim_19CDispersalSimulator_31__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_13pycoalescence_6necsim_6necsim_19CDispersalSimulator_33__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_13pycoalescence_6necsim_6necsim_CDispersalSimulator = {
  PyVarObject_HEAD_INIT(0, 0)
  "pycoalescence.necsim.necsim.CDispersalSimulator", /*tp_name*/
  sizeof(struct __pyx_obj_13pycoalescence_6necsim_6necsim_CDispersalSimulator), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_13pycoalescence_6necsim_6necsim_CDispersalSimulator, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_13pycoalescence_6necsim_6necsim_CDispersalSimulator, /*tp_traverse*/
  __pyx_tp_clear_13pycoalescence_6necsim_6necsim_CDispersalSimulator, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_13pycoalescence_6necsim_6necsim_CDispersalSimulator, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_13pycoalescence_6necsim_6necsim_CDispersalSimulator, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static PyMethodDef __pyx_methods[] = {
  {0, 0, 0, 0}
};

#if PY_MAJOR_VERSION >= 3
#if CYTHON_PEP489_MULTI_PHASE_INIT
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/
static int __pyx_pymod_exec_necsim(PyObject* module); /*proto*/
static PyModuleDef_Slot __pyx_moduledef_slots[] = {
  {Py_mod_create, (void*)__pyx_pymod_create},
  {Py_mod_exec, (void*)__pyx_pymod_exec_necsim},
  {0, NULL}
};
#endif

static struct PyModuleDef __pyx_moduledef = {
    PyModuleDef_HEAD_INIT,
    "necsim",
    0, /* m_doc */
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    0, /* m_size */
  #else
    -1, /* m_size */
  #endif
    __pyx_methods /* m_methods */,
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    __pyx_moduledef_slots, /* m_slots */
  #else
    NULL, /* m_reload */
  #endif
    NULL, /* m_traverse */
    NULL, /* m_clear */
    NULL /* m_free */
};
#endif
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif

static __Pyx_StringTabEntry __pyx_string_tab[] = {
  {&__pyx_n_s_CCommunity, __pyx_k_CCommunity, sizeof(__pyx_k_CCommunity), 0, 0, 1, 1},
  {&__pyx_n_s_CDispersalSimulator, __pyx_k_CDispersalSimulator, sizeof(__pyx_k_CDispersalSimulator), 0, 0, 1, 1},
  {&__pyx_n_s_CLandscapeMetricsCalculator, __pyx_k_CLandscapeMetricsCalculator, sizeof(__pyx_k_CLandscapeMetricsCalculator), 0, 0, 1, 1},
  {&__pyx_n_s_CMetacommunity, __pyx_k_CMetacommunity, sizeof(__pyx_k_CMetacommunity), 0, 0, 1, 1},
  {&__pyx_n_s_CNSESimulation, __pyx_k_CNSESimulation, sizeof(__pyx_k_CNSESimulation), 0, 0, 1, 1},
  {&__pyx_n_s_CPNSESimulation, __pyx_k_CPNSESimulation, sizeof(__pyx_k_CPNSESimulation), 0, 0, 1, 1},
  {&__pyx_n_s_CPSpatialSimulation, __pyx_k_CPSpatialSimulation, sizeof(__pyx_k_CPSpatialSimulation), 0, 0, 1, 1},
  {&__pyx_n_s_CSpatialSimulation, __pyx_k_CSpatialSimulation, sizeof(__pyx_k_CSpatialSimulation), 0, 0, 1, 1},
  {&__pyx_kp_u_Couldn_t_find_paused_simulation, __pyx_k_Couldn_t_find_paused_simulation, sizeof(__pyx_k_Couldn_t_find_paused_simulation), 0, 1, 0, 0},
  {&__pyx_n_u_DISPERSAL_DISTANCES, __pyx_k_DISPERSAL_DISTANCES, sizeof(__pyx_k_DISPERSAL_DISTANCES), 0, 1, 0, 1},
  {&__pyx_n_u_DISTANCES_TRAVELLED, __pyx_k_DISTANCES_TRAVELLED, sizeof(__pyx_k_DISTANCES_TRAVELLED), 0, 1, 0, 1},
  {&__pyx_n_s_List, __pyx_k_List, sizeof(__pyx_k_List), 0, 0, 1, 1},
  {&__pyx_kp_u_Maps_have_already_been_imported, __pyx_k_Maps_have_already_been_imported, sizeof(__pyx_k_Maps_have_already_been_imported), 0, 1, 0, 0},
  {&__pyx_kp_u_Maps_have_not_been_imported_cann, __pyx_k_Maps_have_not_been_imported_cann, sizeof(__pyx_k_Maps_have_not_been_imported_cann), 0, 1, 0, 0},
  {&__pyx_kp_u_Output_database_has_not_been_set, __pyx_k_Output_database_has_not_been_set, sizeof(__pyx_k_Output_database_has_not_been_set), 0, 1, 0, 0},
  {&__pyx_n_s_RuntimeError, __pyx_k_RuntimeError, sizeof(__pyx_k_RuntimeError), 0, 0, 1, 1},
  {&__pyx_n_s_TypeError, __pyx_k_TypeError, sizeof(__pyx_k_TypeError), 0, 0, 1, 1},
  {&__pyx_n_s_base_setup, __pyx_k_base_setup, sizeof(__pyx_k_base_setup), 0, 0, 1, 1},
  {&__pyx_n_s_check_completed, __pyx_k_check_completed, sizeof(__pyx_k_check_completed), 0, 0, 1, 1},
  {&__pyx_n_s_cline_in_traceback, __pyx_k_cline_in_traceback, sizeof(__pyx_k_cline_in_traceback), 0, 0, 1, 1},
  {&__pyx_n_b_closed, __pyx_k_closed, sizeof(__pyx_k_closed), 0, 0, 0, 1},
  {&__pyx_n_s_coarse_map_file, __pyx_k_coarse_map_file, sizeof(__pyx_k_coarse_map_file), 0, 0, 1, 1},
  {&__pyx_n_s_coarse_map_scale, __pyx_k_coarse_map_scale, sizeof(__pyx_k_coarse_map_scale), 0, 0, 1, 1},
  {&__pyx_n_s_coarse_map_x_offset, __pyx_k_coarse_map_x_offset, sizeof(__pyx_k_coarse_map_x_offset), 0, 0, 1, 1},
  {&__pyx_n_s_coarse_map_x_size, __pyx_k_coarse_map_x_size, sizeof(__pyx_k_coarse_map_x_size), 0, 0, 1, 1},
  {&__pyx_n_s_coarse_map_y_offset, __pyx_k_coarse_map_y_offset, sizeof(__pyx_k_coarse_map_y_offset), 0, 0, 1, 1},
  {&__pyx_n_s_coarse_map_y_size, __pyx_k_coarse_map_y_size, sizeof(__pyx_k_coarse_map_y_size), 0, 0, 1, 1},
  {&__pyx_n_s_cutoff, __pyx_k_cutoff, sizeof(__pyx_k_cutoff), 0, 0, 1, 1},
  {&__pyx_n_s_database, __pyx_k_database, sizeof(__pyx_k_database), 0, 0, 1, 1},
  {&__pyx_n_s_deme, __pyx_k_deme, sizeof(__pyx_k_deme), 0, 0, 1, 1},
  {&__pyx_n_s_dispersal_file, __pyx_k_dispersal_file, sizeof(__pyx_k_dispersal_file), 0, 0, 1, 1},
  {&__pyx_n_s_dispersal_method, __pyx_k_dispersal_method, sizeof(__pyx_k_dispersal_method), 0, 0, 1, 1},
  {&__pyx_n_s_dispersal_relative_cost, __pyx_k_dispersal_relative_cost, sizeof(__pyx_k_dispersal_relative_cost), 0, 0, 1, 1},
  {&__pyx_n_s_fine_map_file, __pyx_k_fine_map_file, sizeof(__pyx_k_fine_map_file), 0, 0, 1, 1},
  {&__pyx_n_s_fine_map_x_offset, __pyx_k_fine_map_x_offset, sizeof(__pyx_k_fine_map_x_offset), 0, 0, 1, 1},
  {&__pyx_n_s_fine_map_x_size, __pyx_k_fine_map_x_size, sizeof(__pyx_k_fine_map_x_size), 0, 0, 1, 1},
  {&__pyx_n_s_fine_map_y_offset, __pyx_k_fine_map_y_offset, sizeof(__pyx_k_fine_map_y_offset), 0, 0, 1, 1},
  {&__pyx_n_s_fine_map_y_size, __pyx_k_fine_map_y_size, sizeof(__pyx_k_fine_map_y_size), 0, 0, 1, 1},
  {&__pyx_n_s_getstate, __pyx_k_getstate, sizeof(__pyx_k_getstate), 0, 0, 1, 1},
  {&__pyx_n_s_import, __pyx_k_import, sizeof(__pyx_k_import), 0, 0, 1, 1},
  {&__pyx_n_s_import_maps_base, __pyx_k_import_maps_base, sizeof(__pyx_k_import_maps_base), 0, 0, 1, 1},
  {&__pyx_n_s_is_sequential, __pyx_k_is_sequential, sizeof(__pyx_k_is_sequential), 0, 0, 1, 1},
  {&__pyx_n_s_landscape_type, __pyx_k_landscape_type, sizeof(__pyx_k_landscape_type), 0, 0, 1, 1},
  {&__pyx_n_s_logger, __pyx_k_logger, sizeof(__pyx_k_logger), 0, 0, 1, 1},
  {&__pyx_n_s_m_prob, __pyx_k_m_prob, sizeof(__pyx_k_m_prob), 0, 0, 1, 1},
  {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
  {&__pyx_n_s_max_speciation_gen, __pyx_k_max_speciation_gen, sizeof(__pyx_k_max_speciation_gen), 0, 0, 1, 1},
  {&__pyx_n_s_max_time, __pyx_k_max_time, sizeof(__pyx_k_max_time), 0, 0, 1, 1},
  {&__pyx_n_s_metacommunity_option, __pyx_k_metacommunity_option, sizeof(__pyx_k_metacommunity_option), 0, 0, 1, 1},
  {&__pyx_n_s_metacommunity_reference, __pyx_k_metacommunity_reference, sizeof(__pyx_k_metacommunity_reference), 0, 0, 1, 1},
  {&__pyx_n_s_metacommunity_size, __pyx_k_metacommunity_size, sizeof(__pyx_k_metacommunity_size), 0, 0, 1, 1},
  {&__pyx_n_s_min_speciation_gen, __pyx_k_min_speciation_gen, sizeof(__pyx_k_min_speciation_gen), 0, 0, 1, 1},
  {&__pyx_n_s_name, __pyx_k_name, sizeof(__pyx_k_name), 0, 0, 1, 1},
  {&__pyx_kp_s_no_default___reduce___due_to_non, __pyx_k_no_default___reduce___due_to_non, sizeof(__pyx_k_no_default___reduce___due_to_non), 0, 0, 1, 0},
  {&__pyx_n_u_none, __pyx_k_none, sizeof(__pyx_k_none), 0, 1, 0, 1},
  {&__pyx_n_s_number_coarse, __pyx_k_number_coarse, sizeof(__pyx_k_number_coarse), 0, 0, 1, 1},
  {&__pyx_n_s_number_fine, __pyx_k_number_fine, sizeof(__pyx_k_number_fine), 0, 0, 1, 1},
  {&__pyx_n_s_number_repeats, __pyx_k_number_repeats, sizeof(__pyx_k_number_repeats), 0, 0, 1, 1},
  {&__pyx_n_s_number_steps, __pyx_k_number_steps, sizeof(__pyx_k_number_steps), 0, 0, 1, 1},
  {&__pyx_n_s_number_workers, __pyx_k_number_workers, sizeof(__pyx_k_number_workers), 0, 0, 1, 1},
  {&__pyx_n_s_out_directory, __pyx_k_out_directory, sizeof(__pyx_k_out_directory), 0, 0, 1, 1},
  {&__pyx_n_s_path_coarse, __pyx_k_path_coarse, sizeof(__pyx_k_path_coarse), 0, 0, 1, 1},
  {&__pyx_n_s_path_fine, __pyx_k_path_fine, sizeof(__pyx_k_path_fine), 0, 0, 1, 1},
  {&__pyx_n_s_pause_directory, __pyx_k_pause_directory, sizeof(__pyx_k_pause_directory), 0, 0, 1, 1},
  {&__pyx_n_s_pycoalescence_system_operations, __pyx_k_pycoalescence_system_operations, sizeof(__pyx_k_pycoalescence_system_operations), 0, 0, 1, 1},
  {&__pyx_n_s_range, __pyx_k_range, sizeof(__pyx_k_range), 0, 0, 1, 1},
  {&__pyx_n_s_rate_coarse, __pyx_k_rate_coarse, sizeof(__pyx_k_rate_coarse), 0, 0, 1, 1},
  {&__pyx_n_s_rate_fine, __pyx_k_rate_fine, sizeof(__pyx_k_rate_fine), 0, 0, 1, 1},
  {&__pyx_n_s_record_fragments, __pyx_k_record_fragments, sizeof(__pyx_k_record_fragments), 0, 0, 1, 1},
  {&__pyx_n_s_reduce, __pyx_k_reduce, sizeof(__pyx_k_reduce), 0, 0, 1, 1},
  {&__pyx_n_s_reduce_cython, __pyx_k_reduce_cython, sizeof(__pyx_k_reduce_cython), 0, 0, 1, 1},
  {&__pyx_n_s_reduce_ex, __pyx_k_reduce_ex, sizeof(__pyx_k_reduce_ex), 0, 0, 1, 1},
  {&__pyx_n_s_restrict_self, __pyx_k_restrict_self, sizeof(__pyx_k_restrict_self), 0, 0, 1, 1},
  {&__pyx_n_s_sample_file, __pyx_k_sample_file, sizeof(__pyx_k_sample_file), 0, 0, 1, 1},
  {&__pyx_n_s_sample_map_x_size, __pyx_k_sample_map_x_size, sizeof(__pyx_k_sample_map_x_size), 0, 0, 1, 1},
  {&__pyx_n_s_sample_map_y_size, __pyx_k_sample_map_y_size, sizeof(__pyx_k_sample_map_y_size), 0, 0, 1, 1},
  {&__pyx_n_s_samples_x, __pyx_k_samples_x, sizeof(__pyx_k_samples_x), 0, 0, 1, 1},
  {&__pyx_n_s_samples_y, __pyx_k_samples_y, sizeof(__pyx_k_samples_y), 0, 0, 1, 1},
  {&__pyx_n_s_seed, __pyx_k_seed, sizeof(__pyx_k_seed), 0, 0, 1, 1},
  {&__pyx_n_s_set_logger, __pyx_k_set_logger, sizeof(__pyx_k_set_logger), 0, 0, 1, 1},
  {&__pyx_n_s_setstate, __pyx_k_setstate, sizeof(__pyx_k_setstate), 0, 0, 1, 1},
  {&__pyx_n_s_setstate_cython, __pyx_k_setstate_cython, sizeof(__pyx_k_setstate_cython), 0, 0, 1, 1},
  {&__pyx_n_s_setup, __pyx_k_setup, sizeof(__pyx_k_setup), 0, 0, 1, 1},
  {&__pyx_n_s_sigma, __pyx_k_sigma, sizeof(__pyx_k_sigma), 0, 0, 1, 1},
  {&__pyx_n_s_speciation_rate, __pyx_k_speciation_rate, sizeof(__pyx_k_speciation_rate), 0, 0, 1, 1},
  {&__pyx_n_s_speciation_rates, __pyx_k_speciation_rates, sizeof(__pyx_k_speciation_rates), 0, 0, 1, 1},
  {&__pyx_n_s_task, __pyx_k_task, sizeof(__pyx_k_task), 0, 0, 1, 1},
  {&__pyx_n_s_tau, __pyx_k_tau, sizeof(__pyx_k_tau), 0, 0, 1, 1},
  {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
  {&__pyx_n_s_time_coarse, __pyx_k_time_coarse, sizeof(__pyx_k_time_coarse), 0, 0, 1, 1},
  {&__pyx_n_s_time_fine, __pyx_k_time_fine, sizeof(__pyx_k_time_fine), 0, 0, 1, 1},
  {&__pyx_n_s_times, __pyx_k_times, sizeof(__pyx_k_times), 0, 0, 1, 1},
  {&__pyx_n_s_typing, __pyx_k_typing, sizeof(__pyx_k_typing), 0, 0, 1, 1},
  {&__pyx_n_s_use_spatial, __pyx_k_use_spatial, sizeof(__pyx_k_use_spatial), 0, 0, 1, 1},
  {&__pyx_n_s_write_database, __pyx_k_write_database, sizeof(__pyx_k_write_database), 0, 0, 1, 1},
  {&__pyx_n_s_write_to_log, __pyx_k_write_to_log, sizeof(__pyx_k_write_to_log), 0, 0, 1, 1},
  {0, 0, 0, 0, 0, 0, 0}
};
static CYTHON_SMALL_CODE int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_RuntimeError = __Pyx_GetBuiltinName(__pyx_n_s_RuntimeError); if (!__pyx_builtin_RuntimeError) __PYX_ERR(0, 53, __pyx_L1_error)
  __pyx_builtin_TypeError = __Pyx_GetBuiltinName(__pyx_n_s_TypeError); if (!__pyx_builtin_TypeError) __PYX_ERR(1, 2, __pyx_L1_error)
  __pyx_builtin_range = __Pyx_GetBuiltinName(__pyx_n_s_range); if (!__pyx_builtin_range) __PYX_ERR(1, 61, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "pycoalescence/necsim/necsim.pyx":53
 *             self.c_simulation.setup()
 *         else:
 *             raise RuntimeError("Couldn't find paused simulation")             # <<<<<<<<<<<<<<
 * 
 *     def add_gillespie(self, gillespie_threshold):
 */
  __pyx_tuple_ = PyTuple_Pack(1, __pyx_kp_u_Couldn_t_find_paused_simulation); if (unlikely(!__pyx_tuple_)) __PYX_ERR(0, 53, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple_);
  __Pyx_GIVEREF(__pyx_tuple_);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__2 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__2)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2);
  __Pyx_GIVEREF(__pyx_tuple__2);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__3 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__3)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__3);
  __Pyx_GIVEREF(__pyx_tuple__3);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__4 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__4)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__4);
  __Pyx_GIVEREF(__pyx_tuple__4);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__5 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__5)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__5);
  __Pyx_GIVEREF(__pyx_tuple__5);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__6 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__6)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__6);
  __Pyx_GIVEREF(__pyx_tuple__6);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__7 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__7)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__7);
  __Pyx_GIVEREF(__pyx_tuple__7);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__8 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__8)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__8);
  __Pyx_GIVEREF(__pyx_tuple__8);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__9 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__9)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__9);
  __Pyx_GIVEREF(__pyx_tuple__9);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__10 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__10)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__10);
  __Pyx_GIVEREF(__pyx_tuple__10);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__11 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__11)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__11);
  __Pyx_GIVEREF(__pyx_tuple__11);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__12 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__12)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__12);
  __Pyx_GIVEREF(__pyx_tuple__12);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__13 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__13)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__13);
  __Pyx_GIVEREF(__pyx_tuple__13);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__14 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__14)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__14);
  __Pyx_GIVEREF(__pyx_tuple__14);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__15 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__15)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__15);
  __Pyx_GIVEREF(__pyx_tuple__15);

  /* "pycoalescence/necsim/necsim.pyx":419
 *         cdef string coarse_map_file_str = coarse_map_file
 *         if self.has_imported_maps:
 *             raise RuntimeError("Maps have already been imported")             # <<<<<<<<<<<<<<
 *         deref(self.c_sim_parameters).setMapParameters(fine_map_file_str, coarse_map_file_str,
 *                                                       sample_map_x_size,
 */
  __pyx_tuple__16 = PyTuple_Pack(1, __pyx_kp_u_Maps_have_already_been_imported); if (unlikely(!__pyx_tuple__16)) __PYX_ERR(0, 419, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__16);
  __Pyx_GIVEREF(__pyx_tuple__16);

  /* "pycoalescence/necsim/necsim.pyx":518
 *     def check_completed(self):
 *         if self.output_database == "none":
 *             raise RuntimeError("Output database has not been set.")             # <<<<<<<<<<<<<<
 *         if not self.has_imported_maps:
 *             raise RuntimeError("Maps have not been imported - cannot start simulations.")
 */
  __pyx_tuple__17 = PyTuple_Pack(1, __pyx_kp_u_Output_database_has_not_been_set); if (unlikely(!__pyx_tuple__17)) __PYX_ERR(0, 518, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__17);
  __Pyx_GIVEREF(__pyx_tuple__17);

  /* "pycoalescence/necsim/necsim.pyx":520
 *             raise RuntimeError("Output database has not been set.")
 *         if not self.has_imported_maps:
 *             raise RuntimeError("Maps have not been imported - cannot start simulations.")             # <<<<<<<<<<<<<<
 * 
 *     def setup(self):
 */
  __pyx_tuple__18 = PyTuple_Pack(1, __pyx_kp_u_Maps_have_not_been_imported_cann); if (unlikely(!__pyx_tuple__18)) __PYX_ERR(0, 520, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__18);
  __Pyx_GIVEREF(__pyx_tuple__18);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__19 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__19)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__19);
  __Pyx_GIVEREF(__pyx_tuple__19);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__20 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__20)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__20);
  __Pyx_GIVEREF(__pyx_tuple__20);
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitGlobals(void) {
  if (__Pyx_InitStrings(__pyx_string_tab) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_modinit_global_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_import_code(void); /*proto*/

static int __Pyx_modinit_global_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_type_init_code(void) {
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  if (PyType_Ready(&__pyx_type_13pycoalescence_6necsim_6necsim_CSpatialSimulation) < 0) __PYX_ERR(0, 19, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_13pycoalescence_6necsim_6necsim_CSpatialSimulation.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_13pycoalescence_6necsim_6necsim_CSpatialSimulation.tp_dictoffset && __pyx_type_13pycoalescence_6necsim_6necsim_CSpatialSimulation.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_13pycoalescence_6necsim_6necsim_CSpatialSimulation.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_CSpatialSimulation, (PyObject *)&__pyx_type_13pycoalescence_6necsim_6necsim_CSpatialSimulation) < 0) __PYX_ERR(0, 19, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_13pycoalescence_6necsim_6necsim_CSpatialSimulation) < 0) __PYX_ERR(0, 19, __pyx_L1_error)
  __pyx_ptype_13pycoalescence_6necsim_6necsim_CSpatialSimulation = &__pyx_type_13pycoalescence_6necsim_6necsim_CSpatialSimulation;
  if (PyType_Ready(&__pyx_type_13pycoalescence_6necsim_6necsim_CNSESimulation) < 0) __PYX_ERR(0, 68, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_13pycoalescence_6necsim_6necsim_CNSESimulation.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_13pycoalescence_6necsim_6necsim_CNSESimulation.tp_dictoffset && __pyx_type_13pycoalescence_6necsim_6necsim_CNSESimulation.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_13pycoalescence_6necsim_6necsim_CNSESimulation.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_CNSESimulation, (PyObject *)&__pyx_type_13pycoalescence_6necsim_6necsim_CNSESimulation) < 0) __PYX_ERR(0, 68, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_13pycoalescence_6necsim_6necsim_CNSESimulation) < 0) __PYX_ERR(0, 68, __pyx_L1_error)
  __pyx_ptype_13pycoalescence_6necsim_6necsim_CNSESimulation = &__pyx_type_13pycoalescence_6necsim_6necsim_CNSESimulation;
  if (PyType_Ready(&__pyx_type_13pycoalescence_6necsim_6necsim_CPNSESimulation) < 0) __PYX_ERR(0, 117, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_13pycoalescence_6necsim_6necsim_CPNSESimulation.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_13pycoalescence_6necsim_6necsim_CPNSESimulation.tp_dictoffset && __pyx_type_13pycoalescence_6necsim_6necsim_CPNSESimulation.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_13pycoalescence_6necsim_6necsim_CPNSESimulation.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_CPNSESimulation, (PyObject *)&__pyx_type_13pycoalescence_6necsim_6necsim_CPNSESimulation) < 0) __PYX_ERR(0, 117, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_13pycoalescence_6necsim_6necsim_CPNSESimulation) < 0) __PYX_ERR(0, 117, __pyx_L1_error)
  __pyx_ptype_13pycoalescence_6necsim_6necsim_CPNSESimulation = &__pyx_type_13pycoalescence_6necsim_6necsim_CPNSESimulation;
  if (PyType_Ready(&__pyx_type_13pycoalescence_6necsim_6necsim_CPSpatialSimulation) < 0) __PYX_ERR(0, 167, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_13pycoalescence_6necsim_6necsim_CPSpatialSimulation.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_13pycoalescence_6necsim_6necsim_CPSpatialSimulation.tp_dictoffset && __pyx_type_13pycoalescence_6necsim_6necsim_CPSpatialSimulation.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_13pycoalescence_6necsim_6necsim_CPSpatialSimulation.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_CPSpatialSimulation, (PyObject *)&__pyx_type_13pycoalescence_6necsim_6necsim_CPSpatialSimulation) < 0) __PYX_ERR(0, 167, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_13pycoalescence_6necsim_6necsim_CPSpatialSimulation) < 0) __PYX_ERR(0, 167, __pyx_L1_error)
  __pyx_ptype_13pycoalescence_6necsim_6necsim_CPSpatialSimulation = &__pyx_type_13pycoalescence_6necsim_6necsim_CPSpatialSimulation;
  if (PyType_Ready(&__pyx_type_13pycoalescence_6necsim_6necsim_CCommunity) < 0) __PYX_ERR(0, 216, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_13pycoalescence_6necsim_6necsim_CCommunity.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_13pycoalescence_6necsim_6necsim_CCommunity.tp_dictoffset && __pyx_type_13pycoalescence_6necsim_6necsim_CCommunity.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_13pycoalescence_6necsim_6necsim_CCommunity.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_CCommunity, (PyObject *)&__pyx_type_13pycoalescence_6necsim_6necsim_CCommunity) < 0) __PYX_ERR(0, 216, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_13pycoalescence_6necsim_6necsim_CCommunity) < 0) __PYX_ERR(0, 216, __pyx_L1_error)
  __pyx_ptype_13pycoalescence_6necsim_6necsim_CCommunity = &__pyx_type_13pycoalescence_6necsim_6necsim_CCommunity;
  if (PyType_Ready(&__pyx_type_13pycoalescence_6necsim_6necsim_CMetacommunity) < 0) __PYX_ERR(0, 281, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_13pycoalescence_6necsim_6necsim_CMetacommunity.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_13pycoalescence_6necsim_6necsim_CMetacommunity.tp_dictoffset && __pyx_type_13pycoalescence_6necsim_6necsim_CMetacommunity.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_13pycoalescence_6necsim_6necsim_CMetacommunity.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_CMetacommunity, (PyObject *)&__pyx_type_13pycoalescence_6necsim_6necsim_CMetacommunity) < 0) __PYX_ERR(0, 281, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_13pycoalescence_6necsim_6necsim_CMetacommunity) < 0) __PYX_ERR(0, 281, __pyx_L1_error)
  __pyx_ptype_13pycoalescence_6necsim_6necsim_CMetacommunity = &__pyx_type_13pycoalescence_6necsim_6necsim_CMetacommunity;
  if (PyType_Ready(&__pyx_type_13pycoalescence_6necsim_6necsim_CLandscapeMetricsCalculator) < 0) __PYX_ERR(0, 346, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_13pycoalescence_6necsim_6necsim_CLandscapeMetricsCalculator.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_13pycoalescence_6necsim_6necsim_CLandscapeMetricsCalculator.tp_dictoffset && __pyx_type_13pycoalescence_6necsim_6necsim_CLandscapeMetricsCalculator.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_13pycoalescence_6necsim_6necsim_CLandscapeMetricsCalculator.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_CLandscapeMetricsCalculator, (PyObject *)&__pyx_type_13pycoalescence_6necsim_6necsim_CLandscapeMetricsCalculator) < 0) __PYX_ERR(0, 346, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_13pycoalescence_6necsim_6necsim_CLandscapeMetricsCalculator) < 0) __PYX_ERR(0, 346, __pyx_L1_error)
  __pyx_ptype_13pycoalescence_6necsim_6necsim_CLandscapeMetricsCalculator = &__pyx_type_13pycoalescence_6necsim_6necsim_CLandscapeMetricsCalculator;
  if (PyType_Ready(&__pyx_type_13pycoalescence_6necsim_6necsim_CDispersalSimulator) < 0) __PYX_ERR(0, 370, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_13pycoalescence_6necsim_6necsim_CDispersalSimulator.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_13pycoalescence_6necsim_6necsim_CDispersalSimulator.tp_dictoffset && __pyx_type_13pycoalescence_6necsim_6necsim_CDispersalSimulator.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_13pycoalescence_6necsim_6necsim_CDispersalSimulator.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_CDispersalSimulator, (PyObject *)&__pyx_type_13pycoalescence_6necsim_6necsim_CDispersalSimulator) < 0) __PYX_ERR(0, 370, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_13pycoalescence_6necsim_6necsim_CDispersalSimulator) < 0) __PYX_ERR(0, 370, __pyx_L1_error)
  __pyx_ptype_13pycoalescence_6necsim_6necsim_CDispersalSimulator = &__pyx_type_13pycoalescence_6necsim_6necsim_CDispersalSimulator;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}


#ifndef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#elif PY_MAJOR_VERSION < 3
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" void
#else
#define __Pyx_PyMODINIT_FUNC void
#endif
#else
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" PyObject *
#else
#define __Pyx_PyMODINIT_FUNC PyObject *
#endif
#endif


#if PY_MAJOR_VERSION < 3
__Pyx_PyMODINIT_FUNC initnecsim(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC initnecsim(void)
#else
__Pyx_PyMODINIT_FUNC PyInit_necsim(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit_necsim(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
static CYTHON_SMALL_CODE int __Pyx_check_single_interpreter(void) {
    #if PY_VERSION_HEX >= 0x030700A1
    static PY_INT64_T main_interpreter_id = -1;
    PY_INT64_T current_id = PyInterpreterState_GetID(PyThreadState_Get()->interp);
    if (main_interpreter_id == -1) {
        main_interpreter_id = current_id;
        return (unlikely(current_id == -1)) ? -1 : 0;
    } else if (unlikely(main_interpreter_id != current_id))
    #else
    static PyInterpreterState *main_interpreter = NULL;
    PyInterpreterState *current_interpreter = PyThreadState_Get()->interp;
    if (!main_interpreter) {
        main_interpreter = current_interpreter;
    } else if (unlikely(main_interpreter != current_interpreter))
    #endif
    {
        PyErr_SetString(
            PyExc_ImportError,
            "Interpreter change detected - this module can only be loaded into one interpreter per process.");
        return -1;
    }
    return 0;
}
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name, int allow_none) {
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        if (allow_none || value != Py_None) {
            result = PyDict_SetItemString(moddict, to_name, value);
        }
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static CYTHON_SMALL_CODE PyObject* __pyx_pymod_create(PyObject *spec, CYTHON_UNUSED PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    if (__Pyx_check_single_interpreter())
        return NULL;
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__", 0) < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static CYTHON_SMALL_CODE int __pyx_pymod_exec_necsim(PyObject *__pyx_pyinit_module)
#endif
#endif
{
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m) {
    if (__pyx_m == __pyx_pyinit_module) return 0;
    PyErr_SetString(PyExc_RuntimeError, "Module 'necsim' has already been imported. Re-initialisation is not supported.");
    return -1;
  }
  #elif PY_MAJOR_VERSION >= 3
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  #if CYTHON_REFNANNY
__Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
if (!__Pyx_RefNanny) {
  PyErr_Clear();
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
  if (!__Pyx_RefNanny)
      Py_FatalError("failed to import 'refnanny' module");
}
#endif
  __Pyx_RefNannySetupContext("__Pyx_PyMODINIT_FUNC PyInit_necsim(void)", 0);
  if (__Pyx_check_binary_version() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pxy_PyFrame_Initialize_Offsets
  __Pxy_PyFrame_Initialize_Offsets();
  #endif
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_AsyncGen_USED
  if (__pyx_AsyncGen_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  #ifdef WITH_THREAD /* Python build with threading support? */
  PyEval_InitThreads();
  #endif
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_m = __pyx_pyinit_module;
  Py_INCREF(__pyx_m);
  #else
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("necsim", __pyx_methods, 0, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  #endif
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = PyImport_AddModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_b);
  __pyx_cython_runtime = PyImport_AddModule((char *) "cython_runtime"); if (unlikely(!__pyx_cython_runtime)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_cython_runtime);
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_pycoalescence__necsim__necsim) {
    if (PyObject_SetAttr(__pyx_m, __pyx_n_s_name, __pyx_n_s_main) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "pycoalescence.necsim.necsim")) {
      if (unlikely(PyDict_SetItemString(modules, "pycoalescence.necsim.necsim", __pyx_m) < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code();
  (void)__Pyx_modinit_variable_export_code();
  (void)__Pyx_modinit_function_export_code();
  if (unlikely(__Pyx_modinit_type_init_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  (void)__Pyx_modinit_type_import_code();
  (void)__Pyx_modinit_variable_import_code();
  (void)__Pyx_modinit_function_import_code();
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif

  /* "pycoalescence/necsim/necsim.pyx":4
 * # cython: c_string_type=unicode, c_string_encoding=utf8
 * 
 * from typing import List             # <<<<<<<<<<<<<<
 * from cython.operator cimport dereference as deref
 * from libcpp.memory cimport shared_ptr, make_shared
 */
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_List);
  __Pyx_GIVEREF(__pyx_n_s_List);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_List);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_typing, __pyx_t_1, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_List); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_List, __pyx_t_1) < 0) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pycoalescence/necsim/necsim.pyx":8
 * from libcpp.memory cimport shared_ptr, make_shared
 * 
 * from pycoalescence.system_operations import write_to_log             # <<<<<<<<<<<<<<
 * 
 * from pycoalescence.necsim.necsim cimport SpatialTree, Tree, GenericTree, ProtractedSpatialTree, ProtractedTree, \
 */
  __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 8, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_write_to_log);
  __Pyx_GIVEREF(__pyx_n_s_write_to_log);
  PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_write_to_log);
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_pycoalescence_system_operations, __pyx_t_2, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 8, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_write_to_log); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 8, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_write_to_log, __pyx_t_2) < 0) __PYX_ERR(0, 8, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pycoalescence/necsim/necsim.pyx":1
 * # distutils: language = c++             # <<<<<<<<<<<<<<
 * # cython: c_string_type=unicode, c_string_encoding=utf8
 * 
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_1) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "vector.to_py":60
 * 
 * @cname("__pyx_convert_vector_to_py_unsigned_long")
 * cdef object __pyx_convert_vector_to_py_unsigned_long(vector[X]& v):             # <<<<<<<<<<<<<<
 *     return [v[i] for i in range(v.size())]
 * 
 */

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  if (__pyx_m) {
    if (__pyx_d) {
      __Pyx_AddTraceback("init pycoalescence.necsim.necsim", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    Py_CLEAR(__pyx_m);
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init pycoalescence.necsim.necsim");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  return (__pyx_m != NULL) ? 0 : -1;
  #elif PY_MAJOR_VERSION >= 3
  return __pyx_m;
  #else
  return;
  #endif
}

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule(modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, "RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* PyObjectGetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStr(__pyx_b, name);
    if (unlikely(!result)) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* RaiseDoubleKeywords */
static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

/* ParseKeywords */
static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    while (PyDict_Next(kwds, &pos, &key, &value)) {
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
            continue;
        }
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (**name == key) ? 0 :
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (__Pyx_PyUnicode_GET_LENGTH(**name) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key);
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (__Pyx_PyUnicode_GET_LENGTH(**argname) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    return -1;
}

/* RaiseArgTupleInvalid */
static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* PyDictVersioning */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    return likely(dict) ? __PYX_GET_DICT_VERSION(dict) : 0;
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj) {
    PyObject **dictptr = NULL;
    Py_ssize_t offset = Py_TYPE(obj)->tp_dictoffset;
    if (offset) {
#if CYTHON_COMPILING_IN_CPYTHON
        dictptr = (likely(offset > 0)) ? (PyObject **) ((char *)obj + offset) : _PyObject_GetDictPtr(obj);
#else
        dictptr = _PyObject_GetDictPtr(obj);
#endif
    }
    return (dictptr && *dictptr) ? __PYX_GET_DICT_VERSION(*dictptr) : 0;
}
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    if (unlikely(!dict) || unlikely(tp_dict_version != __PYX_GET_DICT_VERSION(dict)))
        return 0;
    return obj_dict_version == __Pyx_get_object_dict_version(obj);
}
#endif

/* GetModuleGlobalName */
#if CYTHON_USE_DICT_VERSIONS
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value)
#else
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name)
#endif
{
    PyObject *result;
#if !CYTHON_AVOID_BORROWED_REFS
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1
    result = _PyDict_GetItem_KnownHash(__pyx_d, name, ((PyASCIIObject *) name)->hash);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    } else if (unlikely(PyErr_Occurred())) {
        return NULL;
    }
#else
    result = PyDict_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
#endif
#else
    result = PyObject_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
    PyErr_Clear();
#endif
    return __Pyx_GetBuiltinName(name);
}

/* PyFunctionFastCall */
#if CYTHON_FAST_PYCALL
static PyObject* __Pyx_PyFunction_FastCallNoKw(PyCodeObject *co, PyObject **args, Py_ssize_t na,
                                               PyObject *globals) {
    PyFrameObject *f;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject **fastlocals;
    Py_ssize_t i;
    PyObject *result;
    assert(globals != NULL);
    /* XXX Perhaps we should create a specialized
       PyFrame_New() that doesn't take locals, but does
       take builtins without sanity checking them.
       */
    assert(tstate != NULL);
    f = PyFrame_New(tstate, co, globals, NULL);
    if (f == NULL) {
        return NULL;
    }
    fastlocals = __Pyx_PyFrame_GetLocalsplus(f);
    for (i = 0; i < na; i++) {
        Py_INCREF(*args);
        fastlocals[i] = *args++;
    }
    result = PyEval_EvalFrameEx(f,0);
    ++tstate->recursion_depth;
    Py_DECREF(f);
    --tstate->recursion_depth;
    return result;
}
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs) {
    PyCodeObject *co = (PyCodeObject *)PyFunction_GET_CODE(func);
    PyObject *globals = PyFunction_GET_GLOBALS(func);
    PyObject *argdefs = PyFunction_GET_DEFAULTS(func);
    PyObject *closure;
#if PY_MAJOR_VERSION >= 3
    PyObject *kwdefs;
#endif
    PyObject *kwtuple, **k;
    PyObject **d;
    Py_ssize_t nd;
    Py_ssize_t nk;
    PyObject *result;
    assert(kwargs == NULL || PyDict_Check(kwargs));
    nk = kwargs ? PyDict_Size(kwargs) : 0;
    if (Py_EnterRecursiveCall((char*)" while calling a Python object")) {
        return NULL;
    }
    if (
#if PY_MAJOR_VERSION >= 3
            co->co_kwonlyargcount == 0 &&
#endif
            likely(kwargs == NULL || nk == 0) &&
            co->co_flags == (CO_OPTIMIZED | CO_NEWLOCALS | CO_NOFREE)) {
        if (argdefs == NULL && co->co_argcount == nargs) {
            result = __Pyx_PyFunction_FastCallNoKw(co, args, nargs, globals);
            goto done;
        }
        else if (nargs == 0 && argdefs != NULL
                 && co->co_argcount == Py_SIZE(argdefs)) {
            /* function called with no arguments, but all parameters have
               a default value: use default values as arguments .*/
            args = &PyTuple_GET_ITEM(argdefs, 0);
            result =__Pyx_PyFunction_FastCallNoKw(co, args, Py_SIZE(argdefs), globals);
            goto done;
        }
    }
    if (kwargs != NULL) {
        Py_ssize_t pos, i;
        kwtuple = PyTuple_New(2 * nk);
        if (kwtuple == NULL) {
            result = NULL;
            goto done;
        }
        k = &PyTuple_GET_ITEM(kwtuple, 0);
        pos = i = 0;
        while (PyDict_Next(kwargs, &pos, &k[i], &k[i+1])) {
            Py_INCREF(k[i]);
            Py_INCREF(k[i+1]);
            i += 2;
        }
        nk = i / 2;
    }
    else {
        kwtuple = NULL;
        k = NULL;
    }
    closure = PyFunction_GET_CLOSURE(func);
#if PY_MAJOR_VERSION >= 3
    kwdefs = PyFunction_GET_KW_DEFAULTS(func);
#endif
    if (argdefs != NULL) {
        d = &PyTuple_GET_ITEM(argdefs, 0);
        nd = Py_SIZE(argdefs);
    }
    else {
        d = NULL;
        nd = 0;
    }
#if PY_MAJOR_VERSION >= 3
    result = PyEval_EvalCodeEx((PyObject*)co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, kwdefs, closure);
#else
    result = PyEval_EvalCodeEx(co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, closure);
#endif
    Py_XDECREF(kwtuple);
done:
    Py_LeaveRecursiveCall();
    return result;
}
#endif
#endif

/* PyObjectCall */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = func->ob_type->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallMethO */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = PyCFunction_GET_FUNCTION(func);
    self = PyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallNoArg */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, NULL, 0);
    }
#endif
#ifdef __Pyx_CyFunction_USED
    if (likely(PyCFunction_Check(func) || __Pyx_CyFunction_Check(func)))
#else
    if (likely(PyCFunction_Check(func)))
#endif
    {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_NOARGS)) {
            return __Pyx_PyObject_CallMethO(func, NULL);
        }
    }
    return __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL);
}
#endif

/* PyCFunctionFastCall */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject * __Pyx_PyCFunction_FastCall(PyObject *func_obj, PyObject **args, Py_ssize_t nargs) {
    PyCFunctionObject *func = (PyCFunctionObject*)func_obj;
    PyCFunction meth = PyCFunction_GET_FUNCTION(func);
    PyObject *self = PyCFunction_GET_SELF(func);
    int flags = PyCFunction_GET_FLAGS(func);
    assert(PyCFunction_Check(func));
    assert(METH_FASTCALL == (flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)));
    assert(nargs >= 0);
    assert(nargs == 0 || args != NULL);
    /* _PyCFunction_FastCallDict() must not be called with an exception set,
       because it may clear it (directly or indirectly) and so the
       caller loses its exception */
    assert(!PyErr_Occurred());
    if ((PY_VERSION_HEX < 0x030700A0) || unlikely(flags & METH_KEYWORDS)) {
        return (*((__Pyx_PyCFunctionFastWithKeywords)(void*)meth)) (self, args, nargs, NULL);
    } else {
        return (*((__Pyx_PyCFunctionFast)(void*)meth)) (self, args, nargs);
    }
}
#endif

/* PyObjectCallOneArg */
#if CYTHON_COMPILING_IN_CPYTHON
static PyObject* __Pyx__PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_New(1);
    if (unlikely(!args)) return NULL;
    Py_INCREF(arg);
    PyTuple_SET_ITEM(args, 0, arg);
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, &arg, 1);
    }
#endif
    if (likely(PyCFunction_Check(func))) {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_O)) {
            return __Pyx_PyObject_CallMethO(func, arg);
#if CYTHON_FAST_PYCCALL
        } else if (__Pyx_PyFastCFunction_Check(func)) {
            return __Pyx_PyCFunction_FastCall(func, &arg, 1);
#endif
        }
    }
    return __Pyx__PyObject_CallOneArg(func, arg);
}
#else
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_Pack(1, arg);
    if (unlikely(!args)) return NULL;
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
#endif

/* PyErrFetchRestore */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
}
#endif

/* RaiseException */
#if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb,
                        CYTHON_UNUSED PyObject *cause) {
    __Pyx_PyThreadState_declare
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause) {
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
#if CYTHON_COMPILING_IN_PYPY
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#else
        PyThreadState *tstate = __Pyx_PyThreadState_Current;
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

/* BytesEquals */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY
    return PyObject_RichCompareBool(s1, s2, equals);
#else
    if (s1 == s2) {
        return (equals == Py_EQ);
    } else if (PyBytes_CheckExact(s1) & PyBytes_CheckExact(s2)) {
        const char *ps1, *ps2;
        Py_ssize_t length = PyBytes_GET_SIZE(s1);
        if (length != PyBytes_GET_SIZE(s2))
            return (equals == Py_NE);
        ps1 = PyBytes_AS_STRING(s1);
        ps2 = PyBytes_AS_STRING(s2);
        if (ps1[0] != ps2[0]) {
            return (equals == Py_NE);
        } else if (length == 1) {
            return (equals == Py_EQ);
        } else {
            int result;
#if CYTHON_USE_UNICODE_INTERNALS
            Py_hash_t hash1, hash2;
            hash1 = ((PyBytesObject*)s1)->ob_shash;
            hash2 = ((PyBytesObject*)s2)->ob_shash;
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                return (equals == Py_NE);
            }
#endif
            result = memcmp(ps1, ps2, (size_t)length);
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & PyBytes_CheckExact(s2)) {
        return (equals == Py_NE);
    } else if ((s2 == Py_None) & PyBytes_CheckExact(s1)) {
        return (equals == Py_NE);
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
#endif
}

/* UnicodeEquals */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY
    return PyObject_RichCompareBool(s1, s2, equals);
#else
#if PY_MAJOR_VERSION < 3
    PyObject* owned_ref = NULL;
#endif
    int s1_is_unicode, s2_is_unicode;
    if (s1 == s2) {
        goto return_eq;
    }
    s1_is_unicode = PyUnicode_CheckExact(s1);
    s2_is_unicode = PyUnicode_CheckExact(s2);
#if PY_MAJOR_VERSION < 3
    if ((s1_is_unicode & (!s2_is_unicode)) && PyString_CheckExact(s2)) {
        owned_ref = PyUnicode_FromObject(s2);
        if (unlikely(!owned_ref))
            return -1;
        s2 = owned_ref;
        s2_is_unicode = 1;
    } else if ((s2_is_unicode & (!s1_is_unicode)) && PyString_CheckExact(s1)) {
        owned_ref = PyUnicode_FromObject(s1);
        if (unlikely(!owned_ref))
            return -1;
        s1 = owned_ref;
        s1_is_unicode = 1;
    } else if (((!s2_is_unicode) & (!s1_is_unicode))) {
        return __Pyx_PyBytes_Equals(s1, s2, equals);
    }
#endif
    if (s1_is_unicode & s2_is_unicode) {
        Py_ssize_t length;
        int kind;
        void *data1, *data2;
        if (unlikely(__Pyx_PyUnicode_READY(s1) < 0) || unlikely(__Pyx_PyUnicode_READY(s2) < 0))
            return -1;
        length = __Pyx_PyUnicode_GET_LENGTH(s1);
        if (length != __Pyx_PyUnicode_GET_LENGTH(s2)) {
            goto return_ne;
        }
#if CYTHON_USE_UNICODE_INTERNALS
        {
            Py_hash_t hash1, hash2;
        #if CYTHON_PEP393_ENABLED
            hash1 = ((PyASCIIObject*)s1)->hash;
            hash2 = ((PyASCIIObject*)s2)->hash;
        #else
            hash1 = ((PyUnicodeObject*)s1)->hash;
            hash2 = ((PyUnicodeObject*)s2)->hash;
        #endif
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                goto return_ne;
            }
        }
#endif
        kind = __Pyx_PyUnicode_KIND(s1);
        if (kind != __Pyx_PyUnicode_KIND(s2)) {
            goto return_ne;
        }
        data1 = __Pyx_PyUnicode_DATA(s1);
        data2 = __Pyx_PyUnicode_DATA(s2);
        if (__Pyx_PyUnicode_READ(kind, data1, 0) != __Pyx_PyUnicode_READ(kind, data2, 0)) {
            goto return_ne;
        } else if (length == 1) {
            goto return_eq;
        } else {
            int result = memcmp(data1, data2, (size_t)(length * kind));
            #if PY_MAJOR_VERSION < 3
            Py_XDECREF(owned_ref);
            #endif
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & s2_is_unicode) {
        goto return_ne;
    } else if ((s2 == Py_None) & s1_is_unicode) {
        goto return_ne;
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        #if PY_MAJOR_VERSION < 3
        Py_XDECREF(owned_ref);
        #endif
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
return_eq:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_EQ);
return_ne:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_NE);
#endif
}

/* PyObjectCall2Args */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2) {
    PyObject *args, *result = NULL;
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyFunction_FastCall(function, args, 2);
    }
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyCFunction_FastCall(function, args, 2);
    }
    #endif
    args = PyTuple_New(2);
    if (unlikely(!args)) goto done;
    Py_INCREF(arg1);
    PyTuple_SET_ITEM(args, 0, arg1);
    Py_INCREF(arg2);
    PyTuple_SET_ITEM(args, 1, arg2);
    Py_INCREF(function);
    result = __Pyx_PyObject_Call(function, args, NULL);
    Py_DECREF(args);
    Py_DECREF(function);
done:
    return result;
}

/* PyObject_GenericGetAttrNoDict */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject *__Pyx_RaiseGenericGetAttributeError(PyTypeObject *tp, PyObject *attr_name) {
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'%.50s' object has no attribute '%U'",
                 tp->tp_name, attr_name);
#else
                 "'%.50s' object has no attribute '%.400s'",
                 tp->tp_name, PyString_AS_STRING(attr_name));
#endif
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name) {
    PyObject *descr;
    PyTypeObject *tp = Py_TYPE(obj);
    if (unlikely(!PyString_Check(attr_name))) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    assert(!tp->tp_dictoffset);
    descr = _PyType_Lookup(tp, attr_name);
    if (unlikely(!descr)) {
        return __Pyx_RaiseGenericGetAttributeError(tp, attr_name);
    }
    Py_INCREF(descr);
    #if PY_MAJOR_VERSION < 3
    if (likely(PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_HAVE_CLASS)))
    #endif
    {
        descrgetfunc f = Py_TYPE(descr)->tp_descr_get;
        if (unlikely(f)) {
            PyObject *res = f(descr, obj, (PyObject *)tp);
            Py_DECREF(descr);
            return res;
        }
    }
    return descr;
}
#endif

/* PyObject_GenericGetAttr */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name) {
    if (unlikely(Py_TYPE(obj)->tp_dictoffset)) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    return __Pyx_PyObject_GenericGetAttrNoDict(obj, attr_name);
}
#endif

/* PyErrExceptionMatches */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx_PyErr_ExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        if (__Pyx_PyErr_GivenExceptionMatches(exc_type, PyTuple_GET_ITEM(tuple, i))) return 1;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    PyObject *exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
    if (unlikely(PyTuple_Check(err)))
        return __Pyx_PyErr_ExceptionMatchesTuple(exc_type, err);
    return __Pyx_PyErr_GivenExceptionMatches(exc_type, err);
}
#endif

/* PyObjectGetAttrStrNoError */
static void __Pyx_PyObject_GetAttrStr_ClearAttributeError(void) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (likely(__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        __Pyx_PyErr_Clear();
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name) {
    PyObject *result;
#if CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_TYPE_SLOTS && PY_VERSION_HEX >= 0x030700B1
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro == PyObject_GenericGetAttr)) {
        return _PyObject_GenericGetAttrWithDict(obj, attr_name, NULL, 1);
    }
#endif
    result = __Pyx_PyObject_GetAttrStr(obj, attr_name);
    if (unlikely(!result)) {
        __Pyx_PyObject_GetAttrStr_ClearAttributeError();
    }
    return result;
}

/* SetupReduce */
static int __Pyx_setup_reduce_is_named(PyObject* meth, PyObject* name) {
  int ret;
  PyObject *name_attr;
  name_attr = __Pyx_PyObject_GetAttrStr(meth, __pyx_n_s_name);
  if (likely(name_attr)) {
      ret = PyObject_RichCompareBool(name_attr, name, Py_EQ);
  } else {
      ret = -1;
  }
  if (unlikely(ret < 0)) {
      PyErr_Clear();
      ret = 0;
  }
  Py_XDECREF(name_attr);
  return ret;
}
static int __Pyx_setup_reduce(PyObject* type_obj) {
    int ret = 0;
    PyObject *object_reduce = NULL;
    PyObject *object_reduce_ex = NULL;
    PyObject *reduce = NULL;
    PyObject *reduce_ex = NULL;
    PyObject *reduce_cython = NULL;
    PyObject *setstate = NULL;
    PyObject *setstate_cython = NULL;
#if CYTHON_USE_PYTYPE_LOOKUP
    if (_PyType_Lookup((PyTypeObject*)type_obj, __pyx_n_s_getstate)) goto __PYX_GOOD;
#else
    if (PyObject_HasAttr(type_obj, __pyx_n_s_getstate)) goto __PYX_GOOD;
#endif
#if CYTHON_USE_PYTYPE_LOOKUP
    object_reduce_ex = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#else
    object_reduce_ex = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#endif
    reduce_ex = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce_ex); if (unlikely(!reduce_ex)) goto __PYX_BAD;
    if (reduce_ex == object_reduce_ex) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_reduce = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto __PYX_BAD;
#else
        object_reduce = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto __PYX_BAD;
#endif
        reduce = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce); if (unlikely(!reduce)) goto __PYX_BAD;
        if (reduce == object_reduce || __Pyx_setup_reduce_is_named(reduce, __pyx_n_s_reduce_cython)) {
            reduce_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_reduce_cython);
            if (likely(reduce_cython)) {
                ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce, reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
            } else if (reduce == object_reduce || PyErr_Occurred()) {
                goto __PYX_BAD;
            }
            setstate = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_setstate);
            if (!setstate) PyErr_Clear();
            if (!setstate || __Pyx_setup_reduce_is_named(setstate, __pyx_n_s_setstate_cython)) {
                setstate_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_setstate_cython);
                if (likely(setstate_cython)) {
                    ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate, setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                    ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                } else if (!setstate || PyErr_Occurred()) {
                    goto __PYX_BAD;
                }
            }
            PyType_Modified((PyTypeObject*)type_obj);
        }
    }
    goto __PYX_GOOD;
__PYX_BAD:
    if (!PyErr_Occurred())
        PyErr_Format(PyExc_RuntimeError, "Unable to initialize pickling for %s", ((PyTypeObject*)type_obj)->tp_name);
    ret = -1;
__PYX_GOOD:
#if !CYTHON_USE_PYTYPE_LOOKUP
    Py_XDECREF(object_reduce);
    Py_XDECREF(object_reduce_ex);
#endif
    Py_XDECREF(reduce);
    Py_XDECREF(reduce_ex);
    Py_XDECREF(reduce_cython);
    Py_XDECREF(setstate);
    Py_XDECREF(setstate_cython);
    return ret;
}

/* Import */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level) {
    PyObject *empty_list = 0;
    PyObject *module = 0;
    PyObject *global_dict = 0;
    PyObject *empty_dict = 0;
    PyObject *list;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_import;
    py_import = __Pyx_PyObject_GetAttrStr(__pyx_b, __pyx_n_s_import);
    if (!py_import)
        goto bad;
    #endif
    if (from_list)
        list = from_list;
    else {
        empty_list = PyList_New(0);
        if (!empty_list)
            goto bad;
        list = empty_list;
    }
    global_dict = PyModule_GetDict(__pyx_m);
    if (!global_dict)
        goto bad;
    empty_dict = PyDict_New();
    if (!empty_dict)
        goto bad;
    {
        #if PY_MAJOR_VERSION >= 3
        if (level == -1) {
            if ((1) && (strchr(__Pyx_MODULE_NAME, '.'))) {
                module = PyImport_ImportModuleLevelObject(
                    name, global_dict, empty_dict, list, 1);
                if (!module) {
                    if (!PyErr_ExceptionMatches(PyExc_ImportError))
                        goto bad;
                    PyErr_Clear();
                }
            }
            level = 0;
        }
        #endif
        if (!module) {
            #if PY_MAJOR_VERSION < 3
            PyObject *py_level = PyInt_FromLong(level);
            if (!py_level)
                goto bad;
            module = PyObject_CallFunctionObjArgs(py_import,
                name, global_dict, empty_dict, list, py_level, (PyObject *)NULL);
            Py_DECREF(py_level);
            #else
            module = PyImport_ImportModuleLevelObject(
                name, global_dict, empty_dict, list, level);
            #endif
        }
    }
bad:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_import);
    #endif
    Py_XDECREF(empty_list);
    Py_XDECREF(empty_dict);
    return module;
}

/* ImportFrom */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name) {
    PyObject* value = __Pyx_PyObject_GetAttrStr(module, name);
    if (unlikely(!value) && PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Format(PyExc_ImportError,
        #if PY_MAJOR_VERSION < 3
            "cannot import name %.230s", PyString_AS_STRING(name));
        #else
            "cannot import name %S", name);
        #endif
    }
    return value;
}

/* CLineInTraceback */
#ifndef CYTHON_CLINE_IN_TRACEBACK
static int __Pyx_CLineForTraceback(CYTHON_NCP_UNUSED PyThreadState *tstate, int c_line) {
    PyObject *use_cline;
    PyObject *ptype, *pvalue, *ptraceback;
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject **cython_runtime_dict;
#endif
    if (unlikely(!__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
#if CYTHON_COMPILING_IN_CPYTHON
    cython_runtime_dict = _PyObject_GetDictPtr(__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
        __PYX_PY_DICT_LOOKUP_IF_MODIFIED(
            use_cline, *cython_runtime_dict,
            __Pyx_PyDict_GetItemStr(*cython_runtime_dict, __pyx_n_s_cline_in_traceback))
    } else
#endif
    {
      PyObject *use_cline_obj = __Pyx_PyObject_GetAttrStr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback);
      if (use_cline_obj) {
        use_cline = PyObject_Not(use_cline_obj) ? Py_False : Py_True;
        Py_DECREF(use_cline_obj);
      } else {
        PyErr_Clear();
        use_cline = NULL;
      }
    }
    if (!use_cline) {
        c_line = 0;
        PyObject_SetAttr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback, Py_False);
    }
    else if (use_cline == Py_False || (use_cline != Py_True && PyObject_Not(use_cline) != 0)) {
        c_line = 0;
    }
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache */
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, ((size_t)new_max) * sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}

/* AddTraceback */
#include "compile.h"
#include "frameobject.h"
#include "traceback.h"
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyObject *py_srcfile = 0;
    PyObject *py_funcname = 0;
    #if PY_MAJOR_VERSION < 3
    py_srcfile = PyString_FromString(filename);
    #else
    py_srcfile = PyUnicode_FromString(filename);
    #endif
    if (!py_srcfile) goto bad;
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        #else
        py_funcname = PyUnicode_FromString(funcname);
        #endif
    }
    if (!py_funcname) goto bad;
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    Py_DECREF(py_funcname);
    return py_code;
bad:
    Py_XDECREF(py_srcfile);
    Py_XDECREF(py_funcname);
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) goto bad;
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}

/* CIntFromPyVerify */
#define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* CIntFromPy */
static CYTHON_INLINE unsigned long __Pyx_PyInt_As_unsigned_long(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const unsigned long neg_one = (unsigned long) -1, const_zero = (unsigned long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(unsigned long) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(unsigned long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (unsigned long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (unsigned long) 0;
                case  1: __PYX_VERIFY_RETURN_INT(unsigned long, digit, digits[0])
                case 2:
                    if (8 * sizeof(unsigned long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned long) >= 2 * PyLong_SHIFT) {
                            return (unsigned long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned long) >= 3 * PyLong_SHIFT) {
                            return (unsigned long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned long) >= 4 * PyLong_SHIFT) {
                            return (unsigned long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (unsigned long) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(unsigned long) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(unsigned long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(unsigned long) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(unsigned long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (unsigned long) 0;
                case -1: __PYX_VERIFY_RETURN_INT(unsigned long, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(unsigned long,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(unsigned long) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned long) - 1 > 2 * PyLong_SHIFT) {
                            return (unsigned long) (((unsigned long)-1)*(((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(unsigned long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned long) - 1 > 2 * PyLong_SHIFT) {
                            return (unsigned long) ((((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(unsigned long) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned long) - 1 > 3 * PyLong_SHIFT) {
                            return (unsigned long) (((unsigned long)-1)*(((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned long) - 1 > 3 * PyLong_SHIFT) {
                            return (unsigned long) ((((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(unsigned long) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned long) - 1 > 4 * PyLong_SHIFT) {
                            return (unsigned long) (((unsigned long)-1)*(((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned long) - 1 > 4 * PyLong_SHIFT) {
                            return (unsigned long) ((((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(unsigned long) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(unsigned long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(unsigned long) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(unsigned long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            unsigned long val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (unsigned long) -1;
        }
    } else {
        unsigned long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (unsigned long) -1;
        val = __Pyx_PyInt_As_unsigned_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to unsigned long");
    return (unsigned long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to unsigned long");
    return (unsigned long) -1;
}

/* CIntFromPy */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(long) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case  1: __PYX_VERIFY_RETURN_INT(long, digit, digits[0])
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 2 * PyLong_SHIFT) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 3 * PyLong_SHIFT) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 4 * PyLong_SHIFT) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (long) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(long) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case -1: __PYX_VERIFY_RETURN_INT(long, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(long,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(long) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(long) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            long val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (long) -1;
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* CIntFromPy */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int, digit, digits[0])
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 2 * PyLong_SHIFT) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 3 * PyLong_SHIFT) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 4 * PyLong_SHIFT) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(int,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int) -1;
        }
    } else {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_unsigned_long(unsigned long value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const unsigned long neg_one = (unsigned long) -1, const_zero = (unsigned long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(unsigned long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(unsigned long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(unsigned long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(unsigned long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(unsigned long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(unsigned long),
                                     little, !is_unsigned);
    }
}

/* CIntFromPy */
static CYTHON_INLINE size_t __Pyx_PyInt_As_size_t(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const size_t neg_one = (size_t) -1, const_zero = (size_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(size_t) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(size_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (size_t) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (size_t) 0;
                case  1: __PYX_VERIFY_RETURN_INT(size_t, digit, digits[0])
                case 2:
                    if (8 * sizeof(size_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 2 * PyLong_SHIFT) {
                            return (size_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(size_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 3 * PyLong_SHIFT) {
                            return (size_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(size_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 4 * PyLong_SHIFT) {
                            return (size_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (size_t) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(size_t) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(size_t) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (size_t) 0;
                case -1: __PYX_VERIFY_RETURN_INT(size_t, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(size_t,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(size_t) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(size_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                            return (size_t) ((((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(size_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                            return (size_t) ((((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(size_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT) {
                            return (size_t) ((((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(size_t) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(size_t) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            size_t val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (size_t) -1;
        }
    } else {
        size_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (size_t) -1;
        val = __Pyx_PyInt_As_size_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to size_t");
    return (size_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to size_t");
    return (size_t) -1;
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
    }
}

/* FastTypeChecks */
#if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = a->tp_base;
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
#if PY_MAJOR_VERSION == 2
static int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject* exc_type2) {
    PyObject *exception, *value, *tb;
    int res;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&exception, &value, &tb);
    res = exc_type1 ? PyObject_IsSubclass(err, exc_type1) : 0;
    if (unlikely(res == -1)) {
        PyErr_WriteUnraisable(err);
        res = 0;
    }
    if (!res) {
        res = PyObject_IsSubclass(err, exc_type2);
        if (unlikely(res == -1)) {
            PyErr_WriteUnraisable(err);
            res = 0;
        }
    }
    __Pyx_ErrRestore(exception, value, tb);
    return res;
}
#else
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    int res = exc_type1 ? __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type1) : 0;
    if (!res) {
        res = __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
    return res;
}
#endif
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        #if PY_MAJOR_VERSION < 3
        if (likely(exc_type == t)) return 1;
        #endif
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* CheckBinaryVersion */
static int __Pyx_check_binary_version(void) {
    char ctversion[4], rtversion[4];
    PyOS_snprintf(ctversion, 4, "%d.%d", PY_MAJOR_VERSION, PY_MINOR_VERSION);
    PyOS_snprintf(rtversion, 4, "%s", Py_GetVersion());
    if (ctversion[0] != rtversion[0] || ctversion[2] != rtversion[2]) {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compiletime version %s of module '%.100s' "
                      "does not match runtime version %s",
                      ctversion, __Pyx_MODULE_NAME, rtversion);
        return PyErr_WarnEx(NULL, message, 1);
    }
    return 0;
}

/* InitStrings */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION < 3
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        #else
        if (t->is_unicode | t->is_str) {
            if (t->intern) {
                *t->p = PyUnicode_InternFromString(t->s);
            } else if (t->encoding) {
                *t->p = PyUnicode_Decode(t->s, t->n - 1, t->encoding, NULL);
            } else {
                *t->p = PyUnicode_FromStringAndSize(t->s, t->n - 1);
            }
        } else {
            *t->p = PyBytes_FromStringAndSize(t->s, t->n - 1);
        }
        #endif
        if (!*t->p)
            return -1;
        if (PyObject_Hash(*t->p) == -1)
            return -1;
        ++t;
    }
    return 0;
}

static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    return __Pyx_PyUnicode_FromStringAndSize(c_str, (Py_ssize_t)strlen(c_str));
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#if !CYTHON_PEP393_ENABLED
static const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    char* defenc_c;
    PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
    if (!defenc) return NULL;
    defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    {
        char* end = defenc_c + PyBytes_GET_SIZE(defenc);
        char* c;
        for (c = defenc_c; c < end; c++) {
            if ((unsigned char) (*c) >= 128) {
                PyUnicode_AsASCIIString(o);
                return NULL;
            }
        }
    }
#endif
    *length = PyBytes_GET_SIZE(defenc);
    return defenc_c;
}
#else
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
}
#endif
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject* x) {
    int retval;
    if (unlikely(!x)) return -1;
    retval = __Pyx_PyObject_IsTrue(x);
    Py_DECREF(x);
    return retval;
}
static PyObject* __Pyx_PyNumber_IntOrLongWrongResultType(PyObject* result, const char* type_name) {
#if PY_MAJOR_VERSION >= 3
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type %.200s).  "
                "The ability to return an instance of a strict subclass of int "
                "is deprecated, and may be removed in a future version of Python.",
                Py_TYPE(result)->tp_name)) {
            Py_DECREF(result);
            return NULL;
        }
        return result;
    }
#endif
    PyErr_Format(PyExc_TypeError,
                 "__%.4s__ returned non-%.4s (type %.200s)",
                 type_name, type_name, Py_TYPE(result)->tp_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_Check(x) || PyLong_Check(x)))
#else
  if (likely(PyLong_Check(x)))
#endif
    return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  #if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = m->nb_int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = m->nb_long(x);
  }
  #else
  if (likely(m && m->nb_int)) {
    name = "int";
    res = m->nb_int(x);
  }
  #endif
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
    res = PyNumber_Int(x);
  }
#endif
  if (likely(res)) {
#if PY_MAJOR_VERSION < 3
    if (unlikely(!PyInt_Check(res) && !PyLong_Check(res))) {
#else
    if (unlikely(!PyLong_CheckExact(res))) {
#endif
        return __Pyx_PyNumber_IntOrLongWrongResultType(res, name);
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(b);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    const digit* digits = ((PyLongObject*)b)->ob_digit;
    const Py_ssize_t size = Py_SIZE(b);
    if (likely(__Pyx_sst_abs(size) <= 1)) {
        ival = likely(size) ? digits[0] : 0;
        if (size == -1) ival = -ival;
        return ival;
    } else {
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return b ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


#endif /* Py_PYTHON_H */
